//
// C++ Implementation: FitData
//
//
// Author: Alexander Stein <AlexanderStein@t-online.de>, (C) 2008 - 2016
//
// Copyright: See README file that comes with this source code
//
//

#include "fitdata.h"
#include "linetable.h"
#include "elstate.h"
#include "molecule.h"
#include "Spektrum.h"
#include "termtable.h"
#include "utils.h"
#include "duntable.h"
#include "potential.h"

#include <QTableWidget>
#include <QStringList>
#include <QGridLayout>
#include <QListWidget>
#include <QComboBox>
#include <QLabel>
#include <QPushButton>
#include <QPainter>
#include <QFile>
#include <QTextStream>
#include <QCheckBox>

#include <stdio.h>
#include <math.h>


FitData::FitData(ElState* nState, MainWindow* MW, Molecule* M): TableWindow(MDIChild::FitDataSet, MW, M)
{
	State = nState;
	setWindowTitle("Fit Data");
	Tab->setColumnCount(12);
	Tab->setHorizontalHeaderLabels(QStringList() << "Iso" << "v" << "J" << "v'" << "J'" << "Source" << "Prog" 
									<< "File" << "Energy" << "uncert." << "obs-calc" << "DevR");
	Tab->setRowCount(0);
	Sources = 0;
	NMarkedLevels = NSources = NSourceOffset = 0;
	FC = 0;
	lRow = -1;
	SourceOffset = 0;
	SourceOffsetNames = 0;
	connect(Tab, SIGNAL(itemSelectionChanged()), this, SIGNAL(SelChanged()));
	NewPix = new QPixmap(10, 10);
    QPainter P(NewPix);
    P.setPen(QColor(255, 0, 0));
    P.setFont(QFont("Arial", 10));
    P.drawText(0, 10, "N");
}

FitData::~FitData()
{
	if (Sources != 0) delete[] Sources;
	if (FC != 0) delete[] FC;
	if (NSourceOffset > 0)
	{
		delete[] SourceOffset;
		delete[] SourceOffsetNames;
	}
	delete NewPix;
}

int FitData::addMarkedLevel(TermEnergy& TE, Spektrum *Source)
{
	int R = Tab->rowCount();
	NMarkedLevels++;
	Tab->blockSignals(true);
	Tab->setRowCount(R+1);
	Tab->setItem(R, 0, new QTableWidgetItem(*NewPix, QString::number(TE.Iso)));
	Tab->setItem(R, 1, new QTableWidgetItem(*NewPix, QString::number(TE.v)));
	Tab->setItem(R, 2, new QTableWidgetItem(*NewPix, QString::number(TE.J)));
	Tab->setItem(R, 3, new QTableWidgetItem(*NewPix, (TE.FC == -1 ? "TE" : QString::number(-1 - TE.FC))));
	Tab->setItem(R, 4, new QTableWidgetItem(*NewPix, (TE.ef ? QString::number(TE.J + 1) : Tab->item(R, 2)->text())));
	Tab->setItem(R, 5, new QTableWidgetItem(*NewPix, Source->getName()));
	Tab->setItem(R, 6, new QTableWidgetItem(*NewPix, ""));
	Tab->setItem(R, 7, new QTableWidgetItem(*NewPix, Source->getFileName()));
	Tab->setItem(R, 8, new QTableWidgetItem(*NewPix, QString::number(TE.E, 'f', 4)));
	Tab->setItem(R, 9, new QTableWidgetItem(*NewPix, QString::number(TE.err, 'f', 4)));
	Tab->setItem(R, 10, new QTableWidgetItem(*NewPix, QString::number(TE.dev, 'f', 4)));
	Tab->setItem(R, 11, new QTableWidgetItem(*NewPix, QString::number(TE.DevR, 'f', 4)));
	Tab->blockSignals(false);
	setBlockChangeSignal(true);
	Changed();
	setBlockChangeSignal(false);
	return R;
}

void FitData::AddRow()
{
	if (Tab == 0) return;
	int r, c, nr = Tab->rowCount(), nc = Tab->columnCount(), cr = Tab->currentRow(), *FCB = new int[nr + 1];
	LineTable **SB = new LineTable*[nr+1];
	Tab->blockSignals(true);
	Tab->setRowCount(nr + 1);
	if (cr == -1) cr = 0;
	for (r=0; r < cr && r < NSources; r++)
	{
		SB[r] = Sources[r];
		FCB[r] = FC[r];
	}
	for (r = nr; r > cr; r--)
	{
		for (c=0; c < nc; c++) Tab->setItem(r, c, Tab->takeItem(r-1, c));
		if (r <= NSources)
		{
			SB[r] = Sources[r-1];
			FCB[r] = FC[r-1];
		}
		else
		{
			SB[r] = 0;
			FCB[r] = -1;
		}
	}
	SB[cr] = 0;
	FCB[cr] = -1;
	NSources = nr + 1;
	if (Sources != 0)
	{
		delete[] Sources;
		delete[] FC;
	}
	Sources = SB;
	FC = FCB;
	Tab->blockSignals(false);
	Changed();
}


void FitData::clearMarkedLevels()
{
	if (NMarkedLevels > 0)
	{
		Tab->setRowCount(Tab->rowCount() - NMarkedLevels);
		NMarkedLevels = 0;
	}
}

void FitData::DeleteRows()
{
    if (Tab == 0) return;
	Tab->blockSignals(true);
	int j, k, n = Tab->rowCount(), c, nc = Tab->columnCount();
	QList<QTableWidgetSelectionRange> SR = Tab->selectedRanges();
	if (NSources > 0) 
	{
		for (j=k=0; j < SR.count() && (j==0 || Sources[k] == 0) ; j++) 
			for (k = SR[j].topRow(); (k <= SR[j].bottomRow() ? (k < NSources ? Sources[k] == 0 : true) : false); k++) ;
		if (Sources[k] != 0)
		{
			QMessageBox::StandardButton B = QMessageBox::question(this, "MolSpektAnalysis", 
				"Delete the lines/progessions from the linetable, too?", QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel,
				QMessageBox::No);
			if (B == QMessageBox::Cancel) return;
			if (B == QMessageBox::Yes)
			{
				QList<LineTable*> SL;
				QList<int> CL;
				for (j=0; j < SR.count(); j++) for (k = SR[j].topRow(); k <= SR[j].bottomRow(); k++) if (k < NSources)
				{
					for (n=0; (n < SL.count() ? SL[n] != Sources[k] : false); n++) ;
					if (n == SL.count())
					{
						SL.append(Sources[k]);
						CL.append(1);
					}
					else CL[n]++;
				}
				int *PN[SL.count()], *vss[SL.count()], *Jss[SL.count()];
				for (n=0; n < SL.count(); n++)
				{
					PN[n] = new int[CL[n]];
					if (SL[n]->getTransition()->getLowerState() == State)
					{
						vss[n] = new int[CL[n]];
						Jss[n] = new int[CL[n]];
					}
					else vss[n] = Jss[n] = 0;
					CL[n] = 0;
				}
				for (j=0; j < SR.count(); j++) for (k = SR[j].topRow(); k <= SR[j].bottomRow(); k++) if (k < NSources)
				{
					for (n=0; SL[n] != Sources[k]; n++) ;
					if (vss[n] != 0)
					{
						vss[n][CL[n]] = Tab->item(k, 1)->text().toInt();
						Jss[n][CL[n]] = Tab->item(k, 2)->text().toInt();
					}
					PN[n][CL[n]++] = Tab->item(k, 6)->text().toInt();
				}
				for (n=0; n < SL.count(); n++)
				{
					SL[n]->DeleteRows(CL[n], PN[n], vss[n], Jss[n]);
					delete[] PN[n];
					if (vss[n] != 0)
					{
						delete[] vss[n];
						delete[] Jss[n];
					}
				}
			}
		}
	}
	for (j=0; j < SR.size(); j++) for (k=SR[j].topRow(); k <= SR[j].bottomRow(); k++) 
			for (c=0; c < nc; c++) Tab->setItem(k, c, 0);
	for (j=0; (j < n ? Tab->item(j, 0) != 0 : false); j++) ;
    for (k=j; k < n; k++) if (Tab->item(k, 0) != 0)
	{
		for (c=0; c < nc; c++) Tab->setItem(j, c, Tab->takeItem(k, c));
		if (k < NSources)
		{
			Sources[j] = Sources[k];
			FC[j] = FC[k];
		}
		else if (j < NSources)
		{
			Sources[j] = 0;
			FC[j] = -1;
		}
		j++;
	}
	if (j < NSources) NSources = j;
    Tab->setRowCount(j);
	Tab->blockSignals(false);
	Changed();
}

QList< LevelComb > FitData::getAvLevelComb(int iso, int comp, int vmin, int vmax)
{
	QList<LevelComb> R;
	LevelComb c;
	int n, m, N = Tab->rowCount();
	for (n=0; n<N; n++) 
	{
		c.v = Tab->item(n, fdcv)->text().toInt();
		if (c.v < vmin || c.v > vmax) continue;
		c.iso = Tab->item(n, fdcIso)->text().toInt();
		if (iso != -1 && iso != c.iso) continue;
		c.ef = 1 - abs(Tab->item(n, fdcJ)->text().toInt() 
					- Tab->item(n, fdcJs)->text().toInt());
		if (comp < 2 && 1 - comp != c.ef) continue;
		for (m=0; m < R.size() && (R[m].iso != c.iso 
			|| R[m].v != c.v || R[m].ef != c.ef); m++) ;
		if (m == R.size()) R.append(c);
	}
	return R;
}

void FitData::deleteRows(int* rows, int N)
{
	int n, m, c, NR = Tab->rowCount(), NC = Tab->columnCount();
	int mb = NR - NMarkedLevels;
	Tab->blockSignals(true);
	if (NSources > 0) 
	{
		for (n=0; (n<N ? (rows[n] < NSources ? Sources[rows[n]] == 0 : true) : false); n++) ;
		if (n<N ? Sources[rows[n]] != 0 : false)
		{
			QMessageBox::StandardButton B = QMessageBox::question(this, "MolSpektAnalysis", 
				"Delete the lines/progessions from the linetable, too?", QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel,
				QMessageBox::No);
			if (B == QMessageBox::Cancel) return;
			if (B == QMessageBox::Yes)
			{
				QList<LineTable*> SL;
				QList<int> CL;
				for (m=0; m<N; m++) if (rows[m] < NSources)
				{
					for (n=0; (n < SL.count() ? SL[n] != Sources[rows[m]] : false); n++) ;
					if (n == SL.count())
					{
						SL.append(Sources[rows[m]]);
						CL.append(1);
					}
					else CL[n]++;
				}
				int *PN[SL.count()], *vss[SL.count()], *Jss[SL.count()];
				for (n=0; n < SL.count(); n++)
				{
					PN[n] = new int[CL[n]];
					if (SL[n]->getTransition()->getLowerState() == State)
					{
						vss[n] = new int[CL[n]];
						Jss[n] = new int[CL[n]];
					}
					else vss[n] = Jss[n] = 0;
					CL[n] = 0;
				}
				for (m=0; m<N; m++) if (rows[m] < NSources)
				{
					for (n=0; SL[n] != Sources[rows[m]]; n++) ;
					if (vss[n] != 0)
					{
						vss[n][CL[n]] = Tab->item(rows[m], 1)->text().toInt();
						Jss[n][CL[n]] = Tab->item(rows[m], 2)->text().toInt();
					}
					PN[n][CL[n]++] = Tab->item(rows[m], 6)->text().toInt();
				}
				for (n=0; n < SL.count(); n++)
				{
					SL[n]->DeleteRows(CL[n], PN[n], vss[n], Jss[n]);
					delete[] PN[n];
					if (vss[n] != 0)
					{
						delete[] vss[n];
						delete[] Jss[n];
					}
				}
			}
		}
	}
	for (n=0; n<N; n++)
	{
		Tab->setItem(rows[n], 0, 0);
		if (rows[n] >= mb) NMarkedLevels--;
	}
	for (n=0; (n < NR ? Tab->item(n, 0) != 0 : false); n++) ;
	for (m=n+1; m < NR; m++) if (Tab->item(m, 0) != 0) 
	{
		for (c=0; c < NC; c++) Tab->setItem(n, c, Tab->takeItem(m, c));
		if (n < NSources)
		{
			if (m < NSources)
			{
				FC[n] = FC[m];
				Sources[n] = Sources[m];
			}
			else NSources = n;
		}
		n++;
	}
	if (NSources > n) NSources = n;
	Tab->blockSignals(false);
	Tab->setRowCount(n);
	Changed();
}

void FitData::findWrongData()
{
	if (State == 0)
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
						"The data set has to be assigned to an electronic state first!");
		return;
	}
	int r, NR = Tab->rowCount(), l = State->getLambda(), c, I, J;
	/*if (State->getS() > 0.0)
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
			"The search for wrong levels is not implemented for the current kind of electronic state!");
		return;
	}*/
	int NI = molecule->getNumIso(), NC = (l==0 ? 1 : 2);
	int JStep[NI], **JStart = CreateInt(NI, NC);
	Transition *Tr;
	ElState *St;
	QString B;
	if (NR == 0) return;
	for (r=0; r < NI; r++) 
	{
		JStep[r] = molecule->getJStep(r);
		for (c=0; c < NC; c++) JStart[r][c] = State->getJStart(r, 1 - c);
	}
	if (lRow == -1) lRow = NR;
	for (r = lRow + 1; r != lRow; r++) 
	{
		if (r >= NR) if (lRow == (r=0)) break;
		J = Tab->item(r, 2)->text().toInt();
		I = Tab->item(r, 0)->text().toInt();
		c = (Tab->item(r, 4)->text().toInt() == J ? 1 : 0);
		B = Tab->item(r, 3)->text();
		if (c == 1 && l == 0 ) 
			if (B == "TE" || B.toInt() < -1 ? true : (r < NSources ? 
				(Sources[r] != 0 ? ((Tr = Sources[r]->getTransition()) != 0 ? 
				  ((St = Tr->getUpperState()) != 0 ? St->getLambda() == 0 
				    : false) : false) : false) : false)) break;
		if (((J - JStart[I][c]) / JStep[I]) * JStep[I] + JStart[I][c] != J) break;
		if (fabs(Tab->item(r, 11)->text().toDouble()) >= 3.5) break;
	}
	if (r != lRow) 
	{
		Tab->selectRow(lRow = r);
		if (r < NSources && Tab->item(r, 3)->text() != "TE" ? Sources[r] != 0 && Tab->item(r, 3)->text().toInt() >= 0 : false)
		{
			double WN = Tab->item(r, 8)->text().toDouble();
			I = Tab->item(r, 0)->text().toInt();
			J = Tab->item(r, 4)->text().toInt();
			if ((c = (B = Tab->item(r, 3)->text()).toInt()) == 0)
				if (B == "TE" || B == "nA") c = -1;
			Sources[r]->MarkLines(&I, &c, &J, &WN, 1);
		}
	}
	else QMessageBox::information(this, "MolSpektAnalysis", "No bad levels found!");
	Destroy(JStart, NI);
}

QList< int > FitData::getaFC()
{
	int i, j;
	QList<int> R;
	for (i=0; i < NSources; i++)
	{
		for (j=0; (j < R.count() ? FC[i] != R[j] : false); j++) ;
		if (j == R.count()) R.append(FC[i]);
	}
	for (i=0; i < R.count() - 1; i++) for (j=i+1; j < R.count(); j++) if (R[i] > R[j]) R.swap(i, j);
	return R;
}

void FitData::getav(int &nv, int*& v)
{
	int n, b, N = getNumLines(), mv=0, Mv=0;
	for (n=0; n<N; n++)
	{
		if ((b = Tab->item(n, 1)->text().toInt()) < mv) mv = -1;
		if (b > Mv) Mv = b;
	}
	bool *va = new bool[Mv - mv + 1];
	for (n=0; n <= Mv - mv; n++) va[n] = false;
	for (n=0; n<N; n++) va[((b = Tab->item(n, 1)->text().toInt()) >= 0 ? b - mv : 0)] = true;
	for (n = nv = 0; n <= Mv - mv; n++) if (va[n]) nv++;
	v = new int[nv];
	for (n=mv, b=0; n <= Mv; n++) if (va[n]) v[b++] = n;
	delete[] va;
}

void FitData::getavIso(int*& Iso, int& NIso)
{
	int r, i, NR = Tab->rowCount(), N = molecule->getNumIso();
	bool IsoAv[N];
	for (r=0; r<N; r++) IsoAv[r] = false;
	for (r=0; r < NR; r++) 
		if ((i = (Tab->item(r, 0) != 0 ? Tab->item(r, 0)->text().toInt() : -1)) >= 0 ? i<N : false) 
			IsoAv[i] = true;
	for (i = NIso = 0; i<N; i++) if (IsoAv[i]) NIso++;
	if (NIso > 0) Iso = new int[NIso];
	for (i=r=0; i<N; i++) if (IsoAv[i]) Iso[r++] = i;
}

void FitData::getData(TableLine*& Lines, int& NLines, int JD, int F, int v, int mJ, int Iso)
{
	
	if ((NLines = getNumLines(JD, F, v, mJ, Iso)) > 0) 
	{
		Lines = new TableLine[NLines];
		int N = getNumLines();
		int *SA = new int[N], n, *S1 = heapSort(sortByProg);
		for (n=0; n < N; n++) SA[S1[n]] = n;
		delete[] S1;
		getData(Lines, SA, NLines, 0, v, 0, mJ, JD, F, Iso);
		delete[] SA;
	}
	else Lines = 0;
}

void FitData::getData(TableLine*& Lines, int& NLines, int *&RowN,
					  bool sortFuncs(QTableWidget *Tab, int n, int m), int *mv, int mJ)
{
	if ((NLines = (mv != 0 ? getNumLines(mv, mJ) : getNumLines())) != 0)
	{
		int n, N = getNumLines();
		int *SA = new int[N], *S1 = heapSort(sortFuncs);
		RowN = new int[NLines];
		for (n=0; n<N; n++) SA[S1[n]] = n;
		delete[] S1;
		Lines = new TableLine[NLines];
		getData(Lines, SA, NLines, RowN, -2, mv, mJ);
		delete[] SA;
	}
	else Lines = 0;
}

void FitData::getData(TableLine* Lines, int* SA, int NLines, int *RowN, int mv, int *Mv, int mJ, int JD, int F,
					  int Iso)
{
	int n, b, m, MCT = 0, *CompTrans = 0, i, NR = Tab->rowCount();
	QString B;
	ElState *XState = (molecule != 0 ? molecule->getStateP(0) : 0);
	TermTable *TT = (XState != 0 ? XState->getTermTable(false) : 0);
	QList<QTableWidgetSelectionRange> SL = Tab->selectedRanges();
	bool *RV = new bool[NR], *RS = new bool[NR];
	bool WFact = (Tab->columnCount() > 12 ? Tab->horizontalHeaderItem(12)->text() == "WFact" : false);
	getViewnRows(RV);
	for (n=0; n < NR; n++) RS[n] = false;
	for (n=0; n < SL.count(); n++) for (m=SL[n].topRow(); m <= SL[n].bottomRow(); m++) RS[m] = true;
	if (TT != 0) TT->getCompT(MCT, CompTrans);
	for (n=m=0; m < NLines; n++)
	{
		if (Tab->item(SA[n], 0) == 0) continue;
		if (Iso >= 0 && Tab->item(SA[n], 0)->text().toInt() != Iso) continue;
		Lines[m].Iso = Tab->item(SA[n], 0)->text().toInt();
		Lines[m].vss = (Tab->item(SA[n], 1) != 0 ? Tab->item(SA[n], 1)->text().toInt() : -1);
		Lines[m].Jss = (Tab->item(SA[n], 2) != 0 ? Tab->item(SA[n], 2)->text().toInt() : -1);
		if ((mJ > 0 ? Lines[m].Jss > mJ : false) || (Mv != 0 ? (mv >= 0 ? Lines[m].vss < mv : false) 
			 || (Lines[m].Jss <= mJ ? Lines[m].vss > Mv[Lines[m].Jss] : true) 
				: (mv >= 0 ? Lines[m].vss != mv : mv == -1 && Lines[m].vss >= 0)))
			continue;
		Lines[m].Js = (Tab->item(SA[n], 4) != 0 ? Tab->item(SA[n], 4)->text().toInt() : -1);
		if (JD != -1 ? fabs(Lines[m].Js - Lines[m].Jss) != JD : false) continue;
		if (F != -2 ? (SA[n] < NSources ? FC[SA[n]] != F : F != -1): false) continue;
		B = Tab->item(SA[n], 3)->text();
		if (B == "TE") 
		{
			Lines[m].vs = -1;
			Lines[m].isTE = true;
		}
		else if ((b = B.toInt()) < -1)
		{
			Lines[m].vs = b;
			Lines[m].isTE = true;
		}
		else
		{
			Lines[m].isTE = false;
			Lines[m].vs = (B == "nA" ? -1 : B.toInt());
		}
		Lines[m].FC = ((SA[n] < NSources ? FC[SA[n]] >= 0 : false) ? 
			(FC[SA[n]] <= MCT && CompTrans != 0 ? CompTrans[FC[SA[n]]] : FC[SA[n]]) : 0);
		//printf("Lines[%d].FC=%d\n", m, Lines[m].FC);
		B = Tab->item(SA[n], 5)->text();
		Lines[m].LTab = (SA[n] < NSources ? Sources[SA[n]] : 0);
		Lines[m].SourceName = B;
		if (Lines[m].LTab == 0 && molecule != 0 && !B.isEmpty() && !SaveMemory)
		{
			Lines[m].LTab = molecule->getLineTable(B);
			if (SA[n] < NSources) Sources[SA[n]] = Lines[m].LTab;
		}
		Lines[m].PN = (Tab->item(SA[n], 6) != 0 ? Tab->item(SA[n], 6)->text().toInt() : -1);
		Lines[m].File = (Tab->item(SA[n], 7) != 0 ? Tab->item(SA[n], 7)->text() : "");
		Lines[m].WN = (Tab->item(SA[n], 8) != 0 ? Tab->item(SA[n], 8)->text().toDouble() : 0.0);
		if (Lines[m].WN == 0.0) continue;
		if (Tab->item(SA[n], 5) != 0) for (i=0; i < NSourceOffset; i++) 
			if (Lines[m].File == SourceOffsetNames[i] || B == SourceOffsetNames[i]) Lines[m].WN -= SourceOffset[i]; 
		Lines[m].err = (Tab->item(SA[n], 9) != 0 ? (B = Tab->item(SA[n], 9)->text()).toDouble() : 0.0);
		Lines[m].dev = (Tab->item(SA[n], 10) != 0 ? Tab->item(SA[n], 10)->text().toDouble() : 0.0);
		Lines[m].DevR = (Tab->item(SA[n], 11) != 0 ? Tab->item(SA[n], 11)->text().toDouble() : 0.0);
		if (RowN != 0) RowN[m] = SA[n];
		Lines[m].isViewn = RV[SA[n]];
		Lines[m].isSelected = RS[SA[n]];
		Lines[m].nDig = B.length() - B.indexOf('.') - 1;
		Lines[m].WeightFact = (WFact ? Tab->item(SA[n], 12)->text().toDouble() : 1.0);
		Lines[m++].Row = SA[n];
	}
	delete[] RV;
	delete[] RS;
	delete[] CompTrans;
}

bool FitData::checkSourceConnections()
{
    int n, N = Tab->rowCount(), CNSources = NSources;
    QMap<QString, LineTable*>& Translations = MW->getLineTableTranslations();
    bool result = true;
    if (Sources == 0) Sources = new LineTable*[NSources = N];
    else if (CNSources < N)
    {
        LineTable**SB = new LineTable*[NSources = N];
        for (n=0; n < CNSources; ++n) SB[n] = Sources[n];
        while (n < NSources) SB[n++] = 0;
        delete[] Sources;
        Sources = SB;
    }
    for (n=0; n<N; ++n) if (Sources[n] == 0)
    {
        QString Name = Tab->item(n, fdcSource)->text();
        if (Tab->item(n, fdcFile)->text().indexOf(Name) == -1)
        {
            if (Translations.contains(Name)) Sources[n] = Translations[Name];
            else
            {
                LineTable* LTab = molecule->getLineTable(Name);
                if (LTab == 0) LTab = MW->getLineTable(Name, molecule);
                if (LTab != 0)
                {
                    Sources[n] = LTab;
                    Translations[Name] = LTab;
                }
                else result = false;
            }
            if (Sources[n] != 0) Tab->item(n, fdcSource)->setText(Sources[n]->getName());
        }
    }
    return result;
}

ElState* FitData::getElState()
{
	return State;
}

int FitData::getMaxJ()
{
	int MJ, n, J, N = Tab->rowCount();
	for (n = MJ = 0; n<N; n++) if ((J = Tab->item(n, 2)->text().toInt()) > MJ) MJ = J;
	return MJ;
}

int FitData::getMaxv()
{
	int Mv, n, v, N = Tab->rowCount();
	for (n = Mv = 0; n<N; n++) if ((v = Tab->item(n, 1)->text().toInt()) > Mv) Mv = v;
	return Mv;
}

bool FitData::isDataAvailable()
{
	if (Tab->rowCount() > 0) return true;
	return false;
}

bool FitData::readData(QString Filename)
{
    if (!TableWindow::readData(Filename)) 
	{
		QFile F(Filename);
		F.open(QIODevice::ReadOnly);
		QTextStream S(&F);
		int r, c, NR = S.readLine().left(5).toInt(), mv = 0, mJ=0;
		bool Error = false;
		if (NR == 0) return false;
		QStringList Buffer = S.readLine().split(' ', QString::SkipEmptyParts);
		if (Buffer.count() >= 2)
		{
			setName("FitData" + ((r = Buffer[1].lastIndexOf('.')) != -1 ? Buffer[1].left(r) : Buffer[1]));
			setSource("Asen");
			setFileName(((r = Filename.lastIndexOf('.')) != -1 ? Filename.left(r) : Filename) + ".fdat");
		}
		else Error = true;
		if (Sources != 0) delete[] Sources;
		Sources = new LineTable*[NSources = NR];
		if (FC == 0) delete[] FC;
		FC = new int[NR];
		for (r=0; r<4; r++) S.readLine();
		Tab->blockSignals(true);
		Tab->setRowCount(NR);
		for (r=0; r < NR; r++)
		{
			if (S.atEnd())
			{
				Error = true;
				break;
			}
			Buffer = S.readLine().split(' ', QString::SkipEmptyParts);
			if (Buffer.count() < 7)
			{
				Error = true;
				continue;
			}
			Tab->setItem(r, 0, new QTableWidgetItem(QString::number((Buffer[6].toInt() - 1) / 10)));
			Tab->setItem(r, 1, new QTableWidgetItem(Buffer[1]));
			if ((c = Buffer[1].toInt()) > mv) mv = c;
			Tab->setItem(r, 2, new QTableWidgetItem(Buffer[2]));
			if ((c = Buffer[2].toInt()) > mJ) mJ = c;
			Tab->setItem(r, 3, new QTableWidgetItem("TE"));
			Tab->setItem(r, 4, new QTableWidgetItem(Buffer[4] == "f" ? Buffer[2] : QString::number(c+1)));
			for (c=5; c<=7; c++) Tab->setItem(r, c, new QTableWidgetItem(""));
			Tab->setItem(r, 8, new QTableWidgetItem(Buffer[3]));
			Tab->setItem(r, 9, new QTableWidgetItem(Buffer[5]));
			for (c=10; c<=12; c++) Tab->setItem(r, c, new QTableWidgetItem(""));
			Sources[r] = 0;
			FC[r] = ((c = Buffer[4].toInt()) > 0 ? c : -1);
		}
		if (Error) Tab->setRowCount(r);
		Tab->blockSignals(false);
		setvMax(mv);
		setJMax(mJ);
		emit propertiesChanged();
		return !Error;
	}
	else
	{
		int n, m, N = Tab->rowCount();
		if (NSourceOffset > 0)
		{
			delete[] SourceOffset;
			delete[] SourceOffsetNames;
		}
		for (n=0; (n<N ? (Tab->item(n, 0) != 0 ? Tab->item(n, 0)->text() != "SourceOffsets:" : true) : false); n++) ;
		if (n<N)
		{
			NSourceOffset = N-n-1;
			SourceOffset = new double[NSourceOffset];
			SourceOffsetNames = new QString[NSourceOffset];
			for (m=n+1; m<N; m++)
			{
				SourceOffsetNames[m-n-1] = (Tab->item(m, 0) != 0 ? Tab->item(m, 0)->text() : "");
				SourceOffset[m-n-1] = (Tab->item(m, 1) != 0 ? Tab->item(m, 1)->text().toDouble() : 0.0);
			}
			Tab->setRowCount(n);
		}
		else
		{
			NSourceOffset = 0;
			SourceOffsetNames = 0;
			SourceOffset = 0;
		}
	}
	Tab->blockSignals(true);
    int n, N = Tab->rowCount(), b;
    if (Sources != 0) delete[] Sources;
    Sources = new LineTable*[NSources = N];
	if (FC == 0) delete[] FC;
	FC = new int[N];
	for (n=0; n<N; n++)
	{
		if (n>0 ? Tab->item(n, 5)->text() == Tab->item(n-1, 5)->text() : false) Sources[n] = Sources[n-1];
		else Sources[n] = (molecule != 0 && !SaveMemory ? molecule->getLineTable(Tab->item(n, 5)->text()) : 0);
		if ((b = Tab->item(n, 3)->text().toInt()) == -1) Tab->item(n, 3)->setText("TE");
		if (b >= 990)
		{
			FC[n] = (b + 10) / 1000 - 1;
			Tab->item(n, 3)->setText((b -= 1000 * (FC[n] + 1)) == -1 ? "TE" : (b == -10 ? "nA" : QString::number(b)));
		}
		else FC[n] = -1;
        if (molecule != 0)
        {
            QString Path = Tab->item(n, fdcFile)->text(), MolPath = molecule->getFileName();
            Tab->item(n, fdcFile)->setText(getAbsolutePath(Path, MolPath));
        }
	}
	if (Tab->columnCount() > 12) Tab->setHorizontalHeaderItem(12, new QTableWidgetItem("WFact"));
	Tab->blockSignals(false);
	emit propertiesChanged();
	return true;
}

void FitData::Assign_v(double**** TE, int NC, int NI, int NJ, int *Nv, int* IsoTrans, int* CompTrans, double Tolerance, 
					   double DoublAssTol)
{
	int CS = ((State != 0 ? State->getLambda() : 0) == 0 ? 1 : 2), ME, NE = 0, I, c, Jss, Js, n, m, i, j=0, NR = Tab->rowCount();
	int MC = 0;
	for (n=0; n < NC; n++) if (CompTrans[n] > MC) MC = CompTrans[n];
	for (n = ME = 0; n <= MC; n++) ME += Nv[n];
	double E[ME], LE, LLE = 0.0, Dev;
	int va[ME], CA[ME], *S1 = heapSort(sortIefJFreq), *SA = new int[NR], lv = -1, lc = -1, fc, k;
	for (n=0; n < NR; n++) SA[S1[n]] = n;
	int vc[MC+1];
	QString B;
	Tab->blockSignals(true);
	for (m=n=0, I = c = Jss = Js = -1; m <= NR; m++)
	{
		if (m < NR ? I == Tab->item(SA[m], 0)->text().toInt() && Jss == Tab->item(SA[m], 2)->text().toInt() 
					 && Js == Tab->item(SA[m], 4)->text().toInt() : false) continue;
		if (NE > 0) for (j=n, i=0; j<m; j++)
		{
			for (LE = Tab->item(SA[j], 8)->text().toDouble(); (i < NE ? E[i] < LE : false); i++) ;
			if (i>0 ? fabs(E[i] - LE) > fabs(E[i-1] - LE) : false) i--;
			if (fabs(Dev = LE - E[i]) <= Tolerance)
			{
				if (lv == va[i] && lc == CA[i])
				{
					if (LE - LLE < DoublAssTol && Tab->item(SA[j], 7)->text() != Tab->item(SA[j-1], 7)->text()) fc = lc;
					else
					{
						if (fabs(E[i] - LE) < fabs(E[i] - LLE))
						{
							if (j-2 >= n ? lv == Tab->item(SA[j-2], 1)->text().toInt() 
											&& (k = Tab->item(SA[j-2], 3)->text().toInt()) < -1 - MC : false)
							{
								Tab->item(SA[j-1], 3)->setText(QString::number(k-1));
								FC[SA[j-1]] = -k;
							}
							else 
							{
								Tab->item(SA[j-1], 3)->setText(QString::number(-2 - MC));
								FC[SA[j-1]] = 1 + MC;
							}
							fc = lc;
						}
						else
						{
							if ((k = Tab->item(SA[j-1], 3)->text().toInt()) < -1 - MC) fc = -k;
							else
							{
								if (j-2 >= n ? lv == Tab->item(SA[j-2], 1)->text().toInt()
												&& (k = Tab->item(SA[j-2], 3)->text().toInt()) < -1 - MC : false) fc = -k;
								else fc = MC + 1;
							}
						}
					}
				}
				else fc = CA[i];
			}
			else
			{
				va[i] = -1;
				if (lv == va[i] && (k = Tab->item(SA[j-1], 3)->text().toInt()) < -1 - MC)
					fc = -k;
				else fc = MC + 1; 
			}
			Tab->item(SA[j], 1)->setText(QString::number(lv = va[i]));
			if (fc == 0)
			{
				Tab->item(SA[j], 3)->setText("TE");
				FC[SA[j]] = -1;
			}
			else Tab->item(SA[j], 3)->setText(QString::number(-1 - (FC[SA[j]] = fc)));
			Tab->item(SA[j], 10)->setText(QString::number(Dev, 'f', 4));
			Tab->item(SA[j], 11)->setText(QString::number(Dev / Tab->item(SA[j], 9)->text().toDouble(), 'f', 4));
			lc = CA[i];
			LLE = LE;
		}
		if (m < NR)
		{
			I = Tab->item(SA[n=m], 0)->text().toInt();
			Jss = Tab->item(SA[n], 2)->text().toInt();
			if ((I < NI ? IsoTrans[I] >= 0 : false) && Jss < NJ)
			{
				Js = Tab->item(SA[n], 4)->text().toInt();
				c = (MC > 0 ? 1 - fabs(Jss - Js) : 0);
				lv = lc = -1;
				for (i=0; i <= MC; i++) vc[i] = 0;
				for (NE = 0; true; )
				{
					for (i=c, E[NE] = 1e10; i < NC; i += CS) 
						if (vc[k = CompTrans[i]] < Nv[CompTrans[i]] ? (LE = TE[k][IsoTrans[I]][vc[k]][Jss]) < E[NE] : false)
							if (LE != 0.0)
					{
						E[NE] = LE;
						j=k;
					}
					if (E[NE] != 1e10)
					{
						va[NE] = vc[j]++;
						CA[NE++] = j;
					}
					else break;
				}
			}
			else NE = 0;
		}
	}
	Tab->blockSignals(false);
	delete[] S1;
	delete[] SA;
	Changed();
}

void FitData::getNumLevels(int**& LNum, int& NumIso, int& NumComp, int type)
{
	int n, c, I, J, v, NR = Tab->rowCount(), Nv = 0, NJ = 0;
	int **b = CreateInt(NR, 4);
	bool ef = (State != 0 ? (State->getLambda() > 0 ? true : false) : false), ****L;
	for (n = 0, NumIso = NumComp = 1; n < NR; n++)
	{
		if ((b[n][0] = Tab->item(n, 0)->text().toInt()) >= NumIso) NumIso = b[n][0] + 1;
		if ((b[n][3] = (FC[n] >= 0 ? FC[n] : -1 - Tab->item(n, 3)->text().toInt())) >= NumComp) 
			NumComp = b[n][3] + 1;
		if (b[n][3] < 0) b[n][3] = 0;
		if ((b[n][1] = Tab->item(n, 1)->text().toInt()) >= Nv) Nv = b[n][1] + 1;
		if ((b[n][2] = Tab->item(n, 2)->text().toInt()) >= NJ) NJ = b[n][2] + 1;
	}
	if (ef) NumComp *= 2;
	L = CreateBool(NumIso, NumComp, Nv, NJ);
	for (n=0; n < NumComp; n++) for (I=0; I < NumIso; I++) for (v=0; v < Nv; v++)
		for (J=0; J < NJ; J++) L[I][n][v][J] = false;
	for (n=0; n < NR; n++) 
		if (type == 0 || (type == 1 && Sources[n] != 0) || (type == 2 && Sources[n] == 0))
	{
		c = (ef ? (b[n][2] == Tab->item(n, 4)->text().toInt() ? 2 * b[n][3] + 1 
				: 2 * b[n][3]) : b[n][3]);
		L[b[n][0]][c][b[n][1]][b[n][2]] = true;
	}
	Destroy(b, NR);
	LNum = CreateInt(NumIso, NumComp);
	for (I=0; I < NumIso; I++) for (c=0; c < NumComp; c++) 
		for (LNum[I][c] = v = 0; v < Nv; v++) for (J=0; J < NJ; J++) 
			if (L[I][c][v][J]) LNum[I][c]++;
	Destroy(L, NumIso, NumComp, Nv);
}

int FitData::getNumLines(int* mv, int mJ)
{
	int n, R, J, N = Tab->rowCount() - NMarkedLevels;
	for (n=R=0; n<N; n++) if (Tab->item(n, 0) != 0 && Tab->item(n, 8) != 0 ? Tab->item(n, 8)->text().toDouble() != 0 : false) 
		if (Tab->item(n, 2) != 0 ? ((J = Tab->item(n, 2)->text().toInt()) <= mJ ? 
			 Tab->item(n, 1)->text().toInt() <= mv[J] : false) : true) R++;
	return R;
}

int FitData::getNumLines(int JD, int F, int v, int mJ, int Iso)
{
	int N = getNumLines(), n, R;
	for (n=R=0; n<N; n++) if (Tab->item(n, 0) != 0 && Tab->item(n, 8) != 0 ? Tab->item(n, 8)->text().toDouble() != 0 : false)
		if ((mJ > 0 && Tab->item(n, 2) != 0 ? Tab->item(n, 2)->text().toInt() <= mJ : true) && (JD >= 0 ? 
			JD == abs((Tab->item(n, 2) != 0 ? Tab->item(n, 2)->text().toInt() : -1) 
					- (Tab->item(n, 4) != 0 ? Tab->item(n, 4)->text().toInt() : -1)) : true)
			&& (Iso == -1 || Tab->item(n, 0)->text().toInt() == Iso)
			&& (F != -2 ? (n < NSources ? FC[n] == F : F == -1) : true)
		     && (v >= 0 && Tab->item(n, 1) != 0 ? (v == Tab->item(n, 1)->text().toInt()) 
			      : (v == -1 && Tab->item(n, 1) != 0 ? Tab->item(n, 1)->text().toInt() < 0 : true))) R++;
	return R;
}

int FitData::getNumLines()
{
    int n, N=0, NR = Tab->rowCount() - NMarkedLevels;
	for (n=0; n < NR; n++) if (Tab->item(n, 0) != 0 && Tab->item(n, 8) != 0 ? Tab->item(n, 8)->text().toDouble() != 0 : false) N++;
	return N;
}

int FitData::getNumProgressions(int *mv, int mJ)
{
	int n, m, l, b, R, J, N = getNumLines();
	if (N==0) return 0;
	int *S1 = heapSort(sortByProg), *SA = new int[N];
	for (n=0; n<N; n++) SA[S1[n]] = n;
	delete[] S1;
	LineTable *LTab = (NSources > 0 ? Sources[SA[0]] : 0);
	for (l = Tab->item(SA[0], 6)->text().toInt(), n=1, R=m=0; n<N; n++)
	{
		if (mJ > 0 ? ((J = Tab->item(SA[n], 2)->text().toInt()) <= mJ ? 
			 Tab->item(SA[n], 1)->text().toInt() > mv[J] : true) : false) continue;
		if ((b = Tab->item(SA[n], 6)->text().toInt()) == l && (SA[n] < NSources ? LTab == Sources[SA[n]] : true)) m++;
		else
		{
			if (m>0 && l>0)
			{
				R++;
				//printf("%d, ", n-m-1);
			}
			l=b;
			m=0;
			LTab = (SA[n] < NSources ? Sources[SA[n]] : 0);
		}
	}
	if (m>0) R++;
	//printf("\n");
	delete SA;
	return R;
}

void FitData::getSourceOffset(QStringList& Names, double*& Offsets)
{
	int n, m, N = Tab->rowCount();
	QString Buffer;
	for (n=0; n<N; n++)
	{
		for (m=0, Buffer = Tab->item(n, 5)->text(); (m < Names.count() ? Buffer != Names[m] : false); m++) ;
		if (m == Names.count() && !Buffer.isEmpty()) Names.append(Buffer);
		for (m=0, Buffer = Tab->item(n, 7)->text(); (m < Names.count() ? Buffer != Names[m] : false); m++) ;
		if (m == Names.count() && !Buffer.isEmpty()) Names.append(Buffer);
	}
	if (Names.count() > 0) Offsets = new double[Names.count()];
	for (m=0; m < Names.count(); m++) Offsets[m] = 0.0;
	for (n=0; n < NSourceOffset; n++)
	{
		for (m=0; (m < Names.count() ? SourceOffsetNames[n] != Names[m] : false); m++) ;
		if (m < Names.count()) Offsets[m] = SourceOffset[n];
	}
}

QStringList FitData::getSources()
{
	int n, N = Tab->rowCount();
	QStringList RList;
	QString Buffer;
	for (n=0; n<N; n++) if (!RList.contains(Buffer = Tab->item(n, 5)->text())) RList.append(Buffer);
	return RList;
}

void FitData::getUncertaintyStats(QList< double >& Uncert, QList< int >& Numbers)
{
	int i, n, N = Tab->rowCount(), vM = getvMax(), JM = getJMax();
	double U;
	for (n=0; n<N; n++) 
		if (Tab->item(n, 1)->text().toInt() <= vM 
			&& Tab->item(n, 2)->text().toInt() <= JM) 
	{
		U = Tab->item(n, 9)->text().toDouble();
		for (i=0; (i < Uncert.count() ? U > Uncert[i] : false); i++) ;
		if (i < Uncert.count() ? U == Uncert[i] : false) Numbers[i]++;
		else
		{
			Uncert.insert(i, U);
			Numbers.insert(i, 1);
		}
	}
}

double FitData::getUncertaintyOfvibLevel(int v, int I, Spektrum* Source)
{
	int cv = 1000, cI = 1000, av, aI, n, m=0, M, N = Tab->rowCount(), dI, dv;
	double U;
	QList<double> Unc;
	QList<int> Num;
	for (n=0; n<N; n++) if (Tab->item(n, 5)->text() == Source->getName())
	{
		av = Tab->item(n, 1)->text().toInt();
		aI = Tab->item(n, 0)->text().toInt();
		if (abs(aI - I) < abs(cI - I) || (abs(aI - I) == abs(cI - I) 
			&& abs(av - v) < abs(cv - v)))
		{
			cv = av;
			cI = aI;
		}
	}
	if (cv == 1000) return 0.0201;
	dI = abs(cI - I);
	dv = abs(cv - v);
	for (n=0; n<N; n++) 
		if (Tab->item(n, 5)->text() == Source->getName() 
			&& abs(Tab->item(n, 0)->text().toInt() - I) <= dI
			&& abs(Tab->item(n, 1)->text().toInt() - v) <= dv)
	{
		for (m=0, U = Tab->item(n, 9)->text().toDouble(); 
			 m < Unc.count() && Unc[m] != U; m++) ;
		if (m < Unc.count()) Num[m]++;
		else
		{
			Unc.append(U);
			Num.append(1);
		}
	}
	for (n=M=0; n < Num.count(); n++) if (Num[n] > M) M = Num[m=n];
	return Unc[m];
}

void FitData::removeDataFSource()
{
	int n, m, N = Tab->rowCount(), NC = Tab->columnCount();
	QStringList S;
	QDialog *D;
	QGridLayout *L;
	QPushButton *O, *C;
	QComboBox *B;
	QString b;
	int c;
	for (n=0; n<N; n++) if (!S.contains(b = Tab->item(n, 5)->text())) S.append(b);
	D = new QDialog(this);
	D->setWindowTitle("MolSpektAnalysis");
	L = new QGridLayout(D);
	L->addWidget(new QLabel("Please select the source of which the data shall be removed from the table:", this), 0, 0, 1, 2);
	L->addWidget(new QLabel("Source:", this), 1, 0);
	L->addWidget(B = new QComboBox(this), 1, 1);
	L->setRowMinimumHeight(2, 20);
	L->addWidget(O = new QPushButton("OK", this), 3, 0);
	L->addWidget(C = new QPushButton("Cancel", this), 3, 1);
	B->addItems(S);
	B->setEditable(false);
	connect(O, SIGNAL(clicked()), D, SLOT(accept()));
	connect(C, SIGNAL(clicked()), D, SLOT(reject()));
	c = D->exec();
	b = B->currentText();
	delete D;
	if (c == QDialog::Rejected) return;
	for (n=0; Tab->item(n, 5)->text() != b; n++) ;
	Tab->blockSignals(true);
	for (m=n+1; m<N; m++) if (Tab->item(m, 5)->text() != b)
	{
		for (c=0; c < NC; c++) Tab->setItem(n, c, Tab->takeItem(m, c));
		n++;
	}
	Tab->setRowCount(n);
	Tab->blockSignals(false);
}

void FitData::RemoveDoubled()
{
    int NR = Tab->rowCount(), n, m, c, NC = Tab->columnCount();
	int *SA = new int[NR], *S1 = heapSort(sortIefJFreqv), *CompT, MCT;
	TermTable *TT = (State != 0 ? State->getTermTable() : 0);
	if (TT != 0) TT->getCompT(MCT, CompT);
	Tab->blockSignals(true);
	for (n=0; n < NR; n++) SA[S1[n]] = n;
	for (m=0, n=1; n < NR; n++) 
	{
		if (fabs(Tab->item(SA[m], 8)->text().toDouble() - Tab->item(SA[n], 8)->text().toDouble()) < 3e-3
			&& Tab->item(SA[m], 0)->text().toInt() == Tab->item(SA[n], 0)->text().toInt()
			 && (Tab->item(SA[m], 1)->text().toInt() == Tab->item(SA[n], 1)->text().toInt()
			 || (Tab->item(SA[m], 3)->text().toInt() < 0 && Tab->item(SA[n], 3)->text().toInt() < 0))
			 && Tab->item(SA[m], 2)->text().toInt() == Tab->item(SA[n], 2)->text().toInt()
			 && Tab->item(SA[m], 4)->text().toInt() == Tab->item(SA[n], 4)->text().toInt()
			 && (n < NSources ? FC[m] == FC[n] || FC[m] < 0 || FC[n] < 0 || FC[m] > MCT || FC[n] > MCT || CompT[FC[m]] < 0 
						|| CompT[FC[n]] < 0 : true)) 
		{
			if (Tab->item(SA[m], 7)->text() == Tab->item(SA[n], 7)->text() || Tab->item(SA[n], 7)->text() == "")
				Tab->setItem(SA[n], 0, 0);
			else 
			{
				if (Tab->item(SA[m], 7)->text() == "") Tab->setItem(SA[m], 0, 0);
				m=n;
			}
		}
		else m=n;
	}
	for (n=0; (n < NR ? Tab->item(n, 0) != 0 : false); n++) ;
	for (m=n+1; m < NR; m++) if (Tab->item(m, 0) != 0)
	{
		if (n < NSources)
		{
			if (m < NSources)
			{
				Sources[n] = Sources[m];
				FC[n] = FC[m];
			}
			else
			{
				Sources[n] = 0;
				FC[n] = -1;
			}
		}
		for (c=0; c < NC; c++) Tab->setItem(n, c, Tab->takeItem(m, c));
		n++;
	}
	if (n < NSources) NSources = n;
	Tab->setRowCount(n);
	Tab->blockSignals(false);
	Changed();
	delete[] SA;
	delete[] S1;
}

void FitData::removeMarkedLevel(TermEnergy& TE, Spektrum* Source)
{
	int n, c, N = Tab->rowCount(), C = Tab->columnCount();
	Tab->blockSignals(true);
	for (n = N - NMarkedLevels; n<N; n++)
	{
		if (Tab->item(n, 0)->text().toInt() != TE.Iso) continue;
		if (Tab->item(n, 1)->text().toInt() != TE.v) continue;
		if (Tab->item(n, 2)->text().toInt() != TE.J) continue;
		if (Tab->item(n, 5)->text() != Source->getName()) continue;
		if (Tab->item(n, 7)->text() != Source->getFileName()) continue;
		if (fabs(Tab->item(n, 8)->text().toDouble() - TE.E) < 1e-4) break;
	}
	if (n==N)
	{
		printf("FitData::removeMarkedLevel error: Levels cannot be found!");
		return;
	}
	for (n++; n<N; n++) for (c=0; c<C; c++) Tab->setItem(n-1, c, Tab->takeItem(n, c));
	for (n=1; n < NSources; n++)
	{
		FC[n-1] = FC[n];
		Sources[n-1] = Sources[n];
	}
	if (NSources == N) NSources = N-1;
	else 
	{
		Sources[NSources - 1] = 0;
		FC[NSources - 1] = -1;
	}
	NMarkedLevels--;
	Tab->setRowCount(N-1);
	Tab->blockSignals(false);
}

void FitData::removeSingleLines()
{
	int n, m, c, N = Tab->rowCount(), C = Tab->columnCount(), lc;
	int *SA = new int[N], *S1 = heapSort(sortByProg);
	for (n=0; n<N; n++) SA[S1[n]] = n;
	delete[] S1;
	Tab->blockSignals(true);
	for (n = lc = 1; n < N - NMarkedLevels; n++, lc++) 
		if (Tab->item(n, 6)->text().toInt() != Tab->item(n-1, 6)->text().toInt() || Sources[n-1] != Sources[n]) 
	{
		if (lc == 1 && Tab->item(n-1, 3)->text() != "TE" && Tab->item(n-1, 3)->text().toInt() >= -1) Tab->setItem(n-1, 0, 0);
		lc = 0;
	}
	if (lc == 1) Tab->setItem(n-1, 0, 0);
	for (n=0; (n<N ? Tab->item(n, 0) != 0 : false); n++) ;
	for (m=n+1; m<N; m++) if (Tab->item(m, 0) != 0)
	{
		for (c=0; c<C; c++) Tab->setItem(n, c, Tab->takeItem(m, c));
		if (m < NSources)
		{
			FC[n] = FC[m];
			Sources[n++] = Sources[m];
		}
		else if (n < NSources)
		{
			FC[n] = -1;
			Sources[n++] = 0;
		}
		else n++;
	}
	Tab->setRowCount(n);
	Tab->blockSignals(false);
	if (n < NSources) NSources = n;
	Changed();
}

void FitData::selectDataFSource(QString Source)
{
	int n, t, N = Tab->rowCount(), C = Tab->columnCount() - 1;
	QList<QTableWidgetSelectionRange> L = Tab->selectedRanges();
	Tab->blockSignals(true);
	for (n=0; n < L.count(); n++) Tab->setRangeSelected(L[n], false);
	for (n=0; n < N; n++)
	{
		while (n<N ? Tab->item(n, 5)->text() != Source : false) n++;
		t=n;
		while (n<N ? Tab->item(n, 5)->text() == Source : false) n++;
		if (n>t) Tab->setRangeSelected(QTableWidgetSelectionRange(t, 0, n-1, C), true);
	}
	Tab->blockSignals(false);
	emit SelChanged();
}

void FitData::addData(TableLine* Lines, int NLines)
{
	int n, m = Tab->rowCount(), sigDig, i, j;
	TermTable *TT = (State != 0 ? State->getTermTable() : 0);
	int *CompZ = (TT != 0 ? TT->getCompZ() : 0), NFC = (TT != 0 ? TT->getNumComp() : 1);
	if (Sources != 0)
	{
		int *FCB = new int[m + NLines];
		LineTable **SB = new LineTable*[m + NLines];
		for (n=0; n < NSources; n++)
		{
			FCB[n] = FC[n];
			SB[n] = Sources[n];
		}
		while (n<m)
		{
			FCB[n] = -1;
			SB[n++] = 0;
		}
		delete[] Sources;
		delete[] FC;
		FC = FCB;
		Sources = SB;
		NSources = m + NLines;
	}
	else
	{
		Sources = new LineTable*[NSources = NLines + m];
		FC = new int[NSources];
		for (n=0; n<m; n++)
		{
			FC[n] = -1;
			Sources[n] = 0;
		}
	}
	Tab->blockSignals(true);
	Tab->setRowCount(NSources);
	for (n=0; n < NLines; n++, m++)
	{
		FC[m] = (CompZ != 0 && Lines[n].FC >= 0 && Lines[n].FC < NFC ? CompZ[Lines[n].FC] : Lines[n].FC);
		Tab->setItem(m, 0, new QTableWidgetItem(QString::number(Lines[n].Iso)));
		Tab->setItem(m, 1, new QTableWidgetItem(QString::number(Lines[n].vss)));
		Tab->setItem(m, 2, new QTableWidgetItem(QString::number(Lines[n].Jss)));
		Tab->setItem(m, 3, 
				   new QTableWidgetItem((Lines[n].vs != -1 ? QString::number(Lines[n].vs)
											          : (Lines[n].isTE ? "TE" : "nA"))));
		Tab->setItem(m, 4, new QTableWidgetItem(QString::number(Lines[n].Js)));
		if ((Sources[n] = Lines[n].LTab) != 0) Tab->setItem(m, 5, new QTableWidgetItem(Sources[n]->getName()));
		else if (!Lines[n].SourceName.isEmpty()) Tab->setItem(m, 5, new QTableWidgetItem(Lines[n].SourceName)); 
		else
		{
			i = Lines[n].File.lastIndexOf(QRegExp("[\\/]")) + 1;
			j = Lines[n].File.indexOf('.', i);
			Tab->setItem(m, 5, new QTableWidgetItem(Lines[n].File.mid(i, j-i)));
		}
		Tab->setItem(m, 6, new QTableWidgetItem(QString::number(Lines[n].PN)));
		Tab->setItem(m, 7, new QTableWidgetItem(Lines[n].File));
		sigDig = -int(floor(log10(Lines[n].err)));
		if (sigDig < 4) sigDig = 4;
		Tab->setItem(m, 8, new QTableWidgetItem(QString::number(Lines[n].WN, 'f', sigDig)));
		Tab->setItem(m, 9, new QTableWidgetItem(QString::number(Lines[n].err, 'f', sigDig)));
		Tab->setItem(m, 10, new QTableWidgetItem(QString::number(Lines[n].dev, 'f', sigDig)));
		Tab->setItem(m, 11, new QTableWidgetItem(QString::number(Lines[n].DevR, 'f', 4)));
	}
	Tab->blockSignals(false);
	delete[] Lines;
	Changed();
}

void FitData::setData(TableLine* Lines, int NLines)
{
	Tab->blockSignals(true);
	if (Sources != 0)
	{
		delete[] Sources;
		delete[] FC;
		NSources = 0;
		Sources = 0;
		FC = 0;
	}
	Tab->setRowCount(0);
	addData(Lines, NLines);
	setNewCreated();
}

void FitData::setDev(double* dev, double* DevR)
{
	int n, m, N = Tab->rowCount(), mJ = getJMax(), mv = getvMax();
	QString Buffer;
	Tab->blockSignals(true);
	for (n=m=0; n<N; n++) if (Tab->item(n, 1)->text().toInt() <= mv && Tab->item(n, 2)->text().toInt() <= mJ)
	{
		Buffer = Tab->item(n, 9)->text();
		Tab->item(n, 10)->setText(QString::number(dev[m], 'f', Buffer.length() - Buffer.indexOf('.') - 1));
		Tab->item(n, 11)->setText(QString::number(DevR[m++], 'f', 4));
	}
	Tab->blockSignals(false);
	Changed();
}

void FitData::setDev(double* dev, int* RowN, int N)
{
	int n;
	QString Buffer;
	Tab->blockSignals(true);
	for (n=0; n<N; n++)
	{
		Buffer = Tab->item(RowN[n], 9)->text();
		Tab->item(RowN[n], 10)->setText(QString::number(dev[n], 'f', Buffer.length() - Buffer.indexOf('.') - 1));
		Tab->item(RowN[n], 11)->setText(QString::number(dev[n] 
									/ Tab->item(RowN[n], 9)->text().toDouble(), 'f', 4));
	}
	Tab->blockSignals(false);
	Changed();
}

void FitData::setElState(ElState* nState)
{
	State = nState;
}

void FitData::setFC(int nFC)
{
	int n, m;
	QList<QTableWidgetSelectionRange> L = Tab->selectedRanges();
	if (L.count() > 0)
	{
		for (n=0; n < L.count(); n++) for (m = L[n].topRow(); m <= L[n].bottomRow(); m++) if (m < NSources)
			FC[n] = nFC;
	}
	else for (n=0; n < NSources; n++) FC[n] = nFC;
}

void FitData::setRWErr(double* RWErr)
{
	int n, N = Tab->rowCount();
	QString Buffer;
	Tab->blockSignals(true);
	if (Tab->columnCount() == 12)
	{
		Tab->setColumnCount(13);
		Tab->setHorizontalHeaderItem(12, new QTableWidgetItem("RWErr"));
		for (n=0; n<N; n++)
		{
			Buffer = Tab->item(n, 9)->text();
			Tab->setItem(n, 12, new QTableWidgetItem(QString::number(RWErr[n], 'f', Buffer.length() - Buffer.indexOf('.') - 1)));
		}
	}
	else for (n=0; n<N; n++)
	{
		Buffer = Tab->item(n, 9)->text();
		Tab->item(n, 12)->setText(QString::number(RWErr[n], 'f', Buffer.length() - Buffer.indexOf('.') - 1));
	}
	if (Tab->horizontalHeaderItem(12)->text() != "RWErr") Tab->horizontalHeaderItem(12)->setText("RWErr");
	Tab->blockSignals(false);
}

void FitData::setSourceOffset(QStringList& Names, double* Offsets)
{
	int n, m;
	if (NSourceOffset > 0)
	{
		delete[] SourceOffset;
		delete[] SourceOffsetNames;
	}
	for (n=m=0; n < Names.count(); n++) if (Offsets[n] != 0.0) m++;
	if ((NSourceOffset = m) > 0)
	{
		SourceOffset = new double[m];
		SourceOffsetNames = new QString[m];
	}
	else
	{
		SourceOffset = 0;
		SourceOffsetNames = 0;
	}
	for (n=m=0; n < Names.count(); n++) if (Offsets[n] != 0.0)
	{
		SourceOffset[m] = Offsets[n];
		SourceOffsetNames[m++] = Names[n];
	}
	if (Names.count() > 0) delete[] Offsets;
	Changed();
}

void FitData::setUncertainty(double Uncertainty, bool Min)
{
	int n, m, i;
	double cU;
	QList<QTableWidgetSelectionRange> L = Tab->selectedRanges();
	QString U = (Uncertainty > 1e-4 ? QString::number(Uncertainty, 'f', 4) : QString::number(Uncertainty));
	QString mU = QString::number(9.0 + Uncertainty, 'f', 4), MU = QString::number(99.0 + Uncertainty, 'f', 4), nU;
	QList<LineTable*> SL;
	QList<int> CL;
	if (L.count() > 0)
	{
		for (n=0; n < L.count(); n++) for (m = L[n].topRow(); m <= L[n].bottomRow(); m++)
			if (m < NSources) if (Sources[m] != 0) if (Sources[m]->getTransition()->getLowerState() == State)
		{
			for (i=0; (i < SL.count() ? SL[i] != Sources[m] : false); i++) ;
			if (i == SL.count())
			{
				SL.append(Sources[m]);
				CL.append(1);
			}
			else CL[i]++;
		}
		int *PN[SL.count()], *vss[SL.count()], *Jss[SL.count()];
		QString *Err[SL.count()];
		for (n=0; n < SL.count(); n++)
		{
			PN[n] = new int[CL[n]];
			vss[n] = new int[CL[n]];
			Jss[n] = new int[CL[n]];
			Err[n] = new QString[CL[n]];
			CL[n] = 0;
		}
		for (n=0; n < L.count(); n++) for (m = L[n].topRow(); m <= L[n].bottomRow(); m++)
		{
			
			if (!Min || (cU = Tab->item(m, 9)->text().toDouble()) < Uncertainty) Tab->item(m, 9)->setText(nU = U);
			else if (cU >= 99.0) Tab->item(m, 9)->setText(nU = MU);
			else if (cU >= 9.0) Tab->item(m, 9)->setText(nU = mU);
			if (m < NSources ? Sources[m] != 0 : false) if (Sources[m]->getTransition()->getLowerState() == State)
			{
				for (i=0; SL[i] != Sources[m]; i++) ;
				PN[i][CL[i]] = Tab->item(m, 6)->text().toInt();
				vss[i][CL[i]] = Tab->item(m, 1)->text().toInt();
				Jss[i][CL[i]] = Tab->item(m, 2)->text().toInt();
				Err[i][CL[i]++] = nU;
			}
		}
		for (n=0; n < SL.count(); n++)
		{
			SL[n]->SetError(CL[n], PN[n], vss[n], Jss[n], Err[n]);
			delete[] PN[n];
			delete[] vss[n];
			delete[] Jss[n];
			delete[] Err[n];
		}
	}
	else 
	{
		for (m=0; m < NSources; m++) if (Sources[m]->getTransition()->getLowerState() == State)
		{
			for (i=0; (i < SL.count() ? SL[i] != Sources[m] : false); i++) ;
			if (i == SL.count())
			{
				SL.append(Sources[m]);
				CL.append(1);
			}
			else CL[i]++;
		}
		int *PN[SL.count()], *vss[SL.count()], *Jss[SL.count()];
		QString *Err[SL.count()];
		for (n=0; n < SL.count(); n++)
		{
			PN[n] = new int[CL[n]];
			vss[n] = new int[CL[n]];
			Jss[n] = new int[CL[n]];
			Err[n] = new QString[CL[n]];
			CL[n] = 0;
		}
		for (m=0; m < Tab->rowCount(); m++)
		{
			if (!Min || (cU = Tab->item(m, 9)->text().toDouble()) < Uncertainty) Tab->item(m, 9)->setText(nU = U);
			else if (cU >= 99.0) Tab->item(m, 9)->setText(nU = MU);
			else if (cU >= 9.0) Tab->item(m, 9)->setText(nU = mU);
			if (m < NSources) if (Sources[m]->getTransition()->getLowerState() == State)
			{
				for (i=0; SL[i] != Sources[m]; i++) ;
				PN[i][CL[i]] = Tab->item(m, 6)->text().toInt();
				vss[i][CL[i]] = Tab->item(m, 1)->text().toInt();
				Jss[i][CL[i]] = Tab->item(m, 2)->text().toInt();
				Err[i][CL[i]++] = nU;
			}
		}
		for (n=0; n < SL.count(); n++)
		{
			SL[n]->SetError(CL[n], PN[n], vss[n], Jss[n], Err[n]);
			delete[] PN[n];
			delete[] vss[n];
			delete[] Jss[n];
			delete[] Err[n];
		}
	}
}

void FitData::setUncertainty(int* Rows, double* Uncertainties, int N)
{
	int n, i;
	Tab->blockSignals(true);
	QList<LineTable*> SL;
	QList<int> CL;
	for (n=0; n<N; n++) if (Rows[n] < NSources ? Sources[Rows[n]] != 0 : false) 
		if (Sources[Rows[n]]->getTransition()->getLowerState() == State)
	{
		for (i=0; (i < SL.count() ? SL[i] != Sources[Rows[n]] : false); i++) ;
		if (i == SL.count())
		{
			SL.append(Sources[Rows[n]]);
			CL.append(1);
		}
		else CL[i]++;
	}
	int *PN[SL.count()], *vss[SL.count()], *Jss[SL.count()];
	double *Err[SL.count()];
	for (n=0; n < SL.count(); n++)
	{
		PN[n] = new int[CL[n]];
		vss[n] = new int[CL[n]];
		Jss[n] = new int[CL[n]];
		Err[n] = new double[CL[n]];
		CL[n] = 0;
	}
	for (n=0; n<N; n++) 
	{
		Tab->item(Rows[n], 9)->setText(QString::number(Uncertainties[n], 'f', 4));
		Tab->item(Rows[n], 11)->setText(QString::number(
		Tab->item(Rows[n], 10)->text().toDouble() / Uncertainties[n], 'f', 4));
		if (Rows[n] < NSources ? Sources[Rows[n]] != 0 : false) if (Sources[Rows[n]]->getTransition()->getLowerState() == State)
		{
			for (i=0; SL[i] != Sources[Rows[n]]; i++) ;
			PN[i][CL[i]] = Tab->item(Rows[n], 6)->text().toInt();
			vss[i][CL[i]] = Tab->item(Rows[n], 1)->text().toInt();
			Jss[i][CL[i]] = Tab->item(Rows[n], 2)->text().toInt();
			Err[i][CL[i]++] = Uncertainties[n];
		}
	}
	for (n=0; n < SL.count(); n++)
	{
		SL[n]->SetError(CL[n], PN[n], vss[n], Jss[n], Err[n]);
		delete[] PN[n];
		delete[] vss[n];
		delete[] Jss[n];
		delete[] Err[n];
	}
	Tab->blockSignals(false);
	Changed();
}

void FitData::setWeightFactors(int N, int* Rows, double* Factors)
{
	int n, NR = Tab->rowCount();
	QString Buffer = "1.0";
	Tab->blockSignals(true);
	if (Tab->columnCount() < 13) Tab->setColumnCount(13);
	Tab->setHorizontalHeaderItem(12, new QTableWidgetItem("WFact"));
	for (n=0; n < NR; n++) Tab->setItem(n, 12, new QTableWidgetItem(Buffer));
	for (n=0; n<N; n++) Tab->item(Rows[n], 12)->setText(QString::number(Factors[n], 'g', 4));
	Tab->blockSignals(false);
	delete[] Rows;
	delete[] Factors;
	Changed();
}

void FitData::setMCSEnergies(double* MCSE, int* Rows, int NR)
{
	int n, sigDig;
	for (n=0; n < NR; n++)
	{
		sigDig = -int(floor(log10(Tab->item(Rows[n], err)->text().toDouble())));
		if (sigDig < 4) sigDig = 4;
		Tab->item(Rows[n], WN)->setText(QString::number(MCSE[n], 'f', sigDig));
	}
	delete[] MCSE;
	delete[] Rows;
}

void FitData::sortTab(int* S2)
{
    int i, P1=0, n, N = Tab->rowCount(), c, C = Tab->columnCount(), AFC[2];
	QTableWidgetItem *AIt[C][2];
	LineTable *AS[2];
	Tab->blockSignals(true);
	for (i=0; i<N; i++) if (S2[i] != i)
	{
		for (c=0; c<C; c++) AIt[c][P1] = Tab->takeItem(i, c);
		AFC[P1] = FC[i];
		AS[P1] = Sources[i];
		while (S2[i] != i) 
		{
			for (c=0; c<C; c++)
			{
				AIt[c][1-P1] = Tab->takeItem(S2[i], c);
				Tab->setItem(S2[i], c, AIt[c][P1]);
			}
			AFC[1-P1] = FC[S2[i]];
			FC[S2[i]] = AFC[P1];
			AS[1-P1] = Sources[S2[i]];
			Sources[S2[i]] = AS[P1];
			P1 = 1 - P1;
			n = S2[S2[i]];
			S2[S2[i]] = S2[i];
			S2[i] = n;
		}
		for (c=0; c<C; c++) Tab->setItem(i, c, AIt[c][P1]);
		FC[i] = AFC[P1];
		Sources[i] = AS[P1];
	}
	Tab->blockSignals(false);
	delete[] S2;
	Changed();
}

void FitData::updateData()
{
	if (State == 0 || molecule == 0)
	{ 
		printf("FitData::updateData error: This object is not assigned to a molecule!\n");
		return;
	}
	TSDialog Diag(this, molecule, State);
	if (Diag.exec() == QDialog::Rejected) return;
	int rr, r = 0, l, n, m, N = Diag.getNumSelected(), NL[N], MJ = 0, Mv = 0, *FCB;
	if (N == 0 && !Diag.getUpdateSA()) return;
	LineTable *L[N], **SB;
	TermEnergy *E;
	TableLine *TL;
	QString CSource;
	for (n=0; n<N; n++) L[n] = Diag.getSelected(n);
	Tab->blockSignals(true);
	if (Diag.getKeepOther())
	{
		for (n=0; n < NSources; n++) 
		{
			CSource = Tab->item(n, fdcSource)->text();
			for (l=0; (l<N ? Sources[n] != L[l] && CSource != L[l]->getName() : false); l++) ;
			if (l<N) Tab->setItem(n, 0, 0);
		}
		if (Diag.getUpdateSA() && (molecule != 0 ? State != molecule->getStateP(0) : false))
		{
			int NS = 0, NumStates = molecule->getNumStates() - 2;
			TermTable *TT = State->getTermTable();
			int NP = (TT != 0 ? TT->getNPerturbations() : 0), Numv = TT->getMaxv() + 1, nC = TT->getNumComp();
			int NumJ = TT->getMaxJ() + 1, NumI = TT->getNumIso();
			double D, sD, En, ****TData = TT->getData();
			if (NP > 0)
			{
				int NComp[NumStates], Nv[NumStates], s, c, I=0, J=0, v=0, bv = 0, bs, R;
				bool *PComp[NumStates];
				TermTable *PTT[NumStates];
				FitData *PFD[NumStates];
				Perturbation *Pert;
				double ****PTData[NumStates];
				QPixmap Pix(10, 10);
				QPainter P(&Pix);
				P.setPen(QColor(255, 0, 0));
				P.setFont(QFont("Arial", 10));
				P.drawText(0, 10, "N");
				for (s=-1; s < NS; s++)
				{
					if (s>=0) NP = PTT[s]->getNPerturbations();
					for (n=0; n < NP; n++)
					{
						if (s==-1) Pert = TT->getPerturbation(n);
						else Pert = PTT[s]->getPerturbation(n);
						if (Pert->Perturber == TT || Pert->Perturber == 0 || !PComp[Pert->Comp]) continue;
						for (m=0; (m < NS ? Pert->Perturber != PTT[m] : false); m++) ;
						if (m == NS)
						{
							PTT[m] = Pert->Perturber;
							if (PTT[m]->getElState() == 0) continue;
							NComp[m] = PTT[m]->getNumComp();
							Nv[m] = PTT[m]->getMaxv() + 1;
							PComp[m] = new bool[NComp[m]];
							for (c=0; c < NComp[m]; c++) PComp[m][c] = false;
							PTData[m] = PTT[m]->getData();
							PFD[NS++] = 0;
						}
						PComp[m][Pert->PComp] = true;
					}
				}
				for (r=0; r < Tab->rowCount(); r++) 
					if (Tab->item(r, 6)->text().isEmpty() && (I = Tab->item(r, 0)->text().toInt()) < NumI &&
						(v = Tab->item(r, 1)->text().toInt()) < Numv && (J = Tab->item(r, 2)->text().toInt()) < NumJ)
				{
					c = (Tab->item(r, 4)->text().toInt() != J ? 0 : (nC > 1 ? 1 : 0));
					for (sD = fabs((En = Tab->item(r, 8)->text().toDouble()) - TData[c][I][v][J]), s=0, bs = -1; s < NS; s++)
						for (c=0; c < NComp[s]; c++) if (PComp[s][c]) for (v=0; v < Nv[s]; v++) 
							if ((D = fabs(En - PTData[s][c][I][v][J])) < sD)
					{
						sD = D;
						bs = s;
						bv = v;
					}
					if (bs >= 0)
					{
						if (PFD[bs] == 0)
						{
							if ((PFD[bs] = PTT[bs]->getElState()->getFitData()) == 0)
							{
								if ((PFD[bs] = MW->CreateFitData()) == 0) continue;
								PTT[bs]->getElState()->addFitData(PFD[bs]);
							}
							if (!PFD[bs]->isVisible()) PFD[bs]->show();
							PFD[bs]->activateWindow();
							PFD[bs]->setFocus();
						}
						R = PFD[bs]->Tab->rowCount();
						PFD[bs]->NMarkedLevels++;
						PFD[bs]->Tab->blockSignals(true);
						PFD[bs]->Tab->setRowCount(R+1);
						for (c=0; c <= 11; c++) 
							PFD[bs]->Tab->setItem(R, c, new QTableWidgetItem(Pix, Tab->item(r, c)->text()));
						PFD[bs]->Tab->item(R, 1)->setText(QString::number(bv));
						PFD[bs]->Tab->blockSignals(false);
						PFD[bs]->setBlockChangeSignal(true);
						PFD[bs]->Changed();
						PFD[bs]->setBlockChangeSignal(false);
						Tab->setItem(r, 0, 0);
					}
				}
				for (s=0; s < NS; s++) delete[] PComp[s];
			}
		}
		for (r=0; (r < Tab->rowCount() ? Tab->item(r, 0) != 0 : false); r++) ;
		for (l=r+1; l < Tab->rowCount(); l++) if (Tab->item(l, 0) != 0)
		{
			for (n=0; n < Tab->columnCount(); n++) Tab->setItem(r, n, Tab->takeItem(l, n));
			if (l < NSources)
			{
				Sources[r] = Sources[l];
				FC[r] = FC[l];
			}
			else if (r < NSources)
			{
				Sources[r] = 0;
				FC[r] = -1;
			}
			r++;
		}
		if (r < NSources) NSources = r;
		if (FC == 0)
		{
			FC = new int[r];
			for (n=0; n<r; n++) FC[n] = -1;
		}
		else if (NSources < r)
		{
			FCB = FC;
			FC = new int[r];
			for (n=0; n < NSources; n++) FC[n] = FCB[n];
			delete[] FCB;
			while (n<r) FC[n++] = -1;
		}
	}
	else
	{
		NSources = 0;
		delete[] Sources;
		Sources = 0;
	}
	Tab->setRowCount(rr = r);
	for (n=0; n<N; n++) 
	{
		if (L[n]->getTransition()->getUpperState() == State)
		{
			L[n]->getgoodTE(NL[n], E);
			Tab->setRowCount(r + NL[n]);
			if (FC != 0)
			{
				FCB = new int[r + NL[n]];
				for (m=0; m<r; m++) FCB[m] = FC[m];
				delete[] FC;
				FC = FCB;
			}
			else FC = new int[r + NL[n]];
			for (l=0; l < NL[n]; l++) 
			{
				FC[r] = E[l].FC;
				Tab->setItem(r, 0, new QTableWidgetItem(QString::number(E[l].Iso)));
				Tab->setItem(r, 1, new QTableWidgetItem(QString::number(E[l].v)));
				Tab->setItem(r, 2, new QTableWidgetItem(QString::number(E[l].J)));
				Tab->setItem(r, 3, new QTableWidgetItem("TE"));
				Tab->setItem(r, 4, new QTableWidgetItem(E[l].ef ? QString::number(E[l].J + 1) : Tab->item(r, 2)->text()));
				Tab->setItem(r, 5, new QTableWidgetItem(L[n]->getName()));
				Tab->setItem(r, 6, new QTableWidgetItem(QString::number(E[l].PN)));
				Tab->setItem(r, 7, new QTableWidgetItem(E[l].File));
				Tab->setItem(r, 8, new QTableWidgetItem(QString::number(E[l].E, 'f', 4)));
				Tab->setItem(r, 9, new QTableWidgetItem(QString::number(E[l].err, 'f', 4)));
				Tab->setItem(r, 10, new QTableWidgetItem(QString::number(E[l].dev, 'f', 4)));
				Tab->setItem(r++, 11, new QTableWidgetItem(QString::number(E[l].DevR, 'f', 4)));
				if (E[l].v > Mv) Mv = E[l].v;
				if (E[l].J > MJ) MJ = E[l].J;
			}
			delete[] E;
		}
		else
		{
			L[n]->getgoodLines(NL[n], TL);
			Tab->setRowCount(r + NL[n]);
			if (FC != 0)
			{
				FCB = new int[r + NL[n]];
				for (m=0; m<r; m++) FCB[m] = FC[m];
				delete[] FC;
				FC = FCB;
			}
			else FC = new int[r + NL[n]];
			for (l=0; l < NL[n]; l++)
			{
				FC[r] = TL[l].FC;
				Tab->setItem(r, 0, new QTableWidgetItem(QString::number(TL[l].Iso)));
				Tab->setItem(r, 1, new QTableWidgetItem(QString::number(TL[l].vss)));
				Tab->setItem(r, 2, new QTableWidgetItem(QString::number(TL[l].Jss)));
				Tab->setItem(r, 3, 
						new QTableWidgetItem((TL[l].vs != -1 ? QString::number(TL[l].vs)
															 : "nA")));
				Tab->setItem(r, 4, new QTableWidgetItem(QString::number(TL[l].Js)));
				Tab->setItem(r, 5, new QTableWidgetItem(L[n]->getName()));
				Tab->setItem(r, 6, new QTableWidgetItem(QString::number(TL[l].PN)));
				Tab->setItem(r, 7, new QTableWidgetItem(TL[l].File));
				Tab->setItem(r, 8, new QTableWidgetItem(QString::number(TL[l].WN, 'f', 4)));
				Tab->setItem(r, 9, new QTableWidgetItem(QString::number(TL[l].err, 'f', 4)));
				Tab->setItem(r, 10, new QTableWidgetItem(QString::number(TL[l].dev, 'f', 4)));
				Tab->setItem(r++, 11, new QTableWidgetItem(QString::number(TL[l].DevR, 'f', 4)));
				if (TL[l].vss > Mv) Mv = TL[l].vss;
				if (TL[l].Jss > MJ) MJ = TL[l].Jss;
			}
			delete[] TL;
		}
	}
	Tab->blockSignals(false);
	SB = Sources;
	Sources = new LineTable*[r];
	for (n=0; n < NSources; n++) Sources[n] = SB[n];
	if (SB != 0) delete[] SB;
	while (n < rr) Sources[n++] = 0;
	for (n=0, NSources = r; n<N; n++) for (l=0; l < NL[n]; l++) Sources[rr++] = L[n];
	if (getvMax() < Mv) setvMax(Mv);
	if (getJMax() < MJ) setJMax(MJ);
	Changed();
}

void FitData::updateEnergy(int N, int *r, double *Energy)
{
	int n;
	Tab->blockSignals(true);
	for (n=0; n<N; n++) Tab->item(r[n], 8)->setText(QString::number(Energy[n], 'f', 4));
	Tab->blockSignals(false);
	Changed();
}

void FitData::updateLevels(int N, int *r, double *Energy, int *vs, int *vss)
{
	int n;
	Tab->blockSignals(true);
	for (n=0; n<N; n++) 
	{
		Tab->item(r[n], 8)->setText(QString::number(Energy[n], 'f', 4));
		Tab->item(r[n], 1)->setText(QString::number(vss[n]));
		Tab->item(r[n], 3)->setText(vs[n] == -1 ? "TE" : QString::number(vs[n]));
	}
	Tab->blockSignals(false);
	Changed();
}

void FitData::updateRow(TableLine* Line)
{
	int n, N = Tab->rowCount();
	for (n=0; n<N; n++) if (Tab->item(n, 8)->text().toDouble() == Line->WN)
		if (Tab->item(n, 0)->text().toInt() == Line->Iso && Tab->item(n, 1)->text().toInt() == Line->vss
				&& Tab->item(n, 2)->text().toInt() == Line->Jss) 
			Tab->item(n, 9)->setText(QString::number(Line->err, 'f', 4)); 
}

bool FitData::writeData(QString Filename)
{
    bool Success;
	int n, N = Tab->rowCount(), c, C = Tab->columnCount();
	QString B;
	if (NMarkedLevels > 0)
	{
		QPixmap P;
		LineTable **NewSources = new LineTable*[N];
		int *NFC = new int[N];
		for (n=0; n < NSources; n++)
		{
			NewSources[n] = Sources[n];
			NFC[n] = FC[n];
		}
		delete[] Sources;
		delete[] FC;
		Sources = NewSources;
		FC = NFC;
		Tab->blockSignals(true);
		for (n = N - NMarkedLevels; n<N; n++)
		{
			for (c=0; c<C; c++) Tab->item(n, c)->setIcon(P);
			Sources[n] = 0;
			B = Tab->item(n, 3)->text();
			FC[n] = (B == "TE" ? -1 : -1 - B.toInt());
		}
		Tab->blockSignals(false);
		NMarkedLevels = 0;
		Changed();
		emit AssignmentsAccepted(this);
	}
	Tab->blockSignals(true);
    for (n=0; n < NSources; n++)
    {
        if (FC[n] >= 0)
            Tab->item(n, 3)->setText(QString::number(((B = Tab->item(n, 3)->text()) == "TE" ? -1 : (B == "nA" ? -10 : B.toInt()))
                          + 1000 * (FC[n] + 1)));
        if (molecule != 0)
        {
            QString CurPath = Tab->item(n, fdcFile)->text(), MolPath = molecule->getFileName();
            Tab->item(n, fdcFile)->setText(getRelativePath(CurPath, MolPath));
        }
    }
	Success = TableWindow::writeData(Filename);
    for (n=0; n < NSources; n++)
    {
        if (FC[n] >= 0)
            Tab->item(n, 3)->setText((c = Tab->item(n, 3)->text().toInt() - 1000 * (FC[n] + 1)) == -1 ? "TE" : (c == -10 ? "nA"
                                    : QString::number(c)));
        if (molecule != 0)
        {
            QString CurPath = Tab->item(n, fdcFile)->text(), MolPath = molecule->getFileName();
            Tab->item(n, fdcFile)->setText(getAbsolutePath(CurPath, MolPath));
        }
    }
	Tab->blockSignals(false);
	if (NSourceOffset > 0 && Success)
	{
		QFile F(getFileName());
		F.open(QIODevice::Append);
		QTextStream S(&F);
		S << "SourceOffsets:\n";
		for (n=0; n < NSourceOffset; n++) S << SourceOffsetNames[n] << '\t' << QString::number(SourceOffset[n], 'f', 12) << '\n';
	}
	return Success;
}

bool FitData::writeExPotFitInput(QString Filename)
{
	if (Filename.isEmpty()) return false;
	if (molecule == 0) return false;
	IsoTab *Iso = molecule->getIso();
	if (Iso == 0) return false;
	QFile File(Filename);
	int nIso = Iso->numIso, n, N = Tab->rowCount(), I;
	QString IsoStr[nIso], B;
	if (!File.open(QIODevice::WriteOnly))
	{
		delete Iso;
		return false;
	}
	QTextStream WS(&File);
	for (n=0; n < nIso; n++) 
		IsoStr[n] = ("    " + QString::number(Iso->mNumIso1[n])).right(5)
				  + ("    " + QString::number(Iso->mNumIso2[n])).right(5);
	for (n=0; n<N; n++)
	{
		I = Tab->item(n, 0)->text().toInt();
		B = Tab->item(n, 3)->text();
		WS << IsoStr[I] + ("     " + Tab->item(n, 1)->text()).right(5) 
			  + ("     " + Tab->item(n, 2)->text()).right(5) 
			  + ("     " + (B.indexOf("TE") >= 0 || B.indexOf("nA") >= 0 ? "-1" : B)).right(5)
			  + ("     " + Tab->item(n, 4)->text()).right(5) + IsoStr[I]
			  + "    0    0    0    0\n" + ("               " + Tab->item(n, 8)->text()).right(15)
			  + ("               " + Tab->item(n, 9)->text()).right(15) + "    2\n";
	}
	File.close();
	delete Iso;
	return true;
}

bool FitData::writeTFGS(QString Filename)
{
	if (Filename.isEmpty()) return false;
	if (molecule == 0) return false;
	IsoTab *Iso = molecule->getIso();
	if (Iso == 0) return false;
	Atom *atom1 = molecule->getAtom1(), *atom2 = molecule->getAtom2();
	if (atom1 == 0 || atom2 == 0) return false;
	QFile File(Filename);
	if (!File.open(QIODevice::WriteOnly))
	{
		delete Iso;
		return false;
	}
	int nIso = Iso->numIso, n, N = Tab->rowCount(), I, vo = 0, *S1 = heapSort(sortforTFGS), *SA = new int[N], vs, Js, lvs = -2, lJs = -2;
	int lI = -1, SI = 0, lSI = -1, PN, lPN = -1, wv, nDig;
	double WN, err;
	QString IsoStr[nIso], B;
	QStringList SourceList;
	QTextStream S(&File);
	for (n=0; n<N; n++) SA[S1[n]] = n;
	delete[] S1;
	S << ("    " + QString::number(N)).right(5) << "   0" << ("    " + QString::number(atom1->getnIso())).right(5)
	  << ("    " + QString::number(atom2->getnIso())).right(5) << ("    " + QString::number(State->getOmega())).right(5) 
	  << "    0    1     levels, Dunham parameters, isotope A and B, omega, select dublicates (=0 no, =1 best, = 2 average)\n";
	for (n=0; n < atom1->getnIso(); n++) 
		S << ("    " + QString::number(atom1->getnNuc(n))).right(5) 
		  << ("            " + QString::number(atom1->getIsoMass(n), 'f', 8)).right(13) << "           0.0\n";
	for (n=0; n < atom2->getnIso(); n++) 
		S << ("    " + QString::number(atom2->getnNuc(n))).right(5) 
		  << ("            " + QString::number(atom2->getIsoMass(n), 'f', 8)).right(13) << "           0.0\n";
	S << "    0" 
	  << ("       " + (State->getDunTable() != 0 ? QString::number(State->getDunTable()->getwe(), 'f', 1) : QString("60.0"))).right(8)
	  << "  -0.0          0" << ("    " + QString::number(getvMax())).right(5) << "    0"
	  << ("    " + QString::number(getJMax())).right(5) + "      v Verschiebung, vibr. spacing, level limits\n"
	  << "  0.0              1           constant zero, switch of quantum numbers\n";
	for (n=0; n < nIso; n++)
		IsoStr[n] = ("    " + QString::number(Iso->mNumIso1[n])).right(5) + ("    " + QString::number(Iso->mNumIso2[n])).right(5);
	delete Iso;
	for (n=0; n<N; n++)
	{
		WN = Tab->item(SA[n], 8)->text().toDouble();
		B = Tab->item(SA[n], 9)->text();
		nDig = B.length() - B.indexOf('.') - 1;
		err = B.toDouble();
		PN = Tab->item(SA[n], 6)->text().toInt();
		Js = Tab->item(SA[n], 4)->text().toInt();
		I = Tab->item(SA[n], 0)->text().toInt();
		B = Tab->item(SA[n], 3)->text();
		vs = (B != "nA" ? B.toInt() : -1);
		if (B == "TE")
		{
			wv = 9999;
			WN *= -1.0;
			SI = -1;
		}
		else 
		{
			if (PN != lPN)
			{
				if (SI == lSI && lvs == vs && Js == lJs && lI == I) vo += 100;
				else vo = 0;
			}
			B = Tab->item(SA[n], 5)->text();
			SI = SourceList.indexOf(B);
			if (SI == -1)
			{
				SI = SourceList.count();
				SourceList << B;
			}
			wv = 1000 * (SI + 1) + vo + vs;
		}
		B = Tab->item(SA[n], 2)->text();
		S << IsoStr[I] << ("    " + QString::number(wv)).right(5) << ("    " + (wv != 9999 ? QString::number(Js) : B)).right(5)
		  << ("     " + Tab->item(SA[n], 1)->text()).right(5)  << ("     " + B).right(5) + IsoStr[I] << "    0    0    0    0\n";
		S << ("              " + QString::number(WN, 'f', nDig)).right(15) 
		  << ("              " + QString::number(err, 'f', nDig)).right(15) << "    2\n";
		lvs = vs;
		lJs = Js;
		lI = I;
		lSI = SI;
		lPN = PN;
	}
	delete[] SA;
	return true;
}


TSDialog::TSDialog(QWidget* parent, Molecule* M, ElState* S) : QDialog(parent)
{
	int i, n, m, N = M->getNumTransitions(), nGS = 0;
	Transition *T, **GT = new Transition*[MaxTransitions];
	LineTable *TabB;
	QGridLayout *L = new QGridLayout(this);
	QLabel *L1 = new QLabel("Selected tables:", this);
	L->addWidget(L1, 0, 0);
	QLabel *L2 = new QLabel("Available tables:", this);
	L->addWidget(L2, 0, 1);
	List = new QListWidget(this);
	L->addWidget(List, 1, 0, 3, 1);
	Box = new QComboBox(this);
	L->addWidget(Box, 1, 1);
	Add = new QPushButton("Add table", this);
	L->addWidget(Add, 2, 1);
	Remove = new QPushButton("Remove table", this);
	L->addWidget(Remove, 3, 1);
	QGridLayout *La2 = new QGridLayout;
	KeepOther = new QCheckBox("Keep other data", this);
	KeepOther->setChecked(true);
	La2->addWidget(KeepOther, 0, 0);
	UpdateSA = new QCheckBox("Update state assignment", this);
	La2->addWidget(UpdateSA, 0, 1);
	L->addLayout(La2, 4, 0, 2, 1);
	L->setRowMinimumHeight(5, 20);
	OKB = new QPushButton("OK", this);
	L->addWidget(OKB, 6, 0);
	Cancel = new QPushButton("Cancel", this);
	L->addWidget(Cancel, 6, 1);
	for (n = NTables = 0; n<N; n++) 
	{
		T = M->getTransitionP(n);
		if (T->getLowerState() == S || T->getUpperState() == S) NTables += (GT[nGS++] = T)->getNumLineTables();
	}
	if (NTables > 0) Tables = new LineTable*[NTables];
	else Tables = 0;
	for (n=m=0; n<N; n++)
	{
		T = M->getTransitionP(n);
		if (T->getLowerState() == S || T->getUpperState() == S) for (i=0; i < T->getNumLineTables(); i++)
			if ((TabB = T->getLineTable(i)) != 0) Box->addItem((Tables[m++] = TabB)->getName());
	}
	NTables = m;
	Box->setEditable(false);
	delete[] GT;
	connect(KeepOther, SIGNAL(toggled(bool)), this, SLOT(KeepOtherChanged(bool)));
	connect(Add, SIGNAL(clicked()), this, SLOT(addTable()));
	connect(Remove, SIGNAL(clicked()), this, SLOT(removeTable()));
	connect(OKB, SIGNAL(clicked()), this, SLOT(OK()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
}

TSDialog::~TSDialog()
{
	if (Tables != 0) delete[] Tables;
}

void TSDialog::addTable()
{
	int n;
	if ((n = Box->currentIndex()) != -1) List->addItem(Tables[n]->getName());
}

bool TSDialog::getKeepOther()
{
	return KeepOther->isChecked();
}

int TSDialog::getNumSelected()
{
	return NTables;
}

LineTable* TSDialog::getSelected(int n)
{
	return Tables[n];
}

void TSDialog::OK()
{
	int n, m, N = List->count();
	bool S;
	for (n=0; n < NTables; n++)
	{
		for (m=0, S = false; m<N; m++) if (Tables[n]->getName() == List->item(m)->text()) S = true;
		if (!S) Tables[n] = 0;
	}
	for (n=0; (n < NTables ? Tables[n] != 0 : false); n++) ;
	for (m=n+1; m < NTables; m++) if (Tables[m] != 0) Tables[n++] = Tables[m];
	NTables = n;
	accept();
}

void TSDialog::removeTable()
{
	int n;
	if ((n = List->currentRow()) != -1) delete List->takeItem(n);
}


vAssignDialog::vAssignDialog(MainWindow* parent, ElState* S) : QDialog(parent)
{
	int n, NP = S->getNumPotentials();
	QGridLayout *L = new QGridLayout(this);
	NT = (State = S)->getNumTermTables();
	ND = State->getNumDunTables();
	setWindowTitle("Please select data source for v assignment");
	L->addWidget(new QLabel("Data Source:", this), 0, 0);
	L->addWidget(Box = new QComboBox(this), 0, 1);
	L->addWidget(List = new QListWidget(this), 1, 0, 3, 1);
	L->addWidget(Add = new QPushButton("<-Add", this), 1, 1);
	L->setRowMinimumHeight(2, 20);
	L->addWidget(Romove = new QPushButton("<-Remove", this), 3, 1);
	L->addWidget(new QLabel("Tolerance:", this), 4, 0);
	L->addWidget(Tolerance = new QLineEdit("10.0", this), 4, 1);
	L->addWidget(new QLabel("Tol f. dbl. obs.:", this), 5, 0);
	L->addWidget(DATolerance = new QLineEdit("0.03", this), 5, 1);
	L->setRowMinimumHeight(6, 20);
	L->addWidget(OK = new QPushButton("OK", this), 7, 0);
	L->addWidget(Cancel = new QPushButton("Cancel", this), 7, 1);
	for (n=0; n < NT; n++)
	{
		Box->addItem(S->getTermTableName(n));
		L1.append(n);
	}
	for (n=0; n < ND; n++)
	{
		Box->addItem(S->getDunTableName(n));
		L1.append(NT + n);
	}
	for (n=0; n < NP; n++)
	{
		Box->addItem(S->getPotentialName(n));
		L1.append(NT + ND + n);
	}
	Box->setEditable(false);
	connect(Add, SIGNAL(clicked()), this, SLOT(addSource()));
	connect(Romove, SIGNAL(clicked()), this, SLOT(removeSource()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
}

void vAssignDialog::addSource()
{
	int n = Box->currentIndex();
	List->addItem(Box->currentText());
	Box->removeItem(n);
	L2.append(L1.takeAt(n));
}

bool vAssignDialog::getSelectedTables(TermTable**& T, int& N, double &Tol, double &DATol)
{
	int n;
	DunTable *D;
	Potential *P;
	if ((N = List->count()) == 0) return false;
	T = new TermTable*[N];
	Tol = Tolerance->text().toDouble();
	DATol = DATolerance->text().toDouble();
	for (n=0; n<N; n++)
	{
		if (L2[n] < NT) T[n] = State->getTermTable(L2[n]);
		else 
		{
			T[n] = 0;
			if ((L2[n] -= NT) < ND)
			{
				if ((D = State->getDunTable(L2[n])) != 0) D->calcTermEnergies(T[n], false);
			}
			else if ((P = State->getPotential(L2[n] - ND)) != 0) P->calcTermEnergies(T[n], 0, 0, false);
		}
		if (T[n] == 0)
		{
			delete[] T;
			return false;
		}
	}
	return true;
}

void vAssignDialog::removeSource()
{
	int n = List->currentRow();
	if (n<0) return;
	Box->addItem(List->currentItem()->text());
	delete List->takeItem(n);
	L1.append(L2.takeAt(n));
}


SourceOffsetDialog::SourceOffsetDialog(QStringList& Names, double* Offsets, MainWindow* parent): QDialog(parent)
{
	int n;
	offsets = Offsets;
	lastIndex = 0;
	QGridLayout *L = new QGridLayout(this);
	QComboBox *NameB = new QComboBox(this);
	QPushButton *SaveB = new QPushButton("Save", this), *Abort = new QPushButton("Abort", this);
	setWindowTitle("Change source offset");
	L->addWidget(new QLabel("Source:", this), 0, 0);
	L->addWidget(NameB, 0, 1);
	NameB->setEditable(false);
	for (n=0; n < Names.count(); n++) NameB->addItem(Names[n]);
	L->addWidget(new QLabel("Offset:", this), 1, 0);
	L->addWidget(OffsetE = new QLineEdit(QString::number(Offsets[0], 'f', 12), this), 1, 1);
	L->setRowMinimumHeight(2, 20);
	L->addWidget(SaveB, 3, 0);
	L->addWidget(Abort, 3, 1);
	connect(NameB, SIGNAL(currentIndexChanged(int)), this, SLOT(ChangeIndex(int)));
	connect(SaveB, SIGNAL(clicked()), this, SLOT(Save()));
	connect(Abort, SIGNAL(clicked()), this, SLOT(reject()));
}

void SourceOffsetDialog::ChangeIndex(int Index)
{
	offsets[lastIndex] = OffsetE->text().toDouble();
	OffsetE->setText(QString::number(offsets[lastIndex = Index], 'f', 12));
}

void SourceOffsetDialog::Save()
{
	offsets[lastIndex] = OffsetE->text().toDouble();
	accept();
}


bool sortIvJFreqF(QTableWidget *Tab, int n, int m)
{
	if (n==-1) return false;
	if (m==-1) return true;
	int in = Tab->item(n, 0)->text().toInt(), Jsn = Tab->item(n, 2)->text().toInt(), Jsm = Tab->item(m, 2)->text().toInt();
	int im = Tab->item(m, 0)->text().toInt(), Jssn = Tab->item(n, 4)->text().toInt(), Jssm = Tab->item(m, 4)->text().toInt();
	if (in < im) return true;
	if (in > im) return false;
	in = Tab->item(n, 1)->text().toInt(), im = Tab->item(m, 1)->text().toInt();
	if (in < im) return true;
	if (in > im) return false;
	if ((in = abs(Jssn - Jsn)) < (im = abs(Jsm - Jssm))) return true;
	if (in > im) return false;
	if (Jsn < Jsm) return true;
	if (Jsn > Jsm) return false;
	if ((in = Tab->item(n, 3)->text().toInt()) < (im = Tab->item(m, 3)->text().toInt())) 
		return true;
	if (in > im) return false;
	if (Jssn < Jssm) return true;
	if (Jssn > Jssm) return false;
	double Intn = Tab->item(n, 8)->text().toDouble(), Intm = Tab->item(m, 8)->text().toDouble();
	if (Intn > Intm) return true;
	return false;
}

bool SortvsIvJ(QTableWidget* Tab, int n, int m)
{
	int in, im;
	if (n==-1) return false;
	if (m==-1) return true;
	if ((in = Tab->item(n, 3)->text().toInt()) > (im = Tab->item(m, 3)->text().toInt()))
		return true;
	if (in < im) return false;
	if ((in = Tab->item(n, 0)->text().toInt()) < (im = Tab->item(m, 0)->text().toInt()))
		return true;
	if (in > im) return false;
	if ((in = Tab->item(n, 1)->text().toInt()) < (im = Tab->item(m, 1)->text().toInt()))
		return true;
	if (in > im) return false;
	if ((in = Tab->item(n, 2)->text().toInt()) < (im = Tab->item(m, 2)->text().toInt()))
		return true;
	if (in > im) return false;
	if ((in = Tab->item(n, 4)->text().toInt()) < (im = Tab->item(m, 4)->text().toInt()))
		return true;
	return false;
}

bool sortIefJFreq(QTableWidget* Tab, int n, int m)
{
	int in, im, Jn, Jm;
	if (n==-1) return false;
	if (m==-1) return true;
	if ((in = Tab->item(n, 0)->text().toInt()) < (im = Tab->item(m, 0)->text().toInt())) return true;
	if (in > im) return false;
	if ((in = fabs((Jn = Tab->item(n, 2)->text().toInt()) - Tab->item(n, 4)->text().toInt()))  
		 < (im = fabs((Jm = Tab->item(m, 2)->text().toInt()) - Tab->item(m, 4)->text().toInt()))) return true;
	if (in > im) return false;
	if (Jn < Jm) return true;
	if (Jn > Jm) return false;
	if (Tab->item(n, 8)->text().toDouble() < Tab->item(m, 8)->text().toDouble()) return true;
	return false;
}

bool sortIefJFreqv(QTableWidget* Tab, int n, int m)
{
	int in, im, Jn, Jm;
	double Im, In;
	if (n==-1) return false;
	if (m==-1) return true;
	if ((in = Tab->item(n, 0)->text().toInt()) < (im = Tab->item(m, 0)->text().toInt())) return true;
	if (in > im) return false;
	if ((in = fabs((Jn = Tab->item(n, 2)->text().toInt()) - Tab->item(n, 4)->text().toInt()))  
		 < (im = fabs((Jm = Tab->item(m, 2)->text().toInt()) - Tab->item(m, 4)->text().toInt()))) return true;
	if (in > im) return false;
	if (Jn < Jm) return true;
	if (Jn > Jm) return false;
	if ((In = Tab->item(n, 8)->text().toDouble()) < (Im = Tab->item(m, 8)->text().toDouble())) return true;
	if (In > Im) return false;
	if ((in = Tab->item(n, 1)->text().toInt()) < (im = Tab->item(m, 1)->text().toInt())) return true;
	if (in > im) return false;
	if ((in = Tab->item(n, 3)->text().toInt()) < (im = Tab->item(m, 3)->text().toInt())) return true;
	return false;
}

bool sortByProg(QTableWidget* Tab, int n, int m)
{
	if (n==-1) return false;
	if (m==-1) return true;
	QString LTn = Tab->item(n, 5)->text(), LTm = Tab->item(m, 5)->text();
	if (LTn < LTm) return true;
	if (LTn > LTm) return false;
	if (Tab->item(n, 6)->text().toInt() < Tab->item(m, 6)->text().toInt()) return true;
	return false;
}

bool sortbyDeviation(QTableWidget* Tab, int n, int m)
{
	if (n==-1) return false;
	if (m==-1) return true;
	if (fabs(Tab->item(n, 10)->text().toDouble()) > fabs(Tab->item(m, 10)->text().toDouble())) return true;
	return false;
}

bool sortbyDevR(QTableWidget* Tab, int n, int m)
{
	if (n==-1) return false;
	if (m==-1) return true;
	if (fabs(Tab->item(n, 11)->text().toDouble()) > fabs(Tab->item(m, 11)->text().toDouble())) return true;
	return false;
}

bool sortforTFGS(QTableWidget* Tab, int n, int m)
{
	if (n==-1) return false;
	if (m==-1) return true;
	QString Bn = Tab->item(n, 3)->text(), Bm = Tab->item(m, 3)->text();
	if (Bn == "TE" && Bm != "TE") return true;
	if (Bn != "TE" && Bm == "TE") return false;
	QString Sn = Tab->item(n, 5)->text(), Sm = Tab->item(m, 5)->text();
	if (Sn < Sm) return true;
	if (Sn > Sm) return false;
	if (Bn == "nA" && Bm != "nA") return true;
	if (Bn != "nA" && Bm == "nA") return false;
	int vn = Bn.toInt(), vm = Bm.toInt();
	if (vn < vm) return true;
	if (vn > vm) return false;
	int Jn = Tab->item(n, 4)->text().toInt(), Jm = Tab->item(m, 4)->text().toInt();
	if (Jn < Jm) return true;
	if (Jn > Jm) return false;
	if (Tab->item(n, 6)->text().toInt() < Tab->item(m, 6)->text().toInt()) return true;
	return false;
}
