//
// C++ Implementation: FitObject
//
//
// Author: Alexander Stein <AlexanderStein@t-online.de>, (C) 2014 - 2016
//
// Copyright: See README file that comes with this source code
//
//

#include "fitobject.h"
#include "utils.h"
#include "tools.h"

#include <cmath>
#include <limits>
#include <stdlib.h>
#include <stdio.h>

using std::isnan;

FitObject::FitObject(int NPar)
{
	nData = 0;
	nPar = NPar;
	X = Y = sig = 0;
}

FitObject::FitObject(int NPar, double* x, double* y, double* Sig, int N)
{
	nPar = NPar;
	X=x;
	Y=y;
	sig = Sig;
	nData = N;
}

FitObject::~FitObject()
{
	destroyData();
}

void FitObject::addDataPoint(double x, double y, double Sig)
{
	double *XB = new double[++nData], *YB = new double[nData], *SB = new double[nData];
	int n;
	for (n=0; (n < nData - 1 ? X[n] < x : false); n++)
	{
		XB[n] = X[n];
		YB[n] = Y[n];
		SB[n] = sig[n];
	}
	XB[n] = x;
	YB[n] = y;
	SB[n] = Sig;
	while (n+1 < nData)
	{
		XB[n+1] = X[n];
		YB[n+1] = Y[n];
		SB[n+1] = sig[n];
		n++;
	}
	destroyData();
	X = XB;
	Y = YB;
	sig = SB;
	
}

void FitObject::destroyData()
{
	if (X != 0)
	{
		delete[] X;
		delete[] Y;
		delete[] sig;
	}
}

void FitObject::getCalcY(double* Ycalc)
{
	int n;
	for (n=0; n < nData; n++) Ycalc[n] = 0.0;
}

void FitObject::getDerivatives(double** deriv)
{
	int n, m;
	for (n=0; n < nData; n++) for (m=0; m < nPar; m++) deriv[n][m] = 0.0;
}

void FitObject::getPar(double* Par)
{
	int n;
	for (n=0; n < nPar; n++) Par[n] = 0;
}

double FitObject::LevenbergMarquardt(int MaxIt, double MinImp)
{
	double **LMEQS = Create(nPar, nPar + 1), **tLMEQS = Create(nPar, nPar + 1), wt, **EQS = Create(nData, nPar);
	double *YC = new double[nData], done, BadCount, lambda, *C = new double[nPar], FQS, aFQS, FQSv, *bPar = new double[nPar];
	int n, i, j, it = 0;
	for (n=0; n < nPar; n++) for (i=0; i <= nPar; i++) LMEQS[n][i] = 0.0;
	getCalcY(YC);
	getDerivatives(EQS);
	getPar(bPar);
	for (FQSv = 0.0, n=0; n < nData; n++) FQSv += sqr(Y[n] - YC[n]) * sig[n];
	for (n=0; n < nData; n++)
	{
		for (i=0; i < nPar; i++)
		{
			wt = EQS[n][i] * sig[n];
			for (j=0; j < nPar; j++) LMEQS[i][j] += wt * EQS[n][j];
			LMEQS[i][nPar] += wt * (Y[n] - YC[n]);
		}
	}
	aFQS = FQS = FQSv;
	for (done = BadCount = 0, lambda = 1e-3; true; it++)
	{
		//printf("bPar[0]=%g, bPar[1]=%g, bPar[2]=%g, aFQS=%g, FQS=%g\n", bPar[0], bPar[1], bPar[2], aFQS, FQS);
		if (done == 4) lambda = 0.0;
		for (n=0; n < nPar; n++)
		{
			for (i=0; i <= nPar; i++) tLMEQS[n][i] = LMEQS[n][i];
			tLMEQS[n][n] *= (1.0 + lambda);
		}
		SolvLinEqS(tLMEQS, nPar);
		for (n=0; n < nPar; n++) C[n] = tLMEQS[n][nPar];
		updatePar(C);
		getCalcY(YC);
		for (FQS = 0.0, n=0; n < nData; n++) FQS += sqr(Y[n] - YC[n]) * sig[n];
		if (it == MaxIt || done == 4 || BadCount == 20) break;
		if (abs(FQS - aFQS) <  1e-10 * aFQS || (FQS < aFQS && aFQS - FQS < MinImp)) done++;
		if (FQS < aFQS && !isnan(FQS))
		{
			BadCount = 0;
			lambda *= 0.1;
			aFQS = FQS;
			getDerivatives(EQS);
			getPar(bPar);
			for (n=0; n < nPar; n++) for (i=0; i <= nPar; i++) LMEQS[n][i] = 0.0;
			for (n=0; n < nData; n++) for (i=0; i < nPar; i++)
			{
				wt = EQS[n][i] * sig[n];
				for (j=0; j < nPar; j++) LMEQS[i][j] += wt * EQS[n][j];
				LMEQS[i][nPar] += wt * (Y[n] - YC[n]);
			}
		}
		else 
		{
			BadCount++;
			lambda *= 10.0;
			setPar(bPar);
		}
	}
	if (FQS > aFQS || isnan(FQS))
	{
		setPar(bPar);
		FQS = aFQS;
		//getCalcY(YC);
	}
	//for (n=0; n < nData; n++) printf("X[%d]=%g, YC=%g, YO=%g\n", n, X[n], YC[n], Y[n]);
	Destroy(LMEQS, nPar);
	Destroy(tLMEQS, nPar);
	Destroy(EQS, nData);
	delete[] YC;
	delete[] C;
	delete[] bPar;
	return FQS;
}

void FitObject::setData(double* x, double* y, double* Sig, int N)
{
	destroyData();
	X = x;
	Y = y;
	nData = N;
	sig = Sig;
}

void FitObject::setPar(double* /*Par*/)
{

}

void FitObject::updatePar(double* /*C*/)
{

}

Lorentzian::Lorentzian() : FitObject(3)
{
	AWF = EWF = GWF = A = E = Gamma = 0.0;
}

Lorentzian::Lorentzian(double* x, double* y, double* Sig, int N) : FitObject(3)
{
	setData(x, y, Sig, N);
}

void Lorentzian::getCalcY(double* YCalc)
{
    int n;
	double AG = A * Gamma;
	for (n=0; n < nData; n++) YCalc[n] = AG / (sqr(E - X[n]) + Gamma);
}

void Lorentzian::getDerivatives(double** deriv)
{
    double ME = 0.0, MA = 0.0, MG = 0.0, D, DSq, Div, AG = A * Gamma;
	int n;
	for (n=0; n < nData; n++)
	{
		D = E - X[n];
		DSq = sqr(D);
		Div = 1.0 / (DSq + Gamma);
		deriv[n][0] = Gamma * Div;
		deriv[n][1] = -2 * AG * D * sqr(Div);
		deriv[n][2] = A * Div * (1.0 - Gamma * Div);
		if (fabs(deriv[n][0]) > MA) MA = fabs(deriv[n][0]);
		if (fabs(deriv[n][1]) > ME) ME = fabs(deriv[n][1]);
		if (fabs(deriv[n][2]) > MG) MG = fabs(deriv[n][2]);
	}
	AWF = 1.0 / MA;
	EWF = 1.0 / ME;
	GWF = 1.0 / MG;
	for (n=0; n < nData; n++)
	{
		deriv[n][0] *= AWF;
		deriv[n][1] *= EWF;
		deriv[n][2] *= GWF;
	}
}

double Lorentzian::getE()
{
	return E;
}

void Lorentzian::getPar(double* Par)
{
    Par[0] = A;
	Par[1] = E;
	Par[2] = Gamma;
}

void Lorentzian::setData(double* x, double* y, double* Sig, int N)
{
    int n, m=0;
	AWF = EWF = GWF = 0.0;
	for (n=0, A=0.0; n<N; n++) if (y[n] > A)
	{
		m=n;
		A = y[n];
	}
	E = x[m];
	if (N>2 && m>0 && m<N-1) Gamma = sqr(x[m+1] - x[m-1]);
	else Gamma = 1;
	FitObject::setData(x, y, Sig, N);
}

void Lorentzian::setPar(double* Par)
{
    A = Par[0];
	E = Par[1];
	Gamma = Par[2];
}

void Lorentzian::updatePar(double* C)
{
    A += AWF * C[0];
	E += EWF * C[1];
	Gamma += GWF * C[2];
}

