//
// C++ Implementation: NaturalSpline
//
//
// Author: Alexander Stein <AlexanderStein@t-online.de>, (C) 2011 - 2016
//
// Copyright: See README file that comes with this source code
//
//

#include <QList>

#include "naturalspline.h"
#include "utils.h"
#include "fit.h"

typedef QList<SplinePoint*> SplinePointArrayList;

NaturalSpline::NaturalSpline()
{
	L = 0;
	points = 0;
	numPoints = 0;
}

NaturalSpline::NaturalSpline(SplinePoint* Npoints, int N)
{
	L = 0;
	points = Npoints;
	numPoints = N;
	CalcLMatrix();
}

NaturalSpline::~NaturalSpline()
{
	if (points != 0) delete[] points;
	if (L != 0) Destroy(L, numPoints);
}

void NaturalSpline::CalcLMatrix()
{
	int N = numPoints, n, i, st = numPoints -1;
	if (N == 0) return;
	if (L==0) L = Create(N, N);
	for (n=0; n<N; n++) for (i=0; i<N; i++) L[n][i] = 0.0;
	points[0].yss = points[st].yss = 0.0;
	if (N<3) return;
	double B[st], F, dx[st];
	for (i=0; i < st; i++) dx[i] = points[i+1].x - points[i].x;
	B[1] = 1.0 / (points[2].x - points[0].x);
	for (n=1; n < st; n++) 
	{
		L[n][n-1] = (n>1 ? L[n-1][n] : 3.0 / dx[0]);
		L[n][n+1] = 3.0 / dx[n];
		L[n][n] = - L[n][n-1] - L[n][n+1];
	}
	for (n=2; n < st; n++)
	{
		F = 0.5 * dx[n-1] * B[n-1];
		B[n] = 1.0 / (points[n+1].x - points[n-1].x - F * 0.5 * dx[n-1]);
		for (i=0; i<=n; i++) L[n][i] -= F * L[n-1][i];
	}
	for (i=0; i<N; i++) L[st-1][i] *= B[st-1];
		//		(L[st][i] - 0.5 * dx[st + 1] * points[st+2].yss) * B[st - 1];
	for (n = st - 2; n>0; n--) for (i=0; i<N; i++) 
			L[n][i] = (L[n][i] - 0.5 * dx[n] * L[n+1][i]) * B[n];
	for (n=1; n < st; n++) for (i=0, points[n].yss = 0.0; i<N; i++) points[n].yss += L[n][i] * points[i].y;
}

double* NaturalSpline::getCurve(double xmin, double xmax, int NPoints)
{
	int n, i, x, N = numPoints, st = numPoints - 1;
	double *y = new double[NPoints];
	if (N<2)
	{
		if (N==0) for (n=0; n < NPoints; n++) y[n] = 0.0;
		else for (n=0; n < NPoints; n++) y[n] = points[0].y;
		return y;
	}
	double dx[st], A, B, h = (xmax - xmin) / double(NPoints - 1), es = 1.0 / 6.0;
	for (n=0; n < st; n++) dx[n] = points[n+1].x - points[n].x;
	for (x = xmin, n=0; x < points[0].x; x+=h, n++) y[n] = 0.0;
	for (i=0; n < NPoints && x <= points[st].x; n++, x+=h)
	{
		while(points[i].x < x) i++;
		if (i>0)
		{
			B = 1.0 - (A = (points[i].x - x) / dx[i-1]);
			y[n] = A * points[i-1].y + B * points[i].y + es * (dx[i-1] * dx[i-1]) * ((A*A*A-A) * points[i-1].yss 
				 + (B*B*B-B) * points[i].yss);
		}
		else y[n] = points[0].y;
	}
	while (n < NPoints) y[n++] = 0.0;
	if (xmax == points[st].x) y[NPoints - 1] = points[st].y; 
	return y;
}

double* NaturalSpline::getDerivation(double Rmin, double Rmax, int NPoints)
{
	if (points == 0) return 0;
	int n, m;
	double r, *R = new double[NPoints], h = (Rmax - Rmin) / (NPoints - 1), xd = 0.0, A, B, eds = 0.1666666666667, T = 0.0, F = 0.0;
	for (n=0, r = Rmin; n < NPoints && r < points[0].x; n++, r+=h) R[n] = 0.0;
	for (m=-1; n < NPoints && r < points[numPoints - 1].x; n++, r+=h)
	{
		if (r > points[m+1].x)
		{
			while (r > points[m+1].x) m++;
			xd = 1.0 / (points[m+1].x - points[m].x);
			T = (points[m+1].y - points[m].y) * xd;
			F = eds * (points[m+1].x - points[m].x);
		}
		B = (r - points[m].x) * xd;
		A = 1.0 - B;
		R[n] = T + F * ((1.0 - 3.0 * A * A) * points[m].yss + (3.0 * B * B - 1.0) * points[m+1].yss);
	}
	while (n < NPoints) R[n] = 0.0;
	return R;
}

double NaturalSpline::getDerivative(double R)
{
	if (points == 0) return 0.0;
	if (R < points[0].x || R > points[numPoints - 1].x) return 0.0;
	int m;
	double xd, A, B, eds = 0.1666666666667, T, F;
	for (m=0; R > points[m+1].x; m++) ;
	xd = 1.0 / (points[m+1].x - points[m].x);
	T = (points[m+1].y - points[m].y) * xd;
	F = eds * (points[m+1].x - points[m].x);
	B = (R - points[m].x) * xd;
	A = 1.0 - B;
	return T + F * ((1.0 - 3.0 * A * A) * points[m].yss + (3.0 * B * B - 1.0) * points[m+1].yss);
}

int NaturalSpline::getNumPoints()
{
	return numPoints;
}

double NaturalSpline::getPoint(double x)
{
	if (numPoints > 0 ? x < points[0].x || x > points[numPoints - 1].x : true) return 0.0;
	int i;
	for (i=0; i < numPoints; i++) if (x == points[i].x) return points[i].y;
	for (i=0; points[i].x < x; i++) ;
	double dx = points[i].x - points[i-1].x;
	double A = (points[i].x - x) / dx;
	double B = 1.0 - A;
	return A * points[i-1].y + B * points[i].y 
	         + dx * dx / 6.0 * ((A*A*A-A) * points[i-1].yss + (B*B*B-B) * points[i].yss); 
}

double** NaturalSpline::getS(double xmin, double xmax, int NPoints)
{
	int n, s, i, x, N = numPoints, st = numPoints - 1;
	if (N==0) return 0;
	double **S = Create(N, NPoints);
	if (N<2)
	{
		for (n=0; n < NPoints; n++) S[0][n] = 0.0;
		return S;
	}
	double dx[st], dxq, A, B, C, D, h = (xmax - xmin) / double(NPoints - 1), es = 1.0 / 6.0;
	for (n=0; n < st; n++) dx[n] = points[n+1].x - points[n].x;
	for (x = xmin, n=0; x < points[0].x; x+=h, n++) for (s=0; s<N; s++) S[s][n] = 0.0;
	for (i=0; n < NPoints && x <= points[st].x; n++, x+=h)
	{
		while(points[i].x < x) i++;
		if (i>0)
		{
			B = 1.0 - (A = (points[i].x - x) / dx[i-1]);
			C = (dxq =  es * (dx[i-1] * dx[i-1])) * (A*A*A-A);
			D = dxq * (B*B*B-B); 
			for (s=0; s<N; s++) S[s][n] = C * L[i-1][s] + D * L[i][s]; 
			S[i-1][n] += A;
			S[i][n] += B;
		}
		else for (s=1, S[0][n] = 1.0; s<N; s++) S[s][n] = 0.0;
	}
	for ( ; n < NPoints; n++) for (s=0; s<N; s++) S[s][n] = 0.0;
	if (xmax == points[st].x) for (s=0, S[st][NPoints - 1] = 1.0; s < st; s++) S[s][NPoints - 1] = 0.0; 
	return S;
}

void NaturalSpline::setPoints(SplinePoint* Npoints, int N)
{
	if (points != 0) delete[] points;
	if (L!=0) 
	{
		Destroy(L, numPoints);
		L=0;
	}
	points = Npoints;
	numPoints = N;
	CalcLMatrix();
}

void NaturalSpline::setyValues(double* yi)
{
	int n, i;
	for (n=0; n < numPoints; n++) points[n].y = yi[n];
	for (n=1; n < numPoints - 1; n++) for (i=0, points[n].yss = 0.0; i < numPoints - 1; i++) 
		points[n].yss += L[n][i] * points[i].y;
}


Spline::Spline(int nSplinePoints, SplinePoint* SPoints, bool natural) 
    : L(0), S(0), S0(0.0), SN(0.0), NDataPoints(0), NSplinePoints(nSplinePoints), 
      DataPoints(0), points(SPoints), Natural(natural), DataOwner(false)
{
}

Spline::~Spline()
{
	delete[] points;
	if (NDataPoints > 0) delete[] DataPoints;
	deleteSL();
}

void Spline::calcS()
{
	int n, m, s = (Natural ? 0 : 1), p = 0;
	double A, B, C, D, DR, d6 = 1.0 / 6, DR2;
	if (L == 0)
	{
		L = Create(NSplinePoints, NSplinePoints + 2*s);
		CalcLMatrix(L, points, 0.0, NSplinePoints, Natural);
	}
	S = Create(NSplinePoints + 2*s, NDataPoints);
	for (n = 0; n < NDataPoints; n++)
	{
		while (DataPoints[n].x > points[p+1].x && p < NSplinePoints - 2) p++;
		DR = points[p+1].x - points[p].x;
		DR2 = DR * DR * d6;
		A = (points[p+1].x - DataPoints[n].x) / DR;
		B = 1.0 - A;
		C = A * (A*A - 1.0) * DR2;
		D = B * (B*B - 1.0) * DR2;
		for (m=0; m < NSplinePoints + 2 * s; m++) S[m][n] = L[p][m] * C + L[p+1][m] * D;
		S[p+s][n] += A;
		S[p+s+1][n] += B;
	}
}

void Spline::calcFQSRes(double* Res, double& FQS)
{
	int n;
	for (n=0, FQS = 0.0; n < NDataPoints; n++)
	{
		Res[n] = DataPoints[n].y - gety(DataPoints[n].x);
		FQS += Res[n] * Res[n] / (DataPoints[n].sig * DataPoints[n].sig);
	}
	printf("FQS=%f\n", FQS);
}

void Spline::calcYss()
{
	int n, m, s = (Natural ? 0 : 1);
	if (Natural && NSplinePoints <= 2) for (n=0; n < NSplinePoints; n++) points[n].yss = 0.0;
	else for (n=0; n < NSplinePoints; n++)
	{
		if (!Natural) points[n].yss = S0 * L[n][0];
		else points[n].yss = 0.0;
		for (m=0; m < NSplinePoints; m++) points[n].yss += L[n][m+s] * points[m].y;
		if (!Natural) points[n].yss += SN * L[n][m+s];
	}
}

void Spline::copyData(int N, Point* dpoints)
{
	int n;
	if (DataPoints != 0) delete[] DataPoints;
	DataPoints = new Point[N];
	for (n=0; n<N; n++) DataPoints[n] = dpoints[n];
	NDataPoints = N;
}

void Spline::deleteSL()
{
	if (L != 0) Destroy(L, NSplinePoints);
	if (S != 0) Destroy(S, NSplinePoints + (Natural ? 0 : 2));
	S = L = 0;
}

void Spline::FitTo(int N, Point* dpoints)
{
	int n = 0;
	if (N == 0)
	{
		for (n=0; n < NSplinePoints; n++) 
			points[n].x = points[n].y = points[n].yss = 0.0;
		return;
	}
	if (points[0].x == points[NSplinePoints - 1].x)
	{
		double Y = 0.0, WS = 0.0, W;
		for (n=0; n<N; n++) 
		{
			W = dpoints[n].sig * dpoints[n].sig;
			Y += W * dpoints[n].y;
			WS += W;
		}
		Y /= WS;
		for (n=0; n < NSplinePoints; n++)
		{
			points[n].y = Y;
			points[n].yss = 0.0;
		}
		copyData(N, dpoints);
		return;
	}
	if (N < NSplinePoints)
	{
		for (n=0; n < NSplinePoints; n++)
		{
			points[n].x = dpoints[0].x;
			points[n].y = dpoints[0].y;
			points[n].yss = 0.0;
		}
		copyData(N, dpoints);
		return;
	}
	if (N == NSplinePoints)
	{
		for (n=0; n<N; n++) 
		{
			points[n].x = dpoints[n].x;
			points[n].y = dpoints[n].y;
			points[n].yss = 0.0;
		}
		copyData(N, dpoints);
		return;
	}
	if (Natural && NSplinePoints == 2)
	{
		double *x = new double[N], *y = new double[N], *unc = new double[N];
		for (n=0; n<N; n++)
		{
			x[n] = dpoints[n].x;
			y[n] = dpoints[n].y;
			unc[n] = dpoints[n].unc;
		}
		FitToStraitLine(N, x, y, unc, points[0].x, points[1].x, points[0].y, points[1].y);
		points[0].yss = points[1].yss = 0.0;
		delete[] x;
		delete[] y;
		delete[] unc;
		copyData(N, dpoints);
		return;
	}
	int m, NC = (Natural ? NSplinePoints : NSplinePoints + 2), s = (Natural ? 0 : 1);
	double C[NC], *Y = new double[N], *Sig = new double[N], FQS, cFQS;
	if (S != 0) 
	{
		if (N == NDataPoints) while (n < N && dpoints[n].x == DataPoints[n].x) n++;
		if (N != NDataPoints || n < N)
		{
			Destroy(S, NSplinePoints + 2 * s);
			S = 0;
		}
	}
	copyData(N, dpoints);
	if (S == 0) calcS();
	double **EQS = Create(N,  NC);
	for (n=0; n < N; n++) 
	{
		Y[n] = dpoints[n].y;
		Sig[n] = dpoints[n].sig;
		for (m=0; m < NC; m++) EQS[n][m] = S[m][n];
	}
	SvdFit(EQS, C, Y, Sig, NC, N, 1e-12);
	if (!Natural) S0 = C[0];
	for (n=0; n < NSplinePoints; n++) points[n].y = C[n+s];
	if (!Natural) SN = C[NC - 1];
	calcYss();
	calcFQSRes(Y, FQS);
	cFQS = 0.9 * FQS; 
	while (FQS - cFQS > 1e-10)
	{
		FQS = cFQS;
		SvdFit(EQS, C, Y, Sig, NC, N, 1e-12);
		UpdateY(C);
		calcYss();
		calcFQSRes(Y, cFQS);
	}
	delete[] Y;
	delete[] Sig;
}

void Spline::GetSFuncs(int i, double** Data, int NPoints)
{
	bool DataEqual = false;
	int n;
	if (NPoints == NDataPoints)
	{
		for (n=0; n < NPoints && Data[n][0] == DataPoints[n].x; n++) ;
		if (n == NPoints) DataEqual = true;
	}
	if (!DataEqual)
	{
		delete[] DataPoints;
		DataPoints = new Point[NDataPoints = NPoints];
		for (n=0; n < NPoints; n++) DataPoints[n].x = Data[n][0];
		calcS();
	}
	for (n=0; n < NPoints; n++) Data[n][1] = S[i][n];
}

double Spline::gety(double x)
{
	double DR, d6 = 1.0 / 6, DR2, A, B, C, D;
	int p=0;
	if (x < points[0].x || (NSplinePoints > 1 && points[0].x == points[1].x)) 
		return points[0].y;
	if (x > points[NSplinePoints - 1].x) return points[NSplinePoints - 1].y;
	while (x > points[p+1].x) p++;
	DR = points[p+1].x - points[p].x;
	DR2 = DR * DR * d6;
	A = (points[p+1].x - x) / DR;
	B = 1.0 - A;
	C = A * (A*A - 1.0) * DR2;
	D = B * (B*B - 1.0) * DR2;
	return A * points[p].y + B * points[p+1].y + C * points[p].yss + D * points[p+1].yss;
}

double Spline::getyn(int n)
{
	if (Natural) return points[n].y;
	if (n==0) return points[n].yss;
	if (n == NSplinePoints + 1) return points[n-2].yss;
	return points[n-1].y;
}

void Spline::setNatural(bool natural)
{
	if (Natural != natural)
	{
		Natural = natural;
		deleteSL();
	}
}

void Spline::setSplinePoints(int NumSplinePoints, SplinePoint* SplinePoints)
{
	deleteSL();
	delete[] points;
	points = SplinePoints;
	NSplinePoints = NumSplinePoints;
}

void Spline::setyn(int n, double y)
{
	if (Natural) points[n].y = y;
	else if (n==0) points[n].yss = y;
	else if (n == NSplinePoints + 1) points[n-2].yss = y;
	else points[n-1].y = y;
}

void Spline::shifty(double val)
{
	int n;
	for (n=0; n < NSplinePoints; n++) points[n].y += val;
}

void Spline::UpdateY(double* C)
{
	int n, s = (Natural ? 0 : 1);
	if (!Natural) S0 += C[0];
	for (n=0; n < NSplinePoints; n++) points[n].y += C[n+s];
	if (!Natural) SN += C[n+s];
}


ResidualFit::ResidualFit() : NumSplines(0), splines(0), cycleDetectList(new QList<SplinePoint*>)
{
}

ResidualFit::~ResidualFit()
{
	clear();
	clearCycleDetectList();
	delete cycleDetectList;
}

void ResidualFit::clear()
{
	int n;
	if (NumSplines > 0)
	{
		for (n=0; n < NumSplines; n++) delete splines[n];
		delete[] splines;
		NumSplines = 0;
		splines = 0;
	}
}

void ResidualFit::FitTo(int N, Point* points)
{
	if (NumSplines > 0) clear();
	if (N==0 || points == 0) return;
	Point B;
	double step, cp, devm2 = 0.0, devm1 = 0.0, cdev = 0.0;
	double devp1 = 0.0;
	QList<int> jumpList, orderList;
	Point *DataPoints = 0; //new Point[N];
	int n, m, j, k=0, l=0, p, Nsps, i, *pos;
	SplinePoint *sps;
	bool toImprove = true, moved, pointDeleted;
	B.x = 1.0;
	while (B.x != -1.0)
	{
		B.x = -1.0;
		for (n=0; n<N-1; n++) 
			if (points[n].x > points[n+1].x 
				|| (points[n].x == points[n+1].x && n>0 
				 && fabs(points[n].y - points[n-1].y) 
				     > fabs(points[n-1].y - points[n+1].y))) 
		{
			B = points[n];
			points[n] = points[n+1];
			points[n+1] = B;
		}
	}
	for (n=0; n<N; n++) if (points[n].unc > 9.0)
	{
		while (points[n].unc > 9.0) points[n].unc -= 9.0;
		points[n].sig = 1.0 / points[n].unc;
	}
	splines = new Spline*[NumSplines = 1];
	sps = new SplinePoint[2];
	sps[0].x = points[0].x;
	sps[1].x = points[N-1].x;
	splines[0] = new Spline(2, sps, N<10);
	printf("Step 1: Fit spline 1 of 1\n");
	splines[0]->FitTo(N, points);
	for (n=0; n<N; n++)
	{
		devm2 = devm1;
		devm1 = cdev;
		cdev = devp1;
		devp1 = points[n].y - splines[0]->gety(points[n].x);
		if (cdev * devm1 < 0.0 && devm1 * devm2 > 0.0 && cdev * devp1 > 0.0 
				&& (fabs(devm1) > fabs(devm2) || fabs(cdev) > fabs(devp1)) 
				&& (fabs(cdev) > 2.0 * points[n-1].unc 
					|| fabs(devm1) > 2.0 * points[n-2].unc)) 
			jumpList.append(n-1);
	}
	if (jumpList.size() > 0)
	{
		delete splines[0];
		delete splines;
		splines = new Spline*[NumSplines = jumpList.size() + 1];
		for (n=0; n < NumSplines; n++)
		{
			sps = new SplinePoint[2];
			sps[0].x = points[m = (n>0 ? jumpList[n-1] : 0)].x;
			sps[1].x = points[(k = (n < jumpList.size() ? jumpList[n] : N)) - 1].x;
			splines[n] = new Spline(2, sps, k-m<10);
			if ((j=k-m) == 1) 
			{
				sps[0].y = sps[1].y = points[m].y;
				sps[0].yss = sps[1].yss = 0.0;
			}
			else 
			{
				printf("Step 2: Fit spline %d of %d\n", n+1, NumSplines);
				splines[n]->FitTo(j, points + m);
			}
		}
	}
	for (n=0; n < NumSplines; n++) 
	{
		i = (n>0 ? jumpList[n-1] : 0);
		j = (n < jumpList.size() ? jumpList[n] : N);
		pointDeleted = false;
		while (!pointDeleted && 4 * (splines[n]->getNPoints() - 1) < j-i)
		{
			for (m=i, cdev = 0.0, toImprove = false; m<j; m++)
			{
				devm1 = cdev;
				cdev = points[m].y - splines[n]->gety(points[m].x);
				if (cdev * devm1 > 0.0) k++;
				else k=l=0;
				if (fabs(cdev) > points[m].unc) l++;
				//printf("k=%d, NPoints=%d\n", k, splines[n]->getNPoints());
				if (k>=4 && l>=3) 
				{
					toImprove = true;
					break;
				} 
			}
			if (!toImprove) break;
			if (splines[n]->isNatural()) splines[n]->setNatural(false);
			else
			{
				Nsps = splines[n]->getNPoints() + 1;
				/*for (m=i, devm1 = 0.0; m<j; m++) 
					devm1 += fabs(points[m].y - splines[n]->gety(points[m].x));
				devm1 /= Nsps;*/
				sps = new SplinePoint[Nsps];
				pos = new int[Nsps];
				sps[0].x = points[pos[0] = i].x;
				sps[Nsps - 1].x = points[pos[Nsps - 1] = j-1].x;
				/*for (m=i, k=1, devp1 = devm1, cdev = 0.0; k < Nsps - 1; m++)
				{
					cdev += (devm2 = fabs(points[m].y - splines[n]->gety(points[m].x)));
					if (cdev >= devp1)
					{
						sps[k].x = points[pos[k] = (sps[k-1].x < points[m-1].x 
										&& (points[m].x == sps[Nsps - 1].x
										 || devp1 - cdev + devm2 < cdev - devp1) ?
										      m-1 : m)].x;
						devp1 += devm1;
						k++;
					}
				}
				for (m=1; m < Nsps - 1 && pos[m-1] < pos[Nsps - 1] - 5 * (Nsps - m - 1); 
																					m++)
					if (pos[m] < pos[m-1] + 5) 
						sps[m].x = points[pos[m] = pos[m-1] + 5].x;
				for (m = Nsps - 2; m>0 && pos[m+1] > pos[0] + 5*m; m--)
					if (pos[m] > pos[m+1] - 5) 
						sps[m].x = points[pos[m] = pos[m+1] - 5].x;*/
				for (m=i, devm2 = devm1 = cdev = 0.0; m<j; m++)
				{
					devp1 = cdev;
					devm1 += fabs(cdev = points[m].y - splines[n]->gety(points[m].x));
					if (cdev * devp1 > 0.0) k++;
					else 
					{
						if (k>=4 && l>=3) devm2 += devm1;
						devm1 = 0.0;
						k=l=0;
					}
					if (fabs(cdev) > points[m].unc) l++;
				}
				cp = devm2 / (Nsps - 1);
				for (m=i, devm2 = devm1 = cdev = 0.0, p=1; m<j && p < Nsps - 1; m++)
				{
					devp1 = cdev;
					devm1 += fabs(cdev = points[m].y - splines[n]->gety(points[m].x));
					if (cdev * devp1 > 0.0) k++;
					else 
					{
						if (k>=4 && l>=3) devm2 += devm1;
						devm1 = 0.0;
						k=l=0;
					}
					if (devm2 + (k>=4 && l>=3 ? devm1 : 0.0) >= cp * double(p))
					{
						sps[p].x = points[pos[p]=m].x;
						p++;
					}
					if (fabs(cdev) > points[m].unc) l++;
				}
				if (p < Nsps - 1 || pos[p] >= j - 2)
				{
					for (k=p; k < Nsps - 1; k++) sps[k].x = points[pos[k] = j + 2 * (k - Nsps)].x;
					for (k=p-1; pos[k] >= pos[k+1] - 1; k--) sps[k].x = points[pos[k] = j + 2 * (k - Nsps)].x;
				}
				clearCycleDetectList();
				for (moved = true; moved && !inCycle(Nsps, sps); ) for (m=1, moved = false; m < Nsps - 1; m++)
				{
					k = pos[m];
					if (sps[m].x - sps[m-1].x > 2 * (sps[m+1].x - sps[m].x))
					{
						//printf("sps[%d].x = %f\n", m, points[pos[m]-1].x);
						for (sps[m].x = points[--pos[m]].x;
							sps[m].x - sps[m-1].x > 2 * (sps[m+1].x - sps[m].x);
							sps[m].x = points[--pos[m]].x) ;// printf("sps[%d].x = %f\n", m, sps[m].x);
						if (2 * (sps[m].x - sps[m-1].x) < sps[m+1].x - sps[m].x
							&& sps[m+1].x - points[pos[m]+1].x > sps[m].x - sps[m-1].x)
						{
							sps[m].x = points[++pos[m]].x;
							//printf("sps[%d].x = %f\n", m, sps[m].x);
						}
					}
					else if (2 * (sps[m].x - sps[m-1].x) < sps[m+1].x - sps[m].x)
					{
						//printf("sps[%d].x = %f\n", m, points[pos[m]+1].x);
						for (sps[m].x = points[++pos[m]].x;
							2 * (sps[m].x - sps[m-1].x) < sps[m+1].x - sps[m].x;
							sps[m].x = points[++pos[m]].x) ; //printf("sps[%d].x = %f\n", m, sps[m].x);
						if (sps[m].x - sps[m-1].x > 2 * ( sps[m+1].x - sps[m].x)
							&& sps[m+1].x - sps[m].x < points[pos[m]-1].x - sps[m-1].x)
						{
							sps[m].x = points[--pos[m]].x;
							//printf("sps[%d].x = %f\n", m, sps[m].x);
						}
					}
					if (k != pos[m]) moved = true;
				}
				if (cycleDetected) setAverage(Nsps, sps);
				for (m=1; m < Nsps; m++) 
				{
					for (k=m-1; sps[k].x == -1.0; k--) ;
					for (p=k+1; p<m && sps[p].x == sps[k].x; p++) ;
					if (sps[p].x == sps[m].x) 
					{
						sps[m].x = -1.0;
						pointDeleted = true;
					}
				}
				if (sps[Nsps - 1].x == -1.0) sps[Nsps - 2].x = splines[n]->getxN(); 
				delete[] pos;
				if (pointDeleted)
				{
					for (m=1; sps[m].x != -1.0; m++) ;
					for (p=m+1; p < Nsps; p++) if (sps[p].x != -1.0) sps[m++] = sps[p];
					Nsps = m;
				}
				splines[n]->setSplinePoints(Nsps, sps);
			}
			printf("Step3: Fit spline %d of %d\n", n+1, NumSplines);
			splines[n]->FitTo(j-i, points + i);
		}
	}
	for (n=m=0; n<N; n++)
	{
		if (m < jumpList.size() && n == jumpList[m])
		{
			k = (m>0 ? jumpList[m-1] : 0);
			printf("Step4: Fit spline %d of %d\n", m, jumpList.size() + 1);
			splines[m]->FitTo(jumpList[m] - k, points + k);
			m++;
		}
		points[n].SplineNum = m;
		if ((cdev = fabs(points[n].y - splines[m]->gety(points[n].x))) 
				> 4.0 * points[n].unc)
		{
			while (cdev > points[n].unc) points[n].unc += 9.0;
			points[n].sig = 1.0 / points[n].unc;
		}
	}
	k = (m>0 ? jumpList[m-1] : 0);
	printf("Step4: Fit spline %d of %d\n", m+1, jumpList.size() + 1);
	splines[m]->FitTo(N-k, points + k);
	return;
	
	
	
	
	for (n=0; n <= jumpList.size(); n++) orderList.append(n);
	for (n=0; n < jumpList.size(); n++)
	{
		for (m=0; orderList[m] != n; m++) ;
		for (k=n+1; k < jumpList.size() 
			&& points[jumpList[k]].x - points[jumpList[n]-1].x <= 6.0
			&& fabs(points[jumpList[k]].y - points[jumpList[n]-1].y) 
			  > 4.0 * MAX(points[jumpList[k]].sig, points[jumpList[n]-1].sig); k++) ;
		if (k < jumpList.size() 
			&& points[jumpList[k]].x - points[jumpList[n]-1].x <= 6.0)
		{
			for (j=m+1; j < orderList.size() && orderList[j] != k+1; j++) ;
			if (j < orderList.size() && j > m+1) orderList.move(j, m+1);
		}
	}
	for (n=m=0; n < orderList.size(); n++) 
		for (j=(orderList[n] > 0 ? jumpList[orderList[n]-1] : 0); 
				j < (orderList[n] < jumpList.size() ? jumpList[orderList[n]] : N); j++) 
			DataPoints[m++] = points[j];
	jumpList.clear();
	for (n=1; n<N; n++) 
		if (DataPoints[n].x - DataPoints[n-1].x > 6.0 
		  || fabs(DataPoints[n].y - DataPoints[n-1].y) 
				> 4.0 * MAX(DataPoints[n].sig, DataPoints[n-1].sig))
			jumpList.append(n);
	splines = new Spline*[NumSplines = jumpList.size() + 1];
	for (n=0; n < NumSplines; n++) 
	{
		m = (k = (n == NumSplines - 1 ? N : jumpList[n])) 
		  - (j = (n==0 ? 0 : jumpList[n-1]));
		if (m==1)
		{
			sps = new SplinePoint[Nsps = 1];
			sps[0].x = DataPoints[j].x;
		}
		else if (m < 15)
		{
			sps = new SplinePoint[Nsps = 2];
			sps[0].x = DataPoints[j].x;
			sps[1].x = DataPoints[k-1].x;
		}
		else
		{
			sps = new SplinePoint[Nsps = (m + 15) / 10];
			step = double(m) / (Nsps - 1);
			for (i=0, cp = j; i < Nsps; cp += step, i++) 
				sps[i].x = DataPoints[(i < Nsps - 1 ? int(cp + 1e-5) : k-1)].x;
		}
		splines[n] = new Spline(Nsps, sps, false);
		splines[n]->FitTo(m, DataPoints + j);
	}
	delete[] DataPoints;
}

void ResidualFit::clearCycleDetectList()
{
	int n, N = cycleDetectList->size();
	for (n=0; n<N; n++) delete[] (*cycleDetectList)[n];
	cycleDetectList->clear();
	cycleDetected = false;
}

bool ResidualFit::inCycle(int N, SplinePoint* spoints)
{
	int n, l, L = cycleDetectList->size();
	for (l=0; !cycleDetected && l<L; l++) for (cycleDetected = true, n=0; cycleDetected && n<N; n++) if (spoints[n] != (*cycleDetectList)[l][n]) 
		cycleDetected = false;
	if (!cycleDetected)
	{
		SplinePoint *deepCopy = new SplinePoint[N];
		for (n=0; n<N; n++) deepCopy[n] = spoints[n];
		cycleDetectList->append(deepCopy);
	}
	else cycleIndex = l;
	return cycleDetected;
}

void ResidualFit::setAverage(int N, SplinePoint* spoints)
{
	int l, L = cycleDetectList->size(), cycleLength = L - cycleIndex, n;
	if (!cycleDetected) return;
	for (n=0; n<N; n++)
	{
		for (l = cycleIndex + 1; l<L; l++) spoints[n] += (*cycleDetectList)[l][n];
		spoints[n] /= cycleLength;
	}
}
