//
// C++ Interface: NaturalSpline
//
//
// Author: Alexander Stein <AlexanderStein@t-online.de>, (C) 2011 - 2016
//
// Copyright: See README file that comes with this source code
//
//

#ifndef NATURALSPLINE_H
#define NATURALSPLINE_H


template<class T> class QList;

struct SplinePoint
{
	double x, y, yss;
	bool variable;
	
	inline SplinePoint& operator+=(SplinePoint &other)
	{
		x += other.x;
		y += other.y;
		yss += other.yss;
		return *this;
	}
	
	inline SplinePoint& operator/=(double div)
	{
		x /= div;
		y /= div;
		yss /= div;
		return *this;
	}
	
	inline bool operator!=(SplinePoint &other)
	{
		return x != other.x || y != other.y || yss != other.yss;
	}
};

struct Point
{
	double x, y, sig, unc;
	int SplineNum, MarkerNum;
};

class NaturalSpline
{
	public:
		NaturalSpline();
		NaturalSpline(SplinePoint *points, int N);
		~NaturalSpline();
		void setPoints(SplinePoint *points, int N);
		void setyValues(double *yi);
		int getNumPoints();
		double getPoint(double R);
		double getDerivative(double R);
		double *getDerivation(double Rmin, double Rmax, int NPoints);
		double *getCurve(double Rmin, double Rmax, int NPoints);
		double **getS(double Rmin, double Rmax, int NPoints);
		
	private:
		void CalcLMatrix();
		
		int numPoints;
		double **L;
		SplinePoint *points;
};


class Spline
{
public:
	Spline(int NSplinePoints, SplinePoint* SPoints, bool natural);
	virtual ~Spline();
	void FitTo(int N, Point *points);
	double gety(double x);
	double getyn(int n);
	void setyn(int n, double y);
	void calcYss();
	void GetSFuncs(int n, double **Data, int NPoints);
	void setNatural(bool natural);
	void setSplinePoints(int NumSplinePoints, SplinePoint* SplinePoints);
	void shifty(double val);
		
	inline double getx0()
	{
		return points[0].x;
	}
	
	inline double getxN()
	{
		return points[NSplinePoints - 1].x;
	}
	
	inline int getNSFuncs()
	{
		return (NSplinePoints > 1 ? (Natural ? NSplinePoints : NSplinePoints + 2) : 0);
	}
	
	inline bool isNatural()
	{
		return Natural;
	}
	
	inline int getNPoints()
	{
		return NSplinePoints;
	}
	
	inline int getNPar()
	{
		return (Natural ? NSplinePoints : NSplinePoints + 2);
	}
	
private:
	void calcS();
	void calcFQSRes(double *Res, double &FQS);
	void UpdateY(double *C);
	void copyData(int N, Point *points);
	void deleteSL();
	
	double **L, **S, S0, SN;
	int NDataPoints, NSplinePoints;
	Point *DataPoints;
	SplinePoint *points;
	bool Natural, DataOwner;
};


class ResidualFit
{
public:
	ResidualFit();
	virtual ~ResidualFit();
	void FitTo(int N, Point *points);
	
	inline int getNumSplines()
	{
		return NumSplines;
	}
	
	inline Spline* getSpline(int n)
	{
		return splines[n];
	}
	
private:
	void clear();
	bool inCycle(int N, SplinePoint* spoints);
	void clearCycleDetectList();
	void setAverage(int N, SplinePoint* spoints);
	
	bool cycleDetected;
	int NumSplines, cycleIndex;
	Spline** splines;
	QList<SplinePoint*> *cycleDetectList;
};

#endif // NATURALSPLINE_H
