//
// C++ Implementation: potential
//
//
// Author: Alexander Stein <AlexanderStein@t-online.de>, (C) 2006 - 2016
//
// Copyright: See README file that comes with this source code
//
//

#include "potential.h"
#include "tools.h"
#include "MainWindow.h"
#include "potential.h"
#include "termtable.h"
#include "linetable.h"
#include "molecule.h"
#include "constants.h"
#include "fit.h"
#include "inputdialog.h"
#include "fitdata.h"
#include "naturalspline.h"
#include "utils.h"
#include "duntable.h"
#include "fitobject.h"
#include "montecarlosim.h"

#include <math.h>
#include <limits>
#include <stdlib.h>

#include <QObject>
#include <QFileDialog>
#include <QString>
#include <QStringList>
#include <QTextStream>
#include <QFile>
#include <QMessageBox>
#include <QDialog>
#include <QProcess>
#include <QPushButton>
#include <QLineEdit>
#include <QListWidget>
#include <QLabel>
#include <QGridLayout>
#include <QPainter>
#include <QRadioButton>
#include <QCheckBox>
#include <QMutex>
#include <QTextEdit>
#include <QPlainTextEdit>
#include <QProgressBar>

using std::numeric_limits;

PotWorker::PotWorker(PotFit *fit, PotentialType PotType)
{
	Type = PotType;
	SMap = 0;
	NumFreePar = 0;
	NumFreeAdCorr = NumFreeLRC = 0;
	Fit = fit;
	NSpinRGamma = 0;
	NAdCorr = PAdCorr = TAdCorr = 0;
	NFitCoeff = 0;
	AdCorrWF = 0;
	SpinRb = SpinRR1 = SpinRR2 = SpinRRm = 0.0;
	SpinRGamma = 0;
	RIso1AdCorr = RIso2AdCorr = AdCorr_b = AdCorr_Rm = 0.0;
	adCorr = 0;
	adCorrFree = 0;
	S = 0;
	Maxv = MaxJ = 0;
	nEs = Mv = nGS = 0;
	mJ = NEL = nGP = nGL = 0;
 	PS = mv = 0;
	EL = 0;
	GL = 0;
	q_e = q_f = 0.0;
	Omega = 0;
	NIso = 0;
	el_S = 0;
	SFQSU = 0;
	BadListData = 0;
	BadListNR = 0;
	ShowBadList = false;
	points = 0;
	L = 0;
	NLRC = 0;
	PLRC = 0;
	LRC = 0;
	Uinf = 0.0;
	iO = iA = 0.0;
	IsoT = 0;
	numSplinePoints = 0;
	iExp = 6.0;
	Ro = 0.0;
	shortRVariable = true;
	LRCfree = 0;
	Lock = new QMutex;
	BLLock = new QMutex;
	Ndev = 0;
	Row = 0;
	dev = RWErr = 0;
	WFS = WWFS = 0;
	calcDiagFuncs = false;
	ES = 0;
	FQS = WeightSum = 0.0;
	NWeightF = 0;
	WeightFRow = 0;
	WeightF = 0;
	reCalcLC = reCalcSC = false;
	WantedValues = 0;
	NumWantedValues = 0;
}

Potential::Potential(MainWindow *Main, Molecule *Mol, int ThreadNum) : TableWindow(MDIChild::PotData, Main, Mol)
{
	setFilter("Potentials (*.pot)");
	setFileExt(".pot");
	Fit = new PotFit();
	Worker = new PotWorker(Fit);
	Fit->setWorker(Worker);
	NFC = 1;
	FixPix = new QPixmap(10, 10);
	QPainter P(FixPix);
	P.setPen(QColor(255, 0, 0));
	P.setFont(QFont("Arial", 10));
	P.drawText(0, 10, "F");
	State = 0;
	setMolecule(Mol);
	maxSplinePoints = 30;
	LRCTabOffs = adCorrTabOffs = ExcIntATabPos = -1;
	BadList = 0;
	fitN_MI = 20;
	fitN_N = 0;
	fitN_V = 4.0;
	fitN_I = 0;
	fitN_C = 0;
	fitN_A = 0.001;
	fitN_maxr = 800.0;
	fitData = 0;
	WaveFuncs = 0;
	CoupledPot = 0;
	CoupledComp = 0;
	NCoupled = 0;
	getTexTableSplineFitRunning = false;
	threadNum = ThreadNum;
	LogWindow = 0;
	writePotFitTraceTab = false;
	qRegisterMetaType<FitResult>();
	Tab->setColumnCount(4);
	Tab->setHorizontalHeaderLabels(QStringList() << "coefficient" << "value" << "error" 
										<< "sig. digits");
	Tab->setColumnWidth(0, 130);
	Tab->setColumnWidth(1, 200);
	disconnect(Tab, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(Changed()));
	connect(Tab, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(updatePot()));
	connect(Fit, SIGNAL(finished()), this, SLOT(fitFinished()));
	connect(Worker, SIGNAL(badListChanged(double,double,double)), this, SLOT(badListChanged(double,double,double)));
	connect(Worker, SIGNAL(potentialImproved(PotentialData*)), this, SLOT(potentialImproved(PotentialData*)));
	connect(Fit, SIGNAL(terminated()), this, SLOT(fitTerminated()));
	/*DebugFile = new QFile(getFileName() + "DebugOut.dat");
	DebugFile->open(QIODevice::WriteOnly);
	DebugStream = new QTextStream(DebugFile);*/
}

PotWorker::PotWorker(const PotWorker& C) : QObject()
{
	int n;
	Type = C.Type; 
	SMap = 0;
	NumFreePar = 0;
	NumFreeAdCorr = NumFreeLRC = 0;
	NFitCoeff = 0;
	if ((NSpinRGamma = C.NSpinRGamma) > 0)
	{
		SpinRGamma = new double[NSpinRGamma];
		for (n=0; n < NSpinRGamma; n++) SpinRGamma[n] = C.SpinRGamma[n];
	}
	else SpinRGamma = 0;
	if ((NAdCorr = C.NAdCorr) > 0)
	{
		adCorr = new double[NAdCorr];
		adCorrFree = new bool[NAdCorr];
		for (n=0; n < NAdCorr; n++)
		{
			adCorr[n] = C.adCorr[n];
			adCorrFree[n] = C.adCorrFree[n];
		}
	}
	else
	{
		adCorr = 0;
		adCorrFree = 0;
	}
	Fit = 0;
	PAdCorr = C.PAdCorr;
	TAdCorr = C.TAdCorr;
	AdCorrWF = 0;
	SpinRb = C.SpinRb;
	SpinRR1 = C.SpinRR1;
	SpinRR2 = C.SpinRR2;
	SpinRRm = C.SpinRRm;
	RIso1AdCorr = C.RIso1AdCorr;
	RIso2AdCorr = C.RIso2AdCorr;
	AdCorr_b = C.AdCorr_b;
	AdCorr_Rm = C.AdCorr_Rm;
	nEs = Mv = nGS = 0;
	mJ = NEL = nGP = nGL = 0;
	EL = 0;
	GL = 0;
	PS = mv = 0;
	S = 0;
	q_e = C.q_e;
	q_f = C.q_f;
	Omega = C.Omega;
	NIso = 0;
	SFQSU = C.SFQSU;
	el_S = C.el_S;
	BadListData = 0;
	BadListNR = 0;
	ShowBadList = false;
	if ((numSplinePoints = C.numSplinePoints) > 0)
	{
		points = new SplinePoint[numSplinePoints];
		for (n=0; n < numSplinePoints; n++) points[n] = C.points[n];
	}
	else points = 0;
	if ((NLRC = C.NLRC) > 0)
	{
		PLRC = new int[NLRC];
		LRC = new double[NLRC];
		LRCfree = new bool[NLRC];
		for (n=0; n < NLRC; n++)
		{
			PLRC[n] = C.PLRC[n];
			LRC[n] = C.LRC[n];
			LRCfree[n] = C.LRCfree[n];
		}
	}
	else
	{
		PLRC = 0;
		LRC = 0;
		LRCfree = 0;
	}
	//printf("Mitte Potential\n");
	Uinf = C.Uinf;
	L = 0;
	iO = C.iO;
	iA = C.iA;
	iExp = C.iExp;
	IsoT = 0;
	Ro = C.Ro;
	shortRVariable = C.shortRVariable;
	Lock = new QMutex;
	BLLock = new QMutex;
	Ndev = 0;
	Row = 0;
	dev = RWErr = 0;
	WFS = WWFS = 0;
	calcDiagFuncs = false;
	ES = 0;
	FQS = WeightSum = 0.0;
	NWeightF = 0;
	WeightFRow = 0;
	WeightF = 0;
	reCalcSC = C.reCalcSC;
	reCalcLC = C.reCalcLC;
	WantedValues = 0;
	NumWantedValues = 0;
}

Potential::Potential(const Potential &C) : TableWindow(PotData, C.MW, C.molecule)
{
	//printf("Potential::Potential(const Potential &C)\n");
	int n;
	setFilter("Potentials (*.pot)");
	setFileExt(".pot");
	FixPix = new QPixmap(10, 10);
	QPainter P(FixPix);
	P.setPen(QColor(255, 0, 0));
	P.setFont(QFont("Arial", 10));
	P.drawText(0, 10, "F");
	Fit = new PotFit();
	Worker = C.Worker->getCopy();
	Worker->setFit(Fit);
	Fit->setWorker(Worker);
	State = C.State;
	NFC = C.NFC;
	BadList = 0;
	setName(C.getName());
	setFileName(C.getFileName());
	setMolecule(C.molecule);
	//printf("Nach setMolecule\n");
	fitN_MI = 20;
	fitN_N = 0;
	fitN_V = 4.0;
	fitN_I = 0;
	fitN_C = 0;
	fitN_A = 0.001;
	fitN_maxr = 800.0;
	getTexTableSplineFitRunning = false;
	threadNum = 0;
	LogWindow = 0;
	writePotFitTraceTab = false;
	LRCTabOffs = adCorrTabOffs = ExcIntATabPos = -1;
	/*if (C.fitData != 0)
	{
		TableLine *fD;
		fitData = new FitData(State, C.MW, C.molecule);
		fitData->setvMax(C.fitData->getvMax());
		fitData->setJMax(C.fitData->getJMax());
		C.fitData->getData(fD, n);
		fitData->setData(fD, n);
	}
	else fitData = 0;*/
	fitData = C.fitData;
	if ((WaveFuncs = C.WaveFuncs) != 0) WaveFuncs->Assign();
	if ((NCoupled = C.NCoupled) > 0)
	{
		CoupledPot = new Potential*[NCoupled];
		CoupledComp = new int[NCoupled];
		for (n=0; n < NCoupled; n++)
		{
			CoupledPot[n] = C.CoupledPot[n];
			CoupledComp[n] = C.CoupledComp[n];
		}
	}
	else
	{
		CoupledPot = 0;
		CoupledComp = 0;
	}
	Tab->setColumnCount(4);
	Tab->setHorizontalHeaderLabels(QStringList() << "coefficient" << "value" << "error" 
										<< "sig. digits");
	Tab->setColumnWidth(0, 130);
	Tab->setColumnWidth(1, 200);
	//printf("Vor UpdateTab\n");
	UpdateTab();
	disconnect(Tab, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(Changed()));
	connect(Tab, SIGNAL(itemChanged(QTableWidgetItem*)), this, SLOT(updatePot()));
	connect(Fit, SIGNAL(finished()), this, SLOT(fitFinished()));
	connect(Worker, SIGNAL(badListChanged(double,double,double)), this, SLOT(badListChanged(double,double,double)));
	connect(Worker, SIGNAL(potentialImproved(PotentialData*)), this, SLOT(potentialImproved(PotentialData*)));
	connect(Fit, SIGNAL(terminated()), this, SLOT(fitTerminated()));
	//printf("Ende Potential\n");
	
	/*DebugFile = new QFile("Pot" + getName() + "Debug.dat");
	DebugFile->open(QIODevice::WriteOnly);
	DebugStream = new QTextStream(DebugFile);*/
}

PotWorker::~PotWorker()
{
	if (NSpinRGamma > 0) delete[] SpinRGamma;
	if (NAdCorr > 0) 
	{
		delete[] adCorr;
		delete[] adCorrFree;
	}
	if (AdCorrWF != 0) delete[] AdCorrWF;
	if (S!=0) Destroy(S, NumS);
	if (SMap != 0) delete[] SMap;
	if (nEs != 0 || nGS != 0) destroyFitData();
	if (NIso > 0)
	{
		delete[] MU;
		delete[] IsoF;
	}
	if (BadListNR > 0) Destroy(BadListData, BadListNR);
	if (NLRC > 0)
	{
		delete[] PLRC;
		delete[] LRC;
		delete[] LRCfree;
	}
	if (points != 0) delete[] points;
	//printf("nach del aPot\n"); 
	if (IsoT != 0) delete IsoT;
	if (L!=0) Destroy(L, numSplinePoints);
	delete Lock;
	delete BLLock;
	if (dev != 0)
	{
		delete[] dev;
		delete[] Row;
		if (RWErr != 0) delete[] RWErr;
	}
	if (WFS != 0) delete[] WFS;
	if (WWFS != 0) delete[] WWFS;
	if (ES != 0) delete[] ES;
	if (NWeightF > 0)
	{
		delete[] WeightFRow;
		delete[] WeightF;
	}
	if (NumWantedValues > 0) delete[] WantedValues;
}

Potential::~Potential()
{
	delete FixPix;
	if (NCoupled > 0)
	{
		delete[] CoupledPot;
		delete[] CoupledComp;
		if (WaveFuncs != 0 ? WaveFuncs->Delete() : false) delete WaveFuncs;
	}
	delete Fit;
	delete Worker;
	//printf("Ende ~Potential\n");
	
	/*DebugFile->close();
	delete DebugStream;
	delete DebugFile;*/
}

PotWorker* PotWorker::getCopy()
{
	switch (Type)
	{
		case NoPotential:
			return new PotWorker(*this);
		case analyticalPotential:
			return new AnaPot(*dynamic_cast<AnaPot*>(this));
		case SplinePotential:
			return new SplinePot(*dynamic_cast<SplinePot*>(this));
		case TangToenniesPotential:
			return new TangToenniesPot(*dynamic_cast<TangToenniesPot*>(this));
		case ModifiedTangToenniesPotential:
			return new MTTPot(*dynamic_cast<MTTPot*>(this));
		case MorseLongRangePotential:
			return new MLRPot(*dynamic_cast<MLRPot*>(this));
			break;
	}
	return 0;
}

void Potential::setWorker(PotWorker* NWorker)
{
	if (Fit->isRunning()) return;
	double *MU, *IsoF;
	IsoTab *IsoT;
	disconnect(Worker, SIGNAL(badListChanged(double,double,double)), this, SLOT(badListChanged(double,double,double)));
	disconnect(Worker, SIGNAL(potentialImproved(PotentialData*)), this, SLOT(potentialImproved(PotentialData*)));
	Worker->takeMolData(IsoT, MU, IsoF);
	delete Worker;
	Worker = NWorker;
	if (IsoT != 0) Worker->setMolData(IsoT, MU, IsoF);
	connect(Worker, SIGNAL(badListChanged(double,double,double)), this, SLOT(badListChanged(double,double,double)));
	connect(Worker, SIGNAL(potentialImproved(PotentialData*)), this, SLOT(potentialImproved(PotentialData*)));
	Fit->setWorker(NWorker);
	Worker->setFit(Fit);
}

bool PotWorker::setOmega(int nOmega, double S)
{
	if (Fit->isRunning()) return false;
	Omega = nOmega;
	el_S = S;
	return true;
}

bool PotWorker::setUinf(double nUinf)
{
	if (Fit->isRunning()) return false;
	Uinf = nUinf;
	return true;
}

void PotWorker::stopFit(bool Stop)
{
	if (Fit->isRunning() || StopFit) StopFit = Stop;
}

void Potential::cutRows(int& numRows, int& numColumns, QString**& Data)
{
    setBlockChangeSignal(true);
	TableWindow::cutRows(numRows, numColumns, Data);
	setBlockChangeSignal(false);
	updatePot();
}

void Potential::DeleteRows()
{
    setBlockChangeSignal(true);
	TableWindow::DeleteRows();
	setBlockChangeSignal(false);
	updatePot();
}

void Potential::insertRows(int numRows, int numColumns, QString** Data)
{
	setBlockChangeSignal(true);
    TableWindow::insertRows(numRows, numColumns, Data);
	setBlockChangeSignal(false);
	updatePot();
}

Potential* Potential::scalePotential(double newRe, double newDe)
{
	//printf("Potential::scalePotential()\n");
	int n;
	double ARe, ADe;
	getMinimum(ARe, ADe);
	double Uinf = Worker->getUinf(), xSF = newRe / ARe, ySF = newDe / (Uinf - ADe);
	double yssSF = ySF / (xSF * xSF);
	Potential *RPot = MW->CreatePotential();
	Uinf *= ySF;
	if (Worker->isSplinePot())
	{
		SplinePoint *points, *SPoints;
		double *LRC, iA, iO, iExp, *SLRC = 0;
		int numSplinePoints, NLRC, *pLRC, *SpLRC = 0;
		Worker->getSplinePotForReading(numSplinePoints, points, NLRC, pLRC, LRC, iA, iO, iExp);
		SPoints = new SplinePoint[numSplinePoints];
		for (n=0; n < numSplinePoints; n++) 
		{
			SPoints[n].x = xSF * points[n].x;
			SPoints[n].y = ySF * points[n].y;
			SPoints[n].yss = yssSF * points[n].yss;
		}
		if (NLRC > 0)
		{
			SpLRC = new int[NLRC];
			SLRC = new double[NLRC];
			for (n=0; n < NLRC; n++) SLRC[n] = LRC[n] * pow(xSF, SpLRC[n] = pLRC[n]) * ySF;
		}
		Worker->unlock();
		iO *= ySF;
		iA *= ySF * pow(xSF, iExp) * iA;
		RPot->Worker->setSplinePot(numSplinePoints, SPoints, NLRC, SpLRC, SLRC, iA, iO, iExp);
	}
	if (Worker->isAnaPotAv() || Worker->isLPot() || Worker->isMTTPot()) 
	{
		RPot->setWorker(Worker->scalePotential(newRe, newRe));
		RPot->Worker->setFit(RPot->Fit);
	}
	RPot->Worker->setUinf(Uinf);
	RPot->setName("Scaled" + getName());
	RPot->setSource("Scaled " + getSource());
	RPot->UpdateTab();
	RPot->Changed();
	return RPot;
}

void PotWorker::getSplinePotForReading(int& rNSplinePoints, SplinePoint*& rpoints, int& rNLRC, int*& rpLRC, double*& rLRC, double& riA, 
									   double& riO, double& riExp)
{
	printf("Lock->lock() PotWorker::getSplinePotForReading\n");
	Lock->lock();
	rNSplinePoints = numSplinePoints;
	rpoints = points;
	rNLRC = NLRC;
	rpLRC = PLRC;
	rLRC = LRC;
	riA = iA;
	riO = iO;
	riExp = iExp;
}

void PotWorker::getSplinePotForWriting(int& rNSplinePoints, SplinePoint*& rpoints, int& rNLRC, int*& rpLRC, double*& rLRC, double& riA,
									   double& riO, double& riExp)
{
	if (Fit->isRunning())
	{
		rNSplinePoints = rNLRC = 0;
		rpLRC = 0;
		rLRC = 0;
		rpoints = 0;
		return;
	}
	rNSplinePoints = numSplinePoints;
	rpoints = points;
	rNLRC = NLRC;
	rpLRC = PLRC;
	rLRC = LRC;
	riA = iA;
	riO = iO;
	riExp = iExp;
}

bool PotWorker::setSplinePot(int nNSplinePoints, SplinePoint* npoints, int nNLRC, int* npLRC, double* nLRC, double niA, double niO, 
							 double niExp, double LongRangeF)
{
	if (Fit->isRunning()) return false;
	int n, i;
	if (points != npoints && npoints != 0) DestroyPotentials();
	numSplinePoints = nNSplinePoints;
	points = npoints;
	NLRC = nNLRC;
	if (nLRC != LRC)
	{
		if (PLRC != 0)
		{
			delete[] PLRC;
			delete[] LRC;
			delete[] LRCfree;
		}
		LRCfree = new bool[NLRC];
		for (n=0; n < NLRC; n++) LRCfree[n] = true;
	}
	PLRC = npLRC;
	LRC = nLRC;
	iA = niA;
	iO = niO;
	iExp = niExp;
	if (LongRangeF != 0.0)
	{
		calcLMatrix();
		for (n=0; n < numSplinePoints; n++)
		{
			for (i=0, points[n].yss = iA * L[n][0]; i < numSplinePoints ; i++) points[n].yss += points[i].y * L[n][i+1];
			points[n].yss += LongRangeF * L[n][numSplinePoints + 1];
			points[n].variable = true;
		}
	}
	return true;
}

AnaPot* PotWorker::getAnaPotForReading()
{
	printf("Lock->lock() PotWorker::getAnaPotForReading\n");
	if (Type == analyticalPotential) Lock->lock();
	return dynamic_cast<AnaPot*>(this);
}

AnaPot* PotWorker::getAnaPotForWriting()
{
	if (Fit->isRunning()) return 0;
	return dynamic_cast<AnaPot*>(this);
}

void PotWorker::unlock()
{
	printf("PotWorker::unlock\n");
	Lock->unlock();
}

double* PotWorker::getMCSE(int& N, int*& Rows)
{
	int s, n;
	for (s=N=0; s < nEs; s++) N += NEL[s];
	for (s=0; s < nGS; s++) N += nGL[s];
	double *R = new double[N];
	Rows = new int[N];
	for (s=N=0; s < nEs; s++) for (n=0; n < NEL[s]; n++)
	{
		Rows[N] = EL[s][n].Row;
		R[N++] = EL[s][n].E;
	}
	for (s=0; s < nGS; s++) for (n=0; n < nGL[s]; n++)
	{
		Rows[N] = GL[s][n].Row;
		R[N++] = GL[s][n].WN;
	}
	return R;
}

void PotWorker::DestroyPotentials()
{
	if (points != 0)
	{
		delete[] points;
		points = 0;
		numSplinePoints = 0;
		if (NLRC != 0)
		{
			delete[] LRC;
			delete[] LRCfree;
			LRC = 0;
			LRCfree = 0;
			NLRC = 0;
		}
		DestroyLS();
	}
}

FitData* Potential::getFitData()
{
	fitData = State->getFitData();
	if (fitData == 0)
	{
		QString F = getFileName();
		QFile Fi(F = F.left(F.lastIndexOf('.')) + ".fdat");
		if(Fi.exists()) fitData = MW->getFitData(F.left(F.lastIndexOf('.')) + ".fdat", molecule);
	}
	if (fitData != 0) connect(fitData, SIGNAL(propertiesChanged()), this, SLOT(Changed()));
	return fitData;
}

void Potential::showFitData()
{
	if (fitData == 0)
	{
		QString F = getFileName();
		QFile Fi(F = F.left(F.lastIndexOf('.')) + ".fdat");
		if(Fi.exists()) fitData = MW->getFitData(F, molecule);
		if (fitData == 0)
		{
			fitData = State->getFitData();
			if (fitData == 0)
			{
				fitData = MW->CreateFitData();
				fitData->setName(getName() + "FitData");
				fitData->setFileName(F);
				State->addFitData(fitData);
			}
		}
	}
	if (!fitData->isDataAvailable()) 
	{
		fitData->updateData();
		if (!fitData->isDataAvailable())
		{
			fitData = 0;
			return;
		}
		Changed();
	}
	connect(fitData, SIGNAL(propertiesChanged()), this, SLOT(Changed()));
	if (!fitData->isVisible()) fitData->show();
	fitData->setWindowTitle("Fit data of " + getTypeString() + " " + getName());
}

void Potential::updateFitData()
{
	if (fitData == 0)
	{
		fitData = new FitData(State, MW, molecule);
		QString F = getFileName();
		QFile Fi(F = F.left(F.lastIndexOf('.')) + ".fdat");
		if (Fi.exists()) fitData->readData(F);
		connect(fitData, SIGNAL(propertiesChanged()), this, SLOT(Changed()));
		fitData->setWindowTitle("Fit data of " + getTypeString() + " " + getName());
		MW->showMDIChild(fitData);
	}
	fitData->updateData();
}

bool Potential::setElState(ElState *S)
{
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
								"The electronic state cannot be changed because there is currently a potential fit running!");
		return false;
	}
	State = S;
	Worker->setOmega(S->getOmega(), S->getS());
	return true;
}

void Potential::calcTermEnergies(TermTable *&TT, int NumWFPoints, int Mv, int MJ, bool show)
{
	//printf("CalcTermEnergies\n");
	int Nv, NJ, Nc, NI;
	double ****E;
	int *IsoZ = 0, *CompZ = 0;
	if (molecule == 0)
	{
		printf("Potential::calcTermEnergies: Error: no molecule available!\n");
		return;
	} 
	if (MW == 0)	
	{
		printf("Potential::calcTermEnergies: Error: no MainWindow available!\n");
		return;
	}
	if (WaveFuncs != 0)
	{
		//int C, L = State->getLambda(), n, m;
		//ElState *St;
		int *Z, n;
		E = WaveFuncs->getEnergy();
		WaveFuncs->getIso(NI, Z);
		IsoZ = new int[NI];
		for (n=0; n < NI; n++) IsoZ[n] = Z[n];
		CompZ = new int[Nc = 1];
		CompZ[0] = 0;
		Nv = WaveFuncs->getNv();
		NJ = WaveFuncs->getNJ();
		/*if (L==0)
		{
			for (n=m=0; n < Nc; n++)
			{
				WaveFuncs->getStateComp(n, St, C);
				if (St->getLambda() == 0 || C%2 == 0) 
				{
					CompZ[m] = m;
					m++;
				}
				else
				{
					Destroy(E[n], NI, Nv);
					E[n] = 0;
				}
			}
			if (m<n)
			{
				for (n=0; E[n] != 0; n++) ;
				for (m=n+1; m < Nc; m++) if (E[m] != 0) E[n++] = E[m];
				Nc = n;
			}
		}
		else for (n=m=0; n < Nc; n++)
		{
			WaveFuncs->getStateComp(n, St, C);
			if (St->getLambda() == 0 ? m%2 == 1 : (m*C)%2 == 1) m++; 
			CompZ[n] = m++;
		}*/
	}
	else
	{
		Nc = ((State != 0 ? State->getLambda() : 1) > 0 ? 2 : 1);
		if (Worker->getNSpinRGamma() > 0) Nc *= (int(2.0 * State->getS()) + 1);
		E = new double***[Nc];
		CompZ = new int[Nc];
		TermTable *cTT = (State != 0 ? State->getTermTable(false) : 0);
		int n;
		if (cTT != 0 ? cTT->getNumComp() == Nc : false)
		{
			int *cCompZ = cTT->getCompZ();
			for (n=0; n < Nc; n++) CompZ[n] = cCompZ[n];
		}
		else for (n=0; n < Nc; n++) CompZ[n] = n;
		if (Mv < 0 || MJ < 0)
		{
			InputDialog Dialog(70, 350); 
			Dialog.setLabel("Maximum v", "Maximum J");
			if (Dialog.exec() == QDialog::Rejected) return;
			Dialog.getData(Mv, MJ);
		}
        getEV(E, Nc, NI = Worker->getNIso(), Nv = Mv + 1, NJ = MJ + 1, NumWFPoints);
	}
	if (E == 0)
	{
		delete[] E;
		return;
	}
	if (TT == 0)
	{
		TT = MW->CreateTermTable();
		if (State != 0) State->addTermTable(TT);
		TT->setName("Term" + getName());
		TT->setFileName("Term" + getFileName());
		//printf("Vor setSource\n");
		TT->setSource("Calculated from " + getSource());
	}
	//printf("Nach getEV\n");
	TT->setData(E, Nc, NI, Nv - 1, NJ - 1, CompZ);
	if (IsoZ != 0) TT->setIsoZ(IsoZ);
	if (show) TT->show();
	//printf("Ende CalcTermEnergies, Nv=%d\n", Nv);
}

void Potential::getFCF(Potential *uPot, int Iso, int Js, int Jss, int &Nvs, int &Nvss, double ***&FCF, int NumWFPoints)
{
	//printf("Potential::getFCF, Iso=%d, Js=%d, Jss=%d\n", Iso, Js, Jss);
	int *Pu, *Px, *rix, *rax, *riu, *rau, UWI = -1, n;
	double *Eu, *Ex, **WFu, **WFx, *Fu, *Fx, *Nu, *Nx, Min = getMinimumE(), uMin = uPot->getMinimumE();
	CoupledSineWaveFunc *UWF = uPot->getCoupledSineWaveFunc();
    getWaveFunc(Iso, Jss, Nvss, Ex, WFx, rix, rax, Px, Fx, Nx, NumWFPoints);
	if (UWF != 0 ? UWF->getNJ() > Js : false)
	{
		int N, *I;
		UWF->getIso(N, I);
		for (n=0; n<N; n++) if (I[n] == Iso) UWI = n;
	}
	if (UWI >= 0)
	{
		int N, *v, c;
		ElState *Stu = uPot->getElState(), *Stc = 0;
		UWF->getv(UWI, Js, N, v);
		Pu = new int[Nvs = v[N-1] + 1];
		riu = new int[Nvs];
		rau = new int[Nvs];
		Eu = new double[Nvs];
		WFu = new double*[Nvs];
		Fu = new double[Nvs];
		Nu = new double[Nvs];
		for (n=0, N = UWF->getNChannels(); Stc != Stu && n<N; n++) UWF->getStateComp(n, Stc, c);
		c=n-1;
		for (n=0; n < Nvs; n++)
		{
			riu[n] = Pu[n] = 0;
            rau[n] = NumWFPoints - 1;
			Fu[n] = Nu[n] = 1.0;
            WFu[n] = new double[NumWFPoints];
		}
        if (!UWF->getWaveFuncs(UWI, Js, c, rmin, rmax, NumWFPoints, WFu, Eu)) for (n=0; n < Nvs; n++)
		{
            for (N=0; N < NumWFPoints; N++) WFu[n][N] = 0.0;
			Eu[n] = 0.0;
		}
	}
    else uPot->getWaveFunc(Iso, Js, Nvs, Eu, WFu, riu, rau, Pu, Fu, Nu, NumWFPoints);
	//printf("getFCF: Nvss=%d\n", Nvss);
	//printf("Vor Create FCF\n");
	getFCF(Nvss, Nvs, Ex, Eu, WFx, WFu, rix, riu, rax, rau, Px, Pu, Fx, Fu, Nx, Nu, FCF, Min, uMin);
	//printf("Name=%s, uName=%s\n", getName().ascii(), uPot->getName().ascii());
}

void Potential::getFCF(int &Nvss, int &Nvs, double* Ex, double* Eu, double** WFx, double** WFu, 
					   int* rix, int* riu, int* rax, int* rau, int* Px, int* Pu, double* Fx, 
                       double* Fu, double* Nx, double* Nu, double ***&FCF, double Min, double uMin)
{
	//printf("Ex[1]=%f, Eu[1]=%f\n", Ex[1], Eu[1]);
	int Mvs = Nvs, Mvss = Nvss, i, vs, vss, P1, P2, ra;
	double GS, S, F1, F2;
	//printf("Min=%f, uMin=%f\n", Min, uMin);
	FCF = Create(Nvs, Nvss, 2);
	//printf("Nach Create FCF\n");
	for (vss = 0; vss < Nvss; vss++) for (vs = 0; vs < Nvs; vs++)
	{
		if (Pu[vs] < Px[vss])
		{
			P1 = Pu[vs];
			P2 = Px[vss];
			F1 = Fu[vs];
			F2 = F1 * Fx[vss];
		}
		else
		{
			P1 = Px[vss];
			P2 = Pu[vs];
			F1 = Fx[vss];
			F2 = F1 * Fu[vs];
		}
		if ((ra = (rax[vss] < rau[vs] ? rax[vss] : rau[vs])) < P1) P1 = ra - 1;
		if (ra < P2) P2 = ra - 1;
		for (i = (rix[vss] > riu[vs] ? rix[vss] : riu[vs]), GS = 0.0; i <= P1; i++) 
			GS += WFu[vs][i] * WFx[vss][i];
		for (S=0.0; i <= P2; i++) S += WFu[vs][i] * WFx[vss][i];
		GS += S * F1;
		for (S=0.0; i < ra; i++) S += WFu[vs][i] * WFx[vss][i];
		GS += S * F2;
		//if (vs == 250 && vss == 60) 
			//printf("ri=%d, ra=%d, P1=%d, P2=%d, GS=%f, Nu=%f, Nx=%f\n", ri, ra, P1);
		FCF[vs][vss][0] = GS * GS * Nu[vs] * Nx[vss];
		FCF[vs][vss][1] = Eu[vs] - Ex[vss];
		if (Min < 0.0) FCF[vs][vss][1] += Min;
		if (uMin < 0.0) FCF[vs][vss][1] -= uMin;
	}
	//printf("Vor delete E\n");
	delete[] Eu;
	delete[] Ex;
	delete[] rix;
	delete[] rax;
	delete[] riu;
	delete[] rau;
	//printf("Vor delete P\n");
	delete[] Pu;
	delete[] Px;
	//printf("Vor delete F\n");
	delete[] Fu;
	delete[] Fx;
	//printf("Vor delete N\n");
	delete[] Nu;
	delete[] Nx;
	//printf("Vor Destroy WF\n");
	Destroy(WFu, Mvs);
	Destroy(WFx, Mvss);
	//printf("Ende GetFCF, Nvs=%d\n", Nvs);
}

double *Potential::calcSigFuncs(int /*numPoints*/, double /*rMin*/, double /*rMax*/)
{
	return 0;
}

void Potential::getWaveFunc(int I, int J, int &Nv, double *&EL, double **&WF, int *&ri, int *&ra, 
                            int *&CP, double *&F, double *&SQ, int NumWFPoints)
{
	//printf("Potential::getWaveFunc, Nv=%d\n", Nv);
    double h = (rmax - rmin) / (NumWFPoints - 1), *RPot = new double[NumWFPoints], *Y = new double[NumWFPoints];
    double *U = getPoints(rmin, rmax, NumWFPoints), *IsoF = Worker->getIsoF_ForReading(), MF = IsoF[I] * h*h;
	int n, nv = Nv;
    WF = Create(Nv, NumWFPoints);
	//printf("Nach Create WF\n");
	//aPot->GetMin(R, M);
	F = new double[Nv];
	SQ = new double[Nv];
	EL = new double[Nv];
	CP = new int[Nv];
	ri = new int[Nv];
	ra = new int[Nv];
    Worker->CRotPot(J, I, U, RPot, rmin, h, NumWFPoints);
    NumerovCooley(Y, MF, Nv, EL, WF, ri, ra, CP, F, SQ, NumWFPoints, RPot);
	delete[] U;
	delete[] Y;
	delete[] RPot;
	for (n = Nv; n < nv; n++) delete[] WF[n];
	//printf("Ende von getWaveFunc, CP[0]=%d, Nv=%d\n", CP[0], Nv);
}

void Potential::exportWaveFunction(int NumWFPoints)
{
	if (WaveFuncs == 0) return;
	int n, NI, *I, NC = WaveFuncs->getNChannels(), v, J, i, Nv = WaveFuncs->getNv();
    double ***WF = Create(NC, Nv, NumWFPoints), E[Nv], r, h = (rmax - rmin) / ((NumWFPoints - 1) * a0_Angstrom);
	double UF = 0.7270458;
	IsoTab *Iso = molecule->getIso();
	QDialog *D = new QDialog(this);
	QGridLayout *L = new QGridLayout(D);
	QComboBox *IB = new QComboBox(D);
	QLineEdit *vE = new QLineEdit("0", D), *JE = new QLineEdit("0", D);
	QPushButton *OK = new QPushButton("OK", this), *Cancel = new QPushButton("Cancel", this);
	QString FileName;
	D->setWindowTitle("MolSpektAnalysis");
	L->addWidget(new QLabel("Please select the level of which the wave function shall be exported:", D), 0, 0, 1, 2);
	L->setRowMinimumHeight(1, 20);
	L->addWidget(new QLabel("Iso:", D), 2, 0);
	L->addWidget(IB, 2, 1);
	L->addWidget(new QLabel("v:", D), 3, 0);
	L->addWidget(vE, 3, 1);
	L->addWidget(new QLabel("J:", D), 4, 0);
	L->addWidget(JE, 4, 1);
	L->setRowMinimumHeight(5, 20);
	L->addWidget(OK, 6, 0);
	L->addWidget(Cancel, 6, 1);
	WaveFuncs->getIso(NI, I);
	for (n=0; n < NI; n++) IB->addItem(Iso->getIsoName(I[n]));
	IB->setEditable(false);
	connect(OK, SIGNAL(clicked()), D, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), D, SLOT(reject()));
	if (D->exec() == QDialog::Accepted)
	{
		i = IB->currentIndex();
		v = vE->text().toInt();
		J = JE->text().toInt();
		for (n=0; n < Nv; n++) E[n] = 0.0;
        if (v >= 0 && v < Nv && J >= 0 ? WaveFuncs->getWaveFuncs(i, J, rmin, rmax, NumWFPoints, WF, E) : false)
		{
			if (E[v] != 0.0)
			{
				if (!(FileName = QFileDialog::getSaveFileName(this, "Select file name", MW->getDir(), 
																	"Text files (*.dat)")).isEmpty()) 
				{
					QFile F(FileName);
					F.open(QIODevice::WriteOnly);
					QTextStream S(&F);
					S << "Isotopologue " << Iso->getIsoName(I[i]) << ", v'=" << QString::number(v) << ", J'=" 
					  << QString::number(J) << "\n"; 
                    for (n=0, r = rmin / a0_Angstrom; n < NumWFPoints; n++, r+=h)
					{
						S << QString::number(r, 'f', 12);
						for (i=0; i < NC; i++) S << "\t" << QString::number(WF[i][v][n] * UF, 'e', 12);
						S << "\n";
					}
				}
			}
			else QMessageBox::information(this, "MolSpektAnalysis", "Error: the level with v'=" + QString::number(v) +
											+ "could not be found!");
		}
		else QMessageBox::information(this, "MolSpektAnalysis", "Error: the level with v'=" + QString::number(v)
										+ " and J'=" + QString::number(J) + " does not exist!");
	}
	Destroy(WF, NC, Nv);
	delete Iso;
	delete D;
}

void Potential::getExchangeInt(double& A, double& alpha, double& gamma)
{
	AnaPot *aPot = Worker->getAnaPotForReading();
	if (aPot != 0) 
	{
		aPot->getExchangeCoeff(A, alpha, gamma);
		Worker->unlock();
	}
	else A = alpha = gamma = 0.0;
}

void Potential::setExchangeInt(double A, double alpha, double gamma)
{
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
			"The spin rotation coefficients cannot be changed because there is currently a fit of the potential running!");
		return;
	}
	AnaPot *aPot = Worker->getAnaPotForWriting();
	if (aPot != 0) aPot->setExchangeCoeff(A, alpha, gamma);
}

void PotWorker::getAdCorrForWriting(int& N, double*& a, int& T, int& P, double& Iso1, double& Iso2, double &Rm, double &b)
{
	if (Fit->isRunning())
	{
		N = 0;
		return;
	}
	N = NAdCorr;
	a = adCorr;
	T = TAdCorr;
	P = PAdCorr;
	Iso1 = RIso1AdCorr;
	Iso2 = RIso2AdCorr;
	Rm = AdCorr_Rm;
	b = AdCorr_b;
}

void PotWorker::getAdCorrForReading(int& N, double*& a, int& T, int& P, double& Iso1, double& Iso2, double &Rm, double &b, 
									bool **rAdCorrFree, bool alreadyLocked)
{
	int n;
	if (!alreadyLocked)
	{
		Lock->lock();
		a = new double[NAdCorr];
		for (n=0; n < NAdCorr; n++) a[n] = adCorr[n];
	}
	else a = adCorr;
	N = NAdCorr;
	T = TAdCorr;
	P = PAdCorr;
	Iso1 = RIso1AdCorr;
	Iso2 = RIso2AdCorr;
	Rm = AdCorr_Rm;
	b = AdCorr_b;
	if (!alreadyLocked)
	{
		if (rAdCorrFree != 0)
		{
			*rAdCorrFree = new bool[N];
			for (n=0; n<N; n++) (*rAdCorrFree)[n] = adCorrFree[n];
		}
		Lock->unlock();
	}
	else if (rAdCorrFree != 0) *rAdCorrFree = adCorrFree; 
}

void Potential::getAdCorrForWriting(int& NAdCorr, double*& adCorr, int& TAdCorr, int& PAdCorr, double& RIso1, double& RIso2, 
									double& Rm, double& b)
{
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
						"The adiabatic correction cannot be modified because there currentliy is a fit of the potential running!");
		NAdCorr = 0;
	}
	else Worker->getAdCorrForWriting(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b);
}

bool PotWorker::setAdCorr(int N, double* a, int T, int P, double RIso1, double RIso2, double Rm, double b, bool *AdCorrFree)
{
	if (Fit->isRunning()) return false;
	int n;
	if (adCorr != 0) 
	{
		delete[] adCorr;
		delete[] adCorrFree;
	}
	if (AdCorrWF != 0)
	{
		delete[] AdCorrWF;
		AdCorrWF = 0;
	}
	NAdCorr = N;
	adCorr = a;
	TAdCorr = T;
	PAdCorr = P;
	RIso1AdCorr = RIso1;
	RIso2AdCorr = RIso2;
	if (AdCorrFree == 0)
	{
		adCorrFree = new bool[N];
		for (n=0; n<N; n++) adCorrFree[n] = true;
	}
	else adCorrFree = AdCorrFree;
	for (n = NumFreeAdCorr = 0; n<N; n++) if (AdCorrFree[n]) NumFreeAdCorr++;
	if (Type == analyticalPotential)
	{
		double nTm, nDe, *C;
		int N;
		AnaPot *APot = dynamic_cast<AnaPot*>(this);
		APot->getPotCoeff(AdCorr_b, AdCorr_Rm, nTm, nDe, N, C);
		if (b != 1e-30 && (fabs(AdCorr_b - b) > 1e-13 || fabs(AdCorr_Rm - Rm) > 1e-13)) APot->fitAdCorrToRmb();
	}
	else
	{
		AdCorr_b = b;
		AdCorr_Rm = Rm;
	}
	return true;
}

bool Potential::setAdCorr(int NAdCorr, double* adCorr, int TAdCorr, int PAdCorr, double RIso1, double RIso2, double Rm, double b)
{
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
						"The adiabatic correction cannot be modified because there currentliy is a fit of the potential running!");
		return false;
	}
	return Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b);
}

void Potential::getLRCoeffForReading(int& N, int*& pLRC, double*& RLRC)
{
	Worker->getLRCoeffForReading(N, pLRC, RLRC);
}

void PotWorker::getLRCoeffForReading(int& N, int*& pLRC, double*& LRC, double *Ra, bool alreadyLocked)
{
	if (!alreadyLocked) Lock->lock();
	int n, *PLRC;
	double *RLRC, R;
	getLRCoeff(R, N, PLRC, RLRC);
	if (alreadyLocked)
	{
		pLRC = PLRC;
		LRC = RLRC;
	}
	else
	{
		pLRC = new int[N];
		LRC = new double[N];
		for (n=0; n<N; n++) 
		{
			pLRC[n] = PLRC[n];
			LRC[n] = RLRC[n];
		}
		Lock->unlock();
		if (Type == TangToenniesPotential) delete[] PLRC;
	}
	if (Ra != 0) *Ra = R; 
}

void Potential::getLRCoeffForWriting(int& N, int*& pLRC, double*& RLRC, bool **LRCFree)
{
	N=0;
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
			"The long range coefficients cannot be changed because there is currently a fit of the potential running!");
		return;
	}
	Worker->getLRCoeffForWriting(N, pLRC, RLRC, LRCFree);
	
}

void PotWorker::getLRCoeffForWriting(int& N, int*& pLRC, double*& LRC, bool** LRCFree, double *rRa)
{
	if (Fit->isRunning())
	{
		N = 0;
		return;
	}
	double R;
	getLRCoeff(R, N, pLRC, LRC);
	if (LRCFree != 0) *LRCFree = getLRCFree();
	if (rRa != 0) *rRa = R;
}

PotWorker* PotWorker::scalePotential(double, double)
{
	printf("Error: Scaling is not implemented for Potentials of type %d\n", Type);
	return 0;
}

void Potential::getFastWaveFunc(int I, int J, int C, double E, double M, double h, int NP, double *WF, 
								double &Ri, double &Ra, double &ISq, double &F, int &i, int &c, int &a)
{
	//printf("Potential::getFastWaveFunc\n");
	double *U = new double[NP], *P = getPoints(Ri, Ra, NP), PM = M * 0.0833333333333, *A = new double[NP];
	if (Worker->isAdCorrA()) 
	{
		IsoTab *IsoT = Worker->getIsoT();
		Worker->CalcAdCorrPot(P, A, Ri, h, NP, IsoT->mIso1[I], IsoT->mIso2[I]);
		delete[] P;
	}
	else A=P;
	Worker->CRotPot(J, I, A, U, Ri, h, NP, C);
	delete[] A;
	Worker->fastWaveFunc(U, E, PM, h, NP, WF, Ri, Ra, ISq, F, i, c, a);
	delete[] U;
}

bool Potential::getFastWaveFuncs(int C, int I, int J, int Nv, double M, double h, int NP, double** WF, 
								 double ri, double ra, double* WFsq, double* F, int* i, int* cp, int* a)
{
	TermTable *TT = State->getTermTable();
	if (C >= TT->getNumComp() || I >= TT->getNumIso() || J > TT->getMaxJ() 
		|| Nv > TT->getMaxv() + 1) return false;
	double ****EData = TT->getData(), *U = getPoints(ri, ra, NP, C), *A, *R, PM = M * 0.0833333333333, Ri, Ra;
	int fc, c, l, v;
	float S;
	IsoTab *IsoT = Worker->getIsoT();
	if ((S = State->getS()) > 0.0 && (l = State->getLambda()) > 0)
	{
		fc = C/2;
		c = C - 2 * fc;
	}
	else if (S > 0.0)
	{
		fc = C;
		c=0;
	}
	else
	{
		c = C;
		fc = 0;
	}
	if (Worker->isAdCorrA())
	{
		Worker->CalcAdCorrPot(U, A = new double[NP], ri, h, NP, IsoT->mIso1[I], IsoT->mIso2[I]);
		delete[] U;
	}
	else A=U;
	Worker->CRotPot(J, I, A, R = new double[NP], ri, h, NP, c, fc);
	delete[] A;
	for (v=0; v < Nv; v++) 
		Worker->fastWaveFunc(R, EData[C][I][v][J], PM, h, NP, WF[v], Ri = ri, Ra, WFsq[v], F[v], i[v], cp[v], a[v]);
	delete[] R;
	return true;
}

void PotWorker::fastWaveFunc(double *U, double E, double PM, double h, int NP, double *WF, double &Ri, 
						 	 double &Ra, double &ISq, double &F, int &i, int &c, int &a)
{
	double Sq, Y0, Y1, Y2, eins = 1.0, ED, M = 12.0 * PM;
	int n;
	getMinMaxInt(U, E, M, NP, i, a);
	if (a >= NP - 1)
	{
		for (a = NP - 1; (a>i? U[a-1] > U[a] || U[a] < E : false); a--) ;
		if (a==i) 
		{
			printf("E=%f is too high!\n", E);
			c=i;
			return;
		}
	}
	//printf("i=%d, a=%d\n", i, a);
	Ra = Ri + double(a) * h;
	Ri += double(i) * h;
	Y1 = 1e-5;
	Y0 = Y1 * exp(-sqrt(M*(U[a]-E)));
	c=a;
	WF[c] = Y0 / (eins - PM * (U[c] - E));
	Sq = WF[c] * WF[c];
	c--;
	WF[c] = Y1 / (eins - PM * (ED = U[c] - E));
	Sq += WF[c] * WF[c];
	for (c--; c>i; c--)
	{
		Y2 = M * ED * WF[c+1] + 2.0 * Y1 - Y0;
		WF[c] = Y2 / (eins - PM * (ED = U[c] - E));
		if (WF[c] <= WF[c+1]) break;
		Sq += WF[c] * WF[c];
		Y0 = Y1;
		Y1 = Y2;
	}
	F = eins / WF[c];
	Y1 = 1e-5;
	Y0 = Y1 * exp(-sqrt(M*(U[i]-E)));
	WF[n=i] = Y0 / (eins - PM * (U[i] - E));
	ISq = WF[n] * WF[n];
	n++;
	WF[n] = Y1 / (eins - PM * (ED = U[n] - E));
	ISq += WF[n] * WF[n];
	for (n++; n<=c; n++)
	{
		//WF[i-1] = Y[i-1] / (eins - PM * (ED = U[i-1] - E));
		//Y[i] = M * ED * WF[i-1] + 2.0 * Y[i-1] - Y[i-2];
		
		Y2 = M * ED * WF[n-1] + 2.0 * Y1 - Y0;
		WF[n] = Y2 / (eins - PM * (ED = U[n] - E));
		ISq += WF[n] * WF[n];
		//if (n-i==1241) printf("WF[%d]=%g, ISq=%g\n", n, WF[n], ISq);
		Y0 = Y1;
		Y1 = Y2;
		//if (WF[n-2] < WF[n-1] && WF[n-1] >= WF[n]) printf("UMax at R=%f\n", Ri + double(n-1-i) * h);
	}
	F *= WF[c];
	ISq += F*F * Sq;
	//printf("WF[%d]=%g, ISq=%g\n", a, WF[a], ISq);
	//printf("Ende getFastWaveFunc\n");
}

void Potential::getFastFCF(int I, Potential *uPot, int Js, int Comp, double uE, int N, double *lE, int *lJss, 
                           double *FCF, int NumWFPoints, double **uWF, double *TStr, int NWFC)
{
	//printf("Potential::getFastFCF, N=%d, uE=%f\n", N, uE);
	if (NWFC < 1) NWFC = 1;
    int NP = NumWFPoints, i, j, n, ui, ua, li, la, s, uc, c, UWI = -1, vs, wf;
	double Ri = rmin, Ra = rmax, uISq, ISq, Sq, WF, uF, F, fcf[NWFC], fcf2[NWFC], tFCF[NWFC];
	double h = (rmax - rmin) / double(NP - 1), uRi = Ri, uRa = Ra;
	double *U = new double[NP], eins = 1.0, *IsoF = Worker->getIsoF_ForReading(), M = IsoF[I] * h*h, Y0, Y1, Y2 = 0.0;
	double PM = M * 0.0833333333333, ED, *P = getPoints(Ri, Ra, NP);
	//printf("Vor lE\n");
	if ((ED = getMinimumE()) < 0.0) for (i=0; i<N; i++) lE[i] += ED;
	
	/*int Nv = 60, *ri, *ra, *CP;
	double **tWF, *EL, *tF, *SQ;
	printf("h=%g\n", h);
	getWaveFunc(I, lJss[0], Nv, EL, tWF, ri, ra, CP, tF, SQ);
	delete[] ri;
	delete[] ra;
	delete[] CP;
	Destroy(tWF, Nv);
	delete[] tF;
	delete[] SQ;
	for (n=0, i=-1; n < Nv; n++)
	{
		printf("v=%d, Ec=%g, ", n, EL[n]);
		for (i++; (i<N ? lJss[i] != lJss[0] : false); i++) ;
		printf("Ei=%g, D=%g\n", lE[i], EL[n]-lE[i]);
	}
	delete[] EL;
	uPot->getWaveFunc(I, Js, Nv, EL, tWF, ri, ra, CP, tF, SQ);
	delete[] ri;
	delete[] ra;
	delete[] CP;
	Destroy(tWF, Nv);
	delete[] tF;
	delete[] SQ;
	for (n=0; n < Nv; n++) printf("Eu=%g, E[%d]=%g, D=%g\n", uE, n, EL[n], uE-EL[n]);
	delete[] EL;*/
	
	//printf("Vor getFastWF\n");
	if (uWF == 0)
	{
		CoupledSineWaveFunc *UWF = uPot->getCoupledSineWaveFunc();
		if (UWF != 0 ? UWF->getNJ() > Js : false)
		{
			int N, *Iso;
			UWF->getIso(N, Iso);
			for (n=0; n<N; n++) if (Iso[n] == I) UWI = n;
		}
		if (UWI >= 0 ? (vs = UWF->getv(UWI, Js, uE)) >= 0 : false)
		{
			int N, c;
			double EB;
			ElState *Stu = uPot->getElState(), *Stc = 0;
			uWF = Create(1, NP);
			for (n=0, N = UWF->getNChannels(); Stc != Stu && n<N; n++) UWF->getStateComp(n, Stc, c);
			c=n-1;
            UWF->getWaveFunc(UWI, Js, vs, c, rmin, rmax, NumWFPoints, uWF[0], EB, N);
		}
	}
	if (uWF != 0)
	{
		uRi = rmin;
		uRa = rmax;
		uISq = 1.0;
		uF = 1.0;
		ui = uc = 0;
        ua = NumWFPoints - 1;
	}
	else 
	{
		uWF = Create(1, NP);
		uPot->getFastWaveFunc(I, Js, Comp, uE, M, h, NP, uWF[0], uRi, uRa, uISq, uF, ui, uc, ua);
	}
	//printf("Nach getFastWF\n");
	for (i=0; i<N; i++) FCF[i] = -1.0;
	for (i=0; i<N; i++) if (FCF[i] == -1.0)
	{
		Worker->CRotPot(lJss[i], I, P, U, Ri, h, NP);
		for (j=i; j<N; j++) if (lJss[j] == lJss[i])
		{
			//printf("j=%d\n", j);
			getMinMaxInt(U, lE[j], M, NP, li, la);
			//printf("Nach gerMinMaxInt\n");
			if (la < NP - 1 ? U[la] < lE[j] : true)
			{
				for (la = NP - 1; (la > li? U[la - 1] > U[la] || U[la] < lE[j] : false); la--) ;
				if (la == li) 
				{
					printf("E[%d]=%f is too high!\n", j, lE[j]);
					FCF[j] = 0.0;
					continue;
				}
			}
			Y1 = 1e-5;
			Y0 = Y1 * exp(-sqrt(M*(U[la] - lE[j])));
			WF = Y0 / (eins - PM * (U[c = la] - lE[j]));
			Sq = WF * WF;
			if (la < ua) for (wf = 0; wf < NWFC; wf++) 
			{
				fcf[wf] = WF * uWF[wf][c];
				if (la <= uc) fcf[wf] /= uF;
			}
			else for (wf = 0; wf < NWFC; wf++) fcf[wf] = 0.0;
			WF = Y1 / (eins - PM * (ED = U[--c] - lE[j]));
			Sq += WF * WF;
			if (c < ua) for (wf = 0; wf < NWFC; wf++)
			{
				if (c > uc) fcf[wf] += WF * uWF[wf][c];
				else fcf[wf] += WF * uWF[wf][c] / uF;
			}
			for (c--; c > ua; c--)
			{
				Y2 = M * ED * WF + 2.0 * Y1 - Y0;
				WF = Y2 / (eins - PM * (ED = U[c] - lE[j]));
				if (Y2 <= Y1) break;
				Sq += WF * WF;
				Y0 = Y1;
				Y1 = Y2;
			}
			if (Y2 > Y1)
			{
				for (c--; c > uc; c--)
				{
					Y2 = M * ED * WF + 2.0 * Y1 - Y0;
					WF = Y2 / (eins - PM * (ED = U[c] - lE[j]));
					if (Y2 <= Y1) break;
					Sq += WF * WF;
					for (wf = 0; wf < NWFC; wf++) fcf[wf] += WF * uWF[wf][c];
					Y0 = Y1;
					Y1 = Y2;
				}
				for (wf = 0; wf < NWFC; wf++) fcf[wf] *= uF;
				if (Y2 > Y1)
				{
					for (c--; c > ui; c--)
					{
						Y2 = M * ED * WF + 2.0 * Y1 - Y0;
						WF = Y2 / (eins - PM * (ED = U[c] - lE[j]));
						if (Y2 <= Y1) break;
						Sq += WF * WF;
						for (wf = 0; wf < NWFC; wf++) fcf[wf] += WF * uWF[wf][c];
						Y0 = Y1;
						Y1 = Y2;
					}
					if (Y2 > Y1)
					{
						for (c--; c > li; c--)
						{
							Y2 = M * ED * WF + 2.0 * Y1 - Y0;
							WF = Y2 / (eins - PM * (ED = U[c] - lE[j]));
							if (Y2 <= Y1) break;
							Sq += WF * WF;
							Y0 = Y1;
							Y1 = Y2;
						}
					}
				}
			}
			F = eins / WF;
			//printf("ui=%d, li=%d, ua=%d, la=%d, NP=%d\n", ui, li, ua, la, NP);
			Y0 = (Y1 = 1e-5) * exp(-sqrt(M * (U[li] - lE[j])));
			WF = Y0 / (eins - PM * (U[li] - lE[j]));
			ISq = WF * WF;
			if (li >= ui) for (wf = 0; wf < NWFC; wf++)
			{
				tFCF[wf] = WF * uWF[wf][li];
				if (li > uc) tFCF[wf] *= uF;
			}
			else for (wf = 0; wf < NWFC; wf++) tFCF[wf] = 0.0;
			WF = Y1 / (eins - PM * (ED = U[n = li + 1] - lE[j]));
			if (n >= ui) for (wf = 0; wf < NWFC; wf++)
			{
				if (n > uc) tFCF[wf] += WF * uWF[wf][n] * uF;
				else tFCF[wf] += WF * uWF[wf][n];
			}
			ISq += WF * WF;
			for (n++, s = (ui <= c ? ui : c+1); n<s; n++)
			{
				Y2 = M * ED * WF + 2.0 * Y1 - Y0;
				WF = Y2 / (eins - PM * (ED = U[n] - lE[j]));
				ISq += WF * WF;
				Y0 = Y1;
				Y1 = Y2;
			}
			for (s = (uc < c ? uc : c); n<=s; n++)
			{
				Y2 = M * ED * WF + 2.0 * Y1 - Y0;
				WF = Y2 / (eins - PM * (ED = U[n] - lE[j]));
				ISq += WF * WF;
				tFCF[wf] += WF * uWF[wf][n];
				Y0 = Y1;
				Y1 = Y2;
			}
			for (wf = 0; wf < NWFC; wf++) fcf2[wf] = 0.0;
			for (s = (ua <= c ? ua : c+1); n<s; n++)
			{
				Y2 = M * ED * WF + 2.0 * Y1 - Y0;
				WF = Y2 / (eins - PM * (ED = U[n] - lE[j]));
				ISq += WF * WF;
				for (wf = 0; wf < NWFC; wf++) fcf2[wf] += WF * uWF[wf][n];
				Y0 = Y1;
				Y1 = Y2;
			}
			while (n <= c)
			{
				Y2 = M * ED * WF + 2.0 * Y1 - Y0;
				WF = Y2 / (eins - PM * (ED = U[n++] - lE[j]));
				ISq += WF * WF;
				Y0 = Y1;
				Y1 = Y2;
			}
			F *= WF;
			ISq += F*F * Sq;
			for (wf = 0, FCF[j] = 0.0; wf < NWFC; wf++)
			{
				tFCF[wf] += fcf2[wf] * uF + fcf[wf] * F;
				//printf("n=%d, la=%d\n", n, la);
				//printf("FCF=%g, ISq=%g, uISq=%g\n", FCF[j], ISq, uISq);
				tFCF[wf] *= tFCF[wf];
				tFCF[wf] /= (ISq * uISq);
				FCF[j] += tFCF[wf] * TStr[wf];
			}
				//printf("lJss[%d]=%d, lE[%d]=%f, FCF[%d]=%f\n", j, lJss[j], j, lE[j], j, FCF[j]);
		}
	}
	delete[] P;
	Destroy(uWF, NWFC);
	delete[] U;
	//printf("Ende getFastFCF\n");
}

void Potential::exportAsymptoticLevels(QString FileName, int numv, int maxJ, int NumWFPoints)
{
	if (State == 0 || molecule == 0)
	{
	    QMessageBox::information(this, "MolSpektAnalysis", "Error: the data cannot be exported!");
		return;
	}
    double *U = getPoints(rmin, rmax, NumWFPoints);
	if (U==0)
	{
	    QMessageBox::information(this, "MolSpektAnalysis", "Error: the data cannot be exported!");
		return;
	}
	int J, nv, I, v, Nv, NJ, NI, NIso = Worker->getNIso(), JStart[NIso], JStep[NIso], Rm, RM, LCP;
    double h = (rmax - rmin) / (NumWFPoints - 1), Max = getAsymptote(), *RPot = new double[NumWFPoints], MF = 0.0;
    double hsq = h*h, DE, *Y = new double[NumWFPoints], *IsoF = Worker->getIsoF_ForReading();
	for (I=0; I < NIso; I++) 
	{
		if (IsoF[I] > MF) MF = IsoF[I];
		JStart[I] = State->getJStart(I, 0);
		JStep[I] = molecule->getJStep(I);
	}
	MF *= hsq;
    getMinMaxInt(U, Max, MF, NumWFPoints, Rm, RM);
	double *LWF = new double[RM - Rm], LF, LSQ;
	Numerov(MF, Max, DE, Nv, RM - Rm, U + Rm, LWF, LCP, LF, LSQ, Y);
	delete[] LWF;
	if (DE > 0.0) Nv--;
    double F[Nv], SQ[Nv], **WF = Create(Nv, NumWFPoints);
	double ***EL = new double**[NI = NIso];
	int Maxv = Nv - 1, MaxJ = maxJ, CP[Nv], ri[Nv], ra[Nv];
	for (I=0, NJ = 0, Nv = 0; I < NI; I++) 
	{
		EL[I] = new double*[MaxJ+1];
		MF = IsoF[I] * hsq;
		for (J = JStart[I]; J <= MaxJ; J += JStep[I])
		{
			EL[I][J] = new double[Maxv+1];
            Worker->CRotPot(J, I, U, RPot, rmin, h, NumWFPoints);
            NumerovCooley(Y, MF, nv = Maxv + 1, EL[I][J], WF, ri, ra, CP, F, SQ, NumWFPoints, RPot);
			printf("Iso=%d, J=%d, Maxv=%d\n", I, J, nv-1);
			if (nv <= Maxv) EL[I][J][nv] = 0.0;
			if (nv > Nv) Nv = nv;
			else if (nv == 0)
			{
				delete[] EL[I][J];
				EL[I][J] = NULL;
				break;
			}
		}
		if (J > NJ) NJ = J;
	}
	if (NJ > MaxJ + 1) NJ = MaxJ + 1;
	Destroy(WF, Maxv + 1);
	delete[] U;
	delete[] RPot;
	delete[] Y;
	QFile File(FileName);
	File.open(QIODevice::WriteOnly);
	QTextStream S(&File);
	IsoTab *IsoT = Worker->getIsoT();
	for (I=0; I<NI; I++)
	{
		S << "Isotopologue " + IsoT->getIsoName(I) + ":\n";
		for (J=JStart[I]; (J < NJ ? EL[I][J] != NULL : false); J += JStep[I])
		{
			S << "J=" + QString::number(J) + ":\n";
			for (v = Nv - numv; (v < Nv ? EL[I][J][v] != 0.0 : false); v++) 
				S << "v = " + QString::number(v) + " E = " << QString::number(EL[I][J][v] - Max, 'f', 4) + "\n";
			delete[] EL[I][J];
		}
		delete[] EL[I];
		S << "\n";
	}
	delete[] EL;
}

void Potential::getEV(double ****&Ev, int &Nc, int &NI, int &Nv, int &NJ, int NumWFPoints)
{
// 	printf("Beginn getEV\n");
	bool AdCorr = Worker->isAdCorrA();
    double *U = getPoints(rmin, rmax, NumWFPoints);
	if (U==0)
	{
	    Ev = 0;
		NI = Nv = NJ = Nc = 0;
		return;
	}
    double h = (rmax - rmin) / (NumWFPoints - 1), *RPot = new double[NumWFPoints], **WF = Create(Nv, NumWFPoints);
	double *IsoF = Worker->getIsoF_ForReading();
    double F[Nv], SQ[Nv], hsq = h*h, MF, *AdCorrPot = (AdCorr ? new double[NumWFPoints] : U), *Y = new double[NumWFPoints];
	int nc = (Worker->isLambdaDoublingA() && Nc > 1 ? 2 : 1);
	double ****EL = new double***[Nc = ((State != 0 ? State->getS() : 0.0) > 0.0 && Nc >= 2 * nc ? 2 * nc : nc)], R, M;
	int J, nv, I, v, Maxv = Nv - 1, MaxJ = NJ - 1, CP[Nv], ri[Nv], ra[Nv], c, FC, ef;
	for (c=0, NI = Worker->getNIso(); c < Nc; c++) EL[c] = new double**[NI];
	IsoTab *IsoT = Worker->getIsoT();
	printf("Nv=%d, Maxv=%d\n", Nv, Maxv);
	//NI=1;
	getMinimum(R, M);
	/*QFile DDatei("DebugTerm.dat");
	DDatei.open(QIODevice::WriteOnly);
	QTextStream S1(&DDatei);*/
	//printf("NIso=%d, MaxJ=%d, Maxv=%d\n", NI, MaxJ, Maxv);
		
	for (c=0; c < Nc; c++) 
	{
		FC = (Nc > nc ? c / nc + 1 : 0);
		ef = (Nc > nc ? c - nc * (FC - 1) : c);
		for (I=0, NJ = 0, Nv = 0; I < NI; I++) 
		{
			//S1 << "Iso=" << I << "\n";
			EL[c][I] = new double*[MaxJ+1];
			MF = IsoF[I] * hsq;
            if (AdCorr) Worker->CalcAdCorrPot(U, AdCorrPot, rmin, h, NumWFPoints, IsoT->mIso1[I], IsoT->mIso2[I]);
			for (J=0; J <= MaxJ; J++)
			{
				//S1 << "J=" << J << "\n";
				EL[c][I][J] = new double[Maxv+1];
                Worker->CRotPot(J, I, AdCorrPot, RPot, rmin, h, NumWFPoints, ef, FC);
				//printf("Vor Numerov\n");
                NumerovCooley(Y, MF, nv = Maxv + 1, EL[c][I][J], WF, ri, ra, CP, F, SQ, NumWFPoints, RPot);
				printf("Iso=%d, J=%d, Maxv=%d\n", I, J, nv-1);
				if (nv <= Maxv) EL[c][I][J][nv] = 0.0;
				if (nv > Nv) Nv = nv;
				else if (nv == 0)
				{
					delete[] EL[c][I][J];
					EL[c][I][J] = 0;
					break;
				}
				//for (v=0; v<nv; v++) 
					//S1 << "v=" << v << "\tT=" << QString::number(EL[I][J][v] - M, 'g', 10) << "\n";
			}
			if (J > NJ) NJ = J;
		}
	}
	//printf("Nach Schleife, Nv=%d, Maxv=%d\n", Nv, Maxv);
	//DDatei.close();
	
	Destroy(WF, Maxv + 1);
	//printf("Nach Destroy WF\n");
	delete[] U;
	delete[] RPot;
	delete[] Y;
	if (AdCorr) delete[] AdCorrPot;
	//printf("Nach delete U\n");
	Ev = Create(Nc, NI, Nv, NJ);
	for (c=0; c < Nc; c++) 
	{
		for (I=0; I<NI; I++)
		{
			for (J=0; (J < NJ ? EL[c][I][J] != NULL : false); J++)
			{
				for (v=0; (v < Nv ? EL[c][I][J][v] != 0.0 : false); v++) 
					Ev[c][I][v][J] = (M>0 ? EL[c][I][J][v] : EL[c][I][J][v] - M);
				for (; v < Nv; v++) Ev[c][I][v][J] = 0.0;
				//printf("Vor delete EL[%d][%d]\n", I, J);
				delete[] EL[c][I][J];
			}
			for (; J < NJ; J++) for (v=0; v < Nv; v++) Ev[c][I][v][J] = 0.0;
			//printf("Vor delete EL[%d]\n", I);
			delete[] EL[c][I];
		}
		delete[] EL[c];
	}
	//printf("Vor delete EL\n");
	delete[] EL;
	//printf("Ende von getEv, Nv=%d, Min=%f, E[9][1][60]=%f\n", Nv, M, Ev[9][1][60]);
}

void Potential::setFitData(int maxv, int maxJ, int Iso, bool OnlyTerm)
{
	if (Fit->isRunning()) return;
	int *mJ, **mv, Mv, nES, *nEL, nGS, *nGP, *nGL, **PS, MaxJ, Maxv, NIso = Worker->getNIso(); 
	TermEnergy **EL;
	TableLine **GL; 
	nES = nGS = 0;
	int i, e, g, n, m, j, N, DL[MaxLineTables];
	TableLine *Lines;
	LineTable *LineT = 0;
	FitData *fDB = (State != 0 ? State->getFitData() : 0);
	if (fDB != 0) fitData = fDB;
	if (fitData == 0) 
	{
		showFitData();
		if (fitData == 0) return;
	}
	fitData->getData(Lines, N, -1, -2, -2, 0, Iso);
	if (N == 0) return;
	for (i=0; i < MaxLineTables; i++) DL[i] = 0;
 	maxJ = MaxJ = fitData->getJMax();
	maxv = Maxv = fitData->getvMax();
	for (n=m=0; n<N; n++) if (Lines[n].LTab != LineT || (n>0 ? Lines[n-1].isTE != Lines[n].isTE : true))
	{
		LineT = Lines[n].LTab;
		m++;
	}
	bool *ES = new bool[m];
 	for (n=0, m=-1, LineT = 0; n<N; n++)
	{
		if (Lines[n].err == 0.0)
		{
			QMessageBox::information(this, "MolSpektAnalysis", 
									 "Error: The data contains lines/levels with uncertainty equal to 0!");
			nES = nGS = 0;
			return;
		}
		if (Lines[n].LTab != LineT || (n>0 ? Lines[n-1].isTE != Lines[n].isTE : true))
		{
			LineT = Lines[n].LTab;
			if (Lines[n].isTE)
			{
				nES++;
				ES[++m] = true;
			}
			else 
			{
				if (!OnlyTerm) nGS++;
				ES[++m] = false;
			}
		}
		if (Lines[n].vss <= maxv && Lines[n].Jss <= maxJ) DL[m]++;
	}
	//int mav[maxJ + 1];
	//for (n=0; n <= maxJ; n++) mav[n] = maxv; 
	mJ = new int[NIso];
	mv = new int*[NIso];
	if (nES > 0)
	{
		nEL = new int[nES];
		EL = new TermEnergy*[nES];
	}
	else
	{
		nEL = 0;
		EL = 0;
	}
	if (nGS > 0)
	{
		nGP = new int[nGS];
		nGL = new int[nGS];
		PS = new int*[nGS];
		GL = new TableLine*[nGS];
	}
	else
	{
		nGP = 0;
		nGL = 0;
		PS = 0;
		GL = 0;
	}
	int lPN, PStart[MaxProgressions];
	for (n=0; n < NIso; n++) mJ[n] = -1;
	for (n=e=g=j=0; n<=m; n++)
	{
		if (ES[n]) 
		{
			EL[e] = new TermEnergy[DL[n]];
			for (i=0; i < DL[n]; j++) if (Lines[j].vss <= maxv && Lines[j].Jss <= maxJ)
			{
				EL[e][i].DevR = Lines[j].DevR;
				EL[e][i].E = Lines[j].WN;
				EL[e][i].Iso = Lines[j].Iso;
				EL[e][i].J = Lines[j].Jss;
				EL[e][i].dev = Lines[j].dev;
				EL[e][i].ef = (Lines[j].Jss != Lines[j].Js ? true : false);
				EL[e][i].err = Lines[j].err;
				EL[e][i].v = Lines[j].vss;
				EL[e][i].FC = Lines[j].FC;
				EL[e][i].nDig = Lines[j].nDig;
				EL[e][i].Row = Lines[j].Row;
				EL[e][i].WeightFactor = Lines[j].WeightFact;
				if (EL[e][i].J > mJ[EL[e][i].Iso]) mJ[EL[e][i].Iso] = EL[e][i].J;
				i++;
			}
			nEL[e++] = DL[n];
		}
		else if (!OnlyTerm)
		{
			GL[g] = new TableLine[DL[n]];
			for (i=0, lPN = -1, nGP[g] = 0; i < DL[n]; j++) if (Lines[j].vss <= maxv && Lines[j].Jss <= maxJ)
			{
				GL[g][i] = Lines[j];
				if (lPN != GL[g][i].PN)
				{
					lPN = GL[g][i].PN;
					PStart[nGP[g]++] = i;
				}
				if (GL[g][i].Jss > mJ[GL[g][i].Iso]) mJ[GL[g][i].Iso] = GL[g][i].Jss;
				i++;
			}
			PS[g] = new int[nGP[g]];
			for (i=0; i < nGP[g]; i++) PS[g][i] = PStart[i];
			nGL[g++] = DL[n];
		}
	}
	//printf("Ende Schleife\n");
	for (n=0; n < NIso; n++) 
	{
		//printf("NIso=%d, mJ[%d]=%d\n", NIso, n, mJ[n]);
		if (mJ[n] > -1) 
		{
			mv[n] = new int[mJ[n] + 1];
			for (i=0; i <= mJ[n]; i++) mv[n][i] = -1;
		}
		else mv[n] = 0;
	}
	//printf("Vor mv, nES=%d, nGS=%d\n", nES, nGS);
	for (n=Mv=0; n < nES; n++) for (i=0; i < nEL[n]; i++) if (EL[n][i].v > mv[EL[n][i].Iso][EL[n][i].J])
	{
		mv[EL[n][i].Iso][EL[n][i].J] = EL[n][i].v;
		if (EL[n][i].v > Mv) Mv = EL[n][i].v;
	}
	for (n=0; n < nGS; n++) for (i=0; i < nGL[n]; i++) 
			if (GL[n][i].vss > mv[GL[n][i].Iso][GL[n][i].Jss])
	{
		mv[GL[n][i].Iso][GL[n][i].Jss] = GL[n][i].vss;
		if (GL[n][i].vss > Mv) Mv = GL[n][i].vss;
	}
	//printf("Vor Delete\n");
	delete[] Lines;
	//printf("Ende getFitData\n");
	Worker->setFitData(mJ, mv, Mv, nES, nEL, EL, nGS, nGP, nGL, PS, GL, ES, MaxJ, Maxv);
}

void PotWorker::setFitData(int* nmJ, int** nmv, int nMv, int nnEs, int* nNEL, TermEnergy** nEL, int nnGS, int* nnGP, int* nnGL, 
						   int** nPS, TableLine** NGL, bool *nES, int nMaxJ, int nMaxv)
{
	if (Fit->isRunning()) return;
	if (nEs > 0 || nGS > 0) destroyFitData();
	mJ = nmJ;
	mv = nmv;
	Mv = nMv;
	nEs = nnEs;
	NEL = nNEL;
	EL = nEL;
	nGS = nnGS;
	nGP = nnGP;
	nGL = nnGL;
	PS = nPS;
	GL = NGL;
	MaxJ = nMaxJ;
	Maxv = nMaxv;
	if (ES != 0) delete[] ES;
	ES = nES;
}

void Potential::getCoupledPotentials(int &nCoupled, Potential** &potentials, int* &components)
{
	nCoupled = NCoupled;
	potentials = CoupledPot;
	components = CoupledComp;
}

CoupledSineWaveFunc* Potential::getCoupledSineWaveFunc()
{
	return WaveFuncs;
}

void Potential::setCouplingData(CoupledSineWaveFunc* WF, int nCoupled, Potential** coupledPot, int* coupledComp)
{
	if (NCoupled  > 0)
	{
		delete[] CoupledPot;
		delete[] CoupledComp;
		if (WaveFuncs != 0 ? WaveFuncs->Delete() : false) delete WaveFuncs;
	}
	WaveFuncs = WF;
	WF->Assign();
	NCoupled = nCoupled;
	CoupledPot = coupledPot;
	CoupledComp = coupledComp;
}

void Potential::setCouplingData(CoupledSineWaveFunc* WF, int nCoupled, ElState** CoupledStates, int* coupledComp)
{
	int n, m, *CC;
	Potential **Pot = new Potential*[nCoupled];
	if (WaveFuncs != 0 ? WaveFuncs->Delete() : false) delete WaveFuncs;
	WaveFuncs = WF;
	WF->Assign();
	if (NCoupled != nCoupled - 1)
	{
		if (NCoupled > 0) delete[] CoupledComp;
		CoupledComp = new int[nCoupled];
	}
	for (n=0, m=-1; n < nCoupled; n++)
	{
		if (n>0 ? CoupledStates[n] != CoupledStates[n-1] : false) m++;
		Pot[n] = (m>=0 ? CoupledPot[m] : this);
		CoupledComp[n] = coupledComp[n]; 
	}
	NCoupled = nCoupled;
	delete[] CoupledPot;
	CoupledPot = Pot;
	for (n=0; n < NCoupled; n++)
	{
		Pot = new Potential*[NCoupled];
		CC = new int[NCoupled];
		for (m=0; m < nCoupled; m++)
		{
			Pot[m] = CoupledPot[m];
			CC[m] = coupledComp[m];
		}
		if (CoupledPot[n] != this) CoupledPot[n]->setCouplingData(WF, NCoupled, Pot, CC);
		if (!CoupledPot[n]->isAssigned()) CoupledStates[n]->addPotential(CoupledPot[n]);
	}
}

void PotWorker::CreateEK(double ****&E, double *****&K, int *mJ, int **mv, int N)
{
	//printf("Potential::CreateEK\n");
	if (NIso == 0) return;
	int I, J, v, F, NFC = int(2.0 * el_S) + 1;
	E = new double***[NFC];
	K = new double****[NFC];
	for (F=0; F < NFC; F++)
	{
		E[F] = new double **[NIso];
		K[F] = new double ***[NIso];
		for (I=0; I < NIso; I++) if (mJ[I] > -1)
		{
			E[F][I] = new double *[mJ[I] + 1];
			K[F][I] = new double **[mJ[I] + 1];
			for (J=0; J <= mJ[I]; J++) if (mv[I][J] > -1)
			{
				E[F][I][J] = new double[mv[I][J] + 1];
				K[F][I][J] = new double*[mv[I][J] + 1];
				for (v=0; v <= mv[I][J]; v++) K[F][I][J][v] = new double[N];
			}
		}
	}
	//printf("Ende CreateEK\n");
}

void PotWorker::calcE(double ****E, int *mJ, int **mv, int Mv, int NumWFPoints, double ****MinErr, bool ****SFQSU, double SFQSRad)
{
	int I, J, nv, FC;
	int NFC = int(2.0 * el_S) + 1;
	bool AdCorr = (NAdCorr > 0);
    double h = (rmax - rmin) / (NumWFPoints - 1), *RPot = new double[NumWFPoints], *Y = new double[NumWFPoints];
    double *U = getPoints(rmin, rmax, NumWFPoints), MF, *AdCorrPot = (AdCorr ? new double[NumWFPoints] : 0);
    double **WF = Create(Mv + 1, NumWFPoints);
	//printf("Nach Create WF\n");
	//aPot->GetMin(R, M);
	double *F = new double[Mv+1];
	double *SQ = new double[Mv+1];
	int *CP = new int[Mv+1], SFQSP = (SFQSU != 0 ? int((SFQSRad - rmin) / h) : 0);
	int *ri = new int[Mv+1];
	int *ra = new int[Mv+1], v;
	//printf("Mv=%d, mv[6][35]=%d\n", Mv, mv[6][35]);
	for (FC = 0; FC < NFC; FC++) for (I=0; I < NIso; I++) 
	{
		MF = IsoF[I] * h*h;
        if (AdCorr) CalcAdCorrPot(U, AdCorrPot, rmin, h, NumWFPoints, IsoT->mIso1[I], IsoT->mIso2[I]);
		else AdCorrPot = U;
		for (J=0; J <= mJ[I]; J++) if (mv[I][J] > -1)
		{
			//if (mv[I][J] >= 60) *DebugStream << "Iso=" << QString::number(I) << " J=" << QString::number(J) << '\n'; 
			
            CRotPot(J, I, AdCorrPot, RPot, rmin, h, NumWFPoints, 0, (NFC > 1 ? FC + 1 : 0));
            NumerovCooley(Y, MF, nv = mv[I][J] + 1, E[FC][I][J], WF, ri, ra, CP, F, SQ, NumWFPoints, RPot,
						  (MinErr != 0 ? MinErr[FC][I][J] : 0));
			if (nv < mv[I][J] + 1)
			{
				printf("For J=%d and Iso%d v=%d is not existing, highest v is %d\n",
					   J, I, mv[I][J], nv - 1);
				for (v = nv; v <= mv[I][J]; v++) E[FC][I][J][v] = 0.0;
			}
			if (SFQSU != 0) 
			{
				for (v=0; v < nv; v++) SFQSU[FC][I][J][v] = (CP[v] > SFQSP);
				while (v <= mv[I][J]) SFQSU[FC][I][J][v++] = false;
			}
		}
	}
	//for (I=0; I <= mv[6][35]; I++) printf("E[6][35][%d]=%f\n", I, E[6][35][I]);
	//printf("Vor delete\n");
	delete[] F;
	delete[] U;
	Destroy(WF, Mv + 1);
	delete[] SQ;
	delete[] CP;
	delete[] ri;
	delete[] ra;
 	if (AdCorr) delete[] AdCorrPot;
	delete[] RPot;
	delete[] Y;
	//printf("Ende\n");
}

void PotWorker::CalcAdCorrPot(double* InPot, double* OutPot, double Rmin, double Step, int NPoints, double MIso1, double MIso2)
{
	Lock->lock();
	calcAdCorrPot(InPot, OutPot, Rmin, Step, NPoints, MIso1, MIso2, NAdCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, 
				  AdCorr_Rm, AdCorr_b, adCorr);
	Lock->unlock();
}

double Potential::getD0(int NumWFPoints)
{
	int I = Worker->getRefIso(), nv = 1;
    double h = (rmax - rmin) / (NumWFPoints - 1);
    double *U = getPoints(rmin, rmax, NumWFPoints), *IsoF = Worker->getIsoF_ForReading(), MF = IsoF[I] * h*h;
    double **WF = Create(1, NumWFPoints), *Y = new double[NumWFPoints];
	printf("Nach Create WF\n");
	//aPot->GetMin(R, M);
	double F, SQ, E;
	int CP, ri, ra;
    NumerovCooley(Y, MF, nv, &E, WF, &ri, &ra, &CP, &F, &SQ, NumWFPoints, U);
	delete[] U;
	delete[] Y;
	Destroy(WF, 1);
	return Worker->getUinf() - E;
}

int PotWorker::getRefIso()
{
	if (IsoT != 0) return IsoT->refIso;
	return 0;
}

void PotWorker::CreateFitEQS(int &N, double **&EQS, double *&EO, double *&DE, double *&Sig, double *&Err,
								   int **&QN, int NP, int nEs, int *NEL, TermEnergy **EL, int nGS, 
		   						   int *nGP, int *nGL, int **PS, TableLine **GL, int AdRows)
{
	//printf("Potential::CreateSplineFitEQS\n");
	int i, n, m, l, d, NFC = int(2.0 * el_S) + 1;
	//double F;
	for (N=n=0; n < nEs; n++) N += NEL[n];
	//printf("1. N=%d\n", N);
	for (n=0; n < nGS; n++) if (nGL[n] > 0)
	{
		N += nGL[n];
		/*for (p=0; p < nGP[n] - 1; p++)
		{
			l = PS[n][p+1] - PS[n][p];
			N += l * (l+1) / 2 - l;
		}
		l = nGL[n] - PS[n][p];
		N += l * (l+1) / 2 - l;*/
		//printf("n=%d, N=%d\n", n, N);
	}
	//printf("Nach erster Schleife, N=%d\n", N);
	EQS = Create(N + AdRows + NumWantedValues, NP);
	EO = new double[N];
	DE = new double[N + AdRows + NumWantedValues];
	QN = CreateInt(N, 6);
	Sig = new double[N + AdRows + NumWantedValues];
	Err = new double[N];
	for (n=m=0; n < nEs; n++) for (l=0; l < NEL[n]; l++)
	{
		EO[m] = EL[n][l].E;
		Sig[m] = 1.0 / EL[n][l].err;
		Err[m] = EL[n][l].err;
		QN[m][0] = EL[n][l].Iso;
		QN[m][1] = EL[n][l].J;
		QN[m][2] = EL[n][l].v;
		QN[m][5] = (EL[n][l].FC < NFC ? EL[n][l].FC : 0);
		QN[m++][3] = -1;
	}
	//printf("Vor zweiter\n");
	for (n=0; n < nGS; n++) for (l=0; l < nGP[n]; l++)
	{
		d = (l+1 < nGP[n] ? PS[n][l+1] : nGL[n]) - PS[n][l];
		//F = double(d) / (d*(d-1)/2);
		for (i = PS[n][l]; i < PS[n][l] + d; i++) 
			//for (j=i+1; j < PS[n][l] + d; j++)
		{
			EO[m] = GL[n][i].WN;// - GL[n][j].WN;
			//Err[m] = sqrt(GL[n][i].err * GL[n][i].err + GL[n][j].err * GL[n][j].err);
			//Sig[m] = F / Err[m];
			Sig[m] = 1.0 / (Err[m] = GL[n][i].err);
			QN[m][0] = GL[n][i].Iso;
			QN[m][1] = GL[n][i].Jss;//j
			QN[m][2] = GL[n][i].vss;//j
			QN[m][3] = l;
			QN[m][4] = d;
			//QN[m][3] = GL[n][i].Jss;
			//QN[m][4] = GL[n][i].vss;
			QN[m++][5] = (GL[n][i].FC < NFC ? GL[n][i].FC : 0);
		}
	}
	//printf("Ende CreateSplineFitEQS, m=%d\n", m);
}

void PotWorker::getSFQSU(bool****& rSFQSU, int*& rmJ, int**& rmv)
{
	rSFQSU = SFQSU;
	rmJ = mJ;
	rmv = mv;
}

double PotWorker::calcFQS(double ****E, int nES, int *NEL, TermEnergy **EL, int nGS, int *nGP, int *nGL,
                          int **PS, TableLine **GL, int NumWFPoints, bool calcDiag, int MJ, int Mv, double *****K,
						  int *rv, double *RSig, double TEF, double hCi_cP, bool ****SFQSU)
{
	//printf("calcFQS\n");
	double D, FQS = 0.0, SS, OT, Sig, CE;
	int n, m, l, M, NK, k, N, r=0, s=0, d, i, j, NFC = int(2.0 * el_S) + 1, FC;
	double *KS, KM = 0.0, ***W1 = 0, ***W2 = 0, F;
	SFQS = 0.0;
	BLLock->lock();
	if (!calcDiagFuncs) calcDiag = false;
	else if (calcDiag) 
	{
		W1 = Create(NIso, Mv + 1, MJ + 1);
		W2 = Create(NIso, Mv + 1, MJ + 1);
		for (n=0; n < NIso; n++) for (m=0; m <= Mv; m++) for (l=0; l <= MJ; l++)
					W1[n][m][l] = W2[n][m][l] = 0.0;
	}
	if (rv != 0) 
	{
		KS = new double[NK = *rv];
		for (k=0; k < NK; k++) KS[k] = 0.0;
	}
	else
	{
		KS = 0;
		NK = 0;
	}
	//printf("Vor erster Schleife\n");
	for (n=N=0; n < nES; n++) N += NEL[n];
	for (n=0; n < nGS; n++) N += nGL[n];
	if (RWErr != 0 && (RSig == 0 || Ndev != N)) delete[] RWErr;
	if (RSig != 0 && (Ndev != N || RWErr == 0)) RWErr = new double[N];
	else if (RSig == 0) RWErr = 0;
	for (n=0; n < nES; n++) for (l=0; l < NEL[n]; l++) 
	{
		FC = (EL[n][l].FC >= 0 && EL[n][l].FC < NFC ? EL[n][l].FC : 0);
		EL[n][l].DevR = D = (EL[n][l].dev = EL[n][l].E - (CE = E[FC][EL[n][l].Iso][EL[n][l].J][EL[n][l].v])) / EL[n][l].err;
		if (CE != 0.0) 
		{
			if (RSig != 0)
			{
				RWErr[r] = sqrt(EL[n][l].err * EL[n][l].err + 0.3 * EL[n][l].dev * EL[n][l].dev);
				RSig[s] = 1.0 / RWErr[r++];
				FQS += EL[n][l].dev * EL[n][l].dev * RSig[s] * RSig[s] * TEF * EL[n][l].WeightFactor;
				s++;
			}
			else FQS += D*D * TEF * EL[n][l].WeightFactor;
			if (SFQSU != 0 ? SFQSU[FC][EL[n][l].Iso][EL[n][l].J][EL[n][l].v] : false) SFQS += D*D * EL[n][l].WeightFactor;
			if (D > 1E3) printf("Error: D=%g, Iso=%d, J=%d, v=%d, E=%f\n", D, EL[n][l].Iso, EL[n][l].J, EL[n][l].v, EL[n][l].E);
		}
		else 
		{
			EL[n][l].DevR = 0.0;
			if (RSig != 0)
			{
				RWErr[r++] = EL[n][l].err;
				RSig[s++] = 1.0 / EL[n][l].err;
			}
		}
		if (calcDiag)
		{
			W1[EL[n][l].Iso][EL[n][l].v][EL[n][l].J] += 1.0 / EL[n][l].err;
			W2[EL[n][l].Iso][EL[n][l].v][EL[n][l].J] += D;
		}
		for (k=0; k < NK; k++) 
			KS[k] += fabs(D * K[(EL[n][l].FC >= 0 && EL[n][l].FC < NFC ? EL[n][l].FC : 0)][EL[n][l].Iso][EL[n][l].J][EL[n][l].v][k]);
		//printf("v=%d, J=%d, Iso=%d, D=%f\n", EL[n][l].v, EL[n][l].J, EL[n][l].Iso, D);
	}
	//printf("Vor 2.\n");
	for (n=0; n < nGS; n++) for (l=0; l < nGP[n]; l++)
	{
		SS = OT = 0.0;
		M = (l+1 < nGP[n] ? PS[n][l+1] : nGL[n]);
		for (m = PS[n][l]; m < M; m++)
		{
			if ((CE = E[(GL[n][m].FC >= 0 && GL[n][m].FC < NFC ? GL[n][m].FC : 0)][GL[n][m].Iso][GL[n][m].Jss][GL[n][m].vss]) == 0.0) continue;
			SS += Sig = 1.0 / (GL[n][m].err * GL[n][m].err);
			OT += Sig * (CE + GL[n][m].WN);
			//if (l==4) printf("m=%d, Iso=%d, Jss=%d, vss=%d, WN=%f, err=%f, E=%f\n", m, GL[n][m].Iso,
				//				GL[n][m].Jss, GL[n][m].vss, GL[n][m].WN, GL[n][m].err,
	 				//			E[GL[n][m].Iso][GL[n][m].Jss][GL[n][m].vss]);
		}
		OT /= SS;
		for (m = PS[n][l]; m < M; m++)
		{
			FC = (GL[n][m].FC >= 0 && GL[n][m].FC < NFC ? GL[n][m].FC : 0);
			D = (GL[n][m].dev = GL[n][m].WN - OT + (CE = E[FC][GL[n][m].Iso][GL[n][m].Jss][GL[n][m].vss])) / GL[n][m].err;
			if (RSig != 0) RWErr[r] = sqrt(GL[n][m].err * GL[n][m].err + 0.03 * GL[n][m].dev * GL[n][m].dev);
			if (CE == 0.0) 
			{
				GL[n][m].DevR = 0.0;
				if (RSig != 0) RWErr[r++] = GL[n][m].err;
				continue;
			}
			GL[n][m].DevR = D;
			//if (fabs(D) > 10.0)
				//printf("v=%d, J=%d, Iso=%d, D=%f\n", GL[n][m].vss, GL[n][m].Jss, GL[n][m].Iso, D);
			FQS += (RSig != 0 ? GL[n][m].dev * GL[n][l].dev / (RWErr[r] * RWErr[r]) : D*D) * GL[n][m].WeightFact;
			if (SFQSU != 0 ? SFQSU[FC][GL[n][m].Iso][GL[n][m].Jss][GL[n][m].vss] : false) SFQS += D * D * GL[n][m].WeightFact; 
			r++;
			if (calcDiag)
			{
				W1[GL[n][m].Iso][GL[n][m].vss][GL[n][m].Jss] += 1.0 / GL[n][m].err;
				W2[GL[n][m].Iso][GL[n][m].vss][GL[n][m].Jss] += D;
			}
			for (k=0; k < NK; k++) 
				KS[k] += fabs(D * K[(GL[n][m].FC >= 0 && GL[n][m].FC < NFC ? GL[n][m].FC : 0)][GL[n][m].Iso][GL[n][m].Jss][GL[n][m].vss][k]);
		}
		if (RSig != 0)
		{
			d = M - PS[n][l];
			F = double(d) / (d*(d-1)/2);
			for (i = PS[n][l]; i<M; i++) for (j=i+1; j<M; j++) 
				RSig[s++] = F / sqrt(RWErr[r-M+i] * RWErr[r-M+i] + RWErr[r-M+j] * RWErr[r-M+j]);
		}	
		//printf("nGS=%d, l=%d, nGP[%d]=%d, FQS=%f, SS=%f, OT=%f\n", nGS, l, n, nGP[n], FQS, SS, OT);
		//printf("PS[n][l]=%d, M=%d\n", PS[n][l], M);
	}
	if (calcDiag)
	{
        if (WFS == 0) WFS = new double[NumWFPoints];
        if (WWFS == 0) WWFS = new double[NumWFPoints];
        double *WF = new double[NumWFPoints], *P = getPoints(rmin, rmax, NumWFPoints);
        double *U = new double[NumWFPoints], PM;
        for (n=0; n < NumWFPoints; n++) WFS[n] = WWFS[n] = 0.0;
        double h = (rmax - rmin) / (NumWFPoints - 1), M, Ri, Ra, ISq, F, F1, F2;
		int i, c, a;
		bool RC;
		for (n=0; n < NIso; n++)
		{
			M = IsoF[n] * h*h;
			PM = M * 0.0833333333333;
			for (m=0; m < MJ; m++)
			{
				RC = false;
				for (l=0; l < Mv; l++) if (W1[n][l][m] != 0.0)
				{
					if (!RC) 
					{
                        CRotPot(m, n, P, U, rmin, h, NumWFPoints);
						RC = true;
					}
                    fastWaveFunc(U, E[0][n][m][l], PM, h, NumWFPoints, WF, Ri = rmin, Ra = rmax, ISq, F, i,
								 c, a);
					F1 = F2 = 1.0 / (ISq * ISq);
					F1 *= W1[n][l][m];
					F2 *= W2[n][l][m];
					for (k=i; k <= c; k++) 
					{
						WFS[k] += F1 * WF[k] * WF[k];
						WWFS[k] += F2 * WF[k] * WF[k];
					}
					for (F1 *= F * F, F2 *= F * F; k<a; k++) 
					{
						WFS[k] += F1 * WF[k] * WF[k];
						WWFS[k] += F2 * WF[k] * WF[k];
					}
				}
			}
		}
		delete[] WF;
		delete[] P;
		delete[] U;
		Destroy(W1, NIso, Mv + 1);
		Destroy(W2, NIso, Mv + 1);
	}
	//printf("Vor 3.\n");
	for (k=0; k < NK; k++) if (KS[k] > KM)
	{
		KM = KS[k];
		*rv = k;
	}
	if (KS != 0) delete[] KS;
	int e, g;
	if (Ndev != N)
	{
		if (Ndev > 0)
		{
			delete[] Row;
			delete[] dev;
		}
		Row = new int[Ndev = N];
		dev = new double[N];
	}
	for (n=l=e=g=0; n<N; l++)
	{
		if (ES[l]) 
		{
			for (m=0; m < NEL[e]; m++)
			{
				Row[n] = EL[e][m].Row;
				dev[n++] = EL[e][m].dev;
				//DevR[n++] = EL[e][m].DevR;
			}
			e++;
		}
		else
		{
			for (m=0; m < nGL[g]; m++)
			{
				Row[n] = GL[g][m].Row;
				dev[n++] = GL[g][m].dev;
				//DevR[n++] = GL[g][m].DevR;
			}
			g++;
		}
	}
	BLLock->unlock();
	if (ShowBadList)
	{
		QString **Data;
        showBadList(Data, m, n, NumWFPoints);
		BLLock->unlock();
	}
	if (hCi_cP > 0.0)
	{
		bool *LRCFree = getLRCFree();
		double *LRC, Ra;
		int NLRC, *pLRC;
		getLRCoeff(Ra, NLRC, pLRC, LRC);
		for (n=0; n < NLRC; n++) if (LRCFree[n] && pLRC[n] >= 12)
		{
			CE = pow(LRC[n-1] / LRC[n-2], 3) * LRC[n-3];
			D = (CE - LRC[n]) / (hCi_cP * CE);
			FQS += D*D;
		}
	}
	for (n=0; n < NumWantedValues; n++)
	{
		D = (WantedValues[n].Value - WantedValues[n].CurValue) / WantedValues[n].Precision;
		FQS += D*D;
	}
	if (LevenbergMarquardtRunning) emit badListChanged(FQS, lambda, minLSFQS);
	else emit badListChanged(FQS, -1.0, 0.0);
	return FQS;
}

bool Potential::getCalcDiagFuncs()
{
	return Worker->getCalcDiagFuncs();
}

void Potential::setCalcDiagFuncs(bool calc)
{
	if (calc && !Worker->getCalcDiagFuncs()) 
	{
		Worker->setCalcDiagFuncs(calc);
		getFQS(true);
	}
	else Worker->setCalcDiagFuncs(calc);
}

void Potential::getDiagFuncs(double *&WaveFSum, double *&rWWFS, double &RStart, double &h, int &NP, int NumWFPoints)
{
	//printf("Potential::getDiagFuncs\n");
    Worker->getDiagFuncs(WaveFSum, rWWFS, NumWFPoints);
	RStart = rmin;
    h = (rmax - rmin) / (NumWFPoints - 1);
    NP = NumWFPoints;
}

void PotWorker::getDiagFuncs(double*& rWFS, double*& rWWFS, int NumWFPoints)
{
	int n;
	if (WFS == 0)
	{
		rWFS = rWWFS = 0;
		return;
	}
    rWFS = new double[NumWFPoints];
    rWWFS = new double[NumWFPoints];
	BLLock->lock();
    for (n=0; n < NumWFPoints; n++)
	{
		rWFS[n] = WFS[n];
		rWWFS[n] = WWFS[n];
	}
	BLLock->unlock();
}

void PotWorker::showBadList(QString**& Data, int& NR, int &NC, int NumWFPoints)
{
	BLLock->lock();
	int NG, NE, s, n, m, i;
	if (dev == 0)
	{
		if(!Fit->isRunning()) 
		{
			int N = numSplinePoints + 2, **QN;
			int rv = N, F, I, J, NFC = int(2.0 * el_S) + 1;
			double ****E, *****K, **EQS, *EO, *DE, *Sig, ****MaxErr, *Err;
			estMinErr(MaxErr, NFC, mJ, mv, nEs, NEL, EL, nGS, nGL, GL);
			CreateEK(E, K, mJ, mv, N);
			CreateFitEQS(NE, EQS, EO, DE, Sig, Err, QN, N, nEs, NEL, EL, nGS, nGP, nGL, PS, GL);
            calcE(E, mJ, mv, Mv, NumWFPoints, MaxErr);
            calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints, false, 0, 0, K, &rv);
			for (F=0; F < NFC; F++)
			{
				for (I=0; I < NIso; I++)
				{
					for (J=0; J <= mJ[I]; J++) if (MaxErr[F][I][J] != 0) delete[] MaxErr[F][I][J];
					if (MaxErr[F][I] != 0) delete[] MaxErr[F][I];
				}
				delete[] MaxErr[F];
			}
			delete[] MaxErr;
			DestroyFitData(E, K, mJ, mv, NE, EQS, EO, DE, Sig, Err, QN);
		}
		else
		{
			Data = 0;
			NR = NC = 0;
			return;
		}
	}
	for (s = NG = 0; s < nGS; s++) for (n=0; n < nGL[s]; n++) 
			if (fabs(GL[s][n].DevR) > 3.0 || (GL[s][n].DevR == 0.0 && GL[s][n].dev != 0.0)) NG++;
	for (s = NE = 0; s < nEs; s++) for (n=0; n < NEL[s]; n++)
			if (fabs(EL[s][n].DevR) > 3.0 || EL[s][n].DevR == 0.0) NE++;
	TermEnergy **EP = 0, *EB = EL[0];
	TableLine **GP = 0, *GB = GL[0];
	//printf("Vor NE\n");
	if (NE > 0)
	{
		TermEnergy **EP1 = new TermEnergy*[NE];
		EP = new TermEnergy*[NE];
		for (s=m=0; s < nEs; s++) for (n=0; n < NEL[s]; n++) 
				if (fabs(EL[s][n].DevR) > 3.0 || EL[s][n].DevR == 0.0) EP1[m++] = EL[s] + n;
		//printf("NE=%d, NEL[0]=%d\n", NE, NEL[0]);
		while (EB != 0) for (n=0, m=2, EB = 0; m <= NE; n++, m+=2) 
		{
			//printf("n=%d, m=%d\n", n, m);
			if (m < NE ? ((EP1[m-1]->DevR == 0.0 && EP1[m]->DevR != 0.0) || (EP1[m-1]->DevR != 0.0 
						 && EP1[m]->DevR != 0.0 && fabs(EP1[m-1]->DevR) > fabs(EP1[m]->DevR))) : true)
				i=m-1;
			else i=m;
			//printf("i=%d\n", i);
			if ((EP1[i]->DevR == 0.0 && EP1[n]->DevR != 0.0) 
						 || (EP1[i]->DevR != 0.0 && EP1[n]->DevR != 0.0 
						 && fabs(EP1[i]->DevR) > fabs(EP1[n]->DevR)))
			{
				EB = EP1[i];
				EP1[i] = EP1[n];
				EP1[n] = EB;
			}
		}
		//printf("Nach S1\n");
		for (i=0; i < NE; i++)
		{
			EP[i] = EP1[0];
			for (n=0, m=2; EP1[n] != 0; m=2*((n=m)+1))
			{
				if (m < NE)
				{
					if (EP1[m-1] != 0 ? (EP1[m] != 0 ? (EP1[m-1]->DevR == 0.0 && EP1[m]->DevR != 0.0) 
						|| (EP1[m-1]->DevR != 0.0 && EP1[m]->DevR != 0.0 
						&& fabs(EP1[m-1]->DevR) > fabs(EP1[m]->DevR)) : true) : false) m--;
				}
				else if (m==NE) m--;
				else break;
				EP1[n] = EP1[m];
			}
			if (m >= NE) EP1[n] = 0;
		}
		//printf("Nach S2\n");
		delete[] EP1;
		if (NG == 0)
		{
			Data = CreateQString(NE, 7);
			for (n=0; n < NE; n++)
			{
				Data[n][0] = QString::number(EP[n]->Iso);
				Data[n][1] = QString::number(EP[n]->v);
				Data[n][2] = QString::number(EP[n]->J);
				Data[n][3] = QString::number(EP[n]->E, 'f', EP[n]->nDig);
				Data[n][4] = QString::number(EP[n]->err, 'f', EP[n]->nDig);
				if (EP[n]->DevR != 0.0)
				{
					Data[n][5] = QString::number(EP[n]->dev, 'f', EP[n]->nDig);
					Data[n][6] = QString::number(EP[n]->DevR, 'f', 2);
				}
				else Data[n][5] = "not found";
			}
			BadListNC = NC = 7;
		}
	}
	//printf("Vor NG\n");
	if (NG > 0)
	{
		GP = new TableLine*[NG];
		TableLine **GP1 = new TableLine*[NG];
		for (s=m=0; s < nGS; s++) for (n=0; n < nGL[s]; n++) 
				if ((GL[s][n].DevR == 0.0 && GL[s][n].dev != 0.0) || fabs(GL[s][n].DevR) > 3.0) 
				  GP1[m++] = GL[s] + n;
		while(GB != 0) for (n=0, m=2, GB = 0; m <= NG; n++, m+=2) 
		{
			//printf("s=%d\n", s);
			if (m < NG ? (GP1[m-1]->DevR == 0.0 && GP1[m]->DevR != 0.0) || (GP1[m-1]->DevR != 0.0 
						 && GP1[m]->DevR != 0.0 && fabs(GP1[m-1]->DevR) > fabs(GP1[m]->DevR)) : true)
				i=m-1;
			else i=m;
			if ((GP1[i]->DevR == 0.0 && GP1[n]->DevR != 0.0) || 
						  (GP1[i]->DevR != 0.0 && GP1[n]->DevR != 0.0
						  && fabs(GP1[i]->DevR) > fabs(GP1[n]->DevR)))
			{
				//printf("GP[%d]=%f, GP[%d]=%f, NG=%d\n", i, GP1[i]->DevR, n, GP1[n]->DevR, NG);
				GB = GP1[i];
				GP1[i] = GP1[n];
				GP1[n] = GB;
			}
		}
		for (i=0; i < NG; i++)
		{
			//printf("i=%d, NG=%d\n", i, NG);
			GP[i] = GP1[0];
			for (n=0, m=2; GP1[n] != 0; m=2*((n=m)+1))
			{
				if (m < NG)
				{
					if (GP1[m-1] != 0 ? (GP1[m] != 0 ? (GP1[m-1]->DevR == 0.0 && GP1[m]->DevR != 0.0) 
						|| (GP1[m-1]->DevR != 0.0 && GP1[m]->DevR != 0.0 
								   && fabs(GP1[m-1]->DevR) > fabs(GP1[m]->DevR)) : true) : false) m--;
				}
				else if (m==NG) m--;
				else break;
				GP1[n] = GP1[m];
			}
			if (m >= NG) GP1[n] = 0;
		}
		delete[] GP1;
		//printf("Nach Sortieren\n");
		Data = CreateQString(NG + NE, 9);
		for (s=m=n=0; n < NG + NE; n++) 
		{
			//printf("m=%d, NE=%d, s=%d, NG=%d, n=%d\n", m, NE, s, NG, n);
			if (s < NG ? (m < NE ? ((fabs(GP[s]->DevR) > fabs(EP[m]->DevR) && EP[m]->DevR != 0.0)
						 || GP[s]->DevR == 0.0) : true) : false)
			{
				Data[n][0] = QString::number(GP[s]->Iso);
				Data[n][1] = QString::number(GP[s]->vss);
				Data[n][2] = QString::number(GP[s]->Jss);
				Data[n][3] = QString::number(GP[s]->vs);
				Data[n][4] = QString::number(GP[s]->Js);
				Data[n][5] = QString::number(GP[s]->WN, 'f', GP[s]->nDig);
				Data[n][6] = QString::number(GP[s]->err, 'f', GP[s]->nDig);
				if (GP[s]->DevR != 0.0)
				{
					Data[n][7] = QString::number(GP[s]->dev, 'f', GP[s]->nDig);
					Data[n][8] = QString::number(GP[s]->DevR, 'f', 2);
				}
				else Data[n][7] = "not found";
				s++;
			}
			else
			{
				Data[n][0] = QString::number(EP[m]->Iso);
				Data[n][1] = QString::number(EP[m]->v);
				Data[n][2] = QString::number(EP[m]->J);
				Data[n][5] = QString::number(EP[m]->E, 'f', EP[m]->nDig);
				Data[n][6] = QString::number(EP[m]->err, 'f', EP[m]->nDig);
				if (EP[m]->DevR != 0.0)
				{
					Data[n][7] = QString::number(EP[m]->dev, 'f', EP[m]->nDig);
					Data[n][8] = QString::number(EP[m]->DevR, 'f', 2);
				}
				else Data[n][7] = "not found";
				m++;
			}
		}
		BadListNC = NC = 9;
	}
	if (BadListNR > 0) Destroy(BadListData, BadListNR);
	BadListData = Data;
	BadListNR = NR = NE + NG;
	//printf("Nach NG\n");
	if (EP != 0) delete[] EP;
	if (GP != 0) delete[] GP;
}

void Potential::showBadList(int NumWFPoints)
{
	int NR, NC;
	QString **Data;
	if (!Fit->isRunning() && !Worker->isFitResultAv()) setFitData();
    Worker->showBadList(Data, NR, NC, NumWFPoints);
	if (NR == 0 && !Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
		"The badlist is empty, because all lines are explained sufficiently well by the current potential.");
		Worker->unlockBadList();
		return;
	}
	if (BadList == 0) 
	{
		BadList = MW->createTableWindow();
		BadList->setType(TextTable1);
		BadList->setWindowTitle("Badlist for the potential " + getName());
		connect(Worker, SIGNAL(badListChanged()), this, SLOT(badListChanged()));
	}
	if (NC == 7) BadList->setHorizontalHeader(QStringList() << "Iso" << "v" << "J" << "Energy" 
					<< "Uncertainty" << "Obs-calc" << "Dev ratio");
	else BadList->setHorizontalHeader(QStringList() << "Iso" << "v''" << "J''" << "v'" << "J'" 
				<< "Energy" << "Uncertainty" << "Obs-calc" << "Dev ratio");
	BadList->setData(Data, NR, NC);
	BadList->setEditable(false);
	Worker->unlockBadList();
	if (!BadList->isVisible()) BadList->show();
	connect(BadList, SIGNAL(TabRowChanged(QString*,int)), this, SLOT(badListChanged(QString*,int)));
}

void PotWorker::unlockBadList()
{
	BLLock->unlock();
}

void PotWorker::getBadList(QString**& Data, int& NR, int& NC)
{
	BLLock->lock();
	Data = BadListData;
	NR = BadListNR;
	NC = BadListNC;
}

void PotWorker::getDeviations(int& N, int*& rRow, double*& rdev, double*& rRWErr)
{
	BLLock->lock();
	N = Ndev;
	rRow = Row;
	rdev = dev;
	rRWErr = RWErr;
}

void Potential::startFit()
{
	if (Worker->isSplinePot()) FitSplinePot();
	else if (Worker->isAnaPotAv()) FitAnaPot();
}

void Potential::badListChanged(double aFQS, double aLambda, double mLSFQS)
{
	if (FQSv == -1.0) 
	{
		bFQS = FQSv = aFQS;
		NFitIt = 0;
	}
	else if (bFQS > aFQS) bFQS = aFQS;
	if (LogWindow != 0)
	{
		QString Text = QString("%1 %4: FQS=%2, bFQS=%3").arg(NFitIt++).arg(aFQS, 0, 'g', 12).arg(bFQS, 0, 'g', 12);
		LogWindow->addTextRow(aLambda > 0 ? Text.arg("LM").append(", l=%1, LS=%2").arg(aLambda, 0, 'g', 1).arg(mLSFQS, 0, 'g', 3) 
										  : Text.arg("SVD"));
	}
}

void Potential::badListChanged(QString* Items, int N)
{
	if (Fit->isRunning()) return;
	TableLine L;
	L.Iso = Items[0].toInt();
	L.Jss = Items[2].toInt();
	L.vss = Items[1].toInt();
	if (N==7)
	{
		L.dev = Items[5].toDouble();
		L.DevR = Items[6].toDouble();
		L.err = Items[4].toDouble();
		L.WN = Items[3].toDouble();
	}
	else
	{
		L.dev = Items[7].toDouble();
		L.DevR = Items[8].toDouble();
		L.err = Items[6].toDouble();
		L.WN = Items[5].toDouble();
		if (!Items[3].isEmpty()) 
		{
			L.Js = Items[4].toInt();
			L.vs = Items[3].toInt();
		}
	}
	fitData->updateRow(&L);
	Worker->badListChanged(Items, N, L);
}

void PotWorker::badListChanged(QString* Items, int, TableLine &L)
{
	if (Fit->isRunning()) return;
	int n, m;
	if (nGS == 0 || Items[3].isEmpty()) 
	{
		for (n=0; n < nEs; n++) for (m=0; m < NEL[n]; m++) 
			if (EL[n][m].Iso == L.Iso && EL[n][m].v == L.vss && EL[n][m].J == L.Jss && EL[n][m].E == L.WN)
				EL[n][m].err = L.err;
	}
	else for (n=0; n < nGS; n++) for (m=0; m < nGL[n]; m++)
		if (GL[n][m].Iso == L.Iso && GL[n][m].vss == L.vss && GL[n][m].Jss == L.Jss && GL[n][m].vs == L.vs
				&& GL[n][m].Js == L.Js && GL[n][m].WN == L.WN) GL[n][m].err = L.err;
}

void PotWorker::destroyFitData()
{
	//printf("destroyFitData\n");
	int I, n;
	for (I=0; I < NIso; I++) if (mJ[I] > -1) delete[] mv[I];
	//printf("Nach 0.\n");
	delete[] mv;
	delete[] mJ;
	for (n=0; n < nEs; n++) delete[] EL[n];
	//printf("Nach 1.\n");
	delete[] EL;
	delete[] NEL;
	for (n=0; n < nGS; n++) if (nGP[n] > 0)
	{
		delete[] PS[n];
		delete[] GL[n];
	}
	//printf("Nach zweiter Schleife\n");
	delete[] PS;
	delete[] GL;
	delete[] nGL;
	delete[] nGP;
	nEs = nGS = 0;
	//printf("Ende destroyFitData\n");
}

void PotWorker::DestroyFitData(double ****E, double *****K, int *mJ, int **mv, int NL, double **EQS, double *EO,
							   double *DE, double *Sig, double *Err, int *QN[5])
{
	//printf("Potential::DestroyFitData\n");
	int I, J, v, F, NFC = int(2.0 * el_S) + 1;
	for (F=0; F < NFC; F++)
	{
		for (I=0; I < NIso; I++) if (mJ[I] > -1)
		{
			//printf("I=%d, NIso=%d, mJ[I]=%d\n", I, NIso, mJ[I]);
			for (J=0; J <= mJ[I]; J++) if (mv[I][J] > -1)
			{
				//printf("J=%d, mJ[I]=%d, mv[I][J]=%d\n", J, mJ[I], mv[I][J]);
				delete[] E[F][I][J];
				for (v=0; v <= mv[I][J]; v++) delete[] K[F][I][J][v];
				delete[] K[F][I][J];
			}
			delete[] E[F][I];
			delete[] K[F][I];
		}
		delete[] E[F];
		delete[] K[F];
	}
	//printf("Nach erster Schleife\n");
	delete[] E;
	delete[] K;
	Destroy(EQS, NL + NumWantedValues);
	delete[] EO;
	delete[] DE;
	delete[] Sig;
	delete[] Err;
	Destroy(QN, NL);
	//printf("Ende\n");
}

void PotWorker::estMinErr(double****& MinErr, int NFC, int* mJ, int** mv, int nEs, int* NEL, TermEnergy** EL, 
						  int nGS, int* nGL, TableLine** GL)
{
	int F, I, J, v, s, l;
	double ErrF = 0.1, MaxErr = 0.001, EB;
	for (F = 0, MinErr = new double***[NFC]; F < NFC; F++) for (I=0, MinErr[F] = new double**[NIso]; I < NIso; I++)
	{
		if (mJ[I] >= 0) for (J=0, MinErr[F][I] = new double*[mJ[I] + 1]; J <= mJ[I]; J++) 
			for (v=0, MinErr[F][I][J] = new double[mv[I][J] + 1]; v <= mv[I][J]; v++) MinErr[F][I][J][v] = MaxErr;
		else MinErr[F][I] = 0;
	}
	for (s=0; s < nGS; s++) for (l=0; l < nGL[s]; l++)
	{
		F = (GL[s][l].FC < NFC && GL[s][l].FC >= 0 ? GL[s][l].FC : 0);
		if ((EB = ErrF * GL[s][l].err) < MinErr[F][GL[s][l].Iso][GL[s][l].Jss][GL[s][l].vss])
			MinErr[F][GL[s][l].Iso][GL[s][l].Jss][GL[s][l].vss] = EB;
	}
	for (s=0; s < nEs; s++) for (l=0; l < NEL[s]; l++)
	{
		F = (EL[s][l].FC < NFC && EL[s][l].FC >= 0 ? EL[s][l].FC : 0);
		if ((EB = ErrF * EL[s][l].err) < MinErr[F][EL[s][l].Iso][EL[s][l].J][EL[s][l].v])
			MinErr[F][EL[s][l].Iso][EL[s][l].J][EL[s][l].v] = EB;
	}
}

double Potential::getFQS(int NumWFPoints, bool calc, double* StdDev, double *sigma, double *FQS_PAL)
{
	if (molecule == 0) return -1.0;
	if (!Fit->isRunning()) setFitData();
    double rFQS = Worker->getFQS(NumWFPoints, calc, StdDev, false, 0.0, sigma, FQS_PAL), *dev, *RWErr;
	int N, *Row;
	Worker->getDeviations(N, Row, dev, RWErr);
	fitData->setDev(dev, Row, N);
	Worker->unlockBadList();
	return rFQS;
}

void Potential::calcFQS_SFQS(bool ****SFQSU, double SFQSRad)
{
	if (molecule == 0 || Fit->isRunning()) 
	{
		printf("calcFQS_SFQS: invalid conditions for starting!\n");
		return;
	}
	if (SFQSU != 0) Worker->setSFQSU(SFQSU);
	setFitData();
	showFitResult = false;
	Fit->CalcFQS(SFQSRad);
}

double PotWorker::getFQS(int NumWFPoints, bool calc, double *StdDev, bool Thread, double SFQSRad, double *sigma, double *FQS_PAL)
{
	if (mJ == 0) return -1.0;
	int n, N, m, I, J, NFC = int(el_S * 2.0) + 1, F, MJ = 0;
	double ****E = new double***[NFC], ****MaxErr;
	if (SFQSRad > 0.0) SFQSU = new bool***[NFC];
	if (!Fit->isRunning() || Thread)
	{
		LevenbergMarquardtRunning = false;
		for (F = 0; F < NFC; F++)
		{
			E[F] = new double**[NIso];
			if (SFQSRad > 0.0) SFQSU[F] = new bool**[NIso];
			for (I=0; I < NIso; I++) 
			{
				if (mJ[I] > -1)
				{
					E[F][I] = new double*[mJ[I] + 1];
					if (SFQSRad > 0.0) SFQSU[F][I] = new bool*[mJ[I] + 1];
					for (J=0; J <= mJ[I]; J++)
					{
					    E[F][I][J] = new double[mv[I][J] + 1];
						if (SFQSRad > 0.0) SFQSU[F][I][J] = new bool[mv[I][J] + 1];
					}
				}
				else E[F][I] = 0;
			}
		}
		//printf("Vor calcE\n");
		estMinErr(MaxErr, NFC, mJ, mv, nEs, NEL, EL, nGS, nGL, GL);
        calcE(E, mJ, mv, Mv, NumWFPoints, MaxErr, (SFQSRad > 0.0 ? SFQSU : 0), SFQSRad);
		//printf("Vor calcFQS\n");
		if (calc) for (n=0; n < NIso; n++) if (mJ[n] > MJ) MJ = mJ[n];
        FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints, calc, MJ, Mv, 0, 0, 0, 1.0, 0.0, SFQSU);
		//printf("Vor Delete\n");
		if (StdDev != 0)
		{
			for (n=N=0, *StdDev = 0.0; n < nEs; n++) for (m=0; m < NEL[n]; m++, N++) *StdDev += EL[n][m].dev * EL[n][m].dev;
			for (n=0; n < nGS; n++) for (m=0; m < nGL[n]; m++, N++) *StdDev += GL[n][m].dev * GL[n][m].dev;
			*StdDev = sqrt(*StdDev / (N - (NumFreePar != 0 ? NumFreePar : 1)));
		}
		if (FQS_PAL != 0)
		{
			for (n=0, *FQS_PAL = 0.0; n < nEs; n++) for (m=0; m < NEL[n]; m++) if (EL[n][m].err < 1e-5) 
				*FQS_PAL += EL[n][m].DevR * EL[n][m].DevR;
			for (n=0; n < nGS; n++) for (m=0; m < nGL[n]; m++) if (GL[n][m].err < 1e-5) *FQS_PAL += GL[n][m].DevR * GL[n][m].DevR;
		}
		for (F=0; F < NFC; F++)
		{
			for (I=0; I < NIso; I++)
			{
				//printf("I=%d, NIso=%d\n", I, NIso);
				for (J=0; J<= mJ[I]; J++) 
				{
					if (E[F][I][J] != 0) delete[] E[F][I][J];
					if (MaxErr[F][I][J] != 0) delete[] MaxErr[F][I][J];
				}
				if (E[F][I] != 0) delete[] E[F][I];
				if (MaxErr[F][I] != 0) delete[] MaxErr[F][I];
				if (mv[I] != 0 && SFQSRad <= 0.0) 
				{
					delete[] mv[I];
					mv[I] = 0;
				}
			}
			delete[] E[F];
			delete[] MaxErr[F];
		}
		delete[] E;
		delete[] MaxErr;
		if (SFQSRad <= 0.0)
		{
			delete[] mJ;
			delete[] mv;
		}
		for (n=0; n < nEs; n++) delete[] EL[n];
		delete[] EL;
		delete[] NEL;
		for (n=0; n < nGS; n++) if (nGL[n] > 0)
		{
			delete[] PS[n];
			delete[] GL[n];
		}
		delete[] PS;
		delete[] GL;
		delete[] nGL;
		delete[] nGP;
		nEs = nGS = 0;
	}
	else if (StdDev != 0)
	{
		BLLock->lock();
		for (n=0, *StdDev = 0.0; n < Ndev; n++) *StdDev += dev[n] * dev[n];
		*StdDev = sqrt(*StdDev / (Ndev - NumFreePar));
		BLLock->unlock();
	}
	if (sigma != 0) *sigma = sqrt(FQS / (Ndev - NumFreePar));
	return FQS;
}

void PotWorker::getLRCoeff(double &R, int &numCoefficients, int *&Exponents, double *&Coefficients)
{
	R = (numSplinePoints > 0 ? points[numSplinePoints - 1].x : 0.0);
	numCoefficients = NLRC;
	Exponents = PLRC;
	Coefficients = LRC;
}

void PotWorker::setLRCoeff(double R, int numCoefficients, int* Exponents, double* Coefficients, bool* LRCFree)
{
	if (R > 0.0 && numSplinePoints > 0 && R < points[numSplinePoints - 1].x) Ro = R;
	NLRC = numCoefficients;
	if (PLRC != 0 && PLRC != Exponents) delete[] PLRC;
	PLRC = Exponents;
	if (LRC != 0 && LRC != Coefficients) delete[] LRC;
	LRC = Coefficients;
	if (LRCfree != 0 && LRCfree != LRCFree) delete[] LRCfree;
	LRCfree = LRCFree;
}

double Potential::CreateSplineFromAnalyticalPot()
{
	Potential *NPot = MW->CreatePotential();
	if (NPot == 0) return 0.0;
	double MinR, MinE, MaxE, E, R, RMin, RMax, *LRC, A, iO, iA, iExp;
	SplinePoint *R1, *R2;
	int n, N=0, N1, N2 = 0, *PLRC, NLRC;
	QString F;
	QDialog *D = new QDialog(this);
	QGridLayout *L = new QGridLayout(D);
	QLineEdit *NPE = new QLineEdit("20", D);
	QPushButton *OK = new QPushButton("OK", D), *Cancel = new QPushButton("Cancel", D);
	D->setWindowTitle("Create spline from analytical potential");
	L->addWidget(new QLabel("Number of points:", D), 0, 0);
	L->addWidget(NPE, 0, 1);
	L->setRowMinimumHeight(1, 20);
	L->addWidget(OK, 2, 0);
	L->addWidget(Cancel, 2, 1);
	connect(OK, SIGNAL(clicked()), D, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), D, SLOT(reject()));
	while (N < 2 || N > 100)
	{
		if (D->exec() == QDialog::Rejected)
		{
			delete D;
			return 0.0;
		}
		N = NPE->text().toInt();
		if (N < 2 || N > 100) QMessageBox::information(this, "MolSpektAnalysis", "You must choose a valid integer between 2 and 100");
	}
	delete D;
	AnaPot *aPot = Worker->getAnaPotForReading();
	Worker->getLRCoeffForReading(NLRC, PLRC, LRC, &RMax, true);
	aPot->getInnerWall(RMin, iO, iExp, iA);
	Worker->unlock();
	aPot->getMinimum(MinR, MinE);
	MaxE = getPoint(RMax);
	if ((E = Worker->getPoint(RMin)) > MaxE) MaxE = E;
	R1 = new SplinePoint[N];
	R2 = new SplinePoint[N];
	R = 0.30 * (RMax - RMin) / (N-1);
	R2[0].x = MinR + R;
	R2[0].y = Worker->getPoint(R2[0].x);
	for (n=0, E = (MaxE - MinE); n<N-1; 
			R2[n].y = Worker->getPoint(R2[n].x = R2[n-1].x + 1.1 * E * R / (0.1 * E + MaxE - R2[n-1].y))) 
	{
		if (R2[n].x < RMax) N2 = n + 2;
		n++;
	}
	E = (MinR - RMin) / R + 0.5;
	N1 = int(ceil(E * double(N) / (E + double(N2))));
	for (R1[n=0].x = MinR - 0.5 * (R = (MinR - RMin) / (double(N1) - 0.5)); n < N1; n++, R1[n].x = R1[n-1].x - R) 
		R1[n].y = Worker->getPoint(R1[n].x);
	for (R = fabs((R1[0].y - (E = R2[0].y)) / (R2[0].y - MinE) * ((A = R2[0].x) - MinR)); 
			fabs(R1[0].y - R2[0].y) > 0.01 * (R1[0].y - MinE); 
			R2[0].y = Worker->getPoint(R2[0].x += (R1[0].y > (E = R2[0].y) ? R : -R))) 
		 if ((E - R1[0].y) * (R2[0].y - R1[0].y) < 0.0) R/=2; 
	for (n=1, N2 = N - N1, A -= R2[0].x; n < N2; n++) R2[n].y = Worker->getPoint(R2[n].x -= A);
	R = ((R2[N2 - 1].x - RMax) * R2[0].y) / (RMax * (R2[N2 - 1].y - R2[0].y));
	for (n=1; n < N2; n++) R2[n].y = Worker->getPoint(R2[n].x *= (R2[0].y / (R2[0].y + R * (R2[n].y - R2[0].y))));
	SplinePoint *points = new SplinePoint[N];
	for (n=0; n < N1; n++) points[n] = R1[N1 - n - 1];
	for ( ; n<N; n++) points[n] = R2[n - N1];
	delete[] R1;
	delete[] R2;
	NPot->setName(getName() + "spline");
	NPot->setSource("Analytical potential from " + getSource());
	F = getFileName();
	NPot->setFileName(F.right(F.length() - F.indexOf('.') - 1) + "spline.pot");
	NPot->setWorker(new SplinePot(NPot->Fit));
	NPot->Worker->setUinf(aPot->getUinf());
	int *NPLRC = new int[NLRC];
	double *nLRC = new double[NLRC];
	for (n=0, R=0.0; n < NLRC; n++) 
	{
		NPLRC[n] = PLRC[n];
		nLRC[n] = LRC[n];
		R += LRC[n] * double(PLRC[n]) * pow(points[N-1].x, -PLRC[n] - 1);
	}
	if (Worker->isAdCorrA())
	{
		double *adCorr, RIso1, RIso2, b, Rm;
		int NAdCorr, PAdCorr, TAdCorr;
		bool *AdCorrFree;
		Worker->getAdCorrForReading(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, &AdCorrFree);
		NPot->Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, AdCorrFree);
	}
	NPot->Tab->setHorizontalHeaderLabels(QStringList() << "R [A]" << "E [cm^-1]" << "d^2E/dR^2");
	NPot->Tab->setColumnCount(3);
	NPot->Worker->setSplinePot(N, points, NLRC, NPLRC, nLRC, iA, iO, iExp, R);
	NPot->UpdateTab();
	State->addPotential(NPot);
	NPot->show();
	//return NPot->FitSplinePot(true, this);
	return 0.0;
}

void PotWorker::createSMap()
{
	int n, m;
	for (NumFreePar = n = 0; n < numSplinePoints; n++) if (points[n].variable) NumFreePar++;
	for (NumFreeLRC = n = 0; n < NLRC; n++) if (LRCfree[n]) NumFreeLRC++;
	//if (State->getStateNum() > 0 && NumFreeLRC > 0) NumFreeLRC = 1;
	for (NumFreeAdCorr = n = 0; n < NAdCorr; n++) if (adCorrFree[n]) NumFreeAdCorr++;
	NumFreePar += (NumFreeLRC + NumFreeAdCorr);
	if (shortRVariable) NumFreePar++;
	SMap = new int[NumFreePar];
	if (shortRVariable) 
	{	
		SMap[0] = 0;
		n=1;
	}
	else n=0;
	for (m=0; m < numSplinePoints; m++) if (points[m].variable) SMap[n++] = m + 1;
	if (NumFreeLRC > 0) 
	{
		/*if (State->getStateNum() > 0) SMap[n++] = numSplinePoints + 1;
		else*/ for (m=0; m < NLRC; m++) if (LRCfree[m]) SMap[n++] = numSplinePoints + 1 + m;
	}
	if (NumFreeAdCorr > 0) for (m=0; m < NAdCorr; m++) if (adCorrFree[m]) 
		SMap[n++] = numSplinePoints + NLRC + m + 1;
}

void PotWorker::saveCoefficients(double*& bC)
{
	int n, m;
	if (bC == 0) bC = new double[2 * numSplinePoints + NLRC + NAdCorr + 2];
	bC[0] = iA;
	bC[1] = iO;
	for (n=0, m=2; n < numSplinePoints; n++)
	{
		bC[m++] = points[n].y;
		bC[m++] = points[n].yss;
	}
	for (n=0; n < NLRC; n++) bC[m++] = LRC[n];
	for (n=0; n < NAdCorr; n++) bC[m++] = adCorr[n];
}

void PotWorker::restorePotential(double* bC)
{
	int n, m;
	Lock->lock();
	iA = bC[0];
	iO = bC[1];
	for (n=0, m=2; n < numSplinePoints; n++)
	{
		points[n].y = bC[m++];
		points[n].yss = bC[m++];
	}
	for (n=0; n < NLRC; n++) LRC[n] = bC[m++];
	for (n=0; n < NAdCorr; n++) adCorr[n] = bC[m++];
	Lock->unlock();
}

void PotWorker::setCurValuesForWantedValues()
{
	int n;
	for (n=0; n < NumWantedValues; n++)
	{
		if (WantedValues[n].CoeffNum == 0) WantedValues[n].CurValue = -iExp * iA * pow(points[0].x, -iExp - 1);
		else if (WantedValues[n].CoeffNum <= numSplinePoints) WantedValues[n].CurValue = points[WantedValues[n].CoeffNum - 1].y;
		else if (WantedValues[n].CoeffNum <= numSplinePoints + NLRC) WantedValues[n].CurValue = LRC[WantedValues[n].CoeffNum - numSplinePoints - 1];
		else WantedValues[n].CurValue = adCorr[WantedValues[n].CoeffNum - numSplinePoints - NLRC - 1];
	}
}

void PotWorker::FitSplinePot(int NumWFPoints, int MaxIt, double threshFakt)
{
	int n, r, NFC, F, I, J;
	double FQSv = -1.0, FQSn, minR, maxR, ****MaxErr;
	bool RPoints[numSplinePoints];
	LevenbergMarquardtRunning = false;
	emit fitRunning();
	if (SMap == 0) createSMap();
	for (n=0; n < NumFreePar; n++) printf("SMap[%d]=%d\n", n, SMap[n]);
	estMinErr(MaxErr, NFC = int(2.0 * el_S) + 1, mJ, mv, nEs, NEL, EL, nGS, nGL, GL);
	//for (n=0; n < nEs; n++) NLines += NEL[n];
	//for (n=0; n < nGS; n++) NLines += nGL[n];

	while ((r = splineFit(RPoints, FQSv, FQSn, mJ, mv, Mv, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, 
            minR, maxR, NumWFPoints, MaxErr, MaxIt, threshFakt)) != -1)
	{
		break;
		
		/*if (FQSn > 0.999 * lFQS) break;
		lFQS = FQSn;
		printf("Optimize points\n");
		r=-2;
		for (n=N=0; n < numSplinePoints; n++) if (RPoints[n]) N++;
		if (r>=0) N++;
		PB = new SplinePoint[N];
		for (n=m=0; n < numSplinePoints; n++)
		{
			if (n==r)
			{
				switch (r)
				{
					case 0:
						PB[0].x = ((X = 2.0 * points[0].x - points[1].x) > minR ? X : minR);
						PB[m++].y = getPoint(PB[0].x);
						break;
					case 1:
						PB[0].x = ((X = 1.5 * points[0].x - 0.5 * points[1].x) > minR ? X : minR);
						PB[0].y = getPoint(PB[0].x);
						PB[1].x = PB[0].x + 0.75 * (points[0].x - X);
						PB[m++].y = getPoint(PB[1].x);
						break;
					default:
						PB[m-1].x = 0.25 * points[n-2].x + 0.75 * points[n-1].x;
						PB[m-1].y = getPoint(PB[m-1].x);
						PB[m].x = 0.75 * points[n-1].x + 0.25 * points[n].x;
						PB[m].y = getPoint(PB[m].x);
						m++;
						break;
				}
			}
			if (RPoints[n]) PB[m++] = points[n];
		}
		//printf("m=%d, N=%d\n", m, N);
		if (r==n)
		{
			PB[N-1].x = ((X = 1.5 * points[numSplinePoints - 1].x 
					        - 0.5 * points[numSplinePoints - 2].x) < maxR ? X : maxR);
			PB[N-1].y = getPoint(PB[N-1].x);
			PB[N-2].x = PB[N-1].x + 0.75 * (points[numSplinePoints - 1].x - X);
			PB[N-2].y = getPoint(PB[N-2].x);
		}
		else if (r==n+1)
		{
			PB[N-1].x = ((X = 2.0 * points[numSplinePoints - 1].x - points[numSplinePoints - 2].x)
					      < maxR ? X : maxR);
			PB[N-1].y = getPoint(PB[N-1].x);
		}
		Ch = true;
		//printf("Vor zweiter\n");
		while (Ch) for (Ch = false, n=1; n<N-1; n++) 
		{
			if (PB[n].x - PB[n-1].x < 0.5 * (PB[n+1].x - PB[n].x))
			{
				PB[n].x = 0.66 * PB[n-1].x + 0.34 * PB[n+1].x;
				PB[n].y = getPoint(PB[n].x);
				Ch = true;
			}
			else if (PB[n].x - PB[n-1].x > 2.0 * (PB[n+1].x - PB[n].x))
			{
				PB[n].x = 0.34 * PB[n-1].x + 0.66 * PB[n+1].x;
				PB[n].y = getPoint(PB[n].x);
				Ch = true;
			}
		}
		if (N==0)
		{
			printf("Error: removed all points!\n");
			delete[] PB;
			break;
		}
		//printf("Vor delete, numSplinePoints=%d\n", numSplinePoints);
		Destroy(L, numSplinePoints);
		//printf("Vor delete S, K=%d\n", numSplinePoints+2);
		Destroy(S, numSplinePoints + 2);
		L = 0;
		S = 0;
		iA = 0.0;
		delete[] points;
		points = PB;
		numSplinePoints = N;
		delete[] LRC;
		delete[] PLRC;
		NLRC = 0;
		//printf("Vor calcyss\n");
		calcyss();
		emit propertiesChanged();
		printf("Ende optimization\n");
		if (!DoubleWell)
		{
			getMinimum(MR, MV);
			for (C = false, n=0; (n < numSplinePoints ? points[n].x < MR : false); n++) ;
			for (LV = MV; n < numSplinePoints; n++)
			{
				if ((AV = points[n].y) < LV)
				{
					for (m=n-1; points[m].y > AV; m--) points[m+1].y = points[m].y;
					points[m+1].y = AV;
					C = true;
				}
				else LV = AV;
			}
			for (n--; (n >= 0 ? points[n].x > MR : false); n--) ;
			for (LV = MV; n >= 0; n--)
			{
				if ((AV = points[n].y) < LV)
				{
					for (m=n+1; points[m].y > AV; m++) points[m-1].y = points[m].y;
					points[m-1].y = AV;
					C = true;
				}
				else LV = AV;
			}
			if (C) 
			{
				calcyss();
				emit propertiesChanged();
			}
		}*/
	}
	for (F=0; F < NFC; F++)
	{
		for (I=0; I < NIso; I++)
		{
			for (J=0; J <= mJ[I]; J++) if (MaxErr[F][I][J] != 0) delete[] MaxErr[F][I][J];
			if (MaxErr[F][I] != 0) delete[] MaxErr[F][I];
		}
		delete[] MaxErr[F];
	}
	delete[] MaxErr;
}

void Potential::FitSplinePot(bool showResult, int MaxIt, double threshFakt)
{
	if (!Worker->isSplinePot()) 
	{
		if (Worker->isAnaPotAv())
		{
			CreateSplineFromAnalyticalPot();
			return;
		}
		QMessageBox::information(this, tr("QT4MolSpektAn"), 
								 tr("Error: no spline potential available!"));
		return;
	}
	if (molecule == 0) 
	{
		QMessageBox::information(this, "QT4MolSpektAn", 
								 "Error: potential is not assigned to a molecule!");
		return;
	}
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", 
						"A fit of the potential cannot be started while there is already a fit of the same potential running!");
		return;
	}
	setFitData();
	if (!Worker->isFitDataAv()) return; 
	showFitResult = showResult;
	FQSv = -1.0;
	if (LogWindow == 0) initLogWindow();
	LogWindow->setWindowTitle("Fit of spline potential " + getName());
	Fit->FitSplinePot(MaxIt, threshFakt);
}

void PotWorker::updatePotential(double* C)
{
	Lock->lock();
	int i;
	double Rb, ys, Ra = points[numSplinePoints - 1].x;
	for (i=0; i < NumFreePar; i++) if (SMap[i] > 0 && SMap[i] <= numSplinePoints) 
		points[SMap[i] - 1].y += C[i];
	if (SMap[0] == 0) iA += (C[0] * pow(points[0].x, iExp));
	/*}
	else
	{
		double Par[numSplinePoints + NLRC + 1];
		for (n=0, Par[0] = iA; n < numSplinePoints; n++) Par[n+1] = points[n].y;
		for (n=0; n < NLRC; n++) Par[numSplinePoints + n + 1] = LRC[n];
		FQSn = fitSplinePot(this, E, K, mJ, mv, Mv, N, NE, QN, Sig, EQS, EO, DE, minR, maxR, 
			        		numSplinePoints + NLRC + 1, Par);
		for (iA = Par[n=0]; n < numSplinePoints; n++) points[n].y = Par[n+1];
		for (n=0; n < NLRC; n++) LRC[n] = Par[numSplinePoints + n + 1];
	}*/
	/*if (GS) 
	{
		for (n=0, i=1; i < numSplinePoints; i++) if (points[i].y < points[n].y) n=i;
		if (points[numSplinePoints - 1].x > maxR) Uinf -= points[n].y;
		for (i=0; i < numSplinePoints; i++) points[i].y -= points[n].y;
	}*/
	iO = points[0].y - iA * pow(points[0].x, -iExp);
	if (NLRC == 0)
	{
		LRC = new double[NLRC = 2];
		PLRC = new int[NLRC];
		PLRC[0] = 6;
		PLRC[1] = 8;
		LRC[0] = 0.0;
		LRC[1] = 0.0;
		Rb = pow(Ra, PLRC[0]);
		ys = PLRC[0] * LRC[0] * pow(Ra, -1 - PLRC[0]) 
		   + PLRC[1] * LRC[1] * pow(Ra, -1 - PLRC[1]) 
		   + (SMap[NumFreePar - 1] > numSplinePoints ? C[NumFreePar - 1] : 0.0);
		LRC[0] = Rb * (0.5 * double(2 + PLRC[0]) * (Uinf - points[numSplinePoints - 1].y) 
					   - 0.5 * (Ra) * ys);
		Rb *= Ra * Ra;
		LRC[1] = Rb * (0.5 * double(PLRC[0]) * (points[numSplinePoints - 1].y - Uinf) 
				+ 0.5 * Ra * ys);
	}
	else 
	{
		for (i=0; i < NumFreePar; i++)
		{
			if (SMap[i] > numSplinePoints + NLRC) adCorr[SMap[i] - numSplinePoints - NLRC - 1] += C[i];
			else if (SMap[i] > numSplinePoints) 
				LRC[SMap[i] - numSplinePoints - 1] += (C[i] * pow(Ra, PLRC[SMap[i] - numSplinePoints - 1]));
		}
	}
	calcyss();
	Lock->unlock();
}

void PotWorker::calcyss()
{
	double ED, Rb, ys, Ra = points[numSplinePoints - 1].x;
	int i, n;
	if (L == 0) calcLMatrix();
	if (Ro == 0.0)
	{
		for (i=0, ED = points[numSplinePoints - 1].y, ys = 0.0; i < NLRC; i++)
		{
			ED += (Rb = LRC[i] * pow(Ra, -PLRC[i]));
			ys += double(PLRC[i]) / Ra * Rb;
		}
		ED -= Uinf;
		//if (GS)
		//{
		for (n=0, iO -= ED; n < numSplinePoints; n++) points[n].y -= ED;
		//Uinf = 0.0;
		//}
		for (n=0; n < numSplinePoints; n++) 
		{
			for (i=0, points[n].yss = iA * L[n][0]; i < numSplinePoints; i++) 
				points[n].yss += points[i].y * L[n][i+1];
			points[n].yss += ys * L[n][numSplinePoints + 1];
			//printf("YSS[%d]=%f\n", n, F);
		}
	}
	else for (n=0; n < numSplinePoints; n++) 
		for (i=0, points[n].yss = 0.0; i < numSplinePoints; i++)
			points[n].yss += points[i].y * L[n][i];
}

int PotWorker::splineFit(bool *RPoints, double &FQSv, double &FQSn, int *&mJ, int **&mv, int &Mv, 
						 int &nEs, int *&NEL, TermEnergy **&EL, int &nGS, int *&nGP, int *&nGL, 
                         int **&PS, TableLine **&GL, double &minR, double &maxR, int NumWFPoints, double ****MaxErr, int MaxIt, double threshFakt)
{
	//printf("Potential::splineFit\n");
	int i, it, n, **QN;
	int N = NumFreePar, NE, BadC;
	int rv = N, MJ;
	for (n = MJ = 0; n < NIso; n++) if (mJ[n] > MJ) MJ = mJ[n];
	double ****E, *****K, **EQS, *EO, *DE, *Sig, C[N], aFQS = 0.0, *Err;
	FQS = FQSv = -1.0;
	//bool GS = (Uinf == 0);
	//for (n=0; n < nEs; n++) if (NEL[n] > 0) GS = false;
	CreateEK(E, K, mJ, mv, N);
	CreateFitEQS(NE, EQS, EO, DE, Sig, Err, QN, N, nEs, NEL, EL, nGS, nGP, nGL, PS, GL);
	if (Type == SplinePotential && L==0) calcLMatrix();
	else if (Type != SplinePotential && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
    if (S==0) calcS(NumWFPoints, rmin, rmax);
	for (it = BadC = 0; it < MaxIt && BadC < (FQS < FQSv ? 3 : 5); it++)
	{
		aFQS = FQS;
		Lock->lock();
        calcEK(E, K, Points(rmin, rmax, NumWFPoints), el_S, IsoT, IsoF, NumFreePar, SMap, S, NAdCorr, TAdCorr, PAdCorr,
			   RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorr, NumFreeAdCorr, Omega, q_e, q_f, SpinRRm, SpinRb, 
               NSpinRGamma, SpinRGamma, SpinRR1, SpinRR2, mJ, mv, Mv, N, NE, QN, Sig, EQS, EO, DE, minR, maxR, NumWFPoints, MaxErr);
		Lock->unlock();
		//printf("Nach calcEK\n");
        FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints);
		if (FQSv == -1.0) FQSv = FQS;
		if (FQS < threshFakt * aFQS) BadC = 0;
		else BadC++;
		//if (NE < 20000)
		//{
		SvdFit(EQS, C, DE, Sig, N, NE, 1e-2);//-2
			//printf("Nach SvdFit\n");
			/*
			for (i=0; points[i].x < minR; i++) ;
			for (i--; i >= 0; i--)
			{
				if (C[i+1] * C[i+2] <= 0.0) C[i+1] = 0.0;
				else if (fabs(C[i+1] * points[i+1].y) > fabs(C[i+2] * points[i].y))
				C[i+1] = C[i+2] * points[i].y / points[i+1].y;
			}
			for (i = numSplinePoints - 1; points[i].x > maxR; i--) ;
			for (i++; i < numSplinePoints; i++)
			{
				if (C[i+1] * C[i] <= 0.0) C[i+1] = 0.0;
				else if (fabs(C[i+1] * (Uinf - points[i-1].y)) > fabs(C[i] * (Uinf - points[i].y)))
					C[i+1] = C[i] * (Uinf - points[i].y) / (Uinf - points[i-1].y);
			}*/
		updatePotential(C);
		//emit propertiesChanged();
		/*if ((D1 = getPoint(D2 = 0.99 * points[numSplinePoints - 1].x 
				   + 0.01 * points[numSplinePoints - 2].x))
				  > (D3 = points[numSplinePoints - 1].y)) printf("RemP\n");*/
		//printf("x1=%f, y1=%f, x2=%f, y2=%f\n", D2, D1, points[numSplinePoints-1].x, D3);
	}
	//printf("Nach Fitschleife, minR=%f, maxR=%f\n", minR, maxR);
	//if (NE < 20000)
	//{
    calcE(E, mJ, mv, Mv, NumWFPoints, MaxErr);
    FQS = FQSn = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints, true, MJ, Mv, K, &rv);
	//}

	if (FQSn < NE) rv = -1;
	for (i=0; (i < numSplinePoints ? points[i].x < minR : false); i++) RPoints[i] = false;
	while (i < numSplinePoints ? points[i].x <= maxR : false) RPoints[i++] = true;
	/*if (FQS < 0.9999 * aFQS) 
	{
		if (i == numSplinePoints) i--;
		rv = -2;
	}
	printf("i=%d, numSplinePoints=%d\n", i, numSplinePoints);
	while (i < numSplinePoints) RPoints[i++] = false;*/
	DestroyFitData(E, K, mJ, mv, NE, EQS, EO, DE, Sig, Err, QN);
	if (rv == -1) for (n=0; n < numSplinePoints; n++) if (!RPoints[n]) return -2;
	return rv;
}

void PotWorker::FitSplinePotToFunction(double* E, int NPoints, double MinR, double MaxR)
{
	int n, m, N;
	double R, h = (MaxR - MinR) / (NPoints - 1), *D = new double[NPoints], **EQS, aFQS = 1e99 , FQS = 1e98, *C, *Sig = new double[NPoints];
	if (L == 0) calcLMatrix();
	calcS(NPoints, MinR, MaxR);
	if (SMap == 0) createSMap();
	C = new double[NumFreePar];
	for (N=0; (N < NumFreePar ? SMap[N] <= numSplinePoints + NLRC : false); N++) ;
	EQS = Create(NPoints, N);
	for (n=0; n < NPoints; n++) for (m=0, Sig[n] = 1e4; m < N; m++) EQS[n][m] = S[SMap[m]][n];
	while (FQS < aFQS)
	{
		aFQS = FQS;
		for (n=0, FQS = 0.0, R = MinR; n < NPoints; n++, R += h)
		{
			D[n] = E[n] - Point(R);
			FQS += D[n] * D[n];
		}
		printf("FSPtF: FQS=%f\n", FQS * 1E8);
		SvdFit(EQS, C, D, Sig, N, NPoints, 1e-2);
		for (n=N; n < NumFreePar; n++) C[n] = 0.0;
		updatePotential(C);
	}
	delete[] D;
	Destroy(EQS, NPoints);
	Destroy(S, NumS);
	S=0;
	delete[] Sig;
	delete[] C;
}

void PotWorker::SetSplinePar(double* Par, double ****E, double *****K, int *mJ, int **mv, int Mv, int N, 
							 int NL, int **QN, double *Sig, double **EQS, double *EO, double *DE, 
                             double &minR, double &maxR, int NumWFPoints)
{
	int n, m;
	double ys, Rb;
	iA = Par[0];
	iO = Par[1] - pow(points[0].x, -6.0);
	for (n=0; n < numSplinePoints; n++) points[n].y = Par[n+1];
	for (n=0, Uinf = ys = 0.0; n < NLRC; n++)
	{
		Uinf += (Rb = (LRC[n] = Par[numSplinePoints + n + 1]) * pow(points[numSplinePoints - 1].x, -PLRC[n]));
		ys += double(PLRC[n]) * Rb / points[numSplinePoints - 1].x;
	}
	for (n=0; n < numSplinePoints; n++)
	{
		for (m=0, points[n].yss = iA * L[n][0]; m < numSplinePoints; m++)
			points[n].yss += points[m].y  * L[n][m+1];
		points[n].yss += ys * L[n][m+1];
	}
	if (Type == SplinePotential && L==0) calcLMatrix();
	else if (Type != SplinePotential && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
    if (S==0) calcS(NumWFPoints, rmin, rmax);
	Lock->lock();
    calcEK(E, K, Points(rmin, rmax, NumWFPoints), el_S, IsoT, IsoF, NumFreePar, SMap, S, NAdCorr, TAdCorr, PAdCorr, RIso1AdCorr,
		   RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorr, NumFreeAdCorr, Omega, q_e, q_f, SpinRRm, SpinRb, NSpinRGamma, SpinRGamma,
           SpinRR1, SpinRR2, mJ, mv, Mv, N, NL, QN, Sig, EQS, EO, DE, minR, maxR, NumWFPoints);
	Lock->unlock();
}

void Potential::MonteCarloSimIteration(double *MCSTab, double UncFact)
{
	if (Fit->isRunning()) return;
	setFitData();
	if (LogWindow == 0) initLogWindow();
	LogWindow->setWindowTitle("MCS Fit Potential " + getName());
	showFitResult = false;
	FQSv = -1.0;
	Fit->RunMonteCarloIteration(MCSTab, UncFact);
}

void PotWorker::MonteCarloSimIteration(double* MCSTab, double UncFact)
{
	int e, n;
	srand(time(0));
	for (e=0; e < nEs; e++) for (n=0; n < NEL[e]; n++) EL[e][n].E -= EL[e][n].dev + MCSTab[rand() % 32768] * EL[e][n].err * UncFact;
	for (e=0; e < nGS; e++) for (n=0; n < nGL[e]; n++) GL[e][n].WN -= GL[e][n].dev + MCSTab[rand() % 32768] * GL[e][n].err * UncFact;
	/*if (points != 0) FitSplinePot(100);
	else if (Type == analyticalPotential)*/ 
	FitAnaPot(false, true, (Type == SplinePotential ? 0.0 : 1.0), false);
}

void Potential::CreateFirstPotFitTraceTabRows(FitResult Result, double FQS)
{
	QStringList Row;
	int c, NC = Tab->rowCount() - 2;
	for (c=0; c < NC; c++) Row << Tab->item(c, 0)->text();
	Row << "FQS" << "FQS_Bad" << "N_Bad" << "N_Bad_PAL" << "Sigma";
	PotFitTraceTab.clear();
	PotFitTraceTab << Row.join("\t");
	for (c=0, Row.clear(); c < NC; c++) Row << Tab->item(c, 1)->text();
	Row << QString::number(FQS, 'g', 10) << QString::number(Result.FQS_Bad, 'g', 10) << QString::number(Result.NBad) << QString::number(Result.NBadPAL) 
		<< QString::number(Result.Sigma, 'f', 10);
	PotFitTraceTab << Row.join("\t");
}

void Potential::potentialImproved(PotentialData *Data)
{
	QString FName = getFileName();
	QString TFile = FName.left(FName.lastIndexOf('.') + 1) + "temp.pot";
	if (fitData != 0)
	{
		int N, *Row;
		double *dev, *RWErr;
		Worker->getDeviations(N, Row, dev, RWErr);
		fitData->setDev(dev, Row, N);
		if (RWErr != 0) fitData->setRWErr(RWErr);
		Worker->unlockBadList();
	}
	if (BadList != 0)
	{
		int NR, NC;
		QString **Data;
		Worker->getBadList(Data, NR, NC);
		if (BadList->getNumColumns() != NC)
		{
			if (NC == 7) BadList->setHorizontalHeader(QStringList() << "Iso" << "v" << "J" << "Energy" 
														<< "Uncertainty" << "Obs-calc" << "Dev ratio");
			else BadList->setHorizontalHeader(QStringList() << "Iso" << "v''" << "J''" << "v'" << "J'" 
												<< "Energy" << "Uncertainty" << "Obs-calc" << "Dev ratio");
		}
		BadList->setData(Data, NR, NC);
		BadList->setEditable(false);
		Worker->unlockBadList();
	}
	FitResult Result = Data->fitResult;
	double FQS = Data->FQS;
	UpdateTab(Data);
	if (writePotFitTraceTab)
	{
		QStringList Row;
		int c, NC = Tab->rowCount() - 2;
		for (c=0, Row.clear(); c < NC; c++) Row << Tab->item(c, 1)->text();
		Row << QString::number(FQS, 'g', 10) << QString::number(Result.FQS_Bad, 'g', 10) << QString::number(Result.NBad) << QString::number(Result.NBadPAL) 
			<< QString::number(Result.Sigma, 'f', 10);
		PotFitTraceTab << Row.join("\t");
	}
	writeData(TFile, false);
	setFileName(FName);
	Changed();
}

void PotWorker::getFitResult(double& rFQS, double& rWeightSum, int& rNWeightF, int*& rWeightFRow, double*& rWeightF, int &numFreePar)
{
	BLLock->lock();
	rFQS = FQS;
	rWeightSum = WeightSum;
	rWeightF = WeightF;
	rNWeightF = NWeightF;
	rWeightFRow = WeightFRow;
	numFreePar = NumFreePar;
}

int PotWorker::cAdRows()
{
	int n, AdRows;
	for (n = AdRows = 0; n < NLRC; n++) if (LRCfree[n] && PLRC[n] >= 12) AdRows++;
	return AdRows;
}

void PotWorker::fillWantedValueEQS(int NEQ, double* Sig, double** EQS, double* DE)
{
	int n, m, c;
	setCurValuesForWantedValues();
	for (m = NEQ, n=0; n < NumWantedValues; n++, m++)
	{
		DE[m] = WantedValues[n].Value - WantedValues[n].CurValue;
		Sig[m] = 1.0 / WantedValues[n].Precision;
		//if (Sig[m] * DE[m] > 100.0) Sig[m] = 100.0 / DE[m];
		for (c=0; c < NumFreePar; c++) EQS[m][c] = (SMap[c] == WantedValues[n].CoeffNum ? 1.0 : 0.0);
	}
}

double PotWorker::FitAnaPot(int NumWFPoints, bool robustWeighting, bool UseSVD, double threshhold, bool useLevenbergMarquardt, int MaxIt,
							bool adjustTEWeightFact, double hCi_cP, bool FitRestarted)
{
	int n, m, e, NLevels = 0, NEQ, **QN, F, I, J, BadCount, i, it=1, done, AdRows;
	double ****MaxErr, ****E, *****K, **EQS, *EObs, *EDiff, *Sig, *Err;
	double aFQS =9e99, minR, maxR, FQSv = -1.0, **LMEQS, **tLMEQS, NFC;

	//writePoints("DebugPointsStart.Dat", 3.0, 50.0, 4701, 8, false, true, 0);
	
	StopFit = FitStopped = false;
	if (!FitRestarted) 
	{
		LevenbergMarquardtRunning = false;
		FQS = 8e99;
	}
	emit fitRunning();
	minLSFQS = -1.0;
	if (SMap != 0) delete[] SMap;
	createSMap();
	double C[NumFreePar];// aimp;
	double *bC = 0;
	cdConnectLR1C();
	cdConnectSR();
	if (getType() == SplinePotential) calcyss();
	saveCoefficients(bC);
	estMinErr(MaxErr, NFC = int(2.0 * el_S) + 1, mJ, mv, nEs, NEL, EL, nGS, nGL, GL);
	for (n=0; n < nEs; n++) NLevels += NEL[n];
	if (WeightF != 0) delete[] WeightF;
	WeightF = (adjustTEWeightFact ? new double[NLevels] : 0);
	if (nEs + nGS == 0) return 0.0; 
	if (hCi_cP > 0.0) AdRows = cAdRows();
	else AdRows = 0;
	CreateEK(E, K, mJ, mv, NumFreePar);
	CreateFitEQS(NEQ, EQS, EObs, EDiff, Sig, Err, QN, NumFreePar, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, AdRows);
	for (e=m=0; e < nEs; e++) for (n=0; n < NEL[e]; n++) Sig[m++] *= sqrt(EL[e][n].WeightFactor);
	if (Type == SplinePotential && L==0) calcLMatrix();
	else if (Type != SplinePotential && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
    if (S==0) calcS(NumWFPoints, rmin, rmax);
	Lock->lock();
    calcEK(E, K, Points(rmin, rmax, NumWFPoints), el_S, IsoT, IsoF, NumFreePar, SMap, S, NAdCorr, TAdCorr, PAdCorr, RIso1AdCorr,
		   RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorr, NumFreeAdCorr, Omega, q_e, q_f, SpinRRm, SpinRb, NSpinRGamma, SpinRGamma,
           SpinRR1, SpinRR2, mJ, mv, Mv, NumFreePar, NEQ, QN, Sig, EQS, EObs, EDiff, minR, maxR, NumWFPoints, MaxErr);
	Lock->unlock();
	if (NumWantedValues > 0) fillWantedValueEQS(NEQ, Sig, EQS, EDiff);
    FQSv = aFQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints, false, 0, 0, 0, 0, (robustWeighting ? Sig : 0), 1.0, hCi_cP);
	calcFitResult();
	emit firstFCFCalculated(Result, FQSv);
	
	//writePoints("DebugPointsStart2.Dat", 3.0, 50.0, 4701, 8, false, true, 0);
	
	if (UseSVD && !(LevenbergMarquardtRunning && FitRestarted)) for (BadCount =0; BadCount < 3; it++)  
	{
		SvdFit(EQS, C, EDiff, Sig, NumFreePar, NEQ + NumWantedValues, 1e-2);
		updatePotential(C);
		if (it == MaxIt || isStopFit())
		{
			if (it == 1) FQS = FQSv;
			useLevenbergMarquardt = false;
			break;
		}
		if (Type == SplinePotential && L==0) calcLMatrix();
		else if (Type != SplinePotential && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
        if (S==0) calcS(NumWFPoints, rmin, rmax);
		Lock->lock();
        calcEK(E, K, Points(rmin, rmax, NumWFPoints), el_S, IsoT, IsoF, NumFreePar, SMap, S, NAdCorr, TAdCorr, PAdCorr,
			   RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorr, NumFreeAdCorr, Omega, q_e, q_f, SpinRRm, SpinRb, NSpinRGamma,
               SpinRGamma, SpinRR1, SpinRR2, mJ, mv, Mv, NumFreePar, NEQ, QN, Sig, EQS, EObs, EDiff, minR, maxR, NumWFPoints, MaxErr);
		Lock->unlock();
		if (NumWantedValues > 0) fillWantedValueEQS(NEQ, Sig, EQS, EDiff);
        FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints, false, 0, 0, 0, 0, (robustWeighting ? Sig : 0), 1.0, hCi_cP);
		if (FQS < aFQS && FQS > 0.0 && !isnan(FQS))
		{
			if (FQS < 0.9 * aFQS || (!useLevenbergMarquardt && aFQS - FQS > threshhold)) BadCount = 0;
			else BadCount++;
			aFQS = FQS;
			saveCoefficients(bC);
		}
		else
		{
			BadCount++;
			//if (stopOnFirstStepFailure && FQS > FQSv) return 0.0;
		}
	}
	else FQS = aFQS;
	if (FQS > aFQS || FQS == 0.0 || isnan(FQS))
	{
		restorePotential(bC);
		if (Type == SplinePotential && L==0) calcLMatrix();
		else if (Type != SplinePotential && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
        if (S==0) calcS(NumWFPoints, rmin, rmax);
		Lock->lock();
        calcEK(E, K, Points(rmin, rmax, NumWFPoints), el_S, IsoT, IsoF, NumFreePar, SMap, S, NAdCorr, TAdCorr, PAdCorr,
			   RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorr, NumFreeAdCorr, Omega, q_e, q_f, SpinRRm, SpinRb, NSpinRGamma,
               SpinRGamma, SpinRR1, SpinRR2, mJ, mv, Mv, NumFreePar, NEQ, QN, Sig, EQS, EObs, EDiff, minR, maxR, NumWFPoints, MaxErr);
		Lock->unlock();
		if (NumWantedValues > 0) 
		{
			for (n=0; n < NumFreePar; n++) C[n] = bC[SMap[n]];
			fillWantedValueEQS(NEQ, Sig, EQS, EDiff);
		}
	}
	if (useLevenbergMarquardt && !isStopFit())
	{
		printf("Now Levenberg-Marquardt:\n");  // Using knowledge from the book "Numerical Recipes, Third Edition"
		LMEQS = Create(NumFreePar, NumFreePar + 1);
		tLMEQS = Create(NumFreePar, NumFreePar + 1);
		calcLMEQS(NEQ + NumWantedValues, EQS, Sig, EDiff, LMEQS, hCi_cP);
		if (!(FitRestarted && LevenbergMarquardtRunning)) lambda = 1e-3;
		LevenbergMarquardtRunning = true;
		for (done = BadCount = 0; true; it++)
		{
			if (done == 4)
			{
				if (adjustTEWeightFact) for (m=e=0; e < nEs; e++) for (n=0; n < NEL[e]; n++) 
					WeightF[m++] = (fabs(EL[e][n].DevR) < 4.0 ? (fabs(EL[e][n].DevR) < 1.0 ? 0.5 : 1.0) : 2.0);
				lambda = 0.0;
				
				//writePoints("DebugPointsFin.Dat", 3.0, 50.0, 4701, 8, false, true, 0);
			}
			if (done < 5)
			{
				for (n=0; n < NumFreePar; n++)
				{
					for (i=0; i <= NumFreePar; i++) tLMEQS[n][i] = LMEQS[n][i];
					tLMEQS[n][n] *= (1.0 + lambda);
				}
				SolvLinEqSwItImp(tLMEQS, NumFreePar, &minLSFQS);
				for (n=0; n < NumFreePar; n++) C[n] = tLMEQS[n][NumFreePar];
				updatePotential(C);
			}
			if (it == MaxIt || isStopFit())
			{
				if (it == 1) FQS = FQSv;
				break;
			}
			if (Type == SplinePotential && L==0) calcLMatrix();
			else if (Type != SplinePotential && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
            if (S==0) calcS(NumWFPoints, rmin, rmax);
			Lock->lock();
            calcEK(E, K, Points(rmin, rmax, NumWFPoints), el_S, IsoT, IsoF, NumFreePar, SMap, S, NAdCorr, TAdCorr, PAdCorr,
			       RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorr, NumFreeAdCorr, Omega, q_e, q_f, SpinRRm, SpinRb, 
                   NSpinRGamma, SpinRGamma, SpinRR1, SpinRR2, mJ, mv, Mv, NumFreePar, NEQ, QN, Sig, EQS, EObs, EDiff, minR, maxR, NumWFPoints,
		           MaxErr);
			Lock->unlock();
			if (NumWantedValues > 0) fillWantedValueEQS(NEQ, Sig, EQS, EDiff);
            FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints, false, 0, 0, 0, 0, (robustWeighting ? Sig : 0), 1.0, hCi_cP);
			if (done == 4 || BadCount == 20) 
			{
				if (adjustTEWeightFact)
				{
					if (FQS < aFQS) for (m=e=0; e < nEs; e++) for (n=0; n < NEL[e]; n++) 
						WeightF[m++] = (fabs(EL[e][n].DevR) < 4.0 ? (fabs(EL[e][n].DevR) < 1.0 ? 0.5 : 1.0) : 2.0);
					for (n=0, done=4; n<m; n++) if (WeightF[n] == 2.0) done++;
					if (done < 5) break;
					for (m=e=0; e < nEs; e++) for (n=0; n < NEL[e]; n++, m++) 
					{
						EL[e][n].WeightFactor *= WeightF[m];
						if (WeightF[m] != 1.0) printf("WeightFactor level with iso=%d, v=%d, J=%d now %g\n", 
												   EL[e][n].Iso, EL[e][n].v, EL[e][n].J, EL[e][n].WeightFactor);
					}
					for (n=0; n<m; n++) Sig[n] *= sqrt(WeightF[n]);
				}
				else break;
			}
			if (abs(FQS - aFQS) <  1e-10 * aFQS || (threshhold != 1e-10 && FQS < aFQS && aFQS - FQS < threshhold)) done++;
			if (FQS < aFQS && FQS > 0.0 && !isnan(FQS))
			{
				calcFitResult();
				emit potentialImproved(getPotentialData());
				if (done > 4) 
				{
					lambda = 1e-3;
                    if (aFQS != 9e99) aFQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints);
					else aFQS = FQS;
					done = 0;
				}
				else 
				{
					lambda *= 0.1;
					aFQS = FQS;
				}
				BadCount = 0;
				calcLMEQS(NEQ + NumWantedValues, EQS, Sig, EDiff, LMEQS, hCi_cP);
				saveCoefficients(bC);
			}
			else 
			{
				restorePotential(bC);
				BadCount++;
				if (done > 4) aFQS = 9e99;
				else lambda *= 10.0;
			}
		}
		if (FQS > aFQS || FQS == 0.0 || isnan(FQS))
		{
			restorePotential(bC);
			FQS = aFQS;
		}
		Destroy(LMEQS, NumFreePar);
		Destroy(tLMEQS, NumFreePar);
	}
	else calcFitResult();
	for (F=0; F < NFC; F++)
	{
		for (I=0; I < NIso; I++)
		{
			for (J=0; J <= mJ[I]; J++) if (MaxErr[F][I][J] != 0) delete[] MaxErr[F][I][J];
			if (MaxErr[F][I] != 0) delete[] MaxErr[F][I];
		}
		delete[] MaxErr[F];
	}
	delete[] MaxErr;
	if (adjustTEWeightFact)
	{
		if (WeightFRow != 0) delete[] WeightFRow;
		WeightFRow = new int[NLevels];
		for (e=m=0, WeightSum = 0.0; e < nEs; e++) for (n=0; n < NEL[e]; n++)
		{
			WeightSum += (WeightF[m] = EL[e][n].WeightFactor);
			Row[m++] = EL[e][n].Row;
		}
		NWeightF = m;
	}
	for (e=0; e < nGS; e++) for (n=0; n < nGL[e]; n++) WeightSum += GL[e][n].WeightFact;
	DestroyFitData(E, K, mJ, mv, NEQ, EQS, EObs, EDiff, Sig, Err, QN);
	destroyFitData();
	delete[] bC;
	return FQS;
}

void Potential::fitFinished()
{
	double FQS, WeightSum, *WeightF;
	int NWeightF, *WeightFRow, NumFreePar;
	Worker->getFitResult(FQS, WeightSum, NWeightF, WeightFRow, WeightF, NumFreePar);
	if (NWeightF > 0) fitData->setWeightFactors(NWeightF, WeightFRow, WeightF);
	Worker->unlockBadList();
	if (Fit->getTask() == PotFit::calcFQS)
	{
		emit SFQScalculated(threadNum, Worker->getSFQS(), FQS);
		return;
	}
	if (!Worker->isLevenbergMarquardtRunning()) potentialImproved(Worker->getPotentialData());
	if (showFitResult) QMessageBox::information(this, tr("Fit result"), 
								 "Initial chiSq=" + QString::number(FQSv, 'g', 6) + ", final chiSq="
								 + QString::number(FQS, 'g', 6) + ", Sigma=" 
								 + QString::number(sqrt(FQS/(WeightSum - double(NumFreePar)))));
	UpdateTab();
	Changed();
	if (Fit->getTask() == PotFit::MonteCarloSimulation) 
	{
		int N, *Rows;
		double *E = Worker->getMCSE(N, Rows);
		fitData->setMCSEnergies(E, Rows, N);
	}
	if (getTexTableSplineFitRunning) getTexTable(FQS);
	emit FitFinished(threadNum, FQS);
}

void PotWorker::getSplineMinMaxR(double& Min, double& Max)
{
	if (points != 0)
	{
		Min = points[0].x;
		Max = points[numSplinePoints - 1].x;
	}
	else Min = Max = 0.0;
}

void Potential::fitTerminated()
{
	emit FitTerminated(threadNum);
}

bool PotWorker::isStopFit()
{
	QMutexLocker Locker(Lock);
	if (!StopFit) return false;
	FitStopped = true;
	emit fitStopped();
	return true;
}

bool PotWorker::isFitStopped()
{
	printf("Lock.lock() PotWorker::isFitStopped\n");
	Lock->lock();
	return FitStopped;
}

void Potential::restartFit()
{
	if (Worker->isFitStopped())
	{
		if (Fit->isRunning()) Fit->wait();
		Fit->restartFit();
	}
	else if (Fit->isRunning()) Worker->stopFit(false);
	Worker->unlock();
}

void Potential::cancelFit(bool wait)
{
	Worker->stopFit(true);
	Fit->terminate();
	if (wait && Fit->isRunning()) Fit->wait();
}

void Potential::initLogWindow()
{
	LogWindow = new PotFitLogWindow();
	connect(LogWindow, SIGNAL(CancelFit(bool)),this, SLOT(cancelFit(bool)));
	connect(LogWindow, SIGNAL(StopFit()), this, SLOT(stopFit()));
	connect(LogWindow, SIGNAL(StartFit()), this, SLOT(startFit()));
	connect(LogWindow, SIGNAL(RestartFit()), this, SLOT(restartFit()));
	connect(Worker, SIGNAL(fitRunning()), LogWindow, SLOT(FitRunning()));
	connect(Worker, SIGNAL(fitStopped()), LogWindow, SLOT(FitStopped()));
	connect(this, SIGNAL(FitFinished(int, double)), LogWindow, SLOT(FitFinished()));
	connect(this, SIGNAL(FitTerminated(int)), LogWindow, SLOT(FitTerminated()));
	MW->showMDIChild(LogWindow);
	MW->setAskForQuit(LogWindow);
}

void Potential::FitAnaPot(bool showResult, bool robustWeighting, bool UseSVD, double threshhold, bool transform, bool ask,
						  bool useLeveMarq, int MaxIt, bool ATEWF, double hCi_cP, bool WritePotFitTraceTab)
{
	int NC, numPoints = 10000, nLRC;
	double Ri, Ra;
	double Rmin, Rmax, iEx = Worker->getIExp(), bmin = -0.1, bmax = -0.59, bstep = -0.01, A_ex = 0.0, beta, gamma;
	bool improvePotential = ((Worker->isAnaPotAv() || Worker->isSplinePot()) && !transform);
	int *pLRC = Worker->getSplinePLRC();
	PotentialType FType;
	if (!WritePotFitTraceTab && writePotFitTraceTab) disconnect(Worker, SIGNAL(firstFCFCalculated(FitResult,double)), this, SLOT(CreateFirstPotFitTraceTabRows(FitResult,double)));
	else if (!writePotFitTraceTab && WritePotFitTraceTab) connect(Worker, SIGNAL(firstFCFCalculated(FitResult,double)), this, SLOT(CreateFirstPotFitTraceTabRows(FitResult,double)));
	writePotFitTraceTab = WritePotFitTraceTab;
	Worker->getSplinePotPointandCoeffNumbers(NC, nLRC);
	Worker->getSplineMinMaxR(Ri, Ra);
	if (Ri == 0.0)
	{
		Ri = rmin + 1.0;
		Ra = 10.0;
	}
	if (ask)
	{
		if (Worker->isAnaPotAv())
		{
			double *aLRC, aiC, aiO;
			int N, n;
			AnaPot *aPot = Worker->getAnaPotForReading();
			bool *LRCFree = aPot->getLRCFree();
			Worker->getLRCoeffForReading(nLRC, pLRC, aLRC, &Ra, true);
			aPot->getInnerWall(Ri, aiO, iEx, aiC);
			aPot->getNC(N, NC, nLRC);
			if (hCi_cP <= 0.0) for (hCi_cP = -1.0, n=0; n < nLRC; n++) if (LRCFree[n] && pLRC[n] >= 12) hCi_cP = 0.0;
			Worker->unlock();
		}
		else NC = 0;
		FitAnaPotDialog *D = new FitAnaPotDialog(improvePotential, Ri, Ra, iEx, NC, nLRC, pLRC, hCi_cP, Worker->getType(), this);
		if (!D->exec(improvePotential, useLeveMarq, UseSVD, Ri, Ra, iEx, NC, nLRC, pLRC, Rmin, Rmax, numPoints, 
			bmin, bmax, bstep, ATEWF, hCi_cP, FType)) 
		{
			delete D;
			return;
		}
		delete D;
		if (bstep == 0.0) bstep = -0.01;
		if (bstep * (bmax - bmin) < 0.0)
		{
			double db = bmax;
			bmax = bmin;
			bmin = db;
		}
	}
	else
	{
		Rmin = Ri;
		Rmax = Ra;
		NC = 15;
	}
	if (!improvePotential)
	{
		int n;
		double *Sig, *U;
		QString F;
		Potential *nPot = (!transform ? MW->CreatePotential() : this);
		AnaPot *TAPot = new AnaPot(nPot->Fit);
		double h = (Rmax - Rmin) / (numPoints - 1), b, Rm, bFQS = 1e50, FQS;
		double uinf = Worker->getUinf();
		U = Worker->getPoints(Rmin, Rmax, numPoints);
		Sig = new double[numPoints];
		if (Worker->isAnaPotAv()) 
		{
			AnaPot *aPot = Worker->getAnaPotForReading();
			aPot->getExchangeCoeff(A_ex, beta, gamma);
			if (A_ex != 0.0) TAPot->setExchangeCoeff(A_ex, beta, gamma, aPot->isExcIntFree());
			Worker->unlock();
		}
		int *PLRC;
		double *LRC;
		bool *LRCFree = Worker->getLRCFree(), *lrcFree = new bool[nLRC];
		for (n=0; n < nLRC; n++) lrcFree[n] = LRCFree[n];
		Worker->getLRCoeffForReading(nLRC, PLRC, LRC);
		TAPot->setLRCoeff(Ra, nLRC, PLRC, LRC, lrcFree);
		for (n=0; n < numPoints; n++) Sig[n] = 1e-4;
		getMinimum(Rm, FQS);
		for (b = bmin; (bstep < 0 ? b > bmax + bstep : b < bmax + bstep); b+= bstep) 
			if ((FQS = TAPot->fitPotential(numPoints, Rmin, h, U, Sig, NC, nLRC, pLRC, Ri, Ra, Rm, b, uinf, iEx)) < bFQS)
		{
			bFQS = FQS;
			nPot->setWorker(new AnaPot(*TAPot));
		}
		if (Worker->isAdCorrA() && !transform)
		{
			int NAdCorr, TAdCorr, PAdCorr;
			double *adCorr, Rm, b, RIso1, RIso2;
			bool *adCorrFree;
			Worker->getAdCorrForReading(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, &adCorrFree);
			nPot->Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, adCorrFree);
		}
		delete[] U;
		delete[] Sig;
		nPot->UpdateTab();
		nPot->setName(getName() + "AnaPot");
		nPot->setSource("Spline potential from " + getSource());
		F = getFileName();
		nPot->setFileName(F.right(F.length() - F.indexOf('.') - 1) + "AnaPot.pot");
		if (!transform)
		{
			if (State != 0) State->addPotential(nPot);
			nPot->show();
		}
		//if (!useLeveMarq && !UseSVD) return bFQS;
		if (!transform) nPot->FitAnaPot(showResult, robustWeighting, UseSVD, threshhold, false, false, useLeveMarq, MaxIt);
		return;
	}
	else
	{
		if (Worker->getNIso() == 0)
		{
			QMessageBox::information(this, "MolSpektAnalysis", "The potential has to be assigned to a molecule with atomic data first!");
			return;
		}
		if (Fit->isRunning())
		{
			QMessageBox::information(this, "MolSpektAnalysis", 
							"A fit of the potential cannot be started while there is already a fit of the same potential running!");
			return;
		}
		if (LogWindow == 0) initLogWindow();
		LogWindow->setWindowTitle("Fit of analytical potential " + getName());
		setFitData(0, 0, -1);
		showFitResult = showResult;
		FQSv = -1.0;
		Fit->FitAnaPot(robustWeighting, UseSVD, threshhold, useLeveMarq, MaxIt, ATEWF, hCi_cP);
		Fit->start();
	}
	
	/*if (State != 0 ? (fitData = State->getFitData()) == 0 : true)
	{
		QMessageBox::information(this, tr("MolSpektAnalysis"), 
							 "Error: No fit data available!");
		return;
	}
	if (aPot == 0 && LPot == 0)
	{
		QMessageBox::information(this, tr("MolSpektAnalysis"), 
							 "Error: No initial potential data available!");
		return;
	}
	int NC, nLRC, NL, i, o, n, *lp, *pLRC, p, q, NS;
	double *C, *lRC, Rm, Ra, b, Tm, De;
	TableLine *data, *data1, dB;
	LineTable *LSource = 0, *Sources[MaxLineTables];
	fitData->getData(data1, NL);
	for (n = NS = i = 0; n < NL; n++) 
	{
		if (!data1[n].isTE)
		{
			if (data1[n].LTab != LSource)
			{
				LSource = data1[n].LTab;
				for (i=0; (i < NS ? LSource != Sources[i] : false); i++) ;
				if (i == NS) Sources[NS++] = LSource;
			}
			data1[n].SourceN = i;
		}
		else data1[n].SourceN = -1;
	}
	for (i=1; i < NL; i++) 
		for (o=((n=i)-1)/2; 
			 (o>=0 ? data1[n].Iso < data1[o].Iso || (data1[n].Iso == data1[o].Iso && data1[n].Jss < data1[o].Jss) 
				   : false); o=((n=o)-1)/2)
	{
		dB = data1[n];
		data1[n] = data1[o];
		data1[o] = dB;
	}
	data = new TableLine[NL];
	for (i=0; i < NL; i++)
	{
		data[i] = data1[0];
		for (o=1, n=0; o < NL; o=2*(n=o)+1)
		{
			if (o+1 < NL)
				if (data1[o].Iso > data1[o+1].Iso || (data1[o].Iso == data1[o+1].Iso && data1[o].Jss > data1[o+1].Jss)) 
					o++;
			if (data1[o].Iso >= 1000) break;
			data1[n] = data1[o];
			data1[o].Iso += 1000;
		}
	}
	delete[] data1;
	if (aPot != 0) 
	{
		aPot->getPotCoeff(b, Rm, Tm, De, NC, C);
		aPot->getLRCoeff(Ra, nLRC, lp, lRC);
	}
	else 
	{
		LPot->getNC(i, NC, nLRC);
		NC += 3;
	}
	MrqLev ML(NC + nLRC, NL);
	if (aPot != 0) 
	{
		for (i=0; i < NC; i++)
		{
			ML.a[i] = C[i];
			ML.ia[i] = true;
		}
		for (n=0; n < nLRC; n++)
		{
			ML.a[i] = lRC[n];
			ML.ia[i++] = true;
		}
	}
	else 
	{
		LPot->getCoefficients(p, q, ML.a[0], ML.a[1], ML.a[2], Rm, NC, C, nLRC, lp, lRC);
		ML.ia[0] = ML.ia[1] = ML.ia[2] = true;
		for (n=0, i=3; n < NC; n++)
		{
			ML.a[i] = C[n];
			ML.ia[i++] = true;
		}
		for (n=0; n < nLRC; n++)
		{
			ML.a[i] = lRC[n];
			ML.ia[i++] = false;
		}
	}
	delete[] lRC;
	delete[] C;
	calcDerivatives(data, NL, ML.E, ML.dyda);
	for (n=0, i=-1; n < NL; n++) 
	{
		ML.sig[n] = data[n].err;
		ML.y[n] = data[n].WN;
		if (!data[n].isTE && data[n].PN > i) i = data[n].PN;
	}
	if (i>=0)
	{
		double **SE = Create(NS, i+1), **WS = Create(NS, i+1), W;
		for (o=0; o < NS; o++) for (n=0; n<=i; n++) SE[o][n] = WS[o][n] = 0.0;
		for (n=0; n < NL; n++) if (!data[n].isTE)
		{
			W = 1.0 / (data[n].err * data[n].err);
			WS[data[n].SourceN][data[n].PN] += W;
			SE[data[n].SourceN][data[n].PN] += W * data[n].WN;
		}
		for (o=0; o < NS; o++) for (n=0; n<=i; n++) if (WS[o][n] != 0.0) SE[o][n] /= WS[o][n];
		for (n=0; n < NL; n++) if (!data[n].isTE) ML.y[n] -= SE[data[n].SourceN][data[n].PN];
		Destroy(SE, NS);
		Destroy(WS, NS);FileName.left(FileName.length() - 4) + "_i.pot"
	}
	ML.init();
	printf("init: %e ma=%d, mfit=%d\n", ML.chisq, ML.ma, ML.mfit);
	for (i=0; i < 1000; i++)
	{
		pLRC = new int[nLRC];
		lRC = new double[nLRC];
		C = new double[NC];
		if (LPot != 0)
		{
			for (n=0, o=3; n < NC; n++) C[n] = ML.atry[o++];
			for (n=0; n < nLRC; n++)
			{
				pLRC[n] = lp[n];
				lRC[n] = ML.atry[o++];
			}
			LPot->setCoefficients(p, q, ML.atry[0], ML.atry[1], ML.atry[2], Rm, NC, C, nLRC, pLRC, lRC);
		}
		else if (aPot != 0)
		{
			for (o=0; o < NC; o++) C[o] = ML.atry[o];
			for (n=0; n < nLRC; n++)
			{
				pLRC[n] = lp[n];
				lRC[n] = ML.atry[o++];
			}
			aPot->setPotCoeff(b, Rm, Tm, De, NC, C);
			aPot->setLRCoeff(Ra, nLRC, pLRC, lRC);
		}
		calcDerivatives(data, NL, ML.E, ML.dyda);
		if (!ML.fit()) 
		{
			printf("finished: chisq=%e\n", ML.chisq);
			break;
		}
		if (ML.alamda > 1e300)
		{
			printf("lamda to large!\n");
			break;
		}
		printf("%d: %e nd=%d, l=%e\n", i, ML.chisq, ML.done, ML.alamda);
	}
	delete[] data;
	UpdateTab();
	Changed();*/
	
    /*int n, NP = NumWFPoints, NK = 5, np, ri, pi, pa, nP, NMax = 0;
	double RMin, RMax, Max, *U, Tm, b, Rm, bb, Err, sErr, De, rMin = rmin, rMax = rmax;
	if (aPot != 0) 
	{
		aPot->getConR(RMin, RMax);
		De = aPot->GetDe();
	}
	else if (points != 0)
	{
		RMin = points[0].x;
		RMax = points[numSplinePoints - 1].x;
		De = points[numSplinePoints - 1].y;
	}
	else
	{
		QMessageBox::information(this, tr("MolSpektAnalysis"), 
							 "Error: No initial potential data available!");
		return;
	}
	double h = (rMax - rMin) / double(NP - 1), r;
	setBlockChangeSignal(true);
	for (r = rMin, pi=0; r <= RMin; r+=h, pi++) ;
	for (RMin = r, pa = pi; r < RMax; r+=h, pa++) ;
	RMax = r-h;
	nP = (pa--) - pi;
	double *R = new double[nP], M, MF = 0.0833333333333 * h * h;
	U = getPoints(rMin, rMax, NP);
	for (R[0] = RMin, n=1; n < nP; n++) R[n] = R[n-1] + h;
	for (Max = U[NP - 1], n = pi; n < pa; n++) if (U[n] > Max && U[n] > U[n-1] && U[n] > U[n+1]) 
			Max = U[n];
	for (n = pi; (n < pa ? U[n] > Max : false); n++) ;
	if (n == pa)
	{
		QMessageBox::information(this, tr("MolSpektAnalysis"), 
							"Error: This fit function should not be used for repulsive potentials!");
		delete[] R;
		delete[] U;
		setBlockChangeSignal(false);
		return;
	}
	ri = n - pi;
	pi = n;
	int *lExp = new int[2];
	double *PotK = new double[NK], *lK = new double[2], *Sig = new double[NP], *WF = new double[NP];
	double *UR = new double[NP];
	int i, j, *mJ, **mv, Mv, nEs, *NEL, nGS, *nGP, *nGL, **PS, I, J, v, ***NL = new int**[NIso];
	TermEnergy **EL;
	TableLine **GL;
	TermTable *TT = State->getTermTable();
	double ****Data = TT->getData(), Ri, Ra, ISq, F, CF;
	int Nv = TT->getMaxv(), NJ = TT->getMaxJ(), Wi, Wc, Wa;
	getFitData(mJ, mv, Mv, nEs, NEL, EL, nGS, nGP, nGL, PS, GL);
	for (I=0; I < NIso; I++) if (mJ[I] != -1)
	{
		NL[I] = new int*[mJ[I] + 1];
		for (J=0; J <= mJ[I]; J++) if (mv[I][J] != -1)
		{
			NL[I][J] = new int[mv[I][J] + 1];
			for (v=0; v <= mv[I][J]; v++) NL[I][J][v] = 0;
		}
	}
	for (i=0; i < nEs; i++) for (j=0; j < NEL[i]; j++) NL[EL[i][j].Iso][EL[i][j].J][EL[i][j].v]++;
	for (i=0; i < nGS; i++) for (j=0; j < nGL[i]; j++) NL[GL[i][j].Iso][GL[i][j].Jss][GL[i][j].vss]++;
	printf("nP=%d, pa=%d\n", nP, pa);
	for (n=pi; n < pa; n++) Sig[n] = 0.0;
	for (I=0; I < NIso; I++) if (mJ[I] != -1) for (M = MF * IsoF[I], J=0; J <= mJ[I]; J++) 
				if (mv[I][J] != -1 && J <= NJ)
	{
		CRotPot(J, I, U, UR, rMin, h, NP);
		printf("I=%d, J=%d, Wi=%d, Wc=%d, Wa=%d\n", I, J, Wi, Wc, Wa);
		for (v=0; v <= mv[I][J]; v++) if (NL[I][J][v] > 0 && v <= Nv && Data[0][I][v][J] != 0.0)
		{
			fastWaveFunc(UR, Data[0][I][v][J], M, h, NP, WF, Ri, Ra, ISq, CF, Wi, Wc, Wa);
			ISq = 1.0 / ISq;
			CF *= CF;
			if (Wa > nP) printf("v=%d, nP=%d, Wa=%d\n", v, nP, Wa);
			for (n = Wi; n < Wa; n++)
			{
				F = WF[n] * WF[n] * (n <= Wc ? ISq : ISq * CF);
				if (F > 1.0) printf("I=%d, J=%d, v=%d, n=%d, F=%g\n", I, J, v, n, F);
				Sig[n] += F * F * double(NL[I][J][v]);
			}
		}
	}
	getMinimum(Rm, Tm);
	printf("pi=%d, pa=%d, nP=%d\n", pi, pa, nP);
	for (n = pi, M = 0.0; n < pa; n++) if (Sig[n] > M) M = Sig[n];// i=n;}
	for (NMax = pa - 1, F = 0.01 * M; Sig[NMax] < F; NMax--) ;
	printf("M=%g\n", M);
	for (n = pi, F = sqrt(1.0 / (M *= 1e-3)); n < pa; n++) 
	{
		if (Sig[n] < M) Sig[n] = F;
		else Sig[n] = sqrt(1.0 / Sig[n]);
	}
	np = pa - pi;
	printf("pi=%d, np=%d\n", pi, np);
	printf("R[ri]=%g, rMin=%g, rMin+pi*h=%g\n", R[ri], rMin, rMin + pi * h);
	for (b=-0.1, sErr = 1e300; b>=-0.6; b-=0.01) 
	{
		if ((Err = fitPotCoeff(np, R + ri, U + pi, NK, Rm, b, PotK, Tm, Sig + pi)) < sErr)
		{
			bb = b;
			sErr = Err;
		}
		printf("b=%e, Err=%g\n", b, Err);
	}
	fitPotCoeff(np, R + ri, U + pi, NK, Rm, bb, PotK, Tm, Sig + pi);
	if (aPot != 0) delete aPot;
	else Tab->setHorizontalHeaderLabels(QStringList() 
					<< "coefficient" << "value" << "error" << "sig. digits");
	aPot = new AnaPot;
	aPot->setPotCoeff(b, Rm, Tm, De, NK, PotK);
	aPot->setLRCoeff(RMax, 0, lExp, lK);
	aPot->sdConnectLR(6);
	aPot->sdConnectIW(6, RMin + double(ri) * h);
	delete[] U;
	U = getPoints(rMin, rMax, NP);
	for (n = NMax, M = U[NP - 1]; (n < NP ? U[n] < M && U[n] > U[n-1] : false); n++) ;
	if ((F = rMin + double(n) * h) < RMax)
	{
		aPot->setLRCoeff(F, 0, lExp, lK);
		aPot->sdConnectLR(6);
	}
	UpdateTab();
	for (n=0; n < numSplinePoints; n++) points[n].y = points[n].yss = 0.0;
	printf("Vor delete\n");
	for (I=0; I < NIso; I++) if (mJ[I] != -1)
	{
		for (J=0; J <= mJ[I]; J++) if (mv[I][J] != -1) delete[] NL[I][J];
		delete[] NL[I];
		delete[] mv[I];
	}
	printf("vor delete NL\n");
	delete[] NL;
	delete[] mJ;
	delete[] mv;
	printf("Vor delete PS\n");
	for (n=0; n < nGS; n++) delete[] PS[n];
	printf("Vor NEL\n");
	delete[] NEL;
	delete[] nGL;
	delete[] nGP;
	delete[] PS;
	printf("Vor delete sig\n");
	delete[] Sig;
	printf("WF\n");
	delete[] WF;
	printf("U\n");
	delete[] U;
	printf("R\n");
	delete[] R;
	printf("UR\n");
	delete[] UR;
	setBlockChangeSignal(false);
	Changed();
	printf("Ende  FitAnapot\n");
	*/
}

void PotWorker::calc_hCi_cP_EQS(int NEQ, double** EQS, double* Sig, double* EDiff, double hCi_cP)
{
	int n, m, l, i, j;
	double CC;
	for (n=i=0; i < NLRC; i++) if (PLRC[i] >= 12 && LRCfree[i]) n++; 
	for (m=i=0; i < NLRC; i++) if (PLRC[i] >= 12 && LRCfree[i])
	{
		Sig[j = NEQ - n + m++] = 1.0 / ((CC = pow(LRC[i-1] / LRC[i-2], 3.0) * LRC[i-3]) * hCi_cP);
		EDiff[j] = CC - LRC[i];
		for (l=0; l < NumFreePar; l++) EQS[j][l] = 0.0;
		for (l=0; (l < NumFreePar ? SMap[l] - numSplinePoints - 1 < i-3 : false); l++) ;
		if (SMap[l] - numSplinePoints == i-2) EQS[j][l++] -= CC / LRC[i-3];
		if (SMap[l] - numSplinePoints == i-1) EQS[j][l++] += 3.0 * CC / LRC[i-2];
		if (SMap[l] - numSplinePoints == i) EQS[j][l++] -= 3.0 * CC / LRC[i-1];
		if (SMap[l] - numSplinePoints - 1 == i) EQS[j][l] = 1.0;
	}
}

void PotWorker::calcLMEQS(int NEQ, double** EQS, double* Sig, double* EDiff, double** LMEQS, double hCi_cP)
{
	if (hCi_cP > 0.0) calc_hCi_cP_EQS(NEQ, EQS, Sig, EDiff, hCi_cP);
	CalcLMEQS(NEQ, EQS, Sig, EDiff, LMEQS, hCi_cP, NumFreePar);
}

double PotWorker::TTFit(double****, double*****, int, int**, double*, double**, double*,
						double*, double****, double*, double&, double)
{
	/*int n;
	double FQS = 8e99, aFQS =9e99, minR, maxR;
	if (b != 0)
	{
		TTPot->setb(b);
		Destroy(S, NumFreePar);
		S = 0;
	}
	while (FQS < 0.9999 * aFQS)
	{
		aFQS = FQS;
		calcEK(E, K, mJ, mv, Mv, NumFreePar, NEQ, QN, Sig, EQS, EObs, EDiff, minR, maxR, MaxErr);
		//printf("Nach calcEK\n");
		FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL);
		if (FQSv == -1.0) FQSv = FQS;
		SvdFit(EQS, C, EDiff, Sig, NumFreePar, NEQ, 1e-7);
		TTPot->updateCoefficients(C);
		for (n=0; n < NAdCorr; n++) adCorr[n] += (C[NumFreePar - NAdCorr + n] * AdCorrWF[n]);
	}
	return FQS;*/
	return 0.0;
}

double PotWorker::FitTangToenniesPot()
{
	/*int n, NC = TTPot->getNFitCoefficients(), NLines = 0, NEQ, **QN, F, I, J;
	double ****MaxErr, ****E, *****K, **EQS, *EObs, *EDiff, *Sig, *Err;
	double FQS = 8e99, FQSv = -1.0, NFC = int(2.0 * el_S) + 1;
	double C[NumFreePar = NC + NAdCorr];
	if (SMap == 0)
	{
		SMap = new int[NumFreePar];
		for (n=0; n < NumFreePar; n++) SMap[n] = n;
	}
	estMinErr(MaxErr, NFC, mJ, mv, nEs, NEL, EL, nGS, nGL, GL);
	for (n=0; n < nEs; n++) NLines += NEL[n];
	for (n=0; n < nGS; n++) NLines += nGL[n];
	if (nEs + nGS == 0) return 0.0; 
	CreateEK(E, K, mJ, mv, NumFreePar);
	CreateFitEQS(NEQ, EQS, EObs, EDiff, Sig, Err, QN, NumFreePar, nEs, NEL, EL, nGS, nGP, nGL, PS, GL);
	//double b1 = 0.0, b2 = TTPot->getb(), b3 = b2 + 0.1;
	FQS = TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv); //FQS2*/
	
	/*double FQS1, FQS3 = TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv, b3);
	while (FQS3 < FQS2 && b3 < 4.5)
	{
		b1 = b2;
		FQS1 = FQS2;
		b2 = b3;
		FQS2 = FQS3;
		FQS3 = TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv, b3 += 0.1);
		}
		if (b1 == 0.0) 
		{
			FQS1 = TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv, b1 = b2 - 0.1);
			while (FQS1 < FQS2 && b1 >= 0.4)
			{
			b3 = b2;
			FQS3 = FQS2;
			b2 = b1;
			FQS2 = FQS1;
			FQS1 = TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv, b1 -= 0.1);
		}
	}
	ParabInterpol(b1, FQS1, b2, FQS2, b3, FQS3, b, FQS);
	for (n=0; n < 25 && fabs(FQS2 - FQS) >= 1e-8; n++)
	{
		if (b > b2)
		{
			b1 = b2;
			FQS1 = FQS2;
		}
		else 
		{
			b3 = b2;
			FQS3 = FQS2;
		}
		FQS2 = TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv, b2 = b);
		ParabInterpol(b1, FQS1, b2, FQS2, b3, FQS3, b, FQS);
	}
	TTFit(E, K, NEQ, QN, Sig, EQS, EObs, EDiff, MaxErr, C, FQSv, b);
	calcE(E, mJ, mv, Mv, MaxErr);
	FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL);*/
	
	/*for (F=0; F < NFC; F++)
	{
		for (I=0; I < NIso; I++)
		{
			for (J=0; J <= mJ[I]; J++) if (MaxErr[F][I][J] != 0) delete[] MaxErr[F][I][J];
			if (MaxErr[F][I] != 0) delete[] MaxErr[F][I];
		}
		delete[] MaxErr[F];
	}
	delete[] MaxErr;
	DestroyFitData(E, K, mJ, mv, NEQ, EQS, EObs, EDiff, Sig, Err, QN);
	destroyFitData();*/
	return 0.0; // FQS;
}

double Potential::FitTangToenniesPot(bool showResult)
{
	if (!Worker->isTTPot())
	{
		int n, NA, NC, numPoints = 10000, CNC, *pCLRC;
		double Re, Te, *CLRC, b;
		bool ExcInt;
		QString F;
		Potential *nPot = MW->CreatePotential();
		TangToenniesPot *TTPot = new TangToenniesPot(nPot->Fit);
		QDialog *D = new QDialog(this);
		QGridLayout *L = new QGridLayout(D);
		QLineEdit *NPE = new QLineEdit("6", D), *NAE = new QLineEdit("7", D);
		QPushButton *OK = new QPushButton("OK", D), *Cancel = new QPushButton("Cancel", D);
		QCheckBox *ExcIntB = new QCheckBox("Use exchange interaction", D);
		D->setWindowTitle("Create Tang-Toennies potential");
		L->addWidget(new QLabel("Number of A_i coefficients:", D), 0, 0);
		L->addWidget(NAE, 0, 1);
		L->addWidget(new QLabel("Number of C_i coefficients:", D), 1, 0);
		L->addWidget(NPE, 1, 1);
		L->addWidget(ExcIntB, 2, 0, 1, 2);
		ExcIntB->setChecked(Worker->isExcIntAvailable());
		ExcIntB->setEnabled(Worker->isExcIntAvailable());
		L->setRowMinimumHeight(3, 20);
		L->addWidget(OK, 4, 0);
		L->addWidget(Cancel, 4, 1);
		connect(OK, SIGNAL(clicked()), D, SLOT(accept()));
		connect(Cancel, SIGNAL(clicked()), D, SLOT(reject()));
		do
		{
			if (D->exec() == QDialog::Rejected)
			{
				delete D;
				return 0.0;
			}
			NA = NAE->text().toInt();
			NC = NPE->text().toInt();
			if (NC < 3 || NC > 20) 
				QMessageBox::information(this, "MolSpektAnalysis", 
					"For NA and NC you must choose valid integer numbers between 1 and 10 and 3 and 20, respectively");
		}
		while (NC < 3 || NC > 20 || NA <= 0 || NA > 10);
		ExcInt = ExcIntB->isChecked();
		delete D;
		getMinimum(Re, Te);
		getLRCoeffForReading(CNC, pCLRC, CLRC);
		if (Worker->isAnaPotAv() && ExcInt)
		{
			double A, b, g;
			AnaPot *aPot = Worker->getAnaPotForReading(); 
			aPot->getExchangeCoeff(A, b, g);
			Worker->unlock();
			TTPot->setExcInt(A, b, g);
		}
		TTPot->setNA(NA);
		if (CNC >= 3 ? !TTPot->calcPotCoeff(CLRC, 3, NC, Re, Te) : true)
		{
			delete TTPot;
			delete[] pCLRC;
			delete[] CLRC;
			QMessageBox::information(this, "MolSpektAnalysis", "Error calculating potential parameters!");
			return 0.0;
		}
		delete[] pCLRC;
		delete[] CLRC;
		double *R = new double[numPoints], *U = new double[numPoints], Uinf = Worker->getUinf(), Ri = getMinR(Uinf, 1e-3), Ra = 50.0;
		double *Sig = new double[numPoints];
		double h = (Ra - Ri) / (numPoints - 1), FQS;
		for (n=0; n < numPoints; n++)
		{
			R[n] = (n==0 ? Ri : R[n-1] + h);
			U[n] = getPoint(R[n]);
			Sig[n] = 1e-4;
		}
		nPot->setWorker(TTPot);
		State->addPotential(nPot);
		double b1 = 0.0, b2 = TTPot->getb(), b3 = b2 + 0.1;
		double FQS2 = TTPot->improvePotential(numPoints, R, U, Sig);
		double FQS1 = 0.0, FQS3 = TTPot->improvePotential(numPoints, R, U, Sig, b3);
		while (FQS3 < FQS2 && b3 < 4.5)
		{
			b1 = b2;
			FQS1 = FQS2;
			b2 = b3;
			FQS2 = FQS3;
			FQS3 = TTPot->improvePotential(numPoints, R, U, Sig, b3 += 0.1);
		}
		if (b1 == 0.0) 
		{
			FQS1 = TTPot->improvePotential(numPoints, R, U, Sig, b1 = b2 - 0.1);
			while (FQS1 < FQS2 && b1 >= 0.4)
			{
				b3 = b2;
				FQS3 = FQS2;
				b2 = b1;
				FQS2 = FQS1;
				FQS1 = TTPot->improvePotential(numPoints, R, U, Sig, b1 -= 0.1);
			}
		}
		ParabInterpol(b1, FQS1, b2, FQS2, b3, FQS3, b, FQS);
		for (n=0; n < 25 && fabs(FQS2 - FQS) >= 1e-8; n++)
		{
			if (b > b2)
			{
				b1 = b2;
				FQS1 = FQS2;
			}
			else 
			{
				b3 = b2;
				FQS3 = FQS2;
			}
			FQS2 = TTPot->improvePotential(numPoints, R, U, Sig, b2 = b);
			ParabInterpol(b1, FQS1, b2, FQS2, b3, FQS3, b, FQS);
		}
		FQS = TTPot->improvePotential(numPoints, R, U, Sig, b);
		//b1 = 0.0; b2 = b; b3 = b2 + 0.01; 
		FQS2 = nPot->FitTangToenniesPot(false);
		/*TTPot->improvePotential(numPoints, R, U, Sig, b3);
		FQS3 = nPot->FitTangToenniesPot(false);
		while (FQS3 < FQS2 && b3 < 4.5)
		{
			b1 = b2;
			FQS1 = FQS2;
			b2 = b3;
			FQS2 = FQS3;
			TTPot->improvePotential(numPoints, R, U, Sig, b3 += 0.01);
			FQS3 = nPot->FitTangToenniesPot(false);
		}
		if (b1 == 0.0) 
		{
			TTPot->improvePotential(numPoints, R, U, Sig, b1 = b2 - 0.01);
			FQS1 = nPot->FitTangToenniesPot(false);
			while (FQS1 < FQS2 && b1 >= 0.4)
			{
				b3 = b2;
				FQS3 = FQS2;
				b2 = b1;
				FQS2 = FQS1;
				TTPot->improvePotential(numPoints, R, U, Sig, b1 -= 0.01);
				FQS1 = nPot->FitTangToenniesPot(false);
			}
		}
		ParabInterpol(b1, FQS1, b2, FQS2, b3, FQS3, b, FQS);
		for (n=0; n < 25 && fabs(FQS2 - FQS) >= 1e-8; n++)
		{
			if (b > b2)
			{
				b1 = b2;
				FQS1 = FQS2;
			}
			else 
			{
				b3 = b2;
				FQS3 = FQS2;
			}
			TTPot->improvePotential(numPoints, R, U, Sig, b2 = b);
			FQS2 = nPot->FitTangToenniesPot(false);
			ParabInterpol(b1, FQS1, b2, FQS2, b3, FQS3, b, FQS);
		}
		TTPot->improvePotential(numPoints, R, U, Sig, b);
		FQS = nPot->FitTangToenniesPot();*/
		delete[] R;
		delete[] U;
		delete[] Sig;
		if (Worker->isAdCorrA())
		{
			int NAdCorr, TAdCorr, PAdCorr;
			double *adCorr, Rm, b, RIso1, RIso2;
			bool *adCorrFree;
			Worker->getAdCorrForReading(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, &adCorrFree);
			nPot->Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, adCorrFree);
		}
		nPot->UpdateTab();
		nPot->setName(getName() + "TTPot");
		nPot->setSource("Potential from " + getSource());
		F = getFileName();
		nPot->setFileName(F.right(F.length() - F.indexOf('.') - 1) + "TTPot.pot");
		nPot->show();
		//return nPot->FitAnaPot();
		return FQS;
	}
	else
	{
		double FQS, WeightSum, *WF;
		int NWF, *WFR, NumFreePar;
		if (molecule->getNumIso() == 0)
		{
			QMessageBox::information(this, "MolSpektAnalysis", "The potential has to be assigned to a molecule with atomic data first!");
			return 0.0;
		}
		setFitData();
		FQS = Worker->FitTangToenniesPot();
		if (showResult)
		{
			Worker->getFitResult(FQS, WeightSum, NWF, WFR, WF, NumFreePar);
			QMessageBox::information(this, tr("Fit result"), 
									 "Initial chiSq=" + QString::number(FQSv, 'g', 6) + ", final chiSq="
									 + QString::number(FQS, 'g', 6) + ", Sigma=" 
									 + QString::number(sqrt(FQS/double(WeightSum - NumFreePar))));
			Worker->unlockBadList();
		}
		UpdateTab();
		Changed();
		return FQS;
	}
}

double Potential::FitIsoMass(int Iso, double InitStepSize, double& RIsoMass, int NumWFPoints)
{
	setFitData();
    return Worker->FitIsoMass(Iso, InitStepSize, RIsoMass, NumWFPoints);
}

double PotWorker::FitIsoMass(int FIso, double ISS, double &RMass, int NumWFPoints)
{
	int n, NLines = 0, J, aNAdcorr = NAdCorr, F, NFC = int(2.0 * el_S) + 1;
	double ****MaxErr, ****E, aIsoF = IsoF[FIso];
	double FQS = 8e99, FQSv = -1.0, UFact = 2e-18 * C_c * C_h * C_u / (C_hq * C_hq);
	NAdCorr = 0;
	estMinErr(MaxErr, NFC, mJ, mv, nEs, NEL, EL, nGS, nGL, GL);
	for (n=0; n < nEs; n++) NLines += NEL[n];
	for (n=0; n < nGS; n++) NLines += nGL[n];
	if (nEs + nGS == 0) return 0.0; 
	E = new double***[NFC];
	for (F=0; F < NFC; F++)
	{
		E[F] = new double**[NIso];
		E[F][FIso] = new double*[mJ[FIso] + 1];
		for (J=0; J <= mJ[FIso]; J++)
		{
			if (mv[FIso][J] >= 0) E[F][FIso][J] = new double[mv[FIso][J] + 1];
			else E[F][FIso][J] = 0;
		}
	}
	double M1 = 0.0, M2 = IsoT->redMass[FIso], M3 = M2 + ISS;
    double FQS2 = IsoMFit(E, MaxErr, FQSv, FIso, UFact, M2, NumWFPoints);
    double FQS1, FQS3 = IsoMFit(E, MaxErr, FQSv, FIso, UFact, M3, NumWFPoints);
	while (FQS3 < FQS2)
	{
		M1 = M2;
		FQS1 = FQS2;
		M2 = M3;
		FQS2 = FQS3;
        FQS3 = IsoMFit(E, MaxErr, FQSv, FIso, UFact, M3 += ISS, NumWFPoints);
	}
	if (M1 == 0.0) 
	{
        FQS1 = IsoMFit(E, MaxErr, FQSv, FIso, UFact, M1 = M2 - ISS, NumWFPoints);
		while (FQS1 < FQS2)
		{
			M3 = M2;
			FQS3 = FQS2;
			M2 = M1;
			FQS2 = FQS1;
            FQS1 = IsoMFit(E, MaxErr, FQSv, FIso, UFact, M1 -= ISS, NumWFPoints);
		}
	}
	ParabInterpol(M1, FQS1, M2, FQS2, M3, FQS3, RMass, FQS);
	for (n=0; n < 25 && fabs(FQS2 - FQS) >= 1e-8; n++)
	{
		if (RMass > M2)
		{
			M1 = M2;
			FQS1 = FQS2;
		}
		else 
		{
			M3 = M2;
			FQS3 = FQS2;
		}
        FQS2 = IsoMFit(E, MaxErr, FQSv, FIso, UFact, M2 = RMass, NumWFPoints);
		ParabInterpol(M1, FQS1, M2, FQS2, M3, FQS3, RMass, FQS);
	}
    FQS = IsoMFit(E, MaxErr, FQSv, FIso, UFact, RMass, NumWFPoints);
	for (F=0; F < NFC; F++)
	{
		for (J=0; J <= mJ[FIso]; J++) if (MaxErr[F][FIso][J] != 0)
		{
			delete[] MaxErr[F][FIso][J];
			delete[] E[F][FIso][J];
		}
		if (MaxErr[F][FIso] != 0) delete[] MaxErr[F][FIso];
		delete[] E[F][FIso];
		delete[] MaxErr[F];
		delete[] E[F];
	}
	delete[] MaxErr;
	delete[] E;
	destroyFitData();
	NAdCorr = aNAdcorr;
	IsoF[FIso] = aIsoF;
	return FQS;
}

double PotWorker::IsoMFit(double**** E, double**** MaxErr, double& FQSv, int Iso, double UFakt, double Mass, int NumWFPoints)
{
	IsoF[Iso] = UFakt * Mass;
    calcE(E, mJ, mv, Mv, NumWFPoints, MaxErr);
    double FQS = calcFQS(E, nEs, NEL, EL, nGS, nGP, nGL, PS, GL, NumWFPoints);
	if (FQSv == -1.0) FQSv = FQS;
	return FQS;
}

void Potential::calcDerivatives(TableLine*, int, double*, double**)
{
    /*double h = (rmax - rmin) / double(NumWFPoints - 1), r, *Ev, **WF = 0, *F, *NF, S;
	int n, m, c, N, NC, NLRC, MaxPN = 0, J = -1, I = -1, Nv, *ri, *ra, *rc, v, NS = 0;
	if (LPot != 0) LPot->getNC(N, NC, NLRC);
	else aPot->getNC(N, NC, NLRC);
    double **DR = Create(NumWFPoints, N);
    if (LPot != 0) for (n=0, r = rmin; n < NumWFPoints; n++, r+=h)
	{
		LPot->getDerivatives(r, S, DR[n] + 3, DR[n] + NC + 3, DR[n][2], DR[n][1]);
		DR[n][0] = 1.0 - DR[n][1];
		DR[n][1] = -DR[n][0];
	}
    else for (n=0, r = rmin; n < NumWFPoints; n++, r+=h)
		aPot->getDerivatives(r, DR[n], DR[n] + NC);
	for (n=0; n < NL; n++)
	{
		if (Lines[n].Iso != I || Lines[n].Jss != J)
		{
			if (WF != 0)
			{
				delete[] Ev;
				Destroy(WF, Nv);
				delete[] F;
				delete[] NF;
				delete[] ri;
				delete[] ra;
				delete[] rc;
			}
			I = Lines[n].Iso;
			J = Lines[n].Jss;
			for (m=n, Nv = 1; (m < NL ? Lines[m].Iso == I && Lines[m].Jss == J : false); m++) if (Lines[m].vss >= Nv)
				Nv = Lines[m].vss + 1;
			getWaveFunc(I, J, Nv, Ev, WF, ri, ra, rc, F, NF);
		}
		if ((v = Lines[n].vss) >= Nv)
		{
			E[n] = -1.0;
			continue;
		}
		for (c=0; c < N; c++)
		{
			for (m = ri[v], S = 0.0; m <= rc[v]; m++) S += WF[v][m] * DR[m][c] * WF[v][m];
			for (S *= F[v] * F[v]; m < ra[v]; m++) S += WF[v][m] * DR[m][c] * WF[v][m];
			dyda[n][c] = NF[v] * S * NF[v];
		}
		E[n] = Ev[v];
		if (!Lines[n].isTE)
		{
			if (Lines[n].PN > MaxPN) MaxPN = Lines[n].PN;
			if (Lines[n].SourceN >= NS) NS = Lines[n].SourceN + 1;
		}
	}
	if (WF != 0)
	{
		delete[] Ev;
		Destroy(WF, Nv);
		delete[] F;
		delete[] NF;
		delete[] ri;
		delete[] ra;
		delete[] rc;
	}
    Destroy(DR, NumWFPoints);
	double W, **SE = Create(NS, MaxPN + 1), **WS = Create(NS, MaxPN + 1);
	//double ***SD = Create(NS, MaxPN + 1, N);
	for (m=0; m < NS; m++) for (n=0; n <= MaxPN; n++)
	{
		SE[m][n] = WS[m][n] = 0.0;
		//for (c=0; c<N; c++) SD[m][n][c] = 0.0;
	}
	for (n=0; n < NL; n++) if (!Lines[n].isTE)
	{
		E[n] *= -1.0;
		W = 1.0 / (Lines[n].err * Lines[n].err);
		SE[Lines[n].SourceN][Lines[n].PN] += W * E[n];
		WS[Lines[n].SourceN][Lines[n].PN] += W;
		//for (c=0; c < N; c++) SD[Lines[n].SourceN][Lines[n].PN][c] += W * dyda[n][c];
	}
	for (m=0; m < NS; m++) for (n=0; n <= MaxPN; n++) if (WS[m][n] != 0.0)
	{
		WS[m][n] = 1.0 / WS[m][n];
		SE[m][n] *= WS[m][n];
		//for (c=0; c<N; c++) SD[m][n][c] *= WS[m][n];
	}
	for (n=0; n < NL; n++) if (!Lines[n].isTE)
	{
		E[n] -= SE[Lines[n].SourceN][Lines[n].PN];
		for (c=0; c<N; c++) dyda[n][c] *= -1.0;// -= SD[Lines[n].SourceN][Lines[n].PN][c];
	}
	Destroy(SE, NS);
	Destroy(WS, NS);
	//Destroy(SD, NS, MaxPN + 1);*/
}

void Potential::setMolecule(Molecule *Mol)
{
	if (Fit->isRunning())
	{
		QMessageBox::information(this, "MolSpektAnalysis", "The molecular assignment cannot be changed while a fit is running!");
		return;
	}
	if (Mol != 0)
	{
		double UFact = 2e-18 * C_c * C_h * C_u / (C_hq * C_hq), *MU, *IsoF;
		int i, NIso;
		IsoTab *IsoT = Mol->getIso();
		if (IsoT != 0)
		{
			NIso = IsoT->numIso;
			MU = new double[NIso];
			IsoF = new double[NIso];
			for (i=0; i < NIso; i++)
			{	
				MU[i] = IsoT->redMass[i];
				IsoF[i] = MU[i] * UFact;
			}
			Worker->setMolData(IsoT, MU, IsoF);
		}
	}
	TableWindow::setMolecule(Mol);
}

bool PotWorker::setMolData(IsoTab* nIsoT, double* nMU, double* nIsoF)
{
	if (Fit->isRunning()) return false;
	if (IsoT != 0) delete IsoT;
	if (NIso > 0)
	{
		delete[] MU;
		delete[] IsoF;
	}
	IsoT = nIsoT;
	NIso = IsoT->numIso;
	MU = nMU;
	IsoF = nIsoF;
	return true;
}

void PotWorker::takeMolData(IsoTab*& rIsoT, double*& rMU, double*& rIsoF)
{
	if (Fit->isRunning())
	{
		rIsoT = 0;
		rMU = 0;
		rIsoF = 0;
		return;
	}
	rIsoT = IsoT;
	IsoT = 0;
	rMU = MU;
	MU = 0;
	rIsoF = IsoF;
	IsoF = 0;
	NIso = 0;
}

double Potential::calcScatWaveFunc(bool write, double **Nodes, double maxR)		
{
	IsoTab *IsoT = Worker->getIsoT();
	if (IsoT == 0)
	{
		printf("Potential::calcScatWaveFunc(): Error: not enough data available!\n");
		return 0.0;
	}
	QTextStream S;
	if (write)
	{
		QString fileName = QFileDialog::getSaveFileName(this, tr("Write scattering wave functions"), 
					MW->getDir(PotData), tr("Text tables (*.dat)"));
		QFile DDatei(fileName);
		if (!DDatei.open(QIODevice::WriteOnly))
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), tr("Error while opening file!"));
			return 0.0;
		}
		S.setDevice(&DDatei);
		S << "Scattering wave functions:\nR[A]";
	}
	int i, I, NI = IsoT->numIso, xi, Np;
	double RMin = 3.0, RMax = 103.0, R, M[NI], h = 0.001;
	double Y[NI][3], WF[NI][3], *U = getPoints(RMin, RMax, Np = int((RMax - RMin) / h) + 1);
	//printf("M=%f, h=%f, I=%f\n", M, h, IsoF[Iso]);
	bool gn[NI];
	int v[NI];
	double LE, AE, NE, *IsoF = Worker->getIsoF_ForReading();
	for (xi=0; U[xi] > 0.0 && xi < Np - 1; xi++) ;
	//printf("xi=%d\n", xi);
	for (LE = 0.0, AE = 1e-5, xi--; xi > 0 && AE < 1e10; xi--)
	{
		NE = AE * (2.0 + M[0] * U[xi]) - LE;
		LE = AE;
		AE = NE;
	}
	for (I=0; I < NI; I++) 
	{
		gn[I] = true;
		v[I] = 0;
		M[I] = IsoF[I] * h * h;
		Y[I][1] = 1e-5;
		Y[I][0] = Y[I][1] * exp(-sqrt(M[I]*U[xi]));
		WF[I][0] = Y[I][0] * PFakt(M[I], U[xi]);
		if (write) S << "\t" << IsoT->mNumIso1[I] << IsoT->chSymb1 << IsoT->mNumIso2[I] << IsoT->chSymb2;
	}
	//printf("xi=%d, U[xi]=%f\n", xi, U[xi]);
	if (write) S << "\n";
	for (i = xi + 1, R = RMin + h * double(xi); 
			R <= maxR && (RMin > 10.0 || i > xi + 50 ? fabs(M[0]*U[i]) > 1e-16 : true); i++)
	{
		//printf("Beginn\n");
		R += h;
		if (write) S << R;
		for (I=0; I < NI; I++)
		{
			//printf("i=%d\n", i);
			NCNF(M[I], U[i-1], Y[I] + 2, WF[I] + 2, -1);
			//printf("Nach NCNF\n");
			if (gn[I] && Y[I][2] < 0.0)
			{
				gn[I] = false;
				if (Nodes != 0 && v[I] <= cMaxv) Nodes[I][v[I]] = R - h * Y[I][2] / (Y[I][2] - Y[I][1]);
				v[I]++;
			}
			else if (!gn[I] && Y[I][2] > 0.0)
			{
				gn[I] = true;
				if (Nodes != 0 && v[I] <= cMaxv) Nodes[I][v[I]] = R - h * Y[I][2] / (Y[I][2] - Y[I][1]);
				v[I]++;
			}
			//printf("Vor write, WF[I]=%x, WF[I][2]=%f\n", WF[I], WF[I][2]);
			if (write) S << "\t" << QString::number(WF[I][1], 'e', 10);
			//printf("Nach write\n");
			Y[I][0] = Y[I][1];
			Y[I][1] = Y[I][2];
			WF[I][0] = WF[I][1];
			WF[I][1] = WF[I][2];
		}
		if (write) S << "\n";
		//if (i==18) break;
		if (i == Np - 1)
		{
			//printf("Vor delete\n");
			delete[] U;
			U = getPoints(RMin += 100.0, RMax += 100.0, Np);
			i=0;
		}
		//printf("Ende\n");
	}
	if (Nodes != 0) for (I=0; I < NI; I++) for (i = v[I]; i <= cMaxv; i++) Nodes[I][i] = 0.0;
	//for (I=0; I<NI; I++) 
		//printf("%d%s%d%s: Maxv=%d\n", IsoT->mNumIso1[I], IsoT->chSymb1.ascii(),
			//   IsoT->mNumIso2[I], IsoT->chSymb2.ascii(), v[I]);
	//printf("Vor delete\n");
	delete[] U;
	return R;
}

void Potential::NCNF(double M, double E, double *Y, double *W, int d)
{
	//printf("W=%x, Y=%x\n", W, Y);
	W[d] = Y[d] * PFakt(M, E);
	Y[0] = M * E * W[d] + 2.0 * Y[d] - Y[2*d];
}

double Potential::PFakt(double M, double E)
{
	return 1.0 / (1.0 - 0.0833333333333 * M * E);
}

bool Potential::readData(QString Filename)
{
	//printf("Potential::readData\n");
	Tab->setColumnCount(4);
	QFile Datei(Filename);
	if (!read(&Datei)) return false;
	Tab->blockSignals(true);
	int n, r, cc, lc = 0;
	bool Success = true;
	QString Buffer;
	QTextStream S(&Datei);
	QString Spacer = " | ";
	QStringList L;
	if ((Buffer = S.readLine()).left(8) != "Source: ") Success = false; 
	else 
	{
		setSource(Buffer.right(Buffer.length() - 8));
		if ((Buffer = S.readLine()).left(6) != "Name: ") Success = false;
		else
		{
			setName(Buffer.right(Buffer.length() - 6));
			for (r=0, cc = Tab->columnCount(); !S.atEnd(); r++)
			{
				if (Tab->rowCount() == r) Tab->setRowCount(r + 100);
				if ((Buffer = S.readLine()).left(15) == "Column titles: ") Buffer = S.readLine();
				if (Buffer.left(21) == "Coupling information:") break;
				L = Buffer.split(Spacer);
				if ((lc = L.count()) > cc) Tab->setColumnCount(cc = lc);
				for (n=0; n < lc; n++) Tab->setItem(r, n, new QTableWidgetItem(L[n]));
				while (n < cc) Tab->setItem(r, n++, new QTableWidgetItem(""));
			}
			if (!S.atEnd())
			{
				r-=2;
				Buffer = S.readLine();
                Filename = Buffer.right(Buffer.length() - 16);
                QString MolFile, CurFile;
                if (molecule != 0) Filename = getAbsolutePath(Filename, MolFile = molecule->getFileName());
				S.readLine();
				S.readLine();
				if (CoupledComp != 0) delete[] CoupledComp;
				CoupledComp = new int[10];
				if (CoupledPot != 0) delete[] CoupledPot;
				CoupledPot = new Potential*[10];
				for (NCoupled = 0; NCoupled < 10 && !S.atEnd(); NCoupled++)
				{
					n = (Buffer = S.readLine()).indexOf(" | ");
                    CoupledPot[NCoupled] = MW->getPotential((molecule != 0 ? getAbsolutePath(CurFile = Buffer.left(n), MolFile)
                                                                           : Buffer.left(n)), molecule);
					CoupledComp[NCoupled] = Buffer.right(Buffer.length() - n - 3).toInt();
					if (CoupledPot[NCoupled] != 0 ? CoupledPot[NCoupled]->getCoupledSineWaveFunc() != 0 : false)
						WaveFuncs = CoupledPot[NCoupled]->getCoupledSineWaveFunc();
				}
				if (WaveFuncs == 0)
				{
					WaveFuncs = new CoupledSineWaveFunc(MW, molecule);
					if (!WaveFuncs->readData(Filename))
					{
						delete WaveFuncs;
						WaveFuncs = 0;
					}
				}
				if (WaveFuncs != 0) WaveFuncs->Assign();
			}
			if (lc == 0) r--;
			Tab->setRowCount(r);
		}
	}
	if (!Success) 
	{
		Tab->blockSignals(false);
		if (!readPointPot(Filename)) if (!readPotData(Filename)) if (!readCoupledPotData(Filename)) return false;
	}
	else 
	{
		//printf("TableWindow::readData successfull\n");
		UpdatePot(-2);
	}
	Tab->blockSignals(false);
	Saved();
	//printf("Ende Potential::readData\n");
	return true;
}

bool Potential::testIfSplinePot()
{
	bool E1, E2;
	double D1 = Tab->item(0, 0)->text().toDouble(&E1), D2 = Tab->item(1, 0)->text().toDouble(&E2);
	return E1 && E2 && D1 < D2;
}

double **Potential::fitNodePos()
{
	AnaPot *aPot = Worker->getAnaPotForWriting();
	if (aPot == 0) 
	{
		QMessageBox::information(this, tr("QT4MolSpektAn"), 
					tr("fitNodePos: fatal error, no analytical potential data available or a different fit is running!"));
		return 0;
	}
	IsoTab *IsoT = Worker->getIsoT();
	if (IsoT == 0)
	{
		QMessageBox::information(this, tr("QT4MolSpektAn"),
								 tr("fitNodePos: fatal error, no isotopic data available!"));
		return 0;
	}
	int NI = IsoT->numIso, NC = Tab->rowCount();
	double D, P, oN, R, F;
	double b, De, iCoeff, iOff, iExp, Ra, Ri, Rm, Tm, *LRCoeff, *PotCoeff;
	double **NodePos = Create(NI, cMaxv+1);
	int nPotCoeff, nLRCoeff, *pLRCoeff, n;
	bool error = true;
	aPot->getPotCoeff(b, Rm, Tm, De, nPotCoeff, PotCoeff);
	aPot->getInnerWall(Ri, iOff, iExp, iCoeff);
	aPot->getLRCoeffForWriting(nLRCoeff, pLRCoeff, LRCoeff, 0, &Ra);
	QDialog *NFDialog = new QDialog(this);
	NFDialog->setMinimumSize(300, 270);
    NFDialog->setMaximumSize(300, 270);
    QLabel *Label1 = new QLabel("Isotopomer:", NFDialog);
    Label1->setGeometry(10, 10, 120, 20);
    QLabel *Label2 = new QLabel("Number of node:", NFDialog);
    Label2->setGeometry(10, 40, 120, 20);
	QLabel *Label3 = new QLabel("Value to fit to [A]:", NFDialog);
	Label3->setGeometry(10, 70, 120, 20);
	QLabel *Label6 = new QLabel("Parameter to vary:", NFDialog);
	Label6->setGeometry(10, 100, 120, 20);
	QLabel *Label4 = new QLabel("Wanted accuracy [A]:", NFDialog);
	Label4->setGeometry(10, 130, 120, 20);
	QLabel *Label5 = new QLabel("Max search radius [A]:", NFDialog);
	Label5->setGeometry(10, 160, 120, 20);
	QLabel *Label7 = new QLabel("Max iterations:", NFDialog);
	Label7->setGeometry(10, 190, 120, 20);
    QComboBox *isotopomer = new QComboBox(NFDialog);
    isotopomer->setGeometry(140, 10, 150, 20);
    QLineEdit *number = new QLineEdit(QString::number(fitN_N), NFDialog);
    number->setGeometry(140, 40, 150, 20);
	QLineEdit *value = new QLineEdit(QString::number(fitN_V, 'f', 1), NFDialog);
	value->setGeometry(140, 70, 150, 20);
	QComboBox *parameter = new QComboBox(NFDialog);
	parameter->setGeometry(140, 100, 150, 20);
	QLineEdit *accuracy = new QLineEdit(QString::number(fitN_A, 'f', 3), NFDialog);
	accuracy->setGeometry(140, 130, 150, 20);
	QLineEdit *maxr = new QLineEdit(QString::number(fitN_maxr, 'f', 1), NFDialog);
	maxr->setGeometry(140, 160, 150, 20);
	QLineEdit *maxIt = new QLineEdit(QString::number(fitN_MI), NFDialog);
	maxIt->setGeometry(140, 190, 150, 20);
    QPushButton *OK = new QPushButton("OK", NFDialog);
    OK->setGeometry(10, 240, 80, 20);
    QPushButton *Cancel = new QPushButton("Cancel", NFDialog);
    Cancel->setGeometry(210, 240, 80, 20);
	for (n=0; n < NI; n++) 
		isotopomer->addItem(QString::number(IsoT->mNumIso1[n]) + *IsoT->chSymb1 
				+ QString::number(IsoT->mNumIso2[n]) + *IsoT->chSymb2);
	isotopomer->setEditable(false);
	isotopomer->setCurrentIndex(fitN_I);
	for (n=0; n < nPotCoeff; n++) parameter->addItem(Tab->item(n, 0)->text());
	parameter->addItem(Tab->item(nPotCoeff + 3, 0)->text());
	for (n = nPotCoeff + 6; n < NC - 3; n++) parameter->addItem(Tab->item(n, 0)->text());
	parameter->setEditable(false);
	parameter->setCurrentIndex(fitN_C);
    connect(OK, SIGNAL(clicked()), NFDialog, SLOT(accept()));
    connect(Cancel, SIGNAL(clicked()), NFDialog, SLOT(reject()));
	while (error)
	{
		if (NFDialog->exec() == QDialog::Rejected) return 0;
		fitN_I = isotopomer->currentIndex();
		fitN_N = number->text().toInt();
		if (fitN_N <= 0 || fitN_N > cMaxv) 
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), 
									 tr("The selected number of the node is invalid!"));
			continue;
		}
		if ((fitN_V = value->text().toDouble()) <= rmin)
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), 
									 tr("The selected value to fit to is invalid!"));
			continue;
		}
		fitN_C = parameter->currentIndex();
		if ((fitN_A = accuracy->text().toDouble()) < 1e-15)
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), 
									 tr("The selected accuracy is invalid!"));
			continue;
		}
		if ((fitN_maxr = maxr->text().toDouble()) < fitN_V)
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), 
								tr("The selected maximum radius is smaller than the value to fit to!"));
			continue;
		}
		fitN_MI = maxIt->text().toInt();
		if ((R = calcScatWaveFunc(false, NodePos, fitN_maxr)) < fitN_V)
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), 
									QString("The selected value to fit to is outside the R range which can be calculated of Rmax=%1!").arg(R, 6, 'f', 4));
			continue;
		}
		if (NodePos[fitN_I][fitN_N-1] == 0.0) 
		{
			QMessageBox::information(this, tr("QT4MolSpektAn"), tr("The selected node does not exist!"));
			continue;
		}
		error = false;
	}
	if (fitN_C < nPotCoeff) P = PotCoeff[fitN_C];
	else if (fitN_C == nPotCoeff) P = Ri;
	else if (fitN_C == nPotCoeff + 1) P = iCoeff;
	else if (fitN_C == nPotCoeff + 2) P = Ra;
	else P = LRCoeff[fitN_C - nPotCoeff - 3];
	P += (D = 0.1 * P);
	printf("fitN_MI=%d, Node=%f, fitN_V=%f, fitN_A=%f\n", fitN_MI, NodePos[fitN_I][fitN_N-1], fitN_V, fitN_A);
	for (n=0; n < fitN_MI && fabs((oN = NodePos[fitN_I][fitN_N-1]) - fitN_V) > fitN_A; n++)
	{
		printf("n=%d, D=%g, P=%g\n", n, D, P);
		printf("Node=%g, V=%g\n", NodePos[fitN_I][fitN_N-1], fitN_V);
		if (fitN_C < nPotCoeff) PotCoeff[fitN_C] = P;
		else if (fitN_C == nPotCoeff) 
		{
			Ri = P;
			aPot->setInnerWall(Ri, iOff, iExp, iCoeff);
		}
		else if (fitN_C == nPotCoeff + 1) 
		{
			iCoeff = P;
			aPot->setInnerWall(Ri, iOff, iExp, iCoeff);
		}		
		else if (fitN_C == nPotCoeff + 2) 
		{
			Ra = P;
			aPot->setLRCoeff(Ra, nLRCoeff, pLRCoeff, LRCoeff);
		}
		else LRCoeff[fitN_C - nPotCoeff - 3] = P;
		aPot->sConnect();
		calcScatWaveFunc(false, NodePos, fitN_maxr);
		if (NodePos[fitN_I][fitN_N-1] == 0.0) D -= (P *= 0.5);
		else 
		{
			if ((F = (fitN_V - NodePos[fitN_I][fitN_N-1]) / (NodePos[fitN_I][fitN_N-1] - oN)) > 10.0) F = 10.0;
			else if (F < -10.0) F = -10.0;
			P += (D *= F);
		}
	}
	if (fabs((oN = NodePos[fitN_I][fitN_N-1]) - fitN_V) > fitN_A) 
		QMessageBox::information(this, tr("QT4MolSpektAn"), 
								 tr("The fit was not successful within the given number of iterations!"));
	else QMessageBox::information(this, tr("QT4MolSpektAn"), 
								 tr("The fit was successful!"));
	UpdateTab();
	delete NFDialog;
	Changed();
	return NodePos;
}

void PotWorker::getSpinRGamma(int& rNSpinRGamma, double*& rSpinRGamma, double& rSpinRR1, double& rSpinRR2, double& rSpinRb, 
							  double& rSpinRRm)
{
	rNSpinRGamma = NSpinRGamma;
	rSpinRGamma = SpinRGamma;
	rSpinRR1 = SpinRR1;
	rSpinRR2 = SpinRR2;
	rSpinRb = SpinRb;
	rSpinRRm = SpinRRm;
}

void PotWorker::setSpinRGamma(int nNSpinRGamma, double* nSpinRGamma, double nSpinRR1, double nSpinRR2, double nSpinRb, double nSpinRRm)
{
	NSpinRGamma = nNSpinRGamma;
	if (SpinRGamma != 0) delete[] SpinRGamma;
	SpinRGamma = nSpinRGamma;
	SpinRR1 = nSpinRR1;
	SpinRR2 = nSpinRR2;
	SpinRb = nSpinRb;
	SpinRRm = nSpinRRm;
}

void Potential::getTexTable(int NumWFPoints, double FQS)
{
	double b, De, iCoeff, iOff, iExp, Ra, Ri, Rm, Tm, *LRCoeff, *PotCoeff, MinR, MinU, U_inf;
	int nPotCoeff, nLRCoeff, *pLRCoeff, n, m, l, p, q, prec = 0, lang, nDig, r;
	QStringList L;
	QString Buffer, eB, B2;
	QTableWidgetItem *I;
	QDialog *D = new QDialog(this);
	QGridLayout *La = new QGridLayout(D);
	D->setWindowTitle("Select language and precision");
	La->addWidget(new QLabel("Precision [cm^-1]:", D), 0, 0);
	QLineEdit *PrecE = new QLineEdit("0.0001", D);
	La->addWidget(PrecE, 0, 1);
	La->addWidget(new QLabel("Language:", D), 1, 0);
	QComboBox *LangB = new QComboBox(D);
	LangB->setEditable(false);
	LangB->addItems(QStringList() << "english" << "german");
	La->addWidget(LangB, 1, 1);
	La->setRowMinimumHeight(2, 20);
	QPushButton *OK = new QPushButton("OK", D), *Cancel = new QPushButton("Cancel", D);
	La->addWidget(OK, 3, 0);
	La->addWidget(Cancel, 3, 1);
	connect(OK, SIGNAL(clicked()), D, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), D, SLOT(reject()));
	if (D->exec() == QDialog::Rejected)
	{
		delete D;
		return;
	}
	double Prec = PrecE->text().toDouble(), x, X;
	if (Prec > 0.0)
	{
		Prec = 1.0 / Prec;
		nDig = int(log10(Prec));
		if (pow(10, nDig) != Prec) nDig++;
	}
	else nDig = -1;
	lang = LangB->currentIndex();
	delete D;
	L << "\\begin{table}" << "\\caption{???}";
	L << "\\label{" + (State != 0 ? State->getName() : QString("???"))  + "Potential}";
	L << "\\centering";
	Tab->blockSignals(true);
	if (Worker->isAnaPotAv())
	{
		AnaPot *aPot = Worker->getAnaPotForReading();
		double q_e, q_f;
		int intIExp;
		Worker->getLambdaDoubling(q_e, q_f);
		aPot->getPotCoeff(b, Rm, Tm, De, nPotCoeff, PotCoeff);
		aPot->getInnerWall(Ri, iOff, iExp, iCoeff);
		aPot->getLRCoeffForReading(nLRCoeff, pLRCoeff, LRCoeff, &Ra, true);
		if ((x = abs((Ri - Rm) / (Ri + b * Rm))) < (X = abs((Ra - Rm) / (Ra + b * Rm)))) x=X;
		L << "\\begin{tabular}{lp{0.1\\textwidth}r}\\hline\\noalign{\\smallskip}";
		for (n=0, X=x; n < nPotCoeff; n++, X*=x) 
		{
			m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() - 1 : -1);
			if (Prec > 0.0 && m<0)
			{
				prec = int(log10(Prec * X * abs(PotCoeff[n])));
				if (prec < 0) prec = 0;
			}
			m = (Buffer = QString::number(PotCoeff[n], 'e', (m>=0 ? m : (Prec > 0.0 ? prec : 16)))).indexOf("e");
			Tab->item(n, 1)->setText(Buffer);
			Buffer = "$a_{" + QString::number(n+1) + "}$   && " + Buffer.left(m) + "$\\times 10^{" 
				+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) + "}$ cm$^{-1}$ \\\\";
			L << (lang == 1 ? Buffer.replace('.', ',') : Buffer);
		}
		L << "\\hline\\noalign{\\smallskip}";
		Buffer = QString::number(b, 'g');
		if (Buffer.indexOf('.') == -1) Buffer += ".0";
		Tab->item(n++, 1)->setText(Buffer);
		L << "$b$   && " 
				+ (lang == 1 ? Buffer.replace('.', ',') : Buffer)
				+ "    \\\\";
		Buffer = QString::number(Rm, 'g', 12);
		if (Buffer.indexOf('.') == -1) Buffer += ".0";
		Tab->item(n++, 1)->setText(Buffer);
		L << "$R_m$ && " 
				+ (lang == 1 ? Buffer.replace('.', ',') : Buffer)
				+ " \\AA \\\\ ";
		Buffer = QString::number(Tm, 'f', ((m = (I = Tab->item(n, 3)) != 0 ? I->text().toInt() : 0) > 0 ? m 
					: (nDig >= 0 ? nDig : 4)));
		Tab->item(n++, 1)->setText(Buffer);
		L << "$T_m$ && " 
				+ (lang == 1 ? Buffer.replace('.', ',') : Buffer)
				+ " cm$^{-1}$ \\\\ \\hline\\noalign{\\smallskip}";
		Buffer = QString::number(Ri, 'g');
		if (Buffer.indexOf('.') == -1) Buffer += ".0";
		Tab->item(n++, 1)->setText(Buffer);
		L << "$R_i$ && " 
				+ (lang == 1 ? Buffer.replace('.', ',') : Buffer)
				+ " \\AA \\\\";
		m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() : 0);
		if (Prec > 0.0 && m<=0)
		{
			prec = int(log10(Prec * iExp * iExp * iCoeff * pow(Ri, -iExp - 1.0)));
			if (prec < 0) prec = 0;
		}
		Buffer = QString::number(iExp, 'g', (m>0 ? m : (Prec > 0.0 ? prec + 1 : 11)));
		Tab->item(n++, 1)->setText(Buffer);
		intIExp = Buffer.toInt();
		L << "$s$   && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + "    \\\\";
		m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() : -1);
		if (Prec > 0.0 && m<=0)
		{
			prec = int(log10(Prec * abs(iOff)));
			if (prec < 0) prec = 0;
		}
		m = (Buffer = QString::number(iOff, 'e', (m>0 ? m : (Prec > 0.0 ? prec : 4)))).indexOf("e");
		Tab->item(n++, 1)->setText(Buffer);
		L << "$A$   && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
				+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) 
				+ "}$ cm$^{-1}$ \\\\";
		m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() : -1);
		if (Prec > 0.0 && m<=0)
		{
			prec = int(log10(abs(Prec * iCoeff * pow(Ri, -iExp))));
			if (prec < 0) prec = 0;
		}
		m = (Buffer = QString::number(iCoeff, 'e', (m>0 ? m : (Prec > 0.0 ? prec : 4)))).indexOf("e");
		Tab->item(n++, 1)->setText(Buffer);
		L << "$B$   && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
				+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) 
				+ "}$ cm$^{-1}$\\AA$^" + (intIExp > 0 ? QString::number(intIExp) : "s") 
				+ "$ \\\\ \\hline\\noalign{\\smallskip}";
		Buffer = QString::number(Ra, 'g');
		Tab->item(n++, 1)->setText(Buffer);
		L << "$R_a$ && " 
				+ (lang == 1 ? Buffer.replace('.', ',') : Buffer) 
				+ " \\AA \\\\";
		for (l=0; l < nLRCoeff; l++)
		{
			m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() - 1 : -1);
			if (Prec > 0.0 && m<0)
			{
				prec = int(log10(abs(Prec * LRCoeff[l] * pow(Ra, -pLRCoeff[l]))));
				if (prec < 0) prec = 0;
			}
			m = (Buffer = QString::number(LRCoeff[l], 'e', (m>=0 ? m : (Prec > 0.0 ? prec : 16)))).indexOf("e");
			Tab->item(n++, 1)->setText(Buffer);
			eB = QString::number(pLRCoeff[l]);
			L << "$C_{" + eB + "}$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
					+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) + "}$ cm$^{-1}$\\AA$^{" 
					+ eB + "}$ \\\\";
		}
		//if (De != 0.0) 
		//{
		Buffer = QString::number(De, 'f', ((m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() : -1)) > 0 ? m 
							: (Prec > 0 ? nDig : 4)));
		if (Buffer.indexOf('.') == -1) Buffer += ".0";
		Tab->item(n++, 1)->setText(Buffer);
		L << "$U_\\infty$ && " 
			+ (lang == 1 ? Buffer.replace('.', ',') : Buffer)
			+ " cm$^{-1}$ \\\\ \\hline\\noalign{\\smallskip}";
		//}
		/*L << "R$_{min}$   && " + QString::number(rmin, 'f', 
									((I = Tab->item(++n, 3)) != 0 ? I->text().toInt() : 1)) + " \\AA \\\\";
		L << "R$_{max}$   && " 
				+ QString::number(rmax, 'f', ((I = Tab->item(++n, 3)) != 0 ? I->text().toInt(): 1)) 
				+ " \\AA \\\\ \\hline";*/
		if (q_e != 0.0 || q_f != 0.0)
		{
			double *IsoF = Worker->getIsoF_ForReading();
			l = (State != 0 ? (State->getFitData() != 0 ? State->getFitData()->getMaxJ() : 300) : 300);
			if (q_e != 0.0)
			{
				m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() - 1 : -1);
				if (Prec > 0.0 && m<0)
				{
					prec = int(log10(abs(Prec * q_e * double(l * (l+1)) / (Ri * Ri * IsoF[0])))) + 1;
					if (prec < 0) prec = 0;
				}
				m = (Buffer = QString::number(q_e, 'g', (m>=0 ? m+1 : (Prec > 0.0 ? prec : 5)))).indexOf("e");
				if (Buffer.indexOf('.') == -1) Buffer += ".0";
				Tab->item(n++, 1)->setText(Buffer);
				if (m>0)  L << "$\\eta_e$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
								+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) + "}$\\\\";
				else L << "$\\eta_e$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + "\\\\";
			}
			if (q_f != 0.0)
			{
				m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() - 1 : -1);
				if (Prec > 0.0 && m<0)
				{
					prec = int(log10(abs(Prec * q_f * double(l * (l+1)) / (Ri * Ri * IsoF[0])))) + 1;
					if (prec < 0) prec = 0;
				}
				m = (Buffer = QString::number(q_f, 'g', (m>=0 ? m+1 : (Prec > 0.0 ? prec : 5)))).indexOf("e");
				if (Buffer.indexOf('.') == -1) Buffer += ".0";
				Tab->item(n++, 1)->setText(Buffer);
				if (m>0) L << "$\\eta_f$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
								+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) + "}$\\\\";
				else L << "$\\eta_f$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + "\\\\";
			}
			L << "\\hline\\noalign{\\smallskip}";
		}
		else l=-1;
		r=n;
		if (Worker->isAdCorrA())
		{
			int NAdCorr, TAdCorr, PAdCorr;
			double *adCorr, Rm, b, RIso1, RIso2;
			Worker->getAdCorrForReading(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, 0, true);
			X = pow(2.0 * Rm / (Ri + Rm), PAdCorr);
			if (NAdCorr > 1) x = (Ri - Rm) / (Ri + b * Rm);
			for (n=0; n < NAdCorr; n++, X*=x)
			{
				m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() - 1 : -1);
				if (Prec > 0.0 && m<0)
				{
					prec = int(log10(abs(Prec * adCorr[n] * X))) + 1;
					if (prec < 0) prec = 0;
				}
				m = (Buffer = QString::number(adCorr[n], 'g', (m>=0 ? m+1 : (Prec > 0.0 ? prec : 10)))).indexOf('e');
				if (Buffer.indexOf('.') == -1) Buffer += ".0";
				Tab->item(r++, 1)->setText(Buffer);
				if (m>0) L << "$v_{" + QString::number(n) + "}$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) 
								+ "$\\times 10^{" + QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) 
									+ "}$ cm$^{-1}$\\\\";
				else L << "$v_{" + QString::number(n) + "}$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) 
							+ " cm$^{-1}$\\\\"; 
			}
			L << "\\hline\\noalign{\\smallskip}";
			r+=4;
			delete[] adCorr;
		}
		if (Worker->getNSpinRGamma() > 0)
		{
			int NSpinRGamma;
			double *SpinRGamma, SpinRR1, SpinRR2, SpinRb, SpinRRm, *IsoF = Worker->getIsoF_ForReading();
			Worker->getSpinRGamma(NSpinRGamma, SpinRGamma, SpinRR1, SpinRR2, SpinRb, SpinRRm);
			if (l==-1) l = (State != 0 ? (State->getFitData() != 0 ? State->getFitData()->getMaxJ() : 300) : 300);
			X = Prec * double(l+1) / (Ri * Ri * IsoF[0]);
			if (NSpinRGamma > 1) x = (Ri - Rm) / (Ri + b * Rm);
			for (n=0; n < NSpinRGamma; n++, X*=x)
			{
				m = ((I = Tab->item(n, 3)) != 0 ? I->text().toInt() - 1 : -1);
				if (Prec > 0.0 && m<0)
				{
					prec = int(log10(abs(X * SpinRGamma[n]))) + 1;
					if (prec < 0) prec = 0;
				}
				m = (Buffer = QString::number(SpinRGamma[n], 'g', (m>=0 ? m+1 : (Prec > 0.0 ? prec : 6)))).indexOf('e');
				if (Buffer.indexOf('.') == -1) Buffer += ".0";
				Tab->item(r++, 1)->setText(Buffer);
				if (m>0) L << "$\\gamma_{" + QString::number(n) + "}$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m)
								+ "$\\times 10^{" + QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) + "}$\\\\";
				else L << "$\\gamma_{" + QString::number(n) + "}$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + "\\\\";
			}
			Buffer = QString::number(SpinRR1, 'g');
			if (Buffer.indexOf('.') == -1) Buffer += ".0";
			L << "$R_1$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + " \\AA \\\\";
			Buffer = QString::number(SpinRR2, 'g');
			if (Buffer.indexOf('.') == -1) Buffer += ".0";
			L << "$R_2$ && " + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + " \\AA \\\\";
			L << "\\hline\\noalign{\\smallskip}";
		}
		L << (lang == 0 ? "\\multicolumn{2}{l}{derived constants:}       \\\\" 
						: "\\multicolumn{2}{l}{Abgeleitete Konstanten:}  \\\\");
		Worker->unlock();
		aPot->getMinimum(MinR, MinU);
		Buffer = (De == 0.0 ? "$D_e$    && " + QString::number(-1.0 * MinU, 'f', 3) :
				"$T_e$    && " + QString::number(MinU, 'f', 4)) + " cm$^{-1}$ \\\\";
		L << (lang == 1 ? Buffer.replace('.', ',') : Buffer);
		L << "$R_e$    && " + (lang == 1 ? QString::number(MinR, 'f', 5).replace('.', ',') 
								: QString::number(MinR, 'f', 5).replace('.', ',')) + " \\AA \\\\ \\hline";
	}
	else if (Worker->isLPot())
	{
		MLRPot *LPot = dynamic_cast<MLRPot*>(Worker);
		LPot->getCoefficients(p, q, U_inf, De, MinR, Rm, nPotCoeff, PotCoeff, nLRCoeff, 
							  pLRCoeff, LRCoeff);
		L << "\\begin{tabular}{lp{0.1\\textwidth}r}\\hline\\noalign{\\smallskip}";
		for (n=0; n < nPotCoeff; n++)
			L << ("$\\beta_{" + QString::number(n) + "}$   ").left(14) + "&&"
				 + ("  " + QString::number(PotCoeff[n], 'f', 12)).right(17) + " \\\\";
		for (n=0; n < nLRCoeff; n++)
			L << ("$C_{" + QString::number(pLRCoeff[n]) + "}$       ").left(14) + "&& "
				 + QString::number(LRCoeff[n], 'e', 3).replace("e", "$\\times 10^{") 
				 + "}$ cm$^{-1}$\\AA$^{" + QString::number(pLRCoeff[n]) + "}$ \\\\";
		if (U_inf > 0.0) 
			L << "$U_\\infty     && " + QString::number(U_inf, 'f', 4) + " cm$^{-1}$ \\\\"; 
		L << "$D_e$         && " + QString::number(De, 'f', 4) + " cm$^{-1}$ \\\\";
		L << "$R_e$         && " + QString::number(MinR, 'f', 7) + " \\AA \\\\";
		L << "$R_{ref}$     && " + QString::number(Rm, 'f', 1) + " \\AA \\\\";
		L << "$p$           && " + QString::number(p) + " \\\\"; 
		L << "$q$           && " + QString::number(q) + " \\\\ \\hline";
	}
	else if (Worker->isSplinePot())
	{
		int numSplinePoints;
		if (!getTexTableSplineFitRunning)
		{
            FQSv = getFQS(NumWFPoints);
			SplinePoint *points = Worker->getSplinePoints(numSplinePoints);
			for (n=0; n < numSplinePoints; n++) 
				Tab->item(n, 0)->setText(QString::number(points[n].x, 'f', 2));
			delete points;
			UpdatePot(1);
			getTexTableSplineFitRunning = true;
			FitSplinePot(false);
			return;
		}
		if (QMessageBox::Yes == QMessageBox::question(this, "MolSpektAnalysis", "The new chisq is " + QString::number(FQS)
				+ " while the old chisq was " + QString::number(FQSv) + ", do you want one more fit iteration?", 
				QMessageBox::Yes | QMessageBox::No))
		{
			FitSplinePot(false);
			return;
		}
		else getTexTableSplineFitRunning = false;
		int NLRC, *pLRC;
		SplinePoint *points;
		double *LRC, iA, iO, iExp, Uinf = Worker->getUinf();
		Worker->getSplinePotForReading(numSplinePoints, points, NLRC, pLRC, LRC, iA, iO, iExp);
		if ((m = numSplinePoints / 2) * 2 < numSplinePoints) m++;
		L << "\\begin{tabular}{ll|ll}";
		L << "\\hline\\noalign{\\smallskip}";
		if (lang == 0)
			L << "R [\\AA] & Energy [cm$^{-1}$] & R [\\AA] & Energy [cm$^{-1}$]\\\\";
		else L << "R [\\AA] & Energie [cm$^{-1}$] & R [\\AA] & Energie [cm$^{-1}$]\\\\";
		L << "\\hline\\noalign{\\smallskip}";
		if (lang == 0) for (n=0; n<m; n++)
		{
			Buffer = QString::number(points[n].y, 'f', nDig);
			Tab->item(n, 1)->setText(Buffer);
			L << QString::number(points[n].x, 'f', 2) + " & " + Buffer;
			Buffer = QString::number(points[n+m].y, 'f', nDig);
			Tab->item(n+m, 1)->setText(Buffer);
			L << (m+n < numSplinePoints ? " & " + QString::number(points[n+m].x, 'f', 2) + " & "
											+ Buffer + " \\\\" 
						    			: " &        &           \\\\");
		}
		else for (n=0; n<m; n++)
		{
			Buffer = QString::number(points[n].y, 'f', nDig);
			Tab->item(n, 1)->setText(Buffer);
			L << QString::number(points[n].x, 'f', 2).replace('.', ',') + " & " 
					+ Buffer.replace('.', ',');
			Buffer = QString::number(points[n+m].y, 'f', nDig);
			Tab->item(n+m, 1)->setText(Buffer);
			L << (m+n < numSplinePoints ? 
					  " & " + QString::number(points[n+m].x, 'f', 2).replace('.', ',') 
					  + " & " + Buffer.replace('.', ',') + " \\\\" 
					: " &        &           \\\\");
		}
		L << "\\hline";
		L << "\\end{tabular}";
		L << "\\begin{tabular}{ll|ll} \\hline\\noalign{\\smallskip}";
		l = NLRC + (Uinf > 0.0 ? 4 : 3);
		QString Buff[l];
		if (Prec > 0.0)
		{
			prec = int(log10(abs(Prec * iO)));
			if (prec < 0) prec = 0;
		}
		m = (Buffer = QString::number(iO, 'e', (Prec > 0.0 ? prec : 4))).indexOf("e");
		Tab->item(numSplinePoints + 1, 1)->setText(Buffer);
		Buff[0] = "$A$ & " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
				+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) 
				+ "}$ cm$^{-1}$";
		if (Prec > 0.0)
		{
			prec = int(log10(Prec * iA * pow(points[0].x, -6.0)));
			if (prec < 0) prec = 0;
		}
		m = (Buffer = QString::number(iA, 'e', (Prec > 0.0 ? prec : 4))).indexOf("e");
		Tab->item(numSplinePoints, 1)->setText(Buffer);
		Buff[1] = "$B$ & " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(m) + "$\\times 10^{" 
				+ QString::number(Buffer.right(Buffer.length() - m - 1).toInt()) 
				+ "}$ cm$^{-1}$\\AA$^6$";
		Buff[n=2] = "$n$ & 6";
		if (Uinf > 0.0)
		{
			Buffer = QString::number(Uinf, 'f', (Prec > 0 ? nDig : 3));
			Tab->item(numSplinePoints + n, 1)->setText(Buffer);
			Buff[++n] = "$U_\\infty $ & " 
						 + (lang == 1 ? Buffer.replace('.', ',') : Buffer) + " cm$^{-1}$";
		}
		for (m=0, Ra = points[numSplinePoints - 1].x; m < NLRC; m++)
		{
			if (Prec > 0.0)
			{
				prec = int(log10(Prec * abs(LRC[m]) * pow(Ra, -pLRC[m])));
				if (prec < 0) prec = 0;
			}
			p = (Buffer = QString::number(LRC[m], 'e', (Prec > 0.0 ? prec : 16))).indexOf("e");
			Tab->item(numSplinePoints + n, 1)->setText(Buffer);
			eB = QString::number(pLRC[m]);
			Buff[++n] = "$C_{" + eB + "}$ & " + (lang == 1 ? Buffer.replace('.', ',') : Buffer).left(p) + "$\\times 10^{" 
					  + QString::number(Buffer.right(Buffer.length() - p - 1).toInt()) + "}$ cm$^{-1}$\\AA$^{" 
					  + eB + "}$";
		}
		if ((m = l / 2) * 2 < l) m++;
		for (n=0; n<m; n++) L << Buff[n] + (n + m < l ? " & " + Buff[n+m] + " \\\\"
													  : " &       &     \\\\");
		L << "\\hline";
		Worker->unlock();
	}
	L << "\\end{tabular}" << "\\end{table}";
	updatePot();
	Tab->blockSignals(false);
	Changed();
	QTextEdit *Window = new QTextEdit();
	Window->setPlainText(L.join("\n"));
	MW->showMDIChild(Window);
}

bool Potential::writeData(QString Filename, bool writeFitData)
{
    QFile Datei(Filename);
	if (!write(&Datei)) return false;
	int r, c, R = Tab->rowCount(), C = Tab->columnCount();
	QTextStream S(&Datei);
	QStringList L;
	QTableWidgetItem *I;
	QString Spacer = " | ";
	S << "Source: " << getSource() << "\n";
	S << "Name: " << getName() << "\n";
	for (c=0; c<C; c++) L << ((I = Tab->horizontalHeaderItem(c)) != 0 ? I->text() : "?");
	S << "Column titles: " << L.join(Spacer) << "\n";
	for (r=0; r < R; r++) 
	{
		L.clear();
		for (c=0; c < C; c++) L << ((I = Tab->item(r, c)) != 0 ? I->text() : "");
		S << L.join(Spacer) << "\n";
	}
	if (WaveFuncs != 0)
	{
		if (!WaveFuncs->isSaved())
		{
			if (WaveFuncs->getFileName().isEmpty())
			{
				Filename = WaveFuncs->getSource();
				Filename = Filename.right(Filename.length() - Filename.lastIndexOf(QRegExp("[\\/]")) - 1);
				Filename = Filename.left(Filename.indexOf('.') + 1) + "SWF";
				WaveFuncs->setFileName(Filename);
			}
			WaveFuncs->writeData();
		}
		if (NCoupled > 0 && WaveFuncs->isSaved())
		{
            QString MolFile;
            S << "\nCoupling information:\n";
            S << "Wave functions: " << (molecule != 0 ? WaveFuncs->getRelativePath(MolFile = molecule->getFileName())
                                                      : WaveFuncs->getFileName()) << '\n';
			S << "\nCoupled potential: | Coupled component:\n";
            for (r=0; r < NCoupled; r++)
                S << (molecule != 0 ? CoupledPot[r]->getRelativePath(MolFile) : CoupledPot[r]->getFileName(true))
                  << " | " << QString::number(CoupledComp[r]) << '\n';
		}
	}
	Saved();
	if (writeFitData && fitData != 0)
	{
		if (Filename.isEmpty()) Filename = getFileName();
		QString FDFile = Filename.left(Filename.lastIndexOf('.')) + ".fdat", CFN = fitData->getFileName();
		QFile FDF(FDFile);
		if (!FDF.exists() || CFN != FDFile || !fitData->isSaved())
		{
			fitData->writeData(FDFile);
			if (CFN != FDFile) fitData->setFileName(CFN);
		}
		if (writePotFitTraceTab && !PotFitTraceTab.isEmpty())
		{
			QFile File(Filename.left(Filename.lastIndexOf('.')) + "_FTT.dat");
			File.open(QIODevice::WriteOnly);
			File.write((PotFitTraceTab.join("\n") + "\n").toLatin1());
			writePotFitTraceTab = false;
		}
	}
	return true;
}

bool Potential::writePotData(QString Filename)
{
	if (!Worker->isAnaPotAv()) return false;
	double b, De, iCoeff, iOff, iExp, Ra, Ri, Rm, Tm, *LRCoeff, *PotCoeff, EI_A, EIalpha, EI_gamma;
	int nPotCoeff, nLRCoeff, *pLRCoeff, pCC = 0, n, intIExp, RIso1, RIso2;
	double Diff, D;
	if (Filename.isEmpty()) 
		if ((Filename = QFileDialog::getSaveFileName(NULL, tr("Select file name"), tr(""), tr(""))).isEmpty())
			return false;
	QFile File(Filename);
	//if (!write(&File)) return false;
	File.open(QIODevice::WriteOnly);
	QTextStream S(&File);
	QString Buffer;
	AnaPot *aPot = Worker->getAnaPotForReading();
	S << getSource() << "\n";
	aPot->getPotCoeff(b, Rm, Tm, De, nPotCoeff, PotCoeff);
	aPot->getInnerWall(Ri, iOff, iExp, iCoeff);
	aPot->getLRCoeffForReading(nLRCoeff, pLRCoeff, LRCoeff, &Ra, true);
	aPot->getExchangeCoeff(EI_A, EI_gamma, EIalpha);
	S << ("               " + QString::number(Tm, 'f', 8)).right(15);
	S << ("               " + QString::number(Rm, 'f', 8)).right(15) << "\n";
	S << ("     " + QString::number(pCC)).right(5);
	S << ("               " + QString::number(De, 'f', 8)).right(15);
	intIExp = int(iExp);
	S << ("     " + QString::number(iExp == double(intIExp) ? intIExp : -intIExp)).right(5) << "    0\n";
	S << ("     " + QString::number(nPotCoeff)).right(5);
	S << ("               " + QString::number(b, 'f', 8)).right(15);
	S << " 0.90000000D+02\n";
	for (n=0; n < nPotCoeff; n++)
	{
		S << ("     " + QString::number(n+1)).right(5);
		S << ("                         " 
				+ QString::number(PotCoeff[n], 'E', 18)).right(25).replace('E', 'D') << "\n";
	}
	if (Worker->isAdCorrA())
	{
		int NAdCorr, TAdCorr, PAdCorr;
		double *adCorr, RIso1AdCorr, RIso2AdCorr, adRm, adb;
		Worker->getAdCorrForReading(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, adRm, adb, 0, true);
		if (molecule != 0)
		{
			Atom *atom = molecule->getAtom1();
			for (n=1, RIso1 = atom->getnNuc(0), Diff = fabs(atom->getIsoMass(0) - RIso1AdCorr); n < atom->getnIso(); n++)
				if ((D = fabs(atom->getIsoMass(n) - RIso1AdCorr)) < Diff)
			{
				Diff = D;
				RIso1 = atom->getnNuc(n);
			}
			atom = molecule->getAtom2();
			for (n=1, RIso2 = atom->getnNuc(0), Diff = fabs(atom->getIsoMass(0) - RIso2AdCorr); n < atom->getnIso(); n++)
				if ((D = fabs(atom->getIsoMass(n) - RIso2AdCorr)) < Diff)
			{
				Diff = D;
				RIso2 = atom->getnNuc(n);
			}
		}
		else
		{
			RIso1 = lround(RIso1AdCorr);
			RIso2 = lround(RIso2AdCorr);
		}
		S << ("    " + QString::number(NAdCorr)).right(5) << ("    " + QString::number(RIso1)).right(5);
		S << ("    " + QString::number(RIso2)).right(5) << ("    " + QString::number(PAdCorr)).right(5);
		S << ("               " + QString::number(RIso1AdCorr, 'f', 9)).right(15);
		S << ("               " + QString::number(RIso2AdCorr, 'f', 9)).right(15);
		S << ("    " + QString::number(TAdCorr)).right(5) << '\n';
		for (n=0; n < NAdCorr; n++)
			S << ("    " + QString::number(n)).right(5)
			  << ("                         " + QString::number(adCorr[n], 'e', 18)).right(25).replace('e', 'D') << '\n';
	}
	else S << "    0\n";
	S << "    0\n";
	S << ("     " + QString::number(nLRCoeff)).right(5) << "\n";
	for (n=0; n < nLRCoeff; n++)
	{
		S << ("     " + QString::number(pLRCoeff[n])).right(5);
		S << ("                         "
				+ QString::number(LRCoeff[n], 'E', 18)).right(25).replace('E', 'D') << "\n";
	}
	if (EI_A == 0.0) EI_A = 1e-30;
	S << ("              " + QString::number(EI_A, 'e', 8)).right(15).replace('e', 'D');
	S << ("              " + QString::number(EIalpha, 'e', 8)).right(15).replace('e', 'D');
	S << ("              " + QString::number(EI_gamma, 'e', 8)).right(15).replace('e', 'D');
	S << "    0    0\n";
	S << (Buffer = ("          " + QString::number(Ri, 'f', 6)).right(10));
	S << ("          " + QString::number(Ra, 'f', 6)).right(10);
	S << "     rmin used:" << Buffer << "\n";
	S << ("                           " + QString::number(iOff, 'E', 20)).right(27).replace('E', 'D');
	S << ("                           " + 
			QString::number(iCoeff, 'E', 20)).right(27).replace('E', 'D');
	S << ("                           " + QString::number(iExp, 'E', 20)).right(27).replace('E', 'D');
	S << "\n   0.000000000000000000D+00   0.000000000000000000D+00\n";
	Worker->unlock();
	return true;
}

void Potential::writePoints(QString FileName, double RMin, double RMax, int NPoints, int NDigits, bool FirstRow, bool AdCorrE, int Iso,
							QString MCSDir)
{
	int n, f;
	double r, h = (RMax - RMin) / double(NPoints - 1), *Dat = getPoints(RMin, RMax, NPoints);
	IsoTab *IsoT = Worker->getIsoT();
	double *adC = (AdCorrE ? getAdCorrEnergy(RMin, h, NPoints, IsoT->mIso1[Iso], IsoT->mIso2[Iso]) : 0);
	double M, E, *Min, *Max, *AMin, *AMax;
	bool MM = !MCSDir.isEmpty();
	QFile Datei(FileName);
	Datei.open(QIODevice::WriteOnly);
	QTextStream S(&Datei);
	if (MM)
	{
		Min = new double[NPoints]; Max = new double[NPoints];
		for (n=0; n < NPoints; n++) Min[n] = Max[n] = Dat[n];
		if (AdCorrE)
		{
			AMin = new double[NPoints]; AMax = new double[NPoints];
			for (n=0; n < NPoints; n++) AMin[n] = AMax[n] = adC[n];
		}
		else AMin = AMax = 0;
		QDir MDir(MCSDir);
		QFileInfoList FList = MDir.entryInfoList(QDir::Dirs | QDir::Files | QDir::NoDotAndDotDot);
		QFile MFile;
		Potential *TPot = new Potential();
		double *MDat, *MadC;
		for (f=0; f < FList.count(); f++)
		{
			if (FList[f].isDir()) 
			{
				MDir.setPath(FList[f].filePath());
				FList << MDir.entryInfoList(QDir::Dirs | QDir::Files | QDir::NoDotAndDotDot);
			}
			else if (FList[f].fileName().right(4) == ".pot")
			{
				TPot->readData(FList[f].filePath());
				MDat = TPot->getPoints(RMin, RMax, NPoints);
				for (n=0; n < NPoints; n++)
				{
					if (MDat[n] < Min[n]) Min[n] = MDat[n];
					else if (MDat[n] > Max[n]) Max[n] = MDat[n];
				}
				delete[] MDat;
				if (AdCorrE)
				{
					MadC = TPot->getAdCorrEnergy(RMin, h, NPoints, IsoT->mIso1[Iso], IsoT->mIso2[Iso]);
					for (n=0; n < NPoints; n++)
					{
						if (MadC[n] < AMin[n]) AMin[n] = MadC[n];
						else if (MadC[n] > AMax[n]) AMax[n] = MadC[n];
					}
					delete[] MadC;
				}
			}
		}
		delete TPot;
	}
	else Min = Max = AMin = AMax = 0;
	if (FirstRow)
	{
		getMinimum(M, E);
		S << QString::number(NPoints);
		S << ("           " + QString::number(M, 'f', 6)).right(12) << "\n";
	}
	for (r = RMin, n=0; n < NPoints; n++, r+=h) 
	{
		S << ("            " + QString::number(r, 'f', 6)).right(12);
		S << ("            " + QString::number(Dat[n], 'f', NDigits)).right(NDigits + 8);
		if (MM)
		{
			S << ("            " + QString::number(Min[n], 'f', NDigits)).right(NDigits + 8);
			S << ("            " + QString::number(Max[n], 'f', NDigits)).right(NDigits + 8);
		}
		if (AdCorrE) 
		{
			S << ("            " + QString::number(adC[n], 'f', NDigits)).right(NDigits + 8);
			if (MM)
			{
				S << ("            " + QString::number(AMin[n], 'f', NDigits)).right(NDigits + 8);
				S << ("            " + QString::number(AMax[n], 'f', NDigits)).right(NDigits + 8);
			}
		}
		S << '\n';
	}
	delete[] Dat;
	if (MM)
	{
		delete[] Min;
		delete[] Max;
		if (AdCorrE)
		{
			delete[] AMin;
			delete[] AMax;
		}
	}
	if (AdCorrE) delete[] adC;
}

bool Potential::readPointPot(QString FileName)
{
	//printf("Potential::readPointPot\n");
	int n=0, N=1000, c, m=0;
	Tab->setRowCount(N);
	QFile File(FileName);
	if (!read(&File)) return false;
	QTextStream S(&File);
	QString Buffer;
	QStringList L;
	Tab->blockSignals(true);
	while (!S.atEnd())
	{
		Buffer = S.readLine();
		L = Buffer.split(" ", QString::SkipEmptyParts);
		c = L.count();
		if (c!=2) 
		{
			if (n>0) if (m++ == 3) return false;
			continue;
		}
		if (L[0].toDouble() == 0.0 || (n==0 && L[1].toDouble() == 0.0)) continue;
		if (n == N) Tab->setRowCount(N+=1000);
		Tab->setItem(n, 0, new QTableWidgetItem(L[0]));
		Tab->setItem(n++, 1, new QTableWidgetItem(L[1]));
	}
	if (n < 3) return false;
	Tab->setRowCount(N = n);
	SplinePoint *points = new SplinePoint[N];
	for (n=0; n<N; n++)
	{
		points[n].x = Tab->item(n, 0)->text().toDouble();
		points[n].y = Tab->item(n, 1)->text().toDouble();
		if (n>0 ? points[n-1].x >= points[n].x : false) 
		{
			printf("read PointPot error: x[%d]=%f, x[%d]=%f\n", n-1, points[n-1].x, n, points[n].x);
			delete[] points;
			return false;
		}
	}
	if (N>2 ? points[N-1].x - points[N-2].x > 10.0 * (points[N-2].x - points[N-3].x) : false)
	{
		Worker->setUinf(points[N-1].y);
		N--;
		Tab->item(N-1, 0)->setText("U_inf");
	}
	if (!Worker->setSplinePot(N, points, 0, 0, 0, 0.0, 0.0, 0.0))
	{
		delete[] points;
		return false;
	}
	Tab->setColumnCount(3);
	Tab->setHorizontalHeaderLabels(QStringList() << "R [A]" << "Energy [cm-1]" << "d^2E/dR^2");
	Tab->blockSignals(false);
	//printf("Vor calcyss\n");
	calcyss();
	//printf("Ende readPointPot\n");
	return true;
}

void Potential::calcyss()
{
	Worker->calcyss();
	UpdateTab();
	Changed();
	//printf("Potential::calcyss\n");
	/*int numSplinePoints, n, NLRC, *pLRC;
	SplinePoint *points;
	double *LRC, iA, iO, iExp;
	Worker->getSplinePotForWriting(numSplinePoints, points, NLRC, pLRC, LRC, iA, iO, iExp);
	if (numSplinePoints == 0) return;
	int N = numSplinePoints, st = N - 3;
	double B[N-3], F, Uinf = Worker->getUinf();
	points[0].yss = 0.0;
	if (N<3) points[N-1].yss = 0.0;
	else
	{
		if (2.0 * (points[N-2].x - points[N-3].x) < points[N-1].x - points[N-2].x 
				  && NLRC == 0 && Uinf == 0.0)
		{
			Uinf = points[N-1].x;
			N--;
			st--;
			//SLIC6 = true;
			//st = N - 4;
			//double R2 = points[N-2].x;
			//double R26 = pow(R2, 6.0), R16 = pow(points[N-1].x, 6.0);
			//points[N-1].yss = (points[N-2].y - points[N-1].y) * R26 * R16 / (R16 - R26);
			//points[N-2].yss = -42.0 * points[N-1].yss / (R26 * R2 * R2);*/
					//0.0;
	/*	}
		points[N-1].yss = 0.0;
		if (st==0)
		{ 
			points[1].yss = 3.0 / (points[2].x - points[0].x) 
						   * ((points[2].y - points[1].y) / (points[2].x - points[1].x)
						     -(points[1].y - points[0].y) / (points[1].x - points[0].x)
							 -(points[2].x - points[1].x) / 6.0 * points[2].yss);
			return;
		}
		else
		{
			F = 0.5 * (points[2].x - points[1].x) / (points[2].x - points[0].x);
			B[0] = points[3].x - points[1].x - F * 0.5 * (points[2].x - points[1].x);
			points[1].yss = 3.0 * ((points[2].y - points[1].y) / (points[2].x - points[1].x)
					              -(points[1].y - points[0].y) / (points[1].x - points[0].x));
			points[2].yss = 3.0 * ((points[3].y - points[2].y) / (points[3].x - points[2].x)
					              -(points[2].y - points[1].y) / (points[2].x - points[1].x))
			      		- F * points[1].yss;
			for (n=1; n < st; n++)
			{
				F = 0.5 * (points[n+2].x - points[n+1].x) / B[n-1];
				B[n] = points[n+3].x - points[n+1].x - F * 0.5 * (points[n+2].x - points[n+1].x);
				points[n+2].yss = 3.0 * ((points[n+3].y - points[n+2].y) 
								/ (points[n+3].x - points[n+2].x)
			    		        -(points[n+2].y - points[n+1].y) / (points[n+2].x - points[n+1].x))
				        		- F * points[n+1].yss;
			}
			for (n = st - 1; n >= 0; n--) 
				points[n+2].yss = 
					(points[n+2].yss - 0.5 * (points[n+3].x - points[n+2].x) * points[n+3].yss) / B[n];
			points[1].yss = (points[1].yss - 0.5 * (points[2].x - points[1].x) * points[2].yss) 
            		      / (points[2].x - points[0].x);
		}
	}
	Tab->blockSignals(true);
	for (n=0; n<N; n++) 
		Tab->setItem(n, 2, new QTableWidgetItem(QString::number(points[n].yss, 'f', 7)));
	Tab->blockSignals(false);
	Worker->recalcSCLC();
	if (N != numSplinePoints) Worker->setSplinePot(N, points, NLRC, pLRC, LRC, iA, iO, iExp);
	//calcLMatrix();
    //calcS(NumWFPoints, points[0].x, points[N-2].x);*/
}

void PotWorker::cdConnectLR1C()
{
	if (Fit->isRunning()) return;
	if (points != 0)
	{
		int n;
		double dx = points[numSplinePoints - 1].x - points[numSplinePoints - 2].x, Lys = 0.0, dE = Uinf - points[numSplinePoints - 1].y;
		double ys = (points[numSplinePoints - 1].y - points[numSplinePoints - 2].y) / dx 
				  + 1.0/6.0 * dx * (points[numSplinePoints - 2].yss + 2.0 * points[numSplinePoints - 1].yss);
		for (n=0; n < NLRC - 1; n++) Lys += double(PLRC[n]) * LRC[n] * pow(points[numSplinePoints - 1].x, -PLRC[n] - 1);
		LRC[NLRC - 1] = (ys - Lys) / double(PLRC[NLRC - 1]) * pow(points[numSplinePoints - 1].x, PLRC[NLRC - 1] + 1);
		for (n=0; n < NLRC; n++) dE -= LRC[n] * pow(points[numSplinePoints - 1].x, -PLRC[n]);
		for (n=0; n < numSplinePoints; n++) points[n].y += dE;
		iO += dE;
	}
}

void PotWorker::cdConnectLR(int p1)
{
	if (Fit->isRunning()) return;
	if (numSplinePoints >= 2)
	{
		int n, i, p2;
		double yd = points[numSplinePoints - 1].y, Ra = points[numSplinePoints - 1].x, Rd = Ra - points[numSplinePoints - 2].x, B;
		double ysd = (yd - points[numSplinePoints - 2].y) / Rd + 1.0 / 6.0 * Rd * points[numSplinePoints - 2].yss 
															   + 1.0 / 3.0 * Rd * points[numSplinePoints - 1].yss; 
		for (i=0; (i < NLRC ? PLRC[i] != p1 : false); i++) ;
		if (i+1 < NLRC) p2 = PLRC[i+1];
		else
		{
			double *nLRC = new double[i+2];
			int *nPLRC = new int[i+2];
			bool *nLRCfree = new bool[i+2];
			p2 = p1 + 2;
			if (NLRC > 0)
			{
				for (n=0; n < NLRC; n++)
				{
					nLRC[n] = LRC[n];
					nPLRC[n] = PLRC[n];
					nLRCfree[n] = LRCfree[n];
				}
				delete[] LRC;
				delete[] PLRC;
				delete[] LRCfree;
			}
			LRC = nLRC;
			PLRC = nPLRC;
			LRCfree = nLRCfree;
			NLRC = i+2;
		}
		for (n=0, yd -= Uinf; n < NLRC; n++) if (n!=i && n!=i+1)
		{
			yd += (B = LRC[n] * pow(Ra, -PLRC[n]));
			ysd -= B * double(PLRC[n]) / Ra;
		}
		LRC[i+1] = pow(Ra, p2) * (B = -(double(p1) * yd + Ra * ysd) / double(p1 - p2));
		LRC[i] = -pow(Ra, p1) * (yd + B);
		PLRC[i] = p1;
		PLRC[i+1] = p2;
		LRCfree[i] = LRCfree[i+1] = true;
	}
}

void Potential::cdConnectLR(int p1)
{
	if (Fit->isRunning()) return;
	Worker->cdConnectLR(p1);
	UpdateTab();
	Changed();
}

void PotWorker::cdConnectSR()
{
	if (Fit->isRunning()) return;
	if (points != 0)
	{
		double xd = points[1].x - points[0].x, ds = 0.16666666666667;
		iA = ds * pow(points[0].x, iExp + 1.0) 
			* (ds * xd * points[1].yss - (points[1].y - points[0].y) / xd);
		if (iA < 0.0)
		{
			points[0].y += points[1].y - points[2].y;
			iA = ds * pow(points[0].x, iExp + 1.0) 
				* (ds * xd * points[1].yss - (points[1].y - points[0].y) / xd);
		}
		iO = points[0].y - iA * pow(points[0].x, -iExp);
	}
}

void Potential::cdConnectSR()
{
	if (Fit->isRunning()) return;
	Worker->cdConnectSR();
	UpdateTab();
	Changed();
}

void PotWorker::calcLMatrix()
{
	int N = numSplinePoints;
	//printf("calcLMatrix: numSplinePoints=%d\n", N);
	int st = N - 1, M;
	if (st==0)
	{
		printf("Potential::calcLMatrix: error, not enough spline points defined!\n");
		return;
	}
	if (L!=0) Destroy(L, numSplinePoints);
	L = Create(N, M=st+3);
	
	CalcLMatrix(L, points, iExp, N, Ro != 0.0);
		
	/*int i, n;  // Uncomment this block and comment the previous line if you don't want to use code from Numerical Recipes
	double B[st], F, dx[st];
	
	//for (n=0; n<N; n++) 
	//{
	//	for (i=0; i<= N+1; i++) printf("L[%d][%d]=%g ", n, i, L[n][i]);
	//	printf("\n");
	//}
	
	for (i=0; i < st; i++) dx[i] = points[i+1].x - points[i].x;
	for (n=0; n <= st; n++) for (i=0; i<M; i++) L[n][i] = 0.0;
	F = 0.5;
	B[0] = 1.0 / (points[2].x - points[0].x - F * 0.5 * dx[0]);
	L[0][0] = 3.0 * iExp * pow(points[0].x, -iExp - 1.0);
	L[0][1] = -3.0 / dx[0];
	L[0][2] = -L[0][1];
	for (n=1; n < st; n++) 
	{
		L[n][n] = L[n-1][n+1];
		L[n][n+2] = 3.0 / dx[n];
		L[n][n+1] = - L[n][n] - L[n][n+2];
	}
	L[n][n] = -1.0 * L[n-1][n+1];
	L[n][n+1] = L[n-1][n+1];
	L[n][n+2] = -3.0;
	for (i=0; i<3; i++) L[1][i] -= F * L[0][i];
	for (n=1; n < st - 1; n++)
	{
		F = 0.5 * dx[n] * B[n-1];
		B[n] = 1.0 / (points[n+2].x - points[n].x - F * 0.5 * dx[n]);
		for (i=0; i<n+3; i++) L[n+1][i] -= F * L[n][i];
	}
	F = -0.5 * dx[n] * B[n-1];
	B[n] = 1.0 / ((-1.0 - F * 0.5) * dx[n]);
	for (i=0; i<n+3; i++) L[n+1][i] -= F * L[n][i];
	for (i=0; i<M; i++) L[st][i] *= B[st -1];
		//		(L[st][i] - 0.5 * dx[st + 1] * points[st+2].yss) * B[st - 1];
	for (n = st - 2; n>=0; n--) for (i=0; i<M; i++) 
			L[n+1][i] = (L[n+1][i] - 0.5 * dx[n+1] * L[n+2][i]) * B[n];
	for (i=0; i<M; i++) L[0][i] = (L[0][i] - 0.5 * dx[0] * L[1][i]) / dx[0];
	
	//for (n=0; n<N; n++) 
	//{
	//	for (i=0; i<= N+1; i++) printf("L[%d][%d]=%g ", n, i, L[n][i]);
	//	printf("\n");
	//}
	
	//double S = 0.0;
	//for (n=0; n < NLRC; n++) S += double(PLRC[n]) * LRC[n] * pow(points[N-1].x, -1 - PLRC[n]);
	*/
}

void PotWorker::calcS(int N, double RMin, double RMax)
{
	int SN = (NLRC == 0 ? 1 : 0) /*(State != 0 ? State->getStateNum() : 1)*/, NL = numSplinePoints + 2;
	int i=0, j, k, K = NL, c;
	//printf("Potential::calcS: K=%d, N=%d\n", K, N);
	/*if (SLIC6) 
	{
		K--;
		if (Uinf == 0.0 && points[numSplinePoints - 1].x > 100.0) Uinf = points[numSplinePoints - 1].y;
	}*/
	double d = (RMax - RMin) / (N-1), xd, R = RMin, R0 = points[numSplinePoints - 1].x, A, B, C, D;
	double F[NLRC], f = 1.0 / 6.0, Off[NLRC], x, iF = 0.0;
	if (RMin >= RMax) 
	{
		printf("Potential::calcS error: invalid radii!\n");
		return;
	}
	//if (NLRC == 0) SN = 1;
	if (SN == 0)
	{
		K = numSplinePoints + NLRC + 1;
		for (i=1, F[0] = pow(R0, j = -PLRC[0] - 1); i < NLRC; i++) 
			for (F[i] = F[i-1]; j >= -PLRC[i]; j--, F[i] /= R0) ;
		for (c=0; c < NLRC; c++) 
		{
			Off[c] = 1.0 / (R0 * F[c]);
			F[c] *= PLRC[c];
		}
	}
	if (NAdCorr > 0) K += NAdCorr;
	if (S!=0) Destroy(S, NumS);
	S = Create(NumS = K, N);
	SRMin = RMin;
	SRMax = RMax;
	//printf("Vor Schleife\n");
	if (R < points[0].x)
	{
		A = pow(points[0].x, -iExp);
		for (i=0; R < points[0].x; R+=d, i++) 
		{
			//printf("R=%f\n", R);
			S[0][i] = (pow(R, -iExp) - A) * (iF = 1.0 / A);
			S[1][i] = 1.0;
			for (k=2; k<K; k++) S[k][i] = 0.0;
			if (SN == 0) for (k=0, S[numSplinePoints][i] = -1.0; k < NLRC; k++) 
				S[numSplinePoints + k + 1][i] = -1.0;
		}
	}
	//printf("Vor 2.\n");
	for (j=1; i<N; i++, R+=d)
	{
		//printf("i=%d, j=%d\n", i, j);
		if (R > points[j].x) j++;
		if (j == numSplinePoints) break;
		xd = points[j].x - points[j-1].x;
		A = (points[j].x - R) / xd;
		B = 1.0 - A;
		C = f*A * (A*A - 1.0) * xd * xd;
		D = f*B * (B*B - 1.0) * xd * xd;
		for (k=0; k < NL; k++) S[k][i] = C * L[j-1][k] + D * L[j][k];
		S[j][i] += A;
		S[j+1][i] += B;
		S[0][i] *= iF;
		if (SN == 0)
		{
			for (c=1; c < NLRC; c++, k++) S[k][i] = S[NL - 1][i] * F[c] * Off[c] - 1.0;
			S[NL - 1][i] *= (F[0] * Off[0]);
			S[NL - 1][i] -= 1.0;
			S[numSplinePoints][i] -= 1.0;
		}
	}
	//printf("Nach Schleife\n");
	if (i<N)
	{
		if (SN == 0)
		{
			for ( ; i<N; R+=d, i++)
			{
				for (j=0; j < NL - 1; j++) S[j][i] = 0.0;
				//S[NL - 2][i] = 1.0;
				for (S[j++][i] = -pow(R0 = 1.0 / R, k = PLRC[0]), c=1; c < NLRC; c++, j++)
					for (S[j][i] = S[j-1][i]; k < PLRC[c]; S[j][i] *= R0, k++) ;
				for (j = NL - 1, c=0; c < NLRC; c++, j++) S[j][i] *= Off[c]; 
			}
		}
		else
		{
			A = 4.0 * (f = pow(R0, 6.0));
			B = 0.5 * (f *= R0);
			C = -3.0 * (f *= R0);
			D = -0.5 * (f *= R0);
			for ( ; i<N; R+=d, i++) 
			{
				for (j=0; j<K-2; j++) S[j][i] = 0.0;
				f = pow(R0 = 1.0 / R, 6.0);
				S[K-2][i] = A*f;
				S[K-1][i] = B*f;
				S[K-2][i] += C * (f *= R0 * R0);
				S[K-1][i] += D*f;
			}
		}
	}
	if (NAdCorr > 0) for (R = RMin, i=0; i<N; i++, R+=d)
	{
		S[j = K - NAdCorr][i] = pow(2.0 * AdCorr_Rm / (R + AdCorr_Rm), PAdCorr);
		if (NAdCorr > 1) for (c=1, j++, x = (R - AdCorr_Rm) / (R + AdCorr_b * AdCorr_Rm); c < NAdCorr; c++, j++)
			S[j][i] = S[j-1][i] * x;
	}
	
	
	/*printf("Vor Test\n");
	double St = 0.0;
	R0 = points[numSplinePoints - 1].x;
	R = RMin + 5.0 * xd;
	d*=1000.0;
	double y;
	for (i=0, R = RMin; i<N; i+=1000, R+=d)
	{
		for (k=1, y = iA * S[0][i]; k<K-1; k++) y += S[k][i] * points[k-1].y;
		y += St * S[k][i];
		if (R > R0) 
		{
			A = R0 / R;
			y += (1.0 + pow(A, 6.0) * (3.0 * A * A - 4.0)) * Uinf;
		}
		printf("R=%f, y=%f=%f\n", R, getPoint(R), y);
	}*/
}

void PotWorker::getS(double &Rmin, double &Rmax, int &NPoints, int& NC, double**& rS)
{
	if (Type == SplinePotential && L==0) calcLMatrix();
	else if (Type != SplinePotential && NAdCorr > 0 && AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
	calcS(NPoints, Rmin, Rmax);
	NC = NumFreePar;
	rS = S;
	S=0;
}

void Potential::setPointData(int N, double *R, double *E)
{
	if (Fit->isRunning()) return;
	//printf("Potential::setPointData\n");
	//printf("R[0][1]=%f\n", R[1]);
	int n;
	Tab->setHorizontalHeaderLabels(QStringList() << "R [A]" << "E [cm^-1]" << "d^2E/dR^2");
	SplinePoint *points = new SplinePoint[N];
	Tab->setColumnCount(3);
	Tab->setRowCount(N);
	Tab->blockSignals(true);
	for (n=0; n<N; n++)
	{
		Tab->setItem(n, 0, new QTableWidgetItem(QString::number(points[n].x = R[n], 'f', 6)));
		Tab->setItem(n, 1, new QTableWidgetItem(QString::number(points[n].y = E[n], 'f', 6)));
		points[n].variable = true;
	}
	Tab->blockSignals(false);
	Worker->setSplinePot(N, points, 0, 0, 0, 0.0, 0.0, 0.0);
	calcyss();
	Changed();
}

void Potential::setMTTCoefficients(double B, double alpha1, int NA, double* A, double alpha, double beta, double gamma, int NLRC,
								   int* pLRC, double* LRC, double Uinf, int NC, double *C, double Calpha, double Cbeta, 
								   int NCLRC, int *pCLRC, double *CLRC, double CInf)
{
	if (Fit->isRunning()) return;
	MTTPot *mTTPot = new MTTPot(Fit);
	mTTPot->setCoefficients(B, alpha1, NA, A, alpha, beta, gamma, NLRC, pLRC, LRC, Uinf, NC, C, Calpha, Cbeta, NCLRC, pCLRC, 
							CLRC, CInf);
	setWorker(mTTPot);
	UpdateTab();
	Changed();
}

bool Potential::readCoupledPotData(QString Filename)
{
	double b, De = 0.0, Ra, Ri, Rm, Tm, *LRCoeff = 0, *PotCoeff, iExp, q_e_si = 0.0, q_e_pi = 0.0, q_f_pi = 0.0, ASO_asym = 0.0, *ASO = 0, alpha = 0.0;
	double beta = 0.0, ASO_scaling = 0.0, xi_scaling = 0.0, *xi = 0, gamma = 0.0, B_xi = 0.0, alpha_xi = 0.0, RFakt = a0_Angstrom;
	int nPotCoeff, nLRCoeff, *pLRCoeff = 0, pCC, n, m, p, N;
	QFile File(Filename);
	if (!read(&File)) return false;
	QTextStream St(&File);
	QString Buffer;
	Buffer = St.readLine();
	N = Buffer.left(5).toInt() + 2;
	if (N <= 0 || N > 5) return false;
	Potential *Pot[N]; 
	AnaPot *aPot;
	MTTPot *couplingFuncs;
	for (p=0; p<N-2; p++)
	{
		if (p>0) 
		{
			Pot[p] = MW->CreatePotential();
			if (Pot[p] == 0) return false;
		}
		else Pot[p] = this;
		aPot = new AnaPot(Pot[p]->Fit);
		Pot[p]->setName((getFName().split('.'))[0] + QString::number(p));
		Pot[p]->setSource(Filename);
		Buffer = St.readLine().replace('D', 'e');
		Tm = Buffer.left(15).toDouble();
		Rm = Buffer.mid(15, 15).toDouble();
		Buffer = St.readLine();
		iExp = Buffer.left(5).toInt();
		De = Buffer.mid(5, 15).toDouble();
		pCC = Buffer.mid(20, 5).toInt();
		Buffer = St.readLine().replace('D', 'e');
		nPotCoeff = Buffer.left(5).toInt();
		b = Buffer.mid(5, 15).toDouble();
		PotCoeff = new double[nPotCoeff];
		for (n=0; n < nPotCoeff; n++)
		{
			Buffer = St.readLine().replace('D', 'e');
			m = Buffer.left(5).toInt();
			if (m > nPotCoeff) return false;
			PotCoeff[m-1] = Buffer.mid(5, 25).toDouble();
		}
		Buffer = St.readLine();
		nLRCoeff = Buffer.toInt();
		pLRCoeff = new int[nLRCoeff + (pCC > 0 ? 2 : 1)];
		LRCoeff = new double[nLRCoeff + (pCC > 0 ? 2 : 1)];
		for (n=0; n < nLRCoeff; n++)
		{
			Buffer = St.readLine().replace('D', 'e');
			pLRCoeff[n] = Buffer.left(5).toInt();
			LRCoeff[n] = Buffer.mid(5, 25).toDouble();
		}
		Buffer = St.readLine().replace('D', 'e');
		Ri = Buffer.left(10).toDouble();
		Ra = Buffer.mid(10, 10).toDouble();
		aPot->setPotCoeff(b, Rm, Tm, De, nPotCoeff, PotCoeff);
		LRCoeff[n] = Buffer.mid(46, 15).toDouble();
		if (LRCoeff[n] != 0.0) nLRCoeff++;
		if (pCC < 0) pLRCoeff[n] = -pCC;
		else
		{
			pLRCoeff[n] = pCC;
			aPot->setLRCoeff(Ra, nLRCoeff, pLRCoeff, LRCoeff);
			double E1 = aPot->getPoint(Ra, 2), E2 = aPot->getPoint(Ra, 3);
			pLRCoeff[nLRCoeff] = (pCC += 2);
			LRCoeff[nLRCoeff++] = (E2 - E1) / pow(Ra, -pCC);
		}
		aPot->setLRCoeff(Ra, nLRCoeff, pLRCoeff, LRCoeff);
		aPot->setInnerWall(Ri, 0.0, iExp, 0.0);
		aPot->cdConnectSR();
		Pot[p]->setWorker(aPot);
	}
	while (!St.atEnd())
	{
		Buffer = St.readLine();
		if (Buffer.indexOf("q_e sigma") >= 0) q_e_si = (Buffer.split(':'))[1].toDouble();
		else if (Buffer.indexOf("q_e pi") >= 0) q_e_pi = (Buffer.split(':'))[1].toDouble();
		else if (Buffer.indexOf("q_f pi") >= 0) q_f_pi = (Buffer.split(':'))[1].toDouble();
		else if (Buffer.indexOf("ASO_asym") >= 0) ASO_asym = (Buffer.split(':'))[1].toDouble() * hartree_cm;
		else if (Buffer.indexOf("ASO_scaling") >= 0) 
		{
			ASO_scaling = (Buffer.split(':'))[1].toDouble();
			if (ASO == 0) return false;
			for (n=0; n<2; n++) ASO[n] *= ASO_scaling;
			if (LRCoeff == 0) return false;
			LRCoeff[0] *= ASO_scaling;
			Pot[p] = MW->CreatePotential();
			if (Pot[p] == 0) return false;
			couplingFuncs = new MTTPot(Pot[p]->Fit);
			couplingFuncs->setCoefficients(0.0, 0.0, 2, ASO, alpha, beta, 0.0, 1, pLRCoeff, LRCoeff, ASO_asym);
			Pot[p++]->setWorker(couplingFuncs);
		}
		else if ((m = Buffer.indexOf("ASO_")) >= 0)
		{
			if ((n = Buffer.mid(m+4, 1).toInt()) < 0 || n>1) return false;
			if (ASO == 0) ASO = new double[2];
			ASO[n] = (Buffer.split(':'))[1].toDouble() * hartree_cm;
			if (n==1) ASO[1] /= RFakt;
		}
		else if (Buffer.indexOf("alpha_xi") >= 0) 
		{
			alpha_xi = (Buffer.split(':'))[1].toDouble() / RFakt; 
			if (xi == 0 || LRCoeff == 0) return false;
			Pot[p] = MW->CreatePotential();
			if (Pot[p] == 0) return false;
			couplingFuncs = new MTTPot(Pot[p]->Fit);
			couplingFuncs->setCoefficients(B_xi, alpha_xi, 3, xi, 0.0, beta, gamma, 1, pLRCoeff, LRCoeff, 0.0);
			Pot[p++]->setWorker(couplingFuncs);
		}
		else if (Buffer.indexOf("alpha") >= 0) alpha = (Buffer.split(':'))[1].toDouble() / RFakt;
		else if (Buffer.indexOf("beta") >= 0) beta = (Buffer.split(':'))[1].toDouble() / RFakt;
		else if (Buffer.indexOf("SO_C6") >= 0)
		{
			pLRCoeff = new int[1];
			pLRCoeff[0] = 6;
			LRCoeff = new double[1];
			LRCoeff[0] = -(Buffer.split(':'))[1].toDouble() * hartree_cm * pow(RFakt, 6);
		}
		else if (Buffer.indexOf("xi_scaling") >= 0) xi_scaling = (Buffer.split(':'))[1].toDouble();
		else if (Buffer.indexOf("gamma") >= 0) gamma = (Buffer.split(':'))[1].toDouble() / (RFakt * RFakt);
		else if (Buffer.indexOf("xi_C4") >= 0)
		{
			LRCoeff = new double[1];
			LRCoeff[0] = -(Buffer.split(':'))[1].toDouble() * hartree_cm * pow(RFakt, 4) * xi_scaling;
			pLRCoeff = new int[1];
			pLRCoeff[0] = 4;
		}
		else if (Buffer.indexOf("B_xi") >= 0) B_xi = (Buffer.split(':'))[1].toDouble() * hartree_cm * xi_scaling;
		else if ((m = Buffer.indexOf("xi_")) >= 0)
		{
			if ((n = Buffer.mid(m+3, 1).toInt()) < 0 || n>2) return false;
			if (xi == 0) xi = new double[2];
			xi[n] = (Buffer.split(':'))[1].toDouble() * xi_scaling * (n==0 ? hartree_cm : hartree_cm / (n==1 ? RFakt : RFakt * RFakt));
		}
	}
	N = p;
	if (q_e_si != 0.0) Worker->setLambdaDoubling(q_e_si, 0.0);
	if (q_e_pi != 0.0 || q_f_pi != 0.0) Pot[1]->Worker->setLambdaDoubling(q_e_pi, q_f_pi);
	if (CoupledPot != 0)
	{
		delete[] CoupledPot;
		delete[] CoupledComp;
	}
	CoupledPot = new Potential*[NCoupled = N - 1];
	CoupledComp = new int[NCoupled];
	for (p=0; p<N; p++)
	{
		Pot[p]->UpdateTab();
		Pot[p]->setImported();
		Pot[p]->Tab->setHorizontalHeaderLabels(QStringList() << "coefficient" << "value" << "error" 
					<< "sig. digits");
		Pot[p]->Changed();
		if (p>0) (CoupledPot[p-1] = Pot[p])->show();
	}
	setImported();
	Changed();
	return true;
}

bool Potential::readPotData(QString FileName)
{
	if (Fit->isRunning()) return false;
	double De = 0.0, iCoeff, iOff, Ra, Ri, Tm, *LRCoeff, *PotCoeff, iExp, ExcInt_A, ExcInt_Alpha, ExcInt_gamma, AdCorr_Rm, AdCorr_b;
	double RIso1AdCorr, RIso2AdCorr, *adCorr;
	bool *adCorrFree;
	int nPotCoeff, nLRCoeff, *pLRCoeff, pCC, n, m, NAdCorr, PAdCorr, TAdCorr;
	QFile File(FileName);
	if (!read(&File)) return false;
	QTextStream St(&File);
	QString Buffer;
	AnaPot *aPot = new AnaPot(Fit);
	Buffer = St.readLine();
	if (Buffer.left(5).toInt() > 0) return false;
	setName((getFName().split('.'))[0]);
	setSource(Buffer);
	Buffer = St.readLine().replace('D', 'e');
	Tm = Buffer.left(17).toDouble();
	AdCorr_Rm = Buffer.right(Buffer.length() - 17).toDouble();
	Buffer = St.readLine();
	pCC = Buffer.left(5).toInt();
	De = Buffer.mid(5, 15).toDouble();
	if ((iExp = Buffer.mid(20, 5).toInt()) > 50.0) iExp -= 50.0;
	Buffer = St.readLine().replace('D', 'e');
	nPotCoeff = Buffer.left(5).toInt();
	AdCorr_b = Buffer.mid(5, 15).toDouble();
	if (nPotCoeff == 0) return false;
	PotCoeff = new double[nPotCoeff];
	for (n=0; n < nPotCoeff; n++)
	{
		Buffer = St.readLine().replace('D', 'e');
		m = Buffer.left(5).toInt();
		PotCoeff[m-1] = Buffer.mid(5, 25).toDouble();
	}
	Buffer = St.readLine();
	NAdCorr = Buffer.left(5).toInt();
	PAdCorr = Buffer.mid(15, 5).toInt();
	RIso1AdCorr = Buffer.mid(20, 15).toDouble();
	RIso2AdCorr = Buffer.mid(35, 15).toDouble();
	TAdCorr = Buffer.mid(50, 5).toInt();
	if (NAdCorr > 0)
	{
		adCorr = new double[NAdCorr];
		adCorrFree = new bool[NAdCorr];
		for (n=0; n < NAdCorr; n++) 
		{
			Buffer = St.readLine().replace('D', 'e');
			adCorr[n] = Buffer.right(Buffer.length() - 5).toDouble();
			adCorrFree[n] = true;
		}
	}
	else
	{
		adCorr = 0;
		adCorrFree = 0;
	}
	St.readLine();
	Buffer = St.readLine();
	nLRCoeff = Buffer.toInt();
	if (nLRCoeff == 0) return false;
	pLRCoeff = new int[nLRCoeff + 2];
	LRCoeff = new double[nLRCoeff + 2];
	for (n=0; n < nLRCoeff; n++)
	{
		Buffer = St.readLine().replace('D', 'e');
		pLRCoeff[n] = Buffer.left(5).toInt();
		LRCoeff[n] = Buffer.mid(5, 25).toDouble();
	}
	Buffer = St.readLine().replace('D', 'e');
	ExcInt_A = Buffer.left(15).toDouble();
	ExcInt_Alpha = Buffer.mid(30, 15).toDouble();
	ExcInt_gamma = Buffer.mid(15, 15).toDouble();
	Buffer = St.readLine().replace('D', 'e');
	Ri = Buffer.left(10).toDouble();
	Ra = Buffer.mid(10, 10).toDouble();
	if (Ri < 0.0 || Ri >= Ra) return false;
	Buffer = St.readLine().replace('D', 'e');
	iOff = Buffer.left(27).toDouble();
	iCoeff = Buffer.mid(27, 27).toDouble();
	if (iExp < 0.0) iExp = Buffer.mid(54, 27).toDouble();
	Buffer = St.readLine().replace('D', 'e');
	pLRCoeff[n] = pCC;
	pLRCoeff[n+1] = pCC + 2;
	LRCoeff[n] = Buffer.left(27).toDouble();
	LRCoeff[n+1] = Buffer.mid(27, 27).toDouble();
	//printf("readPotData: iExp = %f\n", iExp);
	if (LRCoeff[n] != 0.0 && LRCoeff[n+1] != 0.0) nLRCoeff += 2;
	//printf("Buffer=%s, iOff=%f, iCoeff=%f\n", Buffer.ascii(), iOff, iCoeff);
	//printf("Left=%s, Mid=%s\n", Buffer.left(27).ascii(), Buffer.mid(27,27).ascii());
	aPot->setInnerWall(Ri, iOff, iExp, iCoeff);
	aPot->setPotCoeff(AdCorr_b, AdCorr_Rm, Tm, De, nPotCoeff, PotCoeff);
	aPot->setLRCoeff(Ra, nLRCoeff, pLRCoeff, LRCoeff);
	aPot->setExchangeCoeff(ExcInt_A, ExcInt_Alpha, ExcInt_gamma);
	//aPot->sdConnectLR(pCC);
	UpdateTab();
	setWorker(aPot);
	Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b);
	Worker->setLambdaDoubling(0.0, 0.0);
	
	/*double Rmin = 3.3, Rmax = 100.0;
	int v, Np = 100000;
	double *R = aPot->getPointRep(Rmin, Rmax, Np), U[Np];
	if (R==NULL) 
	{
		printf("Fehler: U=NULL!\n");
		return false;
	}
	double h = (Rmax - Rmin) / (Np - 1);//, R;//[Np];
	CRotPot(19, 0, R, U, Rmin, h, Np);
	
	QFile DDatei("DebugPot.dat");
	DDatei.open(QIODevice::WriteOnly);
	QTextStream S1(&DDatei);
	for (n=1, R[0]=Rmin; n<Np; n++) //R[n] = R[n-1] + h;
		S1 << QString::number(R+=h, 'g', 10) << "	" << QString::number(U[n], 'g', 10) << "\n";
	for (n=0, R=Rmin; R <= 10.0; n++) 
	{
		if (R > 9.5) printf("R=%f, V=%f\n", R += h, U[n]);
		R+=h;
	}*/
	//double DE;//, *WF = new double[Np];
	//Numerov(9, h, -500.0, DE, v, Np, U, WF);
	//printf("v=%d\n", v);
	/*double **WF = Create(Maxv, Np), *E = new double[Maxv];
	printf("Vor NumerovCooley\n");
	NumerovCooley(0, h, n, E, WF, Np, U);
	for (m=0; m<n; m++) printf("v=%d, E=%f\n", m, E[m]);
	for (n=0; R[n] < Ri; n++);
	for (m=n; R[m] < Ra; m++);
	m-=n;
	aPot->calcPotCoeff(m, R+n, U+n, nPotCoeff, Rm, b);*/
	/*delete[] E;
	//delete[] U;
	//delete[] WF;
	delete[] R;
	Destroy(WF, Maxv);*/
	setImported();
	Tab->setHorizontalHeaderLabels(QStringList() << "coefficient" << "value" << "error" 
			<< "sig. digits");
	Changed();
	return true;
}

void PotWorker::DestroyLS()
{
	if (L!=0) 
	{
		Destroy(L, numSplinePoints);
		L=0;
	}
	if (S!=0)
	{
		Destroy(S, NumS);
		S=0;
	}
	if (SMap != 0)
	{
		delete[] SMap;
		SMap = 0;
	}
}

void Potential::UpdateTab(PotentialData *D)
{
	//printf("Potential::UpdateTab\n");
	LRCTabOffs = adCorrTabOffs = ExcIntATabPos = -1;
	if (D == 0) D = Worker->getPotentialData();
	Tab->blockSignals(true);
	if (Worker->isAnaPotAv()) 
	{
		int n, r;
		QString Buffer;
		Tab->setRowCount((D->q_f != 0.0 ? 12 : 11) + (D->q_e != 0.0 ? D->nPotCoeff + 1 : D->nPotCoeff) + (D->NAdCorr > 0 ? D->NAdCorr + 4 : 0)
							+ (D->A_ex == 0.0 ? D->nLRCoeff : D->nLRCoeff + 3) + (D->NSpinRGamma > 0 ? D->NSpinRGamma + 2 : 0));
		for (n=1, r=0; n <= D->nPotCoeff; n++, r++)
		{
			Tab->setItem(r, 0, new QTableWidgetItem((D->CFree[r] ? QIcon() : *FixPix), 
													 "a" + QString::number(n) + " [cm^-1]"));
			Tab->setItem(r, 1, new QTableWidgetItem((D->CFree[r] ? QIcon() : *FixPix), 
													QString::number(D->PotCoeff[r], 'E', 16)));
		}
		Tab->setItem(r, 0, new QTableWidgetItem("b"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->b, 'f', 4)));
		Tab->setItem(r, 0, new QTableWidgetItem("Rm [A]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Rm, 'f', 10)));
		Tab->setItem(r, 0, new QTableWidgetItem("Tm [cm^-1]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Tm, 'f', 10)));
		Tab->setItem(r, 0, new QTableWidgetItem("Ri [A]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Ri, 'f', 10)));
		Tab->setItem(r, 0, new QTableWidgetItem("n"));
		Tab->setItem(r++, 1, new QTableWidgetItem(Buffer = QString::number(D->iExp, 'g', 10)));
		//printf("UpdateTab: iExp=%f\n", iExp);
		Tab->setItem(r, 0, new QTableWidgetItem("A [cm^-1]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->iOff, 'f', 10)));
		Tab->setItem(r, 0, new QTableWidgetItem("B [cm^-1 A^" + Buffer + "]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->iCoeff, 'f', 16)));
		Tab->setItem(r, 0, new QTableWidgetItem("Ra [A]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Ra, 'f', 10)));
		LRCTabOffs = r;
		for (n=0; n < D->nLRCoeff; n++)
		{
			Buffer = QString::number(D->pLRCoeff[n]);
			if (n < D->nLRCoeff - 1 ? D->LRCFree[n] : true)
			{
				Tab->setItem(r, 0, new QTableWidgetItem("C" + Buffer + " [cm^-1 A^" + Buffer + "]"));
				Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->LRCoeff[n], 'E', 16)));
			}
			else
			{
				Tab->setItem(r, 0, new QTableWidgetItem(*FixPix, "C" + Buffer + " [cm^-1 A^" + Buffer + "]"));
				Tab->setItem(r++, 1, new QTableWidgetItem(*FixPix, QString::number(D->LRCoeff[n], 'E', 16)));
			}
		}
		Tab->setItem(r, 0, new QTableWidgetItem("U infinity [cm^-1]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->De, 'f', 10)));
		if (D->A_ex != 0.0)
		{
			Tab->setItem(ExcIntATabPos = r, 0, new QTableWidgetItem((D->EIFree ? QIcon() : *FixPix),
						 "A_ex [cm^-1 A^" + QString::number(D->gamma, 'f', 5) + "]"));
			Tab->setItem(r++, 1, new QTableWidgetItem((D->EIFree ? QIcon() : *FixPix), QString::number(D->A_ex, 'E', 8)));
			Tab->setItem(r, 0, new QTableWidgetItem("gamma"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->gamma, 'f', 5)));
			Tab->setItem(r, 0, new QTableWidgetItem("beta [A^-1]"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->beta, 'f', 5)));
		}
		if (D->q_e != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("q_e"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->q_e, 'E', 16)));
		}
		if (D->q_f != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("q_f"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->q_f, 'E', 16)));
		}
		if (D->NAdCorr > 0) 
		{
			adCorrTabOffs = r;
			for (n=0; n < D->NAdCorr; n++)
			{
				Tab->setItem(r, 0, new QTableWidgetItem((D->adCorrFree[n] ? QIcon() : *FixPix), "AdCorr" + QString::number(n)));
				Tab->setItem(r++, 1, new QTableWidgetItem((D->adCorrFree[n] ? QIcon() : *FixPix), QString::number(D->adCorr[n], 'E', 16)));
			}
			Tab->setItem(r, 0, new QTableWidgetItem("Type AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->TAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("Pow AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->PAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso1AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso1AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso2AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso2AdCorr, 'f', 9)));
		}
		if (D->NSpinRGamma > 0)
		{
			for (n=0; n < D->NSpinRGamma; n++)
			{
				Tab->setItem(r, 0, new QTableWidgetItem("SpinRGamma" + QString::number(n)));
				Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->SpinRGamma[n], 'E', 16)));
			}
			Tab->setItem(r, 0, new QTableWidgetItem("SpinRR1"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->SpinRR1, 'E', 10)));
			Tab->setItem(r, 0, new QTableWidgetItem("SpinRR2"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->SpinRR2, 'E', 10)));
		}
		Tab->setItem(r, 0, new QTableWidgetItem("R min [A]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(rmin, 'f', 10)));
		Tab->setItem(r, 0, new QTableWidgetItem("R max [A]"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(rmax, 'f', 10)));
	}
	else if (Worker->isTTPot())
	{
		int r, n;
		Tab->setRowCount(D->N + D->NA + (D->NAdCorr > 0 ? D->NAdCorr + 8 : 2));
		for (n=0; n < D->NA; n++)
		{
			Tab->setItem(n, 0, new QTableWidgetItem("A_" + QString::number(n)));
			Tab->setItem(n, 1, new QTableWidgetItem(QString::number(D->A[n], 'f', 10)));
		}
		Tab->setItem(n, 0, new QTableWidgetItem("b"));
		Tab->setItem(n, 1, new QTableWidgetItem(QString::number(D->b, 'f', 10)));
		LRCTabOffs = r = D->NA + 1;
		for (n=0; n < D->N; n++, r++)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("C_" + QString::number(2*(n+3))));
			Tab->setItem(r, 1, new QTableWidgetItem(QString::number(D->C[n], 'E', 16)));
		}
		Tab->setItem(r, 0, new QTableWidgetItem("U_inf"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Uinf, 'f', 10)));
		if (D->NAdCorr > 0)
		{
			adCorrTabOffs = r;
			for (n=0; n < D->NAdCorr; n++, r++)
			{
				Tab->setItem(r, 0, new QTableWidgetItem("AdCorr" + QString::number(n)));
				Tab->setItem(r, 1, new QTableWidgetItem(QString::number(D->adCorr[n], 'E', 16)));
			}
			Tab->setItem(r, 0, new QTableWidgetItem("Type AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->TAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("Pow AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->PAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso1AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso1AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso2AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso2AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("AdCorr b"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->AdCorr_b, 'f', 4)));
			Tab->setItem(r, 0, new QTableWidgetItem("AdCorr Rm"));
			Tab->setItem(r, 1, new QTableWidgetItem(QString::number(D->AdCorr_Rm, 'f', 10)));
		}
	}
	else if (Worker->isLPot())
	{
		int n, r;
		Tab->setRowCount(D->NCi + D->Nb + (D->NAdCorr > 0 ? D->NAdCorr + 12 : 6));
		if (Tab->item(0, 0) != 0) Tab->item(0, 0)->setText("p");
		else Tab->setItem(0, 0, new QTableWidgetItem("p"));
		if (Tab->item(0, 1) != 0) Tab->item(0, 1)->setText(QString::number(D->p));
		else Tab->setItem(0, 1, new QTableWidgetItem(QString::number(D->p)));
		if (Tab->item(1, 0) != 0) Tab->item(1, 0)->setText("q");
		else Tab->setItem(1, 0, new QTableWidgetItem("q"));
		if (Tab->item(1, 1) != 0) Tab->item(1, 1)->setText(QString::number(D->q));
		else Tab->setItem(1, 1, new QTableWidgetItem(QString::number(D->q)));
		if (Tab->item(2, 0) != 0) Tab->item(2, 0)->setText("R_ref [A]");
		else Tab->setItem(2, 0, new QTableWidgetItem("R_ref [A]"));
		if (Tab->item(2, 1) != 0) Tab->item(2, 1)->setText(QString::number(D->Rref, 'f', 2));
		else Tab->setItem(2, 1, new QTableWidgetItem(QString::number(D->Rref, 'f', 2)));
		if (Tab->item(3, 0) != 0) Tab->item(3, 0)->setText("R_e [A]");
		else Tab->setItem(3, 0, new QTableWidgetItem("R_e [A]"));
		if (Tab->item(3, 1) != 0) Tab->item(3, 1)->setText(QString::number(D->Re, 'f', 8));
		else Tab->setItem(3, 1, new QTableWidgetItem(QString::number(D->Re, 'f', 8)));
		if (Tab->item(4, 0) != 0) Tab->item(4, 0)->setText("D_e [cm^-1]");
		else Tab->setItem(4, 0, new QTableWidgetItem("D_e [cm^-1]"));
		if (Tab->item(4, 1) != 0) Tab->item(4, 1)->setText(QString::number(D->De, 'f', 4));
		else Tab->setItem(4, 1, new QTableWidgetItem(QString::number(D->De, 'f', 4)));
	 	if (Tab->item(5, 0) != 0) Tab->item(5, 0)->setText("U_infinity [cm^-1]");
		else Tab->setItem(5, 0, new QTableWidgetItem("U_infinity [cm^-1]"));
		if (Tab->item(5, 1) != 0) Tab->item(5, 1)->setText(QString::number(D->T, 'f', 4));
		else Tab->setItem(5, 1, new QTableWidgetItem(QString::number(D->T, 'f', 4)));
		for (n=0; n < D->Nb; n++)
		{
			if (Tab->item(n+6, 0) != 0) 
				Tab->item(n+6, 0)->setText("beta_" + QString::number(n));
			else Tab->setItem(n+6, 0, 
							  new QTableWidgetItem("beta_" + QString::number(n)));
			if (Tab->item(n+6, 1) != 0) Tab->item(n+6, 1)->setText(QString::number(D->bA[n], 'f', 8));
			else Tab->setItem(n+6, 1, new QTableWidgetItem(QString::number(D->bA[n], 'f', 8)));
		}
		LRCTabOffs = n + D->Nb + 6;
		for (n=0; n < D->NCi; n++)
		{
			if (Tab->item(n + D->Nb + 6, 0) != 0) 
				Tab->item(n + D->Nb + 6, 0)->setText("C_" + QString::number(D->pCi[n]) + " [cm^-1 A^"
												  + QString::number(D->pCi[n]) + "]");
			else Tab->setItem(n + D->Nb + 6, 0, new QTableWidgetItem("C_" + QString::number(D->pCi[n]) 
													+ " [cm^-1 A^" + QString::number(D->pCi[n]) + "]"));
			if (Tab->item(n + D->Nb + 6, 1) != 0) 
				Tab->item(n + D->Nb + 6, 1)->setText(QString::number(D->Ci[n], 'e', 10));
			else Tab->setItem(n + D->Nb + 6, 1, new QTableWidgetItem(QString::number(D->Ci[n], 'e', 10)));
		}
		if (D->NAdCorr > 0) 
		{
			adCorrTabOffs = r = n + D->Nb + 7;
			for (n=0; n < D->NAdCorr; n++, r++)
			{
				Tab->setItem(r, 0, new QTableWidgetItem("AdCorr" + QString::number(n)));
				Tab->setItem(r, 1, new QTableWidgetItem(QString::number(D->adCorr[n], 'E', 16)));
			}
			Tab->setItem(r, 0, new QTableWidgetItem("Type AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->TAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("Pow AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->PAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso1AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso1AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso2AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso2AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("AdCorr b"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->AdCorr_b, 'f', 4)));
			Tab->setItem(r, 0, new QTableWidgetItem("AdCorr Rm"));
			Tab->setItem(r, 1, new QTableWidgetItem(QString::number(D->AdCorr_Rm, 'f', 10)));
		}
	}
	else if (Worker->isMTTPot())
	{
		int n, r=0;
		Tab->setRowCount(D->NA + 7 + D->NLRC + D->NSA + D->NSLRC);
		if (D->B != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("B"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->B, 'g', 15)));
		}
		if (D->alpha1 != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("alpha1"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->alpha1, 'f', 14)));
		}
		for (n=0; n < D->NA; n++)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("A" + QString::number(n)));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->A[n], 'g', 15)));
		}
		Tab->setItem(r, 0, new QTableWidgetItem("alpha"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->alpha, 'f', 14)));
		Tab->setItem(r, 0, new QTableWidgetItem("beta"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->beta, 'f', 14)));
		if (D->gamma != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("gamma"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->gamma, 'f', 14)));
		}
		LRCTabOffs = r;
		for (n=0; n < D->NLRC; n++)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("C" + QString::number(D->pLRC[n])));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->LRC[n], 'g', 15)));
		}
		Tab->setItem(r, 0, new QTableWidgetItem("U_inf"));
		Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Uinf, 'f', 4)));
		for (n=0; n < D->NSA; n++)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("SA" + QString::number(n)));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Salpha, 'f', 14)));
		}
		if (D->Salpha != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("Salpha"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Salpha, 'f', 14)));
		}
		if (D->Sbeta != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("Sbeta"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Sbeta, 'f', 14)));
		}
		for (n=0; n < D->NSLRC; n++)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("SC" + QString::number(D->pSLRC[n])));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->SLRC[n], 'g', 15)));
		}
		if (D->Sinf != 0.0)
		{
			Tab->setItem(r, 0, new QTableWidgetItem("S_inf"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->Sinf, 'f', 4)));
		}
		if (D->NAdCorr > 0) 
		{
			adCorrTabOffs = r;
			for (n=0; n < D->NAdCorr; n++, r++)
			{
				Tab->setItem(r, 0, new QTableWidgetItem("AdCorr" + QString::number(n)));
				Tab->setItem(r, 1, new QTableWidgetItem(QString::number(D->adCorr[n], 'E', 16)));
			}
			Tab->setItem(r, 0, new QTableWidgetItem("Type AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->TAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("Pow AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->PAdCorr)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso1AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso1AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("RIso2AdCorr"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->RIso2AdCorr, 'f', 9)));
			Tab->setItem(r, 0, new QTableWidgetItem("AdCorr b"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->AdCorr_b, 'f', 4)));
			Tab->setItem(r, 0, new QTableWidgetItem("AdCorr Rm"));
			Tab->setItem(r++, 1, new QTableWidgetItem(QString::number(D->AdCorr_Rm, 'f', 10)));
		}
		Tab->setRowCount(r);
	}
	else if (Worker->isSplinePot())
	{
		int c, n;
		Tab->setRowCount(D->N + D->NLRC + (D->NAdCorr > 0 ? D->NAdCorr + 6 : 0) 
						  + (D->Ro == 0 ? (D->iExp != 6.0 ? 4 : 3) : 2));
		for (n=0; n < D->N; n++)
		{
			if (D->points[n].variable)
			{
				Tab->setItem(n, 0, new QTableWidgetItem(QString::number(D->points[n].x, 'f', 6)));
				Tab->setItem(n, 1, new QTableWidgetItem(QString::number(D->points[n].y, 'f', 6)));
				Tab->setItem(n, 2, new QTableWidgetItem(QString::number(D->points[n].yss, 'f', 6)));
			}
			else
			{
				Tab->setItem(n, 0, new QTableWidgetItem(*FixPix, QString::number(D->points[n].x, 'f', 6)));
				Tab->setItem(n, 1, new QTableWidgetItem(*FixPix, QString::number(D->points[n].y, 'f', 6)));
				Tab->setItem(n, 2, new QTableWidgetItem(*FixPix, QString::number(D->points[n].yss, 'f', 6)));
			}
		}
		if (D->Ro == 0.0)
		{
			if (D->shortRVariable)
			{
				Tab->setItem(n, 0, new QTableWidgetItem("iA"));
				Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->iA, 'e', 12)));
				Tab->setItem(n, 0, new QTableWidgetItem("iO"));
				Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->iO, 'f', 4)));
				if (D->iExp != 6.0)
				{
					Tab->setItem(n, 0, new QTableWidgetItem("iExp"));
					Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->iExp, 'f', 10)));
				}
			}
			else
			{
				Tab->setItem(n, 0, new QTableWidgetItem(*FixPix, "iA"));
				Tab->setItem(n++, 1, new QTableWidgetItem(*FixPix, QString::number(D->iA, 'e', 12)));
				Tab->setItem(n, 0, new QTableWidgetItem(*FixPix, "iO"));
				Tab->setItem(n++, 1, new QTableWidgetItem(*FixPix, QString::number(D->iO, 'f', 4)));
				if (D->iExp != 6.0)
				{
					Tab->setItem(n, 0, new QTableWidgetItem(*FixPix, "iExp"));
					Tab->setItem(n++, 1, new QTableWidgetItem(*FixPix, QString::number(D->iExp, 'f', 10)));
				}
			}
		}
		else
		{
			Tab->setItem(n, 0, new QTableWidgetItem("R_o"));
			Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->Ro, 'f', 4)));
		}
		Tab->setItem(n, 0, new QTableWidgetItem("U_inf"));
		Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->Uinf, 'f', 4)));
		LRCTabOffs = n;
		for (c=0; c < D->NLRC; c++)
		{
			Tab->setItem(n, 0, 
						 new QTableWidgetItem((D->LRCFree[c] ? QIcon() : *FixPix), 
											  "C_" + QString::number(D->PLRC[c])));
			Tab->setItem(n++, 1, 
						 new QTableWidgetItem((D->LRCFree[c] ? QIcon() : *FixPix),
											  QString::number(D->LRC[c], 'e', 12)));
		}
		if (D->NAdCorr > 0) 
		{
			adCorrTabOffs = n;
			for (c=0; c < D->NAdCorr; n++, c++)
			{
				Tab->setItem(n, 0, 
							 new QTableWidgetItem((D->adCorrFree[c] ? QIcon() : *FixPix),
												  "AdCorr" + QString::number(c)));
				Tab->setItem(n, 1, 
							 new QTableWidgetItem((D->adCorrFree[c] ? QIcon() : *FixPix),
												  QString::number(D->adCorr[c], 'E', 16)));
			}
			Tab->setItem(n, 0, new QTableWidgetItem("Type AdCorr"));
			Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->TAdCorr)));
			Tab->setItem(n, 0, new QTableWidgetItem("Pow AdCorr"));
			Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->PAdCorr)));
			Tab->setItem(n, 0, new QTableWidgetItem("RIso1AdCorr"));
			Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->RIso1AdCorr, 'f', 9)));
			Tab->setItem(n, 0, new QTableWidgetItem("RIso2AdCorr"));
			Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->RIso2AdCorr, 'f', 9)));
			Tab->setItem(n, 0, new QTableWidgetItem("AdCorr b"));
			Tab->setItem(n++, 1, new QTableWidgetItem(QString::number(D->AdCorr_b, 'f', 4)));
			Tab->setItem(n, 0, new QTableWidgetItem("AdCorr Rm"));
			Tab->setItem(n, 1, new QTableWidgetItem(QString::number(D->AdCorr_Rm, 'f', 10)));
		}
	}
	//printf("Ende UpdateTab\n");
	Tab->blockSignals(false);
	delete D;
}

void Potential::VaryCoefficients(bool V, int NRows, int *Rows)
{
	if (!(Worker->isAnaPotAv() || Worker->isSplinePot())) 
	{
		printf("Potential::VaryCoefficients is not implemented for the case numSplinePoints == 0 && aPot == 0 !\n");
		return;
	}
	if (Fit->isRunning()) return;
	int i, r, N;
	if (NRows == 0)
	{
		QList<QTableWidgetSelectionRange> SR = Tab->selectedRanges();
		for (i=N=0; i < SR.count(); i++) for (r=SR[i].topRow(); r <= SR[i].bottomRow(); r++) N++;
		Rows = new int[N];
		for (i=N=0; i < SR.count(); i++) for (r=SR[i].topRow(); r <= SR[i].bottomRow(); r++) Rows[N++] = r;
	}
	else N = NRows;
	Tab->blockSignals(true);
	if (Worker->isSplinePot())
	{
		int numSplinePoints, NLRC, *PLRC, NAdCorr, TAdCorr, PAdCorr;
		SplinePoint *points;
		double *LRC, iA, iO, iExp, *adCorr, RIso1, RIso2, Rm, b;
		bool shortRVariable = Worker->getShortRVariable(), *LRCfree = Worker->getLRCFree(), *adCorrFree;
		Worker->getSplinePotForWriting(numSplinePoints, points, NLRC, PLRC, LRC, iA, iO, iExp);
		Worker->getAdCorrForReading(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1, RIso2, Rm, b, &adCorrFree, true);
		int NSR = (iExp == 6.0 ? 2 : 3), c;
		//bool GS = (State != 0 ? State->getStateNum() == 0 : false);
		for (r=0; r<N; r++)
		{
			if (Rows[r] < numSplinePoints) 
			{
				if ((points[Rows[r]].variable = V)) 
				{
					Tab->item(Rows[r], 0)->setIcon(QIcon());
					Tab->item(Rows[r], 1)->setIcon(QIcon());
				}
				else
				{
					Tab->item(Rows[r], 0)->setIcon(*FixPix);
					Tab->item(Rows[r], 1)->setIcon(*FixPix);
				}
			}
			else if (Rows[r] < numSplinePoints + NSR)
			{
				if ((shortRVariable = V))
				{
					Tab->item(numSplinePoints, 0)->setIcon(QIcon());
					Tab->item(numSplinePoints, 1)->setIcon(QIcon());
					Tab->item(numSplinePoints + 1, 0)->setIcon(QIcon());
					Tab->item(numSplinePoints + 1, 1)->setIcon(QIcon());
				}
				else
				{
					Tab->item(numSplinePoints, 0)->setIcon(*FixPix);
					Tab->item(numSplinePoints, 1)->setIcon(*FixPix);
					Tab->item(numSplinePoints + 1, 0)->setIcon(*FixPix);
					Tab->item(numSplinePoints + 1, 1)->setIcon(*FixPix);
				}
				//r = numSplinePoints + NSR;
			}
			else if (Rows[r] < numSplinePoints + NSR + NLRC + 1)
			{
				//if (GS)
				//{
					if ((c = Rows[r] - numSplinePoints - NSR - 1) >= 0)
					{
						if ((LRCfree[c] = V))
						{
							Tab->item(Rows[r], 0)->setIcon(QIcon());
							Tab->item(Rows[r], 1)->setIcon(QIcon());
						}
						else
						{
							Tab->item(Rows[r], 0)->setIcon(*FixPix);
							Tab->item(Rows[r], 1)->setIcon(*FixPix);
						}
					}
				/*}
				else
				{
					if (V) for (r = numSplinePoints + NSR; r < numSplinePoints + NSR + NLRC + 1; r++)
					{
						Tab->item(r, 0)->setIcon(QIcon());
						Tab->item(r, 1)->setIcon(QIcon());
					}
					else for (r = numSplinePoints + 2; r < Tab->rowCount(); r++)
					{
						Tab->item(r, 0)->setIcon(*FixPix);
						Tab->item(r, 1)->setIcon(*FixPix);
					}
					for (c=0; c < NLRC; c++) LRCfree[c] = V;
				}*/
			}
			else if ((c = Rows[r] - numSplinePoints - NSR - NLRC - 1) < NAdCorr)
			{
				if ((adCorrFree[c] = V))
				{
					Tab->item(Rows[r], 0)->setIcon(QIcon());
					Tab->item(Rows[r], 1)->setIcon(QIcon());
				}
				else
				{
					Tab->item(Rows[r], 0)->setIcon(*FixPix);
					Tab->item(Rows[r], 1)->setIcon(*FixPix);
				}
			}
		}
	}
	else
	{
		for (r=0; r<N; r++)
		{
			Tab->item(Rows[r], 0)->setIcon(V ? QIcon() : *FixPix);
			Tab->item(Rows[r], 1)->setIcon(V ? QIcon() : *FixPix);
		}
		updatePot();
	}
	if (N>0) 
	{
		if (NRows == 0) delete[] Rows;
		Worker->DestroySMap();
	}
	Tab->blockSignals(false);
}

void Potential::getFixedCoefficients(int& N, int*& Rows)
{
	int n, NR = Tab->rowCount();
	for (n=N=0; n < NR; n++) if (!Tab->item(n, 0)->icon().isNull()) N++;
	Rows = new int[N];
	for (n=N=0; n < NR; n++) if (!Tab->item(n, 0)->icon().isNull()) Rows[N++] = n;
}

void Potential::updatePot()
{
	UpdatePot(Worker->getType() != NoPotential ? -1 : -2);
	Changed();
}

void Potential::UpdatePot(int type)
{
	if (Fit->isRunning()) return;
	if (type == -2)
	{
		if (testIfSplinePot())
		{
			type = 1;
			/*if (2.0 * (points[N-2].x - points[N-3].x) < points[N-1].x - points[N-2].x 
						 && points[N-1].yss != 0.0) 
			{
				printf("SLIC6!\n");
				SLIC6 = true;
			}*/
		}
		else
		{
			Tab->setHorizontalHeaderLabels(QStringList() 
					<< "coefficient" << "value" << "error" << "sig. digits");
			if (Tab->item(2, 0)->text().indexOf("R_ref") >= 0) type = 2;
			else if (Tab->item(0, 0)->text().indexOf("A0") >= 0) type = 3;
			else type = 0;
		}
	}
	LRCTabOffs = adCorrTabOffs = ExcIntATabPos = -1;
	if (type == 0 || (type == -1 && Worker->isAnaPotAv()))
	{
		int nPotCoeff, nLRCoeff, *pLRCoeff, n, r, NR, NAdCorr, TAdCorr = 0, PAdCorr = 0, NSpinRGamma;
		double Ri, Ra, iExp, iOff, iCoeff, Tm, De, *PotCoeff, *LRCoeff, A_ex = 0.0, beta = 0.0, gamma = 0.0, *adCorr, AdCorr_b, AdCorr_Rm;
		double RIso1AdCorr = 0.0, RIso2AdCorr = 0.0, *SpinRGamma, SpinRR1 = 0.0, SpinRR2 = 0.0, q_e = 0.0, q_f = 0.0;
		bool *LRCFree, *CoeffFree, ExcIntFree = false, *adCorrFree;
		QString Buffer;
		NR = Tab->rowCount();
		QTableWidgetItem *I;
		for (nPotCoeff = 0; (nPotCoeff < NR ? ((I = Tab->item(nPotCoeff, 0)) != 0 ? 
				I->text().left(1) == "a" : false) : false); nPotCoeff++) ;
		//printf("NR = %d, nPotCoeff=%d\n", NR, nPotCoeff);
		PotCoeff = new double[nPotCoeff];
		CoeffFree = new bool[nPotCoeff];
		for (n=0; n < nPotCoeff; n++) 
		{
			PotCoeff[n] = ((I = Tab->item(n, 1)) != 0 ? I->text().toDouble() : 0.0);
			CoeffFree[n] = (I != 0 ? I->icon().isNull() : true);
		}
		if (NR < n + 8) return;
		AdCorr_b = ((I = Tab->item(n, 1)) != 0 ? I->text().toDouble() : 0.0);
		AdCorr_Rm = ((I = Tab->item(n+1, 1)) != 0 ? I->text().toDouble() : 0.0);
		Tm = ((I = Tab->item(n+2, 1)) != 0 ? I->text().toDouble() : 0.0);
		Ri = ((I = Tab->item(n+3, 1)) != 0 ? I->text().toDouble() : 0.0);
		iExp = ((I = Tab->item(n+4, 1)) != 0 ? I->text().toDouble() : 0.0);
		iOff = ((I = Tab->item(n+5, 1)) != 0 ? I->text().toDouble() : 0.0);
		iCoeff = ((I = Tab->item(n+6, 1)) != 0 ? I->text().toDouble() : 0.0);
		Ra = ((I = Tab->item(n+7, 1)) != 0 ? I->text().toDouble() : 0.0);
		for (r=n+8; (r < NR ? ((I = Tab->item(r, 0)) != 0 ? I->text().left(1) == "C" : false) : false);
			 	r++) ;
		LRCoeff = new double[nLRCoeff = r - n - 8];
		pLRCoeff = new int[nLRCoeff];
		LRCFree = new bool[nLRCoeff];
		LRCTabOffs = (r -= nLRCoeff);
		for (n=0; n < nLRCoeff; n++) 
		{
			Buffer = Tab->item(n+r, 0)->text();
			pLRCoeff[n] = Buffer.mid(1, Buffer.indexOf(" ") - 1).toInt();
			LRCoeff[n] = ((I = Tab->item(n+r, 1)) != 0 ? I->text().toDouble() : 0.0);
			LRCFree[n] = Tab->item(n+r, 0)->icon().isNull();
		}
		De = (NR > r + n ? ((I = Tab->item(n+r, 1)) != 0 ? I->text().toDouble() : 0.0) : 0.0);
		if (NR > (r+=n) + 3) if (Tab->item(r+1, 0)->text().left(4) == "A_ex")
		{
			ExcIntATabPos = (++r);
			A_ex = Tab->item(r, 1)->text().toDouble();
			ExcIntFree = Tab->item(r, 0)->icon().isNull();
			gamma = Tab->item(++r, 1)->text().toDouble();
			beta = Tab->item(++r, 1)->text().toDouble();
		}
		if (NR > r+1 ? Tab->item(r+1, 0)->text().left(3) == "q_e" : false) q_e = Tab->item(++r, 1)->text().toDouble();
		if (NR > r+1 ? Tab->item(r+1, 0)->text().left(3) == "q_f" : false) q_f = Tab->item(++r, 1)->text().toDouble();
		for (NAdCorr = 0, r++; (NR > r + NAdCorr ? Tab->item(r + NAdCorr, 0)->text().left(6) == "AdCorr" : false); NAdCorr++) ;
		if (NAdCorr > 0)
		{
			adCorr = new double[NAdCorr];
			adCorrFree = new bool[NAdCorr];
			adCorrTabOffs = n;
			for (n=0; n < NAdCorr; n++, r++) 
			{
				adCorrFree[n] = Tab->item(r, 0)->icon().isNull();
				adCorr[n] = (Tab->item(r, 1) != 0 ? Tab->item(r, 1)->text().toDouble() : 0.0);
			}
			TAdCorr = Tab->item(r++, 1)->text().toInt();
			PAdCorr = Tab->item(r++, 1)->text().toInt();
			RIso1AdCorr = Tab->item(r++, 1)->text().toDouble();
			RIso2AdCorr = Tab->item(r++, 1)->text().toDouble();
		}
		else 
		{
			adCorr = 0;
			adCorrFree = 0;
		}
		for (NSpinRGamma = 0; (NR > r + NSpinRGamma ? 
			Tab->item(r + NSpinRGamma, 0)->text().left(10) == "SpinRGamma" : false); NSpinRGamma++) ;
		if (NSpinRGamma > 0 && NR >= r + NSpinRGamma + 2)
		{
			
			SpinRGamma = new double[NSpinRGamma];
			for (n=0; n < NSpinRGamma; n++) SpinRGamma[n] = Tab->item(r+n, 1)->text().toDouble();
			SpinRR1 = Tab->item(r+n, 1)->text().toDouble();
			SpinRR2 = Tab->item(r+n+1, 1)->text().toDouble();
		}
		else SpinRGamma = 0;
		AnaPot *aPot;
		if (Worker->isAnaPotAv()) aPot = dynamic_cast<AnaPot*>(Worker);
		else setWorker(aPot = new AnaPot(Fit)); 
		aPot->setInnerWall(Ri, iOff, iExp, iCoeff);
		aPot->setPotCoeff(AdCorr_b, AdCorr_Rm, Tm, De, nPotCoeff, PotCoeff, CoeffFree);
		aPot->setLRCoeff(Ra, nLRCoeff, pLRCoeff, LRCoeff, LRCFree);
		if (A_ex != 0.0 || ExcIntFree) aPot->setExchangeCoeff(A_ex, beta, gamma, ExcIntFree);
		Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorrFree);
		Worker->setLambdaDoubling(q_e, q_f);
		Worker->setSpinRGamma(NSpinRGamma, SpinRGamma, SpinRR1, SpinRR2, AdCorr_b, AdCorr_Rm);
	}
	else if (type == -1 ? (Tab->rowCount() > 6 ? (Tab->item(2, 0) != 0 ? 
				Tab->item(2, 0)->text().indexOf("R_ref") >= 0 : false) : false) : type == 2)
	{
		MLRPot *LPot = dynamic_cast<MLRPot*>(Worker);
		if (LPot == 0) setWorker(LPot = new MLRPot(Fit));
		QString B;
		int n, m, r, c, Nb, NCi, p = (Tab->item(0, 1) != 0 ? Tab->item(0, 1)->text().toInt() : 0);
		int q = (Tab->item(1, 1) != 0 ? Tab->item(1, 1)->text().toInt(): 0), NAdCorr;
		double Re = (Tab->item(3, 1) != 0 ? Tab->item(3, 1)->text().toDouble() : 0.0);
		double De = (Tab->item(4, 1) != 0 ? Tab->item(4, 1)->text().toDouble() : 0.0);
		double Rref = (Tab->item(2, 1) != 0 ? Tab->item(2, 1)->text().toDouble() : 0.0);
		double T = (Tab->item(5, 1) != 0 ? Tab->item(5, 1)->text().toDouble() : 0.0);
		for (n=6, Nb = 0; (n < Tab->rowCount() ? (Tab->item(n, 0) != 0 ? 
				Tab->item(n, 0)->text().indexOf("beta") != -1 : false) : false); n++, Nb++) ;
		for (NCi = 0; (n < Tab->rowCount() ? (Tab->item(n, 0) != 0 ?
				Tab->item(n, 0)->text().indexOf("C") != -1 : false) : false); n++, NCi++) ;
		double *b = new double[Nb], *Ci = new double[NCi];
		int *pCi = new int[NCi];
		for (n=0, r=6; n < Nb; n++, r++) 
			b[n] = (Tab->item(r, 1) != 0 ? Tab->item(r, 1)->text().toDouble() : 0.0);
		LRCTabOffs = r;
		for (n=0; n < NCi; n++, r++)
		{
			B = (Tab->item(r, 0) ? Tab->item(r, 0)->text() : "");
			c = B.indexOf('_');
			m = B.indexOf(' ', c+1);
			pCi[n] = B.mid(c+1, m - c - 1).toInt();
			Ci[n] = (Tab->item(r, 1) != 0 ? Tab->item(r, 1)->text().toDouble() : 0.0);
		}
		for (NAdCorr = 0, r++; (Tab->rowCount() > r + NAdCorr ? Tab->item(r + NAdCorr, 0)->text().left(6) == "AdCorr" : false); NAdCorr++) ;
		if (NAdCorr > 0)
		{
			double *adCorr = new double[NAdCorr];
			bool *adCorrFree = new bool[NAdCorr];
			adCorrTabOffs = r;
			for (n=0; n < NAdCorr; n++)
			{
				adCorrFree[n] = Tab->item(r, 0)->icon().isNull();
				adCorr[n] = Tab->item(r++, 1)->text().toDouble();
			}
			int TAdCorr = Tab->item(r++, 1)->text().toInt();
			int PAdCorr = Tab->item(r++, 1)->text().toInt();
			double RIso1AdCorr = Tab->item(r++, 1)->text().toDouble();
			double RIso2AdCorr = Tab->item(r++, 1)->text().toDouble();
			double AdCorr_b = Tab->item(r++, 1)->text().toDouble();
			double AdCorr_Rm = Tab->item(r++, 1)->text().toDouble();
			Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b);
		}
		else Worker->setAdCorr(0, 0, 0, 0, 0.0, 0.0);
		LPot->setCoefficients(p, q, T, De, Re, Rref, Nb, b, NCi, pCi, Ci);
	}
	else if (type == 1 || (type == -1 && testIfSplinePot()))
	{
		double x, iA = 0.0, iO = 0.0, iExp = 6.0, Uinf = 0.0, RIso1AdCorr = 0.0, RIso2AdCorr = 0.0, AdCorr_b = 0.0, AdCorr_Rm = 0, Ro = 0.0, *LRC, *adCorr;
		int i, n, N = Tab->rowCount(), NAdCorr, numSplinePoints, NLRC, TAdCorr = 0, PAdCorr = 0, *PLRC;
		bool *LRCfree, *adCorrFree, calcYss = true, nanfound = false;
		QString Buffer[N];
		if (!Worker->isSplinePot()) setWorker(new SplinePot(Fit));
		Tab->setColumnCount(3);
		Tab->setHorizontalHeaderLabels(QStringList() << "R [A]" << "E [cm^-1]" << "d^2E/dR^2");
		for (n = NAdCorr = numSplinePoints = NLRC = 0; n<N; n++) 
		{
			Buffer[n] = Tab->item(n, 0)->text();
			if (Buffer[n].toDouble() > 0.0) numSplinePoints++;
			else if (Buffer[n].indexOf("C_") != -1) NLRC++;
			else if (Buffer[n].indexOf("iA") != -1) iA = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("iO") != -1) iO = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("iExp") != -1) iExp = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("U_inf") != -1) Uinf = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("Type AdCorr") != -1) TAdCorr = Tab->item(n, 1)->text().toInt();
			else if (Buffer[n].indexOf("Pow AdCorr") != -1) PAdCorr = Tab->item(n, 1)->text().toInt();
			else if (Buffer[n].indexOf("RIso1AdCorr") != -1) RIso1AdCorr = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("RIso2AdCorr") != -1) RIso2AdCorr = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("AdCorr b") != -1) AdCorr_b = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("AdCorr Rm") != -1) AdCorr_Rm = Tab->item(n, 1)->text().toDouble();
			else if (Buffer[n].indexOf("AdCorr") != -1) NAdCorr++;
			else if (Buffer[n].indexOf("R_o") != -1) Ro = Tab->item(n, 1)->text().toDouble();
		}
		SplinePoint *points = new SplinePoint[numSplinePoints];
		if (NLRC > 0)
		{
			LRC = new double[NLRC];
			PLRC = new int[NLRC];
			LRCfree = new bool[NLRC];
		}
		else
		{
			LRC = 0;
			PLRC = 0;
			LRCfree = 0;
		}
		if (NAdCorr > 0) 
		{
			adCorr = new double[NAdCorr];
			adCorrFree = new bool[NAdCorr];
		}
		else 
		{
			adCorr = 0;
			adCorrFree = 0;
		}
		for (n = numSplinePoints = NLRC = 0; n<N; n++)
		{
			if ((x = Buffer[n].toDouble()) > 0.0)
			{
				points[numSplinePoints].x = x;
				points[numSplinePoints].y = Tab->item(n, 1)->text().toDouble();
                if (points[numSplinePoints].y == 0.0)
                {
                    QLocale german(QLocale::German);
                    points[numSplinePoints].y = german.toDouble(Tab->item(n, 1)->text());
                }
				points[numSplinePoints].yss = (Tab->item(n, 2) != 0 ? Tab->item(n, 2)->text().toDouble() : 0.0);
				if (points[numSplinePoints].yss != 0.0) calcYss = false;
				if (isnan(points[numSplinePoints].yss) != 0 || isinf(points[numSplinePoints].yss) != 0) nanfound = true;
				points[numSplinePoints++].variable = Tab->item(n, 0)->icon().isNull();
			}
			else if ((i = Buffer[n].indexOf("C_")) != -1)
			{
				if (NLRC == 0) LRCTabOffs = n;
				PLRC[NLRC] = Buffer[n].right(Buffer[n].count() - i - 2).toInt();
				LRCfree[NLRC] = Tab->item(n, 0)->icon().isNull();
				LRC[NLRC++] = (Tab->item(n, 1) != 0 ? Tab->item(n, 1)->text().toDouble() : 0.0);
			}
			else for (i=0; i < NAdCorr; i++) if (Buffer[n] == "AdCorr" + QString::number(i)) 
			{
				if (i==0) adCorrTabOffs = n;
				adCorrFree[i] = Tab->item(n, 0)->icon().isNull();
				adCorr[i] = (Tab->item(n, 1) != 0 ? Tab->item(n, 1)->text().toDouble() : 0.0);
			}
		}
		if (nanfound) calcYss = true;
		if (NLRC == 0 && numSplinePoints > 2 ? 
				  points[numSplinePoints - 1].x - points[numSplinePoints - 2].x > 
				  	10.0 * (points[numSplinePoints - 2].x - points[numSplinePoints - 3].x) : false)
		{
			Tab->blockSignals(true);
			Uinf = points[numSplinePoints - 1].y;
			Tab->item(--numSplinePoints, 0)->setText("U_inf");
			Tab->item(numSplinePoints, 2)->setText("");
			Tab->blockSignals(false);
		}
		Worker->setSplinePot(numSplinePoints, points, NLRC, PLRC, LRC, iA, iO, iExp);
		Worker->setLRCFree(LRCfree);
		Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorrFree);
		Worker->setUinf(Uinf);
		Worker->setR_o(Ro);
		if (calcYss) calcyss();
	}
	else if (type == -1 ? Tab->item(0, 0)->text().indexOf("A0") >= 0 || Tab->item(0, 0)->text().indexOf("B") >= 0 : type == 3)
	{
		double *A, *C, alpha, beta, gamma, Uinf, *SA = 0, *SC = 0, Salpha = 0.0;
		double Sinf = 0.0, Sbeta = 0.0, B = 0.0, alpha1 = 0.0, *adCorr, RIso1AdCorr = 0.0, RIso2AdCorr = 0.0, AdCorr_Rm = 0.0, AdCorr_b = 0.0;
		int NA, NC, n=0, m, *pC, NSA = 0, NSC = 0, *pSC = 0, NAdCorr, TAdCorr = 0, PAdCorr = 0;
		bool *adCorrFree;
		QString Buffer;
		if (Tab->item(0, 0)->text().indexOf("B") >= 0) B = Tab->item(n++, 1)->text().toDouble();
		if (Tab->item(n, 0)->text().indexOf("alpha1", 0, Qt::CaseInsensitive) >= 0) 
			alpha1 = Tab->item(n++, 1)->text().toDouble();
		for (NA = 0; ((m = Tab->item(n + NA, 0)->text().indexOf("A")) >= 0 ? 
			Tab->item(n + NA, 0)->text().mid(m+1, 1).toInt() == NA : false); NA++) ;
		if (Tab->item(n + NA, 0)->text().indexOf("alpha", 0, Qt::CaseInsensitive) >= 0
			&& Tab->item(n + NA + 1, 0)->text().indexOf("beta", 0, Qt::CaseInsensitive) >= 0
			/*&& Tab->item(NA + 2, 0)->text().indexOf("gamma", 0, Qt::CaseInsensitive) >= 0
			&& Tab->item(NA + 3, 0)->text().indexOf("C") >= 0*/)
		{
			for (NC = 1; Tab->item(n + NA + 3 + NC, 0)->text().indexOf("C") >= 0; NC++) ;
			if (Tab->item(n + NA + 3 + NC, 0)->text().indexOf("U_inf") >= 0
				&& !Tab->item(n + NA + 3 + NC, 1)->text().isEmpty())
			{
				A = new double[NA];
				pC = new int[NC];
				C = new double[NC];
				for (m=0; m < NA; m++) A[m] = Tab->item(n++, 1)->text().toDouble();
				alpha = Tab->item(n++, 1)->text().toDouble();
				beta = Tab->item(n++, 1)->text().toDouble();
				if (Tab->item(n, 0)->text().indexOf("gamma", 0, Qt::CaseInsensitive) >= 0)
					gamma = Tab->item(n++, 1)->text().toDouble();
				else gamma = 0;
				LRCTabOffs = n;
				for (m=0; m < NC; m++)
				{
					Buffer = Tab->item(n, 0)->text();
					pC[m] = Buffer.right(Buffer.length() 
											- Buffer.indexOf("C") - 1).toInt();
					C[m] = Tab->item(n++, 1)->text().toDouble();
				}
				Uinf = Tab->item(n++, 1)->text().toDouble();
				if (Tab->rowCount() >= n+5)
				{
					while ((m = Tab->item(n + NSA, 0)->text().indexOf("SA")) >= 0 ? 
						   Tab->item(n + NSA, 0)->text().mid(m+2, 1).toInt() == NSA : false)
						NSA++;
					if (NSA > 0)
					{
						SA = new double[NSA];
						for (m=0; m < NSA; m++) SA[m] = Tab->item(n++, 1)->text().toDouble();
					}
					if (Tab->item(n, 0)->text().indexOf("Salpha", 0, Qt::CaseInsensitive) >= 0)
						Salpha = Tab->item(n++, 1)->text().toDouble();
					if (Tab->item(n, 0)->text().indexOf("Sbeta", 0, Qt::CaseInsensitive) >= 0)
						Sbeta = Tab->item(n++, 1)->text().toDouble();
					while (Tab->item(n + NSC, 0)->text().indexOf("SC", 0, Qt::CaseInsensitive) >= 0)
						NSC++;
					if (NSC > 0)
					{
						pSC = new int[NSC];
						SC = new double[NSC];
						for(m=0; m < NSC; m++)
						{
							Buffer = Tab->item(n, 0)->text();
							pSC[m] = Buffer.right(Buffer.length() 
											- Buffer.indexOf("C") - 1).toInt();
							SC[m] = Tab->item(n++, 1)->text().toDouble();
						}
					}
					if (Tab->item(n, 0)->text().indexOf("S_inf", 0, Qt::CaseInsensitive) >= 0)
						Sinf = Tab->item(n++, 1)->text().toDouble();
				}
				for (NAdCorr = 0; (Tab->rowCount() > n + NAdCorr + 6 ? Tab->item(n + NAdCorr, 0)->text().left(6) == "AdCorr" : false); NAdCorr++) ;
				if (NAdCorr > 0)
				{
					adCorr = new double[NAdCorr];
					adCorrFree = new bool[NAdCorr];
					adCorrTabOffs = n;
					for (m=0; m < NAdCorr; m++) 
					{
						adCorrFree[m] = Tab->item(n, 0)->icon().isNull();
						adCorr[m] = Tab->item(n++, 1)->text().toDouble();
					}
					TAdCorr = Tab->item(n++, 1)->text().toInt();
					PAdCorr = Tab->item(n++, 1)->text().toInt();
					RIso1AdCorr = Tab->item(n++, 1)->text().toDouble();
					RIso2AdCorr = Tab->item(n++, 1)->text().toDouble();
					AdCorr_b = Tab->item(n++, 1)->text().toDouble();
					AdCorr_Rm = Tab->item(n++, 1)->text().toDouble();
				}
				else 
				{
					adCorr = 0;
					adCorrFree = 0;
				}
				MTTPot *mTTPot = new MTTPot(Fit);
				mTTPot->setCoefficients(B, alpha1, NA, A, alpha, beta, gamma, NC, pC, C, Uinf, 
										NSA, SA, Salpha, Sbeta, NSC, pSC, SC, Sinf);
				setWorker(mTTPot);
				Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorrFree);
			}
		}
	}
	else if (type == -1 ? Tab->item(0, 0)->text()[0] == 'A' && Tab->rowCount() > 5 : type == 4)
	{
		int n, m, N, NA, NR = Tab->rowCount(), NAdCorr, TAdCorr = 0, PAdCorr = 0;
		double *A, b, Uinf, *C, *adCorr, RIso1AdCorr = 0.0, RIso2AdCorr = 0.0, AdCorr_b = 0.0, AdCorr_Rm = 0.0;
		bool *adCorrFree;
		for (NA = 0; (NA < NR ? Tab->item(NA, 0)->text().left(1) == "A" : false); NA++) ;
		for (n=0, A = new double[NA]; n < NA; n++) A[n] = Tab->item(n, 1)->text().toDouble();
		b = Tab->item(NA, 1)->text().toDouble();
		for (N=0; (N + NA + 1 < NR ? Tab->item(N + NA + 1, 0)->text().left(2) == "C_" : false) ; N++) ;
		LRCTabOffs = NA + 1;
		for (n=0, C = new double[N]; n<N; n++) C[n] = Tab->item(n + LRCTabOffs, 1)->text().toDouble();
		if (N + NA + 1 < NR) Uinf = Tab->item(n + NA + 1, 1)->text().toDouble();
		else Uinf = 0.0;
		for (NAdCorr = 0; (NR > n + NA + NAdCorr + 8 ? 
			Tab->item(n + NA + NAdCorr + 2, 0)->text().left(6) == "AdCorr" : false); NAdCorr++) ;
		if (NAdCorr > 0)
		{
			adCorr = new double[NAdCorr];
			adCorrFree = new bool[NAdCorr];
			adCorrTabOffs = n;
			for (m=0, n += (NA + 2); m < NAdCorr; m++) 
			{
				adCorrFree[m] = Tab->item(n, 0)->icon().isNull();
				adCorr[m] = Tab->item(n++, 1)->text().toDouble();
			}
			TAdCorr = Tab->item(n++, 1)->text().toInt();
			PAdCorr = Tab->item(n++, 1)->text().toInt();
			RIso1AdCorr = Tab->item(n++, 1)->text().toDouble();
			RIso2AdCorr = Tab->item(n++, 1)->text().toDouble();
			AdCorr_b = Tab->item(n++, 1)->text().toDouble();
			AdCorr_Rm = Tab->item(n++, 1)->text().toDouble();
		}
		else 
		{
			adCorr = 0;
			adCorrFree = 0;
		}
		TangToenniesPot *TTPot = new TangToenniesPot(Fit);
		TTPot->setPotCoeff(A, NA, b, C, N, Uinf);
		setWorker(TTPot);
		Worker->setAdCorr(NAdCorr, adCorr, TAdCorr, PAdCorr, RIso1AdCorr, RIso2AdCorr, AdCorr_Rm, AdCorr_b, adCorrFree);
	}
	else printf("Potential::UpdatePot error: unknown type=%d\n", type);
	Changed();
}

double PotWorker::getMinR(double E, double Prec)
{
	Lock->lock();
	double R, R1 = rmin, R2 = R1 + 0.1, E1 = Point(R1), E2 = Point(R2);
	int n;
	for (n=0; abs(E2 - E) > Prec && n < 1000; n++)
	{
		R = R2 + (E2 - E) / (E1 - E2) * (R2 - R1);
		R1 = R2;
		E1 = E2;
		R2 = R;
		E2 = Point(R);
	}
	Lock->unlock();
	printf("getMinR: R=%f, n=%d\n", R2, n);
	if (n==1000) return -1.0;
	return R2;
}

double Potential::getMinR()
{
	//if (aPot == 0 && LPot == 0 && mTTPot == 0 && points != 0) return points[0].x;
	return rmin;
}

double Potential::getMaxR()
{
	//if (aPot == 0 && LPot == 0 && points != 0) return points[numSplinePoints - 1].x;
	return rmax;
}

int Potential::getNFC()
{
	if (Worker->isFSA()) 
		return (State != 0 ? int(2.0 * State->getS()) + 1 : 1);
	return 1;
}

double* Potential::getAdCorrEnergy(double Rmin, double Step, int numPoints, double mIso1, double mIso2)
{
	double *R = new double[numPoints];
	int n;
	for (n=0; n < numPoints; n++) R[n] = 0.0;
	Worker->CalcAdCorrPot(R, R, Rmin, Step, numPoints, mIso1, mIso2);
	return R;
}

double *PotWorker::getPoints(double Ri, double Ra, int N, int FC)
{
	//printf("Potential::getPoints, Ri=%f, Ra=%f, N=%d\n", Ri, Ra, N);
	QMutexLocker lock(Lock);
	return Points(Ri, Ra, N, FC);
}

PotentialData* PotWorker::getPotentialData()
{
	int n;
	PotentialData *R = new PotentialData();
	Lock->lock();
	R->N = numSplinePoints;
	R->NLRC = NLRC;
	if (NLRC > 0)
	{
		R->PLRC = new int[NLRC];
		R->LRC = new double[NLRC];
		R->LRCFree = new bool[NLRC];
		for (n=0; n < NLRC; n++)
		{
			R->PLRC[n] = PLRC[n];
			R->LRC[n] = LRC[n];
			R->LRCFree[n] = LRCfree[n];
		}
	}
	R->Ro = Ro;
	R->NAdCorr = NAdCorr;
	R->TAdCorr = TAdCorr;
	R->PAdCorr = PAdCorr;
	R->points = new SplinePoint[numSplinePoints];
	for (n=0; n < numSplinePoints; n++) R->points[n] = points[n];
	R->iA = iA;
	R->iO = iO;
	R->iExp = iExp;
	if (NAdCorr > 0)
	{
		R->adCorr = new double[NAdCorr];
		R->adCorrFree = new bool[NAdCorr];
		for (n=0; n < NAdCorr; n++) 
		{
			R->adCorr[n] = adCorr[n];
			R->adCorrFree[n] = adCorrFree[n];
		}
	}
	R->RIso1AdCorr = RIso1AdCorr;
	R->RIso2AdCorr = RIso2AdCorr;
	R->AdCorr_b = AdCorr_b;
	R->AdCorr_Rm = AdCorr_Rm;
	R->Uinf = Uinf;
	R->shortRVariable = shortRVariable;
	R->fitResult = Result;
	R->FQS = FQS;
	Lock->unlock();
	return R;
}

double* PotWorker::Points(double Ri, double Ra, int N, int FC)
{
	if (Type == NoPotential) return 0;
	double *R = new double[N], r, s = (Ra - Ri) / double(N - 1);
	int n;
	for (n=0, r = Ri; n<N; n++, r+=s) R[n] = Point(r, FC);
	return R;
}

double PotWorker::getPoint(double R, int FC)
{
	Lock->lock();
	double U = Point(R, FC);
	Lock->unlock();
	return U;
}

double PotWorker::Point(double R, int /*FC*/)
{
	//printf("Potential::getPoint(%f)\n", R);
	double U = 0.0;
	if (points == 0) return U;
	int n, N = numSplinePoints;
	for (n=0; (n < N ? points[n].x < R : false); n++) ;
	if (n==0) 
	{
		if (R < points[0].x && Ro == 0.0) 
		{
			if (iA == 0.0 || reCalcSC)
			{
				printf("getPoints: connect SR\n");
				double xd = points[1].x - points[0].x, ds = 0.16666666666667;
				iA = ds * pow(points[0].x, 7.0) 
					* (ds * xd * points[1].yss - (points[1].y - points[0].y) / xd);
				iO = points[0].y - iA * pow(points[0].x, -6.0);
				emit potentialChanged();
				reCalcSC = false;
			}
			return U + iO + iA * pow(R, -iExp);
		}
		return U + points[0].y;
	}
	if (n==N || (Ro > 0.0 && R > Ro)) 
	{
		//printf("Long Range, R=%f, N=%d\n", R, N);
		double Ra = points[N-1].x, Rb, Res;
		if (NLRC == 0 || reCalcLC)
		{
			printf("getPoints: connect LR\n");
			if (NLRC >= 2) cdConnectLR(PLRC[NLRC - 2]);
			else
			{
				if (NLRC == 0)
				{
					LRC = new double[NLRC = 2];
					PLRC = new int[NLRC];
					PLRC[0] = 6;
					PLRC[1] = 8;
				}
				Rb = pow(Ra, PLRC[0]);
				double ys = (points[N-1].y - points[N-2].y) / (points[N-1].x - points[N-2].x)
						+ (points[N-1].x - points[N-2].x) * (points[N-2].yss + 2.0 * points[N-1].yss) / 6.0;
				LRC[0] = Rb * (0.5 * double(2 + PLRC[0]) * (Uinf - points[N-1].y) - 0.5 * (Ra) * ys);
				Rb *= Ra * Ra;
				LRC[1] = Rb * (0.5 * double(PLRC[0]) * (points[N-1].y - Uinf) + 0.5 * Ra * ys);
				emit potentialChanged();
			}
			reCalcLC = false;
		}
		//printf("y=%g, g=%g\n", points[N-1].y, Uinf - LRC[0] * pow(points[N-1].x, -1 * PLRC[0])
			//		- LRC[1] * pow(points[N-1].x, -1 * PLRC[1]));
		int p = PLRC[NLRC-1];
		for (n = NLRC - 2, Res = LRC[NLRC - 1]; n>=0 ; n--)
		{
			//printf("n=%d, Res=%g\n", n, Res);
			while (p > PLRC[n])
			{
				p--;
				Res /= R;
				//printf("R=%g, Res=%g, p=%d\n", R, Res, p);
			}
			Res += LRC[n];
			//printf("Res=%g, PLRC[%d]=%d\n", Res, n, PLRC[n]);
		}
		Res *= pow(R, -1 * PLRC[0]);
		//printf("R=%g, points[N-1].x=%g, Result=%g\n", R, points[N-1].x, Uinf - Res);
		return U + Uinf - Res;
	}
	/*if (n>=N-1 && SLIC6) 
	{
		//printf("SLIC6!\n");
		n = N-1;
		return U + points[n].y - points[n].yss * (pow(points[n].x, -6.0) - pow(R, -6.0));
	}*/
	double xd = points[n].x - points[n-1].x;
	double A = (points[n].x - R) / xd;
	double B = 1.0 - A;
	//printf("R=%f, n=%d, N=%d, xd=%f, A=%f, B=%f\n", R, n, N, xd, A, B);
	return U + A * points[n-1].y + B * points[n].y + xd * xd * (A*(A*A-1.0) * points[n-1].yss
			           + B*(B*B-1.0) * points[n].yss) * 0.16666666666667;
}

void PotWorker::getMaximum(double &R, double &U)
{
	Lock->lock();
	if (points != 0)
	{
		int n, N = numSplinePoints;
		double a, b, c, xd, yssd, r, E, x;
		for (n=0, U = points[0].y, R = points[0].x; n<N-1; n++)
		{
			yssd = points[n+1].yss - points[n].yss;
			a = (points[n+1].x * points[n].yss - points[n].x * points[n+1].yss) / yssd;
			xd = points[n+1].x - points[n].x;
			b = (2.0 * (points[n+1].y - points[n].y) + points[n].x * points[n].x * points[n+1].yss
					 - points[n+1].x * points[n+1].x * points[n].yss) / yssd - xd * xd / 3;
			if ((c=a*a-b) <= 0.0) continue;
			x = (r = sqrt(c)) - a;
			if (x >= points[n].x && x < points[n+1].x && (E = Point(x)) > U)
			{
				U = E;
				R = x;
			}
			x = - r - a;
			if (x >= points[n].x && x < points[n+1].x && (E = Point(x)) > U)
			{
				U = E;
				R = x;
			}
		}
		//printf("Potential::getMinimum: R=%f, U=%f\n", R, U);
	}
	else U = Point(R = rmin);
	Lock->unlock();
}

void PotWorker::getMinimum(double &R, double &U)
{
	Lock->lock();
	if (points != 0)
	{
		int n, N = numSplinePoints;
		double a, b, c, xd, yssd, r, E, x;
		for (n=0, U = points[0].y; n<N-1; n++)
		{
			yssd = points[n+1].yss - points[n].yss;
			a = (points[n+1].x * points[n].yss - points[n].x * points[n+1].yss) / yssd;
			xd = points[n+1].x - points[n].x;
			b = (2.0 * (points[n+1].y - points[n].y) + points[n].x * points[n].x * points[n+1].yss
					 - points[n+1].x * points[n+1].x * points[n].yss) / yssd - xd * xd / 3;
			if ((c=a*a-b) <= 0.0) continue;
			x = (r = sqrt(c)) - a;
			if (x >= points[n].x && x < points[n+1].x && (E = Point(x)) < U)
			{
				U = E;
				R = x;
			}
			//printf("p[%d].x=%f, p[%d].y=%f, x1=%f, E1=%f, ", n, points[n].x, n, points[n].y, x, E);
			x = - r - a;
			if (x >= points[n].x && x < points[n+1].x && (E = Point(x)) < U)
			{
				U = E;
				R = x;
			}
			//printf("x2=%f, E2=%f\n", x, E);
		}
		//printf("Potential::getMinimum: R=%f, U=%f\n", R, U);
	}
	/*else if (points != 0 && aPot != 0)
	{
		double x1, y1, x2, y2=0.0, x3, y3=Point(rmin), r;
		for (r = rmin + 0.1, U = y3; r <= rmax; r+=0.1)
		{
			x1 = x2;
			y1 = y2;
			x2 = x3;
			y2 = y3;
			x3 = r;
			y3 = Point(r);
			if (y1 != 0.0 && y1 >= y2 && y2 < y3 && y2 < U)
			{
				R = x2;
				U = y2;
			}
		}
		x2 = R;
		y2 = U;
		x1 = R - 0.1;
		y1 = Point(x1);
		x3 = R + 0.1;
		y3 = Point(x3);
		while (fabs(r - U) > 0.0001)
		{
			ParabInterpol(x1, y1, x2, y2, x3, y3, R, U);
			r = Point(R);
			if (R < x2)
			{
				x3 = x2;
				y3 = y2;
			}
			else
			{
				x1 = x2;
				y1 = y2;
			}
			x2 = R;
			y2 = r;
		}
		U = r;
	}*/
	else R = U = 0.0;
	Lock->unlock();
}

void Potential::getReDe(double &Re, double &De)
{
	double Te;
	getMinimum(Re, Te);
	if (Worker->isLPot()) De = dynamic_cast<MLRPot*>(Worker)->GetDe();
	else if (Worker->isTTPot()) De = -Te;
	else 
	{
		double Uinf = Worker->getUinf();
		De = Uinf - Te;
	}
}

double Potential::getMinimumE()
{
	double R, E;
	getMinimum(R, E);
	return E;
}

void Potential::setMinimum(double E)
{
	shiftEnergyOffset(E - getMinimumE());
}

double Potential::getAsymptote()
{
	return Worker->getUinf();
	/*if (points != 0)
	{
		int n = numSplinePoints - 1;
		if (SLIC6) return points[n].y - points[n].yss * pow(points[n].x, 6.0);
		return points[n].y;
		return Uinf;
	}
	printf("Potential::getasymptote: error, there is no potential data available!");
	return 0.0;*/
}

void Potential::setCoefficients()
{
	Tab->blockSignals(true);
	QStringList Set;
	int n, N = Tab->rowCount();
	for (n=0; n<N; n++) Set << Tab->item(n, 0)->text();
	CoefficientDialog *D = new CoefficientDialog(this, Set);
	if (D->exec() == QDialog::Rejected) return;
	Set = D->getResults();
	if (Set.count() > 6 ? Set[2].indexOf("R_ref") == -1 : true) Set.sort();
	Tab->setRowCount(N = Set.size());
	for (n=0; n<N; n++) Tab->setItem(n, 0, new QTableWidgetItem(Set[n]));
	Tab->blockSignals(false);
	updatePot();
	delete D;
}

void Potential::setAsymptote(double E)
{
	shiftEnergyOffset(E - getAsymptote());
}

void PotWorker::shift(double E)
{
    Lock->lock();
	int n, N = numSplinePoints;
	for (n=0; n<N; n++) points[n].y += E;
	Uinf += E;
	iO += E;
	Lock->unlock();
}

void Potential::shiftEnergyOffset(double E)
{
	Worker->shift(E);
	UpdateTab();
	Changed();
}

ElState *Potential::getElState()
{
	return State;
}

void Potential::autoCalcScatLengthsPotentialSet(int NumWFPoints)
{
	IsoTab *IsoT = Worker->getIsoT();
	int i, c, n, m, nC, *lExp, NIso = IsoT->numIso;
	double Ri, Ra, *lC, Rmin, Emin, iO, iC, Rm, Tm, b, iExp;
	AnaPot *aPot;
	QString PotFile, IsoS1[NIso], IsoS2[NIso], Buffer, IBuff;
	QString PotDir = QFileDialog::getExistingDirectory(this, 
			"Please select the directory of potentials to use");
	if (PotDir.isEmpty()) return;
	PotDir.replace('/', '\\');
	QString ProgDir = QFileDialog::getOpenFileName(this,
			"Please give the path of the program 'scat_all_asym9.exe'", PotDir, "Programs (*.exe)");
	ProgDir.replace('/', '\\');
	if (ProgDir.isEmpty()) return;
	QString InFile = QFileDialog::getOpenFileName(this,
			"Please give the input file for the program", ProgDir);
	if (InFile.isEmpty()) return;
	InFile.replace('/', '\\');
	QString OutFile = InFile + ".out";
	QString ResFile = QFileDialog::getSaveFileName(this, 
			"Please select the file name for the results", InFile);
	if (ResFile.isEmpty()) return;
	ResFile.replace('/', '\\');
	QFile IFile(InFile);
	QFile OFile(OutFile);
	IFile.open(QIODevice::ReadOnly);
	QTextStream IS(&IFile);
	QTextStream OS(&OFile);
	QStringList InList;
	while ((IBuff = IS.readLine()).left(13) != "  AtomMass_A=") InList << IBuff;
	for (n=0; n<3; n++) IS.readLine();
	IBuff = IS.readAll();
	IFile.close();
	QFile RFile(ResFile);
	QProcess *Calc = new QProcess;
	Calc->setWorkingDirectory(InFile.left(InFile.lastIndexOf('\\')));
	RFile.open(QIODevice::WriteOnly);
	QTextStream RS(&RFile);
	QDir PDir(PotDir);
	QFileInfoList PList = PDir.entryInfoList(QDir::Dirs | QDir::Files | QDir::NoDotAndDotDot);
	RS << "Potential\tC6\tC8\tC10\tRmin\tDe\tD0\tn\tb\tRm\tTm\tnC\tRi\tRa";
	for (n=0; n < NIso; n++) 
	{
		RS << "\t" << IsoT->getIsoName(n);
		IsoS1[n] = "  AtomMass_A=" + QString::number(IsoT->mIso1[n], 'f', 8) + "\n";
		IsoS2[n] = "  AtomMass_B=" + QString::number(IsoT->mIso2[n], 'f', 8) + "\n";
	}
	RS << "\n";
	for (n=0; n < PList.size(); n++) 
	{
		if (PList[n].isDir())
		{
			PDir.setPath(PList[n].filePath());
			PList << PDir.entryInfoList(QDir::Dirs | QDir::Files | QDir::NoDotAndDotDot);
			continue;
		}
		if (!PList[n].isFile()) continue;
		if (!readData(PotFile = PList[n].filePath())) continue;
		aPot = Worker->getAnaPotForReading();
		if (aPot == 0) continue;
		if (!isImported()) writePotData(PotFile = ProgDir.left(ProgDir.lastIndexOf('\\') + 1) + "TempPot.dat");
		aPot->getLRCoeffForReading(nC, lExp, lC, &Ra);
		RS << PList[n].fileName() << "\t" << (nC > 0 ? QString::number(lC[0], 'f', 8) : "0.0") << "\t";
		RS << (nC > 1 ? QString::number(lC[1], 'f', 8) : "0.0");
		RS << "\t" << (nC > 2 ? QString::number(lC[2], 'f', 8) : "0.0") << "\t";
		Worker->unlock();
		aPot->getMinimum(Rmin, Emin);
		RS << QString::number(Rmin, 'f', 8) << "\t" << QString::number(aPot->getUinf() - Emin, 'f', 5);
        RS << "\t" << QString::number(getD0(NumWFPoints), 'f', 5) << "\t";
		aPot->getInnerWall(Ri, iO, iExp, iC);
		RS << QString::number(iExp, 'g', 10) << "\t";
		aPot->getPotCoeff(b, Rm, Tm, Emin, nC, lC);
		RS << QString::number(b, 'f', 2) << "\t" << QString::number(Rm, 'f', 8) << "\t";
		RS << QString::number(Tm, 'f', 6) << "\t" << nC << "\t";
		RS << QString::number(Ri, 'f', 3) << "\t" << QString::number(Ra, 'f', 6);
		for (m=0; m < NIso; m++)
		{
			IFile.open(QIODevice::WriteOnly);
			IS.setDevice(&IFile);
			for (i=0; i < InList.size(); i++) IS << InList[i] << "\n";
			IS << IsoS1[m] << IsoS2[m] << " /\n" << "   11" << PotFile.replace('/', '\\') << "\n" 
					<< IBuff;
			IFile.close();
			Calc->start(ProgDir);
			if (!Calc->waitForStarted(5000))
			{
				QMessageBox::information(this, "MolSpektAnalysis", "0Error starting " + ProgDir);
				return;
			}
			for (c=0, Buffer = ""; c < 10 && Buffer.isEmpty(); Buffer = Calc->readLine(), c++)
			{
				if (!Calc->waitForReadyRead(5000))
				{
					QMessageBox::information(this, "MolSpektAnalysis", "1Error starting " + ProgDir);
					return;
				}
			}
			if (!Buffer.contains("input:"))
			{
				QMessageBox::information(this, "MolSpektAnlalyisis", 
										 "2Error in " +  ProgDir + ": " + Buffer);
				return;
			}
			Calc->write((InFile + "\n").toUtf8());
			if (!Calc->waitForReadyRead(5000))
			{
				QMessageBox::information(this, "MolSpektAnalysis", "3Error starting " + ProgDir);
				return;
			}
			Buffer = Calc->readAll();
			if (!Buffer.contains("output:"))
			{
				QMessageBox::information(this, "MolSpektAnlalyisis", 
										 "4Error in " +  ProgDir + ": " + Buffer);
				return;
			}
			Calc->write((OutFile + "\n").toUtf8());
			/*Calc->waitForReadyRead(5000);
			QMessageBox::information(this, "MolSpektAnalysis", 
										 "Result in " + ProgDir + ": " + Calc->readAll());
			Calc->waitForReadyRead(5000);
			Calc->readAll();*/
			if (!Calc->waitForFinished(5000))
			{
				QMessageBox::information(this, "MolSpektAnalysis", 
										 "5Error in " + ProgDir + ": " + Calc->readAll());
				return;
			}
			OFile.open(QIODevice::ReadOnly);
			OS.setDevice(&OFile);
			while ((Buffer = OS.readLine()).left(48) != 
							 " Elastic and Inelastic Scattering Lengths [bohr]" && !OS.atEnd()) ;
			if (Buffer.indexOf("Scattering Lengths") == -1) 
			{
				QMessageBox::information(this, "MolSpektAnalysis", "1Error reading output file, InFile=" + InFile + ", OutFile=" + OutFile);
				return;
			}
			while ((Buffer = OS.readLine()).left(18) != "  | 0  0>  | 0  0>") ;
			if (Buffer.left(18) != "  | 0  0>  | 0  0>")
			{
				QMessageBox::information(this, "MolSpektAnalysis", "2Error reading output file");
				return;
			}
			OFile.close();
			RS << "\t" << Buffer.right(Buffer.length() - 18);
		}
		RS << "\n";
	}
	delete Calc;
}

bool PotWorker::addPoint(double x, double y)
{
	if (Fit->isRunning()) return false;
	if (points == 0) 
	{
		printf("Potential::addPoint error: points=0!\n");
		return false;
	}
	if (L!=0) 
	{
		Destroy(L, numSplinePoints);
		L = 0;
	}
	if (S!=0) 
	{
		Destroy(S, NumS);
		S = 0;
	}
	if (SMap != 0) 
	{
		delete[] SMap;
		SMap = 0;
	}
	int n;
	//if (aPot != 0) y -= aPot->GetPoint(x);
	SplinePoint *buff = new SplinePoint[++numSplinePoints];
	for (n=0; (n < numSplinePoints - 1 ? points[n].x < x : false); n++) buff[n] = points[n];
	buff[n].x = x;
	buff[n].y = y;
	buff[n].variable = true;
	for (n++; n < numSplinePoints; n++) buff[n] = points[n-1];
	delete[] points;
	points = buff;
	return true;
}

void Potential::addPoint(double x, double y)
{
	if (Worker->addPoint(x, y))
	{
		UpdateTab();
		calcyss();
		Changed();
	}
	else QMessageBox::information(this, "MolSpektAnalysis", "The spline points cannot be changed manually while a fit is running.");
}

bool PotWorker::movePoint(int& n, double x, double y)
{
	if (Fit->isRunning()) return false;
	if (points == 0)
	{
		printf("Potential::movePoint error: points = 0!\n");
		return false;
	}
	if (n < 0 || n >= numSplinePoints)
	{
		printf("Potential::movePoint error: n=%d, numSplinePoints=%d\n", n, numSplinePoints);
		return false;
	}
	if (L!=0) 
	{
		Destroy(L, numSplinePoints);
		L = 0;
	}
	if (S!=0) 
	{
		Destroy(S, NumS);
		S = 0;
	}
	if (SMap != 0) 
	{
		delete[] SMap;
		SMap = 0;
	}
	while (n > 0 ? x < points[n-1].x : false) 
	{
		points[n] = points[n-1];
		n--;
	}
	while (n < numSplinePoints - 1 ? x > points[n+1].x : false) 
	{
		points[n] = points[n+1];
		n++;
	}
	points[n].x = x;
	points[n].y = y;
	return true;
}

void Potential::movePoint(int &n, double x, double y)
{
	if (Worker->movePoint(n, x, y))
	{
		UpdateTab();
		calcyss();
		Changed();
	}
	else QMessageBox::information(this, "MolSpektAnalysis", "The spline points cannot be changed manually while a fit is running.");
}

bool PotWorker::removePoint(int n)
{
	if (Fit->isRunning()) return false;
	if (points == 0)
	{
		printf("Potential::removePoint error: points = 0!\n");
		return false;
	}
	if (n < 0 || n >= numSplinePoints)
	{
		printf("Potential::removePoint error: n=%d, numSplinePoints=%d\n", n, numSplinePoints);
		return false;
	}
	if (L!=0) 
	{
		Destroy(L, numSplinePoints);
		L = 0;
	}
	if (S!=0) 
	{
		Destroy(S, NumS);
		S = 0;
	}
	if (SMap != 0) 
	{
		delete[] SMap;
		SMap = 0;
	}
	for (n++; n < numSplinePoints; n++) points[n-1] = points[n];
	numSplinePoints--;
	return true;
}

void Potential::removePoint(int n)
{
	if (Worker->removePoint(n))
	{
		UpdateTab();
		calcyss();
		Changed();
	}
	else QMessageBox::information(this, "MolSpektAnalysis", "The spline points cannot be changed manually while a fit is running.");
}

SplinePoint *PotWorker::getSplinePoints(int &n)
{
	QMutexLocker lock(Lock);
	if (points == 0)
	{
		n = 0;
		return 0;
	}
	int i;
	SplinePoint *rPoints = new SplinePoint[n = numSplinePoints];
	for (i=0; i<n; i++) rPoints[i] = points[i];
	//if (aPot != 0) for (i=0; i<n; i++) rPoints[i].y -= aPot->GetPoint(points[i].x);
	return rPoints;
}

void Potential::createMLRPot()
{
	//if (LPot != 0) {LPot->test(); return;}
/*	int n, m, nLRC, *pLRC;
	double *lRC;
	QStringList L;
	int p, q, Nbeta;
	double TAs, De, Re, Rref, *beta;
	getLRCoeffForReading(nLRC, pLRC, lRC);
	if (Worker->isLPot()) dynamic_cast<MLRPot*>(Worker)->getCoefficients(p, q, TAs, De, Re, Rref, Nbeta, beta, nLRC, pLRC, lRC);
	for (n=0; n < nLRC; n++) 
	{
		switch (pLRC[n])
		{
			case 6:
				if (L.size() == 0) L << "";
				break;
			case 8:
				for (m = L.size(); m<2; m++) L << "";
				break;
			case 10:
				for (m = L.size(); m<3; m++) L << "";
				break;
			case 12:
				for (m = L.size(); m<4; m++) L << "";
				break;
		}
		L << QString::number(lRC[n], 'e', 6);
	}
	CMLRPotDialog *D = new CMLRPotDialog(&L, this);
	for (n=0; n < nLRC; n++) D->LRC->addItem("C" + QString::number(pLRC[n]) + " = " + QString::number(lRC[n], 'e', 6));
	if (Worker->isLPot())
	{
		D->pE->setText(QString::number(p));
		D->qE->setText(QString::number(q));
		D->RrefE->setText(QString::number(Rref, 'f', 1));
		D->NCE->setText(QString::number(Nbeta));
		delete[] beta;
	}
	if (pLRC != 0)
	{
		delete[] pLRC;
		delete[] lRC;
	}
	if (D->exec() == QDialog::Rejected) 
	{
		delete D;
		return;
	}
	double *R, mR = D->mRE->text().toDouble(), MR = D->MRE->text().toDouble(), r;
	int NP = D->NPE->text().toInt(), NC = D->NCE->text().toInt(), i, o, *PLRC;
	double h = (MR - mR) / double(NP - 1), EMin, UInf;
	nLRC = D->LRC->count();
//	MrqLev ML(NC + nLRC + 3, NP);
//	if ((ML.a[0] = UInf = getAsymptote()) != 0.0) ML.ia[0] = false;
	//ML.ia[0] = ML.ia[1] = ML.ia[2] = false;
//	getMinimum(ML.a[2], EMin);
//	getReDe(ML.a[2], ML.a[1]);
	QString S;
	R = getPoints(mR, MR, NP);
//	fitBeta(R, mR, MR, NP, EMin, beta = new double[NC], NC, p = D->pE->text().toInt(), q = D->qE->text().toInt(), ML.a[1], ML.a[2],
//			Rref = D->RrefE->text().toDouble(), lRC, pLRC, nLRC);
	for (n=0, m=3; n < NC; n++, m++) ML.a[m] = beta[n];
	for (pLRC = new int[nLRC], PLRC = new int[nLRC], lRC = new double[nLRC], n=0; n < nLRC; n++) 
	{
		i = ((i = (S = D->LRC->item(n)->text()).indexOf(' ')) > 0 ? i : S.length());
		PLRC[n] = pLRC[n] = S.mid(1, i-1).toInt();
		ML.a[m] = lRC[n] = ((i = S.indexOf('=')) > 0 ? S.right(S.length() - i - 2).toDouble() : 0.0);
		ML.ia[m++] = (lRC[n] == 0.0 ? true : false);
	}
	Potential *Pot = MW->CreatePotential();
	MLRPot *MP = new MLRPot(Pot->Fit);
	MP->setCoefficients(p, q, ML.a[0], ML.a[1], ML.a[2], Rref, NC, beta, nLRC, pLRC, lRC);
	//for (r = 4.0; r <= 10.0; r+=100*h) printf("R=%f, E=%e\n", r, MP->GetPoint(r)); return;
	beta = new double[NC];
	lRC = new double[nLRC];
	for (n=0, r = mR; n < NP; n++, r+=h)
	{
		ML.y[n] = R[n];
		ML.sig[n] = 0.01;
		ML.dyda[n][0] = 1.0;
		MP->getDerivatives(r, ML.E[n], beta, lRC, ML.dyda[n][2], ML.dyda[n][1]);
		for (m=0, o=3; m < NC; m++) ML.dyda[n][o++] = beta[m];
		for (m=0; m < nLRC; m++) ML.dyda[n][o++] = lRC[m];
		if (isinf(ML.E[n])) 
		{
			ML.E[n] = 0.0;
			for (m=0; m < NC + nLRC + 3; m++) ML.dyda[n][m] = 0.0;
		}
	}
	ML.init();
	printf("init: %e ma=%d, mfit=%d\n", ML.chisq, ML.ma, ML.mfit);
	//return;
	//fitMLRPot(mR, MR, R, NP, MP, fLRC);
	for (i=0; i < 1000; i++)
	{
		//for (n=0; n < ML.ma; n++) printf("%e ", ML.atry[n]); printf("\n");
		pLRC = new int[nLRC];
		for (m=0, o=3; m < NC; m++) beta[m] = ML.atry[o++];
		for (m=0; m < nLRC; m++)
		{
			pLRC[m] = PLRC[m];
			lRC[m] = ML.atry[o++];
		}
		MP->setCoefficients(p, q, ML.atry[0], ML.atry[1], ML.atry[2], Rref, NC, beta, nLRC, pLRC, lRC);
		beta = new double[NC];
		lRC = new double[nLRC];
		for (n=0, r = mR; n < NP; n++, r+=h)
		{
			MP->getDerivatives(r, ML.E[n], beta, lRC, ML.dyda[n][2], ML.dyda[n][1]);
			for (m=0, o=3; m < NC; m++) ML.dyda[n][o++] = beta[m];
			for (m=0; m < nLRC; m++) ML.dyda[n][o++] = lRC[m];
		}
		if (!ML.fit()) 
		{
			printf("finished: chisq=%e\n", ML.chisq);
			break;
		}
		if (ML.alamda > 1e300)
		{
			printf("lamda to large!\n");
			break;
		}
		printf("%d: %e nd=%d, l=%e\n", i, ML.chisq, ML.done, ML.alamda);
	}
	pLRC = new int[nLRC];
	for (m=0, o=3; m < NC; m++) beta[m] = ML.a[o++];
	for (m=0; m < nLRC; m++)
	{
		pLRC[m] = PLRC[m];
		lRC[m] = ML.a[o++];
	}
	MP->setCoefficients(p, q, UInf, ML.a[1], ML.a[2], Rref, NC, beta, nLRC, pLRC, lRC);
	Pot->setWorker(MP);
	Pot->UpdateTab();
	State->addPotential(Pot);
	Pot->show();
	delete[] R;
	delete[] PLRC;
    delete D;*/
}

PotentialType Potential::getPotType()
{
	if (Worker->isSplinePot()) return SplinePotential;
	if (Worker->isAnaPotAv()) return analyticalPotential;
	if (Worker->isTTPot()) return TangToenniesPotential;
	if (Worker->isLPot()) return MorseLongRangePotential;
	if (Worker->isMTTPot()) return ModifiedTangToenniesPotential;
	return NoPotential;
}

FitResult Potential::getFitResult()
{
	FitResult Result = Worker->getFitResult();
	Result.ProcessNum = threadNum;
	return Result;
}

void PotWorker::calcFitResult()
{
	int s, n, N;
	double devq;
	BLLock->lock();
	for (s = N = Result.NBad = Result.NBadPAL = 0, Result.FQS_Bad = Result.Sigma = 0.0; s < nEs; s++) 
		for (n=0, N += NEL[s]; n < NEL[s]; n++)
	{
		Result.Sigma += (devq = EL[s][n].DevR * EL[s][n].DevR);
		if (fabs(EL[s][n].DevR >= 4.0))
		{
			Result.NBad++;
			Result.FQS_Bad += devq;
			if (EL[s][n].err < 1e-4) Result.NBadPAL++;
		}
	}
	for (s=0; s < nGS; s++) for (n=0, N += nGL[s]; n < nGL[s]; n++)
	{
		Result.Sigma += (devq = GL[s][n].DevR * GL[s][n].DevR);
		if (fabs(GL[s][n].DevR >= 4.0))
		{
			Result.NBad++;
			Result.FQS_Bad += devq;
			if (GL[s][n].err < 1e-4) Result.NBadPAL++;	
		}
	}
	BLLock->unlock();
	Result.Sigma = sqrt(Result.Sigma / (N - NumFreePar));
	if (NumWantedValues > 0)
	{
		for (n=1, Result.LRelParDev = fabs(WantedValues[0].Value - WantedValues[0].CurValue) / WantedValues[0].Precision; n < NumWantedValues; n++)
			if ((devq = fabs(WantedValues[n].Value - WantedValues[n].CurValue) / WantedValues[n].Precision) < Result.LRelParDev)
				Result.LRelParDev = devq;
	}
	else Result.LRelParDev = 0.0;
}


TangToenniesPot::TangToenniesPot(PotFit *Fit) : PotWorker(Fit, TangToenniesPotential)
{
	Uinf = b = 0.0;
	C = SWF = 0;
	NC = 0;
	NA = 0;
	A = 0;
	A_ex = beta = gamma = 0.0;
	RMin = UMax = 0.0;
}

TangToenniesPot::TangToenniesPot(const TangToenniesPot& O) : PotWorker(O)
{
	int n;
	Uinf = O.Uinf;
	b = O.b;
	C = new double[NC = O.NC];
	for (n=0; n < NC; n++) C[n] = O.C[n];
	A = new double[NA = O.NA];
	for (n=0; n < NA; n++) A[n] = O.A[n];
	SWF = 0;
	A_ex = O.A_ex;
	beta = O.beta;
	gamma = O.gamma;
	RMin = O.RMin;
	UMax = O.UMax;
}

TangToenniesPot::~TangToenniesPot()
{
	if (C != 0) delete[] C;
	if (SWF != 0) delete[] SWF;
	if (A != 0) delete[] A;
}

void TangToenniesPot::AB(double B, double& Y, double& AA, double* D, double RM)
{
	//Translation of Fortran routine from K.T. Tang and J.P. Toennies, Z. Phys. D 1, 91 (1986)
	double F[NC], G[NC], SU, FR, GR, TK;
	int i, j, k;
	B *= RM;
	for (i=0; i < NC; i++)
	{
		SU = 0.0;
		FR = 1.0;
		for (j=1; j <= 2*(i+3); j++)
		{
			FR *= double(j);
			GR = pow(B, j) / FR;
			SU += GR;
		}
		G[i] = GR * exp(-B);
		F[i] = 1.0 - (1.0 + SU) * exp(-B);
	}
	AA = 0.0;
	Y = 0.0;
	for (k=0; k < NC; k++)
	{
		TK = 2*(k+3);
		AA += (F[k] * D[k] * TK / B - G[k] * D[k]);
		Y += (F[k] * D[k] * (TK / B - 1.0) - G[k] * D[k]);
	}
	AA *= exp(B);
	Y += 1.0;
}

bool TangToenniesPot::calcAb(double Re, double De)
{
	//Translation of Fortran routine from K.T. Tang and J.P. Toennies, Z. Phys. D 1, 91 (1986)
	int i, ik;
	double D[NC], DB, B1, Y1, Y2, A1, A2, R, YO, AO, BB, YY;
	for (i=0; i < NC; i++) D[i] = C[i] / (abs(De) * pow(Re, 2*(i+3)));
	b = 0.4;
	DB = 0.2;
	while(true)
	{
		for (ik = 1; ik <= 11; ik++)
		{
			B1 = b + DB;
			AB(b, Y1, A1, D, Re);
			AB(B1, Y2, A2, D, Re);
			R = Y2 / Y1;
			if (R <= 0.0) break;
			b += DB;
		}
		if (R > 0.0) return false;
		BB = b + DB / (1.0 - R);
		AB(BB, YO, AO, D, Re);
		YY = fabs(YO);
		if (YY <= 1e-12) break;
		DB /= 10.0;
	}
	A[0] = AO * abs(De);
	return true;
}


bool TangToenniesPot::calcPotCoeff(double* CC, int NCC, int FNC, double Re, double De)
{
	//De = 4.929e-3; Re = 8.828; CC[0] = 31.03e2; CC[1] = 37.92e4; CC[2] = 42.15e6;
	int i;
	if (NC != FNC)
	{
		if (C != 0) delete[] C;
		C = new double[NC = FNC];
	}
	if (A == 0) A = new double[1];
	for (i=0; i < NCC; i++) C[i] = CC[i];
	for ( ; i < NC; i++) C[i] = pow(C[i-1] / C[i-2], 3) * C[i-3];
	return calcAb(Re, De);
}

double TangToenniesPot::calcPotCoeff(int numPoints, double* R, double* U, double* Sig, int numCoeff, double nb)
{
	int n, m, NFC = numCoeff + (A_ex != 0.0 ? 2 : 1);
	b = nb;
	if (NC != numCoeff)
	{
		if (C != 0) delete[] C;
		C = new double[NC = numCoeff];
	}
	if (A == 0) A = new double[1];
	double **EQS = Create(numPoints, NFC), x[NFC], Calc, Diff, FQS;
	for (n=0; n <= numPoints; n++) getPreFakt(R[n], EQS[n]);
	SvdFit(EQS, x, U, Sig, NFC, numPoints, 1e-12);
	for (n=0, FQS = 0.0; n < numPoints; n++)
	{
		for (m=0, Calc = 0.0; m < NFC; m++) Calc += EQS[n][m] * x[m];
		Diff = U[n] - Calc;
		FQS += (Diff * Diff);
	}
	for (A[0] = x[n=0]; n < NC; n++) C[n] = x[n+1];
	if (A_ex != 0) A_ex = x[n+1];
	Destroy(EQS, numPoints);
	printf("FQS=%f\n", FQS);
	return FQS;
}

void TangToenniesPot::calcS(int numPoints, double Rmin, double Rmax)
{
	int n, R, r, m, NumCoefficients = getNFitCoefficients();
	double PreFakt[NumS = NumFreePar = NumCoefficients], x = 0.0, h = (Rmax - Rmin) / double(numPoints - 1);
	if (SWF != 0) delete[] SWF;
	SWF = new double[NumCoefficients];
	S = Create(NumCoefficients + NAdCorr, numPoints);
	for (n=0; n < NumCoefficients; n++) SWF[n] = 0.0;
	for (n=0; n < NAdCorr; n++) AdCorrWF[n] = 0.0;
	for (r=0, R = Rmin; r < numPoints; r++, R+=h)
	{
		getPreFakt(R, PreFakt);
		for (n=0; n < NumCoefficients; n++) SWF[n] += fabs(S[n][r] = PreFakt[n]);
		if (NAdCorr > 0)
		{
			if (NAdCorr > 1) x = (R - AdCorr_Rm) / (R + AdCorr_b * AdCorr_Rm);
			AdCorrWF[0] += fabs(S[n][r] = pow(2.0 * AdCorr_Rm / (R + AdCorr_Rm), PAdCorr));
			for (m=1; m < NAdCorr; m++, n++) AdCorrWF[m] += fabs(S[n][r] = S[n-1][r] * x);
		}
	}
	for (n=0; n < NumCoefficients; n++) SWF[n] = 1.0 / SWF[n];
	for (n=0; n < NAdCorr; n++) AdCorrWF[n] = 1.0 / AdCorrWF[n];
	for (r=0; r < numPoints; r++)
	{
		for (n=0; n < NumCoefficients; n++) S[n][r] *= SWF[n];
		for (m=0; m < NAdCorr; m++, n++) S[n][r] *= AdCorrWF[m];
	}
	NumCoefficients += NAdCorr;
}

void TangToenniesPot::getLRCoeff(double& R, int& numCoefficients, int*& Exponents, double*& Coefficients)
{
    int n;
	R = 0.0;
	Exponents = new int[numCoefficients = NC];
	Coefficients = C;
	for (n=1, Exponents[0] = 6; n < NC; n++) Exponents[n] = Exponents[n-1] + 2;
}

void TangToenniesPot::getMinimum(double& R, double& T)
{
	Lock->lock();
	double x1 = rmin, x2 = x1 + 1.0, x3 = x2 + 1.0;
	double y1 = Point(x1), y2 = Point(x2), y3 = Point(x3);
	int n;
	while (y3 < y2 && x3 < rmax)
	{
		x1 = x2;
		y1 = y2;
		x2 = x3;
		y2 = y3;
		y3 = Point(x3 = x2 + 1.0);
	}
	ParabInterpol(x1, y1, x2, y2, x3, y3, R, T);
	for (n=0; n < 1000 && fabs(x2 - R) >= 1e-8; n++)
	{
		if (R > x2)
		{
			x1 = x2;
			y1 = y2;
		}
		else 
		{
			x3 = x2;
			y3 = y2;
		}
		y2 = Point(x2 = R);
		ParabInterpol(x1, y1, x2, y2, x3, y3, R, T);
	}
	T = Point(R);
	Lock->unlock();
}

PotentialData* TangToenniesPot::getPotentialData()
{
	int n;
	PotentialData *R = new PotentialData();
	Lock->lock();
	R->A = new double[NA];
	for (n=0; n < NA; n++) R->A[n] = A[n];
	R->b = b;
	R->C = new double[NC];
	for (n=0; n < NC; n++) R->C[n] = C[n];
	R->Uinf = Uinf;
	if (NAdCorr > 0)
	{
		R->adCorr = new double[NAdCorr];
		for (n=0; n < NAdCorr; n++) R->adCorr[n] = adCorr[n];
	}
	R->RIso1AdCorr = RIso1AdCorr;
	R->RIso2AdCorr = RIso2AdCorr;
	R->AdCorr_b = AdCorr_b;
	R->AdCorr_Rm = AdCorr_Rm;
	R->N = NC;
	R->NA = NA;
	R->NAdCorr = NAdCorr;
	R->TAdCorr = TAdCorr;
	R->PAdCorr = PAdCorr;
	R->fitResult = Result;
	R->FQS = FQS;
	Lock->unlock();
	return R;
}

double TangToenniesPot::Point(double R, int)
{
	if (R <= RMin) return UMax;
	double U = 0.0, F[getNFitCoefficients()];
	int n, m;
	getPreFakt(R, F);
	for (n=0; n < NA; n++) U += A[n] * F[n];
	for (m=0; m < NC; m++) U += C[m] * F[n+m];
	if (A_ex != 0.0) U += A_ex * F[n+m];
	return U;
}

void TangToenniesPot::getPotCoeff(double *&rA, int &rNA, double& rb, double*& rC, int& N, double& rUinf)
{
	rA = A;
	rNA = NA;
	rb = b;
	rC = C;
	N = NC;
	rUinf = Uinf;
}

void TangToenniesPot::getPreFakt(double R, double* F)
{
	int n, m, i;
	double bR = b*R, Q = 1.0, dR = Q / R, dRq = dR * dR;
	for (i=1, F[0] = exp(-bR); i < NA; i++) F[i] = F[i-1] * dR; 
	for (n=0, m=1; n < NC; n++)
		for (F[n+i] = (n>0 ? F[n+i-1] : 1.0); m <= 2*(n+3); m++) F[n+i] += (Q *= (bR / m));
	for (n=i, Q = -pow(R, -6); n < NC + NA; n++, Q*= dRq) F[n] = Q * (1.0 - F[n] * F[0]);
	if (A_ex != 0) F[NC + NA] = pow(R, gamma) * exp(-beta * R);
}

double TangToenniesPot::improvePotential(int numPoints, double* R, double* U, double* Sig, double nb)
{
	int n, m, nc = getNFitCoefficients();
	double **EQS = Create(numPoints, nc), x[nc], Calc, aFQS = 1e99, FQS; 
	double *UD = new double[numPoints], Div, r1 = R[0], r2 = 0.0, r3, u1 = U[0], u2 = 0.0, u3, r, u;
	if (nb != 0.0) b = nb;
	for (n=0; n < NA; n++) x[n] = A[n];
	for (m=0; m < NC; m++) x[n+m] = C[m];
	if (A_ex != 0.0) x[n+m] = A_ex;
	for (n=0, FQS = 0.0; n < numPoints; n++)
	{
		getPreFakt(R[n], EQS[n]);
		for (m=0, UD[n] = U[n]; m < nc; m++) UD[n] -= EQS[n][m] * x[m];
		Div = UD[n] / Sig[n];
		FQS += Div * Div;
	}
	while (FQS < aFQS * 0.999)
	{
		printf("FQS=%g\n", aFQS = FQS);
		SvdFit(EQS, x, UD, Sig, nc, numPoints, 1e-12);
		for (n=0; n < NA; n++) x[n] = (A[n] += x[n]);
		for (m=0; m < NC; m++) x[n+m] = (C[m] += x[n+m]);
		if (A_ex != 0) x[n+m] = (A_ex += x[n+m]);
		for (n=0, FQS = 0.0; n < numPoints; n++)
		{
			for (m=0, Calc = 0.0; m < nc; m++) Calc += EQS[n][m] * x[m];
			Div = (UD[n] = (U[n] - Calc)) / Sig[n];
			FQS += (Div * Div);
		}
	}
	Destroy(EQS, numPoints);
	delete[] UD;
	printf("Finally: FQS=%g\n", FQS);
	do
	{
		r3 = r2;
		u3 = u2;
		r2 = r1;
		u2 = u1;
		r1 -= 0.1;
		u1 = Point(r1);
	}
	while (r1 >= rmin && u1 > u2);
	if (u1 < u2)
	{
		if (r2 == R[0]) u3 = Point(r3 = r2 + 0.1);
		do
		{
			ParabInterpol(r1, u1, r2, u2, r3, u3, r, u);
			if (r < r2)
			{
				r3 = r2;
				u3 = u2;
			}
			else
			{
				r1 = r2;
				u1 = u2;
			}
			r2 = r;
			u2 = Point(r);
		}
		while (r3 - r2 > 1e-12);
		RMin = r3;
		UMax = u3;
	}
	else RMin = UMax = 0.0;
	return FQS;
}

void TangToenniesPot::setPotCoeff(double *nA, int nNA, double nb, double* nC, int N, double nUinf)
{
	if (C != nC)
	{
		if (C != 0) delete[] C;
		C = nC;
	}
	if (A != nA)
	{
		if (A != 0) delete[] A;
		A = nA;
	}
	NA = nNA;
	b = nb;
	NC = N;
	Uinf = nUinf;
}

void TangToenniesPot::updateCoefficients(double* CD)
{
	int n, m; 
	for (n=0; n < NA; n++) A[n] += (SWF[n] * CD[n]);
	for (m=0; m < NC; m++) C[m] += (SWF[n+m] * CD[n+m]);
	if (A_ex != 0.0) A_ex += (SWF[n+m] + CD[n+m]);
}


AnaPot::AnaPot(PotFit *Fit) : PotWorker(Fit, analyticalPotential)
{
	nLRCoeff = nPotCoeff = numFreeCoeff = numFreeLRC = 0;
	b = iCoeff = iOff = Ra = Ri = iExp = A_ex = beta = gamma = Rm = Tm = 0.0;
	LRCoeff = PotCoeff = 0;
	pLRCoeff = 0;
	SWF = 0;
	LRCFree = CoeffFree = 0;
	ExcIntFree = false;
	ADLRC = false;
}

AnaPot::AnaPot(const AnaPot &C) : PotWorker(C)
{
	//printf("AnaPot::AnaPot(const AnaPot &C)\n");
	int n;
	if (C.SWF != 0)
	{
		SWF = new double[C.nPotCoeff + C.nLRCoeff - 1];
		for (n=0; n < C.nPotCoeff + C.nLRCoeff - 1; n++) SWF[n] = C.SWF[n];
	}
	else SWF = 0;
	b = C.b;
	iExp = C.iExp;
	numFreeCoeff = C.numFreeCoeff;
	numFreeLRC = C.numFreeLRC;
	ExcIntFree = C.ExcIntFree;
	if ((nLRCoeff = C.nLRCoeff) > 0)
	{
		LRCoeff = new double[nLRCoeff + 2];
		pLRCoeff = new int[nLRCoeff + 2];
		LRCFree = new bool[nLRCoeff];
		for (n=0; n < nLRCoeff; n++) 
		{
			LRCoeff[n] = C.LRCoeff[n];
			pLRCoeff[n] = C.pLRCoeff[n];
			LRCFree[n] = C.LRCFree[n];
		}
	}
	else
	{
		LRCoeff = 0;
		pLRCoeff = 0;
		LRCFree = 0;
	}
	if ((nPotCoeff = C.nPotCoeff) > 0)
	{
		PotCoeff = new double[nPotCoeff];
		CoeffFree = new bool[nPotCoeff];
		for (n=0; n < nPotCoeff; n++) 
		{
			PotCoeff[n] = C.PotCoeff[n];
			CoeffFree[n] = C.CoeffFree[n];
		}
	}
	iCoeff = C.iCoeff;
	iOff = C.iOff;
	Ra = C.Ra;
	Ri = C.Ri;
	Rm = C.Rm;
	Tm = C.Tm;
	A_ex = C.A_ex;
	beta = C.beta;
	gamma = C.gamma;
	ADLRC = C.ADLRC;
	//printf("Ende AnaPot\n");
}

AnaPot::~AnaPot()
{
	if (nPotCoeff > 0) 
	{
		delete[] PotCoeff;
		delete[] CoeffFree;
	}
	if (nLRCoeff > 0)
	{
		delete[] pLRCoeff;
		delete[] LRCoeff;
		delete[] LRCFree;
	}
	if (SWF != 0) delete[] SWF;
}

AnaPot* AnaPot::scalePotential(double newRe, double newDe)
{
	AnaPot *RPot = new AnaPot(Fit);
	int n;
	double ARe, ADe, xSF, ySF;
	getMinimum(ARe, ADe);
	xSF = newRe / ARe;
	ySF = newDe / (Uinf - ADe);
	RPot->b = b;
	RPot->Uinf = ySF * Uinf;
	if ((RPot->nLRCoeff = nLRCoeff) > 0)
	{
		RPot->LRCoeff = new double[nLRCoeff + 2];
		RPot->pLRCoeff = new int[nLRCoeff + 2];
		for (n=0; n < nLRCoeff; n++) 
			RPot->LRCoeff[n] = LRCoeff[n] * pow(xSF, RPot->pLRCoeff[n] = pLRCoeff[n]) * ySF;
	}
	else
	{
		RPot->LRCoeff = 0;
		RPot->pLRCoeff = 0;
	}
	if ((RPot->nPotCoeff = nPotCoeff) > 0)
	{
		RPot->PotCoeff = new double[nPotCoeff];
		for (n=0; n < nPotCoeff; n++) RPot->PotCoeff[n] = ySF * PotCoeff[n];
	}
	RPot->iCoeff = iCoeff * pow(xSF, RPot->iExp = iExp) * ySF;
	RPot->iOff = ySF * iOff;
	RPot->Ra = xSF * Ra;
	RPot->Ri = xSF * Ri;
	RPot->Rm = xSF * Rm;
	RPot->Tm = ySF * Tm;
	RPot->A_ex = ySF * A_ex * pow(xSF, -gamma);
	RPot->beta = beta / xSF;
	RPot->gamma = gamma;
	return RPot;
}

double *AnaPot::Points(double RMin, double RMax, int nP, int)
{
	//printf("getPointRep: nP=%d\n", nP);
	if (nLRCoeff <= 0 || nPotCoeff <= 0 || Ri == 0.0 || nP <= 0 || RMin >= RMax) return NULL;
	double *RA = new double[nP], h = (RMax - RMin) / double(nP - 1), R, x, edR;
	int i, Ni = int((Ri - RMin) / h), Na = int((Ra - RMin) / h), n, p;
	//printf("Ni=%d, Na=%d, Np=%d\n", Ni, Na, nP);
	if (Ni < 0) Ni = -1;
	if (Na < 0) Na = -1;
	if (Ni >= nP) Ni = nP - 1;
	if (Na >= nP) Na = nP - 1;
	for (i=0, R = RMin; i <= Ni; i++)
	{
		RA[i] = iOff + iCoeff * pow(R, -iExp);
		//printf("RA[%d]=%g\n", i, RA[i]);
		R += h;
	}
	for (i = Ni + 1; i <= Na; i++)
	{
		x = (R - Rm) / (R + b * Rm);
		for (n = nPotCoeff - 1, RA[i] = 0.0; n >= 0; n--)
		{
			RA[i] += PotCoeff[n];
			RA[i] *= x;
		}
		RA[i] += Tm;
		R += h;
	}
	for (i = Na + 1; i < nP; i++)
	{
		edR = 1.0 / R;
		for (n = nLRCoeff - 1, p = pLRCoeff[nLRCoeff-1], RA[i] = 0.0; n >= 0; n--)
		{
			while (p > pLRCoeff[n]) 
			{
				RA[i] *= edR;
				p--;
			}
			RA[i] -= LRCoeff[n];
		}
		RA[i] *= pow(edR, pLRCoeff[0]);
		RA[i] += Uinf;
		if (A_ex != 0.0) RA[i] -= A_ex * pow(R, gamma) * exp(-beta * R);
		R += h;
	}
	//printf("Ende getPointRep\n");
	return RA;
}

double AnaPot::Point(double r, int typ)
{
	if (typ == 0 ? r <= Ri : typ == 1) return iOff + iCoeff * pow(r, -iExp);
	int n, p;
	double R, x;
	if (typ == 0 ? r < Ra : typ == 2)
	{
		if (nPotCoeff <= 0) return 0.0;
		x = (r - Rm) / (r + b * Rm);
		for (n = nPotCoeff - 1, R = 0.0; n >= 0; n--)
		{
			R += PotCoeff[n];
			R *= x;
		}
		return R + Tm;
	}
	if (nLRCoeff <= 0) return 0.0;
	double edR = 1.0 / r;
	for (n = nLRCoeff - 1, p = pLRCoeff[nLRCoeff-1], R = 0.0; n >= 0; n--)
	{
		while (p > pLRCoeff[n])
		{
			R *= edR;
			p--;
		}
		R -= LRCoeff[n];
	}
	if (A_ex == 0.0) return R * pow(edR, pLRCoeff[0]) + Uinf;
	return R * pow(edR, pLRCoeff[0]) + Uinf - A_ex * pow(r, gamma) * exp(-beta * r);
}

void AnaPot::getMinimum(double &R, double &T)
{
	//printf("AnaPot::GetMin()\n");
	Lock->lock();
	double x1, y1, x2, y2=0.0, x3 = rmin, y3=Point(rmin), r;
	for (r = rmin + 0.1, T = y3; r <= rmax; r+=0.1)
	{
		x1 = x2;
		y1 = y2;
		x2 = x3;
		y2 = y3;
		x3 = r;
		y3 = Point(r);
		if (y1 != 0.0 && y1 >= y2 && y2 < y3 && y2 < T)
		{
			R = x2;
			T = y2;
		}
	}
	x2 = R;
	y2 = T;
	x1 = R - 0.1;
	y1 = Point(x1);
	x3 = R + 0.1;
	y3 = Point(x3);
	while (fabs(r - T) > 0.0001)
	{
		ParabInterpol(x1, y1, x2, y2, x3, y3, R, T);
		r = Point(R);
		if (R < x2)
		{
			x3 = x2;
			y3 = y2;
		}
		else
		{
			x1 = x2;
			y1 = y2;
		}
		x2 = R;
		y2 = r;
	}
	T = r;
	//printf("Ende GetMin, T=%f\n", T);
	Lock->unlock();
}

void AnaPot::setInnerWall(double R, double O, double E, double C)
{
	Ri = R;
	iOff = O;
	iExp = E;
	iCoeff = C;
}

void AnaPot::getInnerWall(double &R, double &O, double &E, double &C)
{
	R = Ri;
	O = iOff;
	E = iExp;
	C = iCoeff;
}

void AnaPot::getConR(double &i, double &a)
{
	i = Ri;
	a = Ra;
}

void AnaPot::getDerivatives(double, double*, double*)
{

}

void AnaPot::setLRCoeff(double R, int N, int *E, double *C, bool *lRCFree)
{
	int n, p=1;
	double Co;
	bool Free;
	numFreeCoeff -= numFreeLRC;
	if (nLRCoeff > 0)
	{
		if (E != pLRCoeff) delete[] pLRCoeff;
		if (C != LRCoeff) delete[] LRCoeff;
		if (lRCFree != LRCFree && (lRCFree != 0 || N != nLRCoeff))
		{
			delete[] LRCFree;
			LRCFree = 0;
		}
	}
	Ra = R;
	nLRCoeff = N;
	pLRCoeff = E;
	LRCoeff = C;
	if (lRCFree != 0) LRCFree = lRCFree;
	else if (LRCFree == 0 && N>0) for (n=0, LRCFree = new bool[N]; n<N; n++) LRCFree[n] = true;
	while (p!=0) for (n=1, p=0; n<N; n++) if (pLRCoeff[n] < pLRCoeff[n-1])
	{
		p = pLRCoeff[n];
		pLRCoeff[n] = pLRCoeff[n-1];
		pLRCoeff[n-1] = p;
		Co = LRCoeff[n-1];
		LRCoeff[n-1] = LRCoeff[n];
		LRCoeff[n] = Co;
		Free = LRCFree[n];
		LRCFree[n] = LRCFree[n-1];
		LRCFree[n-1] = Free;
	}
	for (numFreeLRC = n = 0; n < N-1; n++) if (LRCFree[n]) numFreeLRC++;
	numFreeCoeff += numFreeLRC;
	ADLRC = (N>4 ? E[N-1] == 16 && E[N-4] == 10 : false);
}

void AnaPot::getLRCoeff(double &R, int &N, int *&E, double *&C)
{
	R = Ra;
	N = nLRCoeff;
	E = pLRCoeff;
	C = LRCoeff;
}

void AnaPot::getNC(int& N, int& NC, int& NLRC)
{
	N = (NC = nPotCoeff) + (NLRC = nLRCoeff);
}

void AnaPot::setExchangeCoeff(double A, double b, double g, bool Free)
{
	A_ex = A;
	beta = b;
	gamma = g;
	ExcIntFree = Free;
}

void AnaPot::getExchangeCoeff(double& A, double& b, double& g)
{
	A = A_ex;
	b = beta;
	g = gamma;
}

void AnaPot::setPotCoeff(double nb, double nRm, double nTm, double nDe, int N, double *C, bool *Free)
{
	//printf("AnaPot::setPotCoeff: N=%d\n", N);
	int n;
	if (Free != 0) 
	{
		for (n = numFreeCoeff = 0; n<N; n++) if (Free[n]) numFreeCoeff++;
	}
	else if (N != nPotCoeff) numFreeCoeff = N;
	numFreeCoeff += numFreeLRC;
	if (nPotCoeff > 0)
	{
		if (C != PotCoeff) delete[] PotCoeff;
		if (Free != CoeffFree && (Free != 0 || N != nPotCoeff))
		{
			delete[] CoeffFree;
			CoeffFree = 0;
		}
	}
	b = nb;
	Rm = nRm;
	Tm = nTm;
	Uinf = nDe;
	nPotCoeff = N;
	PotCoeff = C;
	if (Free != 0) CoeffFree = Free;
	else if (CoeffFree == 0 && N>0) for (CoeffFree = new bool[N], n=0; n<N; n++) CoeffFree[n] = true;
}

void AnaPot::getPotCoeff(double &nb, double &nRm, double &nTm, double &nDe, int &N, double *&C)
{
	nb = b;
	nRm = Rm;
	nTm = Tm;
	nDe = Uinf;
	N = nPotCoeff;
	C = PotCoeff;
	//printf("AnaPot::getPotCoeff: N=%d\n", N);
}

void AnaPot::cdConnectLR1C()
{
	int n, p;
	double dxdR, z, A = 0.0, dAdR = 0.0, x, L = 0.0, dLdR = 0.0, edR, D, pnC;
	/*if (ADLRC) for (nLRCoeff = 1; pLRCoeff[nLRCoeff - 1] < 10; nLRCoeff++) ;
	else if (nLRCoeff >= 3 ? pLRCoeff[nLRCoeff - 3] == 6 && pLRCoeff[nLRCoeff - 2] == 8 
				&& pLRCoeff[nLRCoeff - 1] == 10 : false)
	{
		ADLRC = true;
		int *pB = new int[nLRCoeff + 3];
		double *LB = new double[nLRCoeff + 3];
		bool *fB = new bool[nLRCoeff + 3];
		for (n=0; n < nLRCoeff; n++)
		{
			pB[n] = pLRCoeff[n];
			LB[n] = LRCoeff[n];
			fB[n] = LRCFree[n];
		}
		delete[] pLRCoeff;
		delete[] LRCoeff;
		delete[] LRCFree;
		pLRCoeff = pB;
		LRCoeff = LB;
		LRCFree = fB;
	}*/
	Lock->lock();
	z = Ra + b * Rm;
	x = (Ra - Rm) / z;
	dxdR = (1.0 - x) / z;
	for (n = nPotCoeff - 1; n >= 0; n--)
	{
		dAdR += double(n+1) * PotCoeff[n];
		if (n>0) dAdR *= x;
		A += PotCoeff[n];
		A *= x;
	}
	dAdR *= dxdR;
	A += Tm;
	edR = 1.0 / Ra;
	if (nLRCoeff > 0)
	{
		for (n = nLRCoeff - 1, p = pLRCoeff[nLRCoeff-1], L = 0.0; n >= 0; n--)
		{
			while (p > pLRCoeff[n]) 
			{
				L *= edR;
				dLdR *= edR;
				p--;
			}
			L -= LRCoeff[n];
			dLdR += LRCoeff[n] * double(p);
		}
		L *= pow(edR, pLRCoeff[0]);
		dLdR *= pow(edR, pLRCoeff[0] + 1);
		if (A_ex != 0)
		{
			z = -A_ex * pow(Ra, gamma) * exp(-beta * Ra);
			L += z;
			dLdR += (gamma * edR - beta) * z;
		}
	}
	L += Uinf;
	/*if (pnC > pLRCoeff[nLRCoeff - 1])
	{
		pLRCoeff[nLRCoeff] = pnC;
		LRCoeff[nLRCoeff++] = 0.0;
	}
	else 
	{*/
	for (n = nLRCoeff -1; (n>=0 ? !LRCFree[n] : false); n--) ;
	pnC = pLRCoeff[(n>=0 ? n : n = nLRCoeff - 1)];
	//}
	LRCoeff[n] += (D = (dAdR - dLdR) / (double(pnC) * (x = pow(Ra, -pnC - 1))));
	/*if (ADLRC)
	{
		double dADdR = 1e99, dRaq = 1.0 / (Ra * Ra), cM, ldADdR, dM = 0.1;
		do
		{
			ldADdR = dADdR;
			for (n = nLRCoeff; n <= nLRCoeff + 2; n++) LRCoeff[n] = pow(LRCoeff[n-1] / LRCoeff[n-2], 3) * LRCoeff[n-3];
			for (n = nLRCoeff + 2, cM = 16, dADdR = 0.0; n >= nLRCoeff; n--, cM -= 2)
			{
				dADdR += cM * LRCoeff[n];
				dADdR *= dRaq;
			}
			if (dADdR > 5.0 * LRCoeff[nLRCoeff - 1])
			{
				printf("Error sdConnectLR: iterative procedure failed, no ADLRC!\n");
				ADLRC = false;
				break;
			}
			LRCoeff[nLRCoeff-1] -= dADdR * dM;
			L += x * dADdR * Ra * dM;
			//printf("dADdR=%g\n", dADdR);
		} 
		while (ldADdR - dADdR  > 1.0);
		if (ADLRC) 
		{
			for (n = nLRCoeff; n <= nLRCoeff + 2; n++) pLRCoeff[n] = pLRCoeff[n-1] + 2;
			for (n = nLRCoeff + 2, dADdR = 0.0; n >= nLRCoeff; n--)
			{
				dADdR += LRCoeff[n];
				dADdR *= dRaq;
				LRCFree[n] = false;
			}
			Tm -= x * dADdR;
			nLRCoeff += 3;
		}
	}*/
	Tm += (L - A - D * Ra * x);
	Lock->unlock();
}

void AnaPot::cdConnectLR(int pnC)
{
	int n, p, m=0;
	double dxdR, z, A = 0.0, dAdR = 0.0, x, L = 0.0, dLdR = 0.0, edR, D, F;
	Lock->lock();
	z = Ra + b * Rm;
	x = (Ra - Rm) / z;
	dxdR = (1.0 - x) / z;
	for (n = nPotCoeff - 1; n >= 0; n--)
	{
		dAdR += double(n+1) * PotCoeff[n];
		if (n>0) dAdR *= x;
		A += PotCoeff[n];
		A *= x;
	}
	dAdR *= dxdR;
	A += Tm;
	edR = 1.0 / Ra;
	if (nLRCoeff > 0)
	{
		if (pnC < pLRCoeff[0]) pnC = pLRCoeff[nLRCoeff - 1] + 2;
		else for (m = nLRCoeff - 1; pnC < pLRCoeff[m]; m--) ;
		for (n=m-1, p = pLRCoeff[n], L = 0.0; n >= 0; n--)
		{
			while (p > pLRCoeff[n]) 
			{
				L *= edR;
				dLdR *= edR;
				p--;
			}
			L -= LRCoeff[n];
			dLdR += LRCoeff[n] * double(p);
		}
		L *= pow(edR, pLRCoeff[0]);
		dLdR *= pow(edR, pLRCoeff[0] + 1);
	}
	else if (pnC < 6) pnC = 6;
	if (m > nLRCoeff - 2)
	{
		if (nLRCoeff > 0)
		{
			int *PB = pLRCoeff;
			double *CB = LRCoeff;
			pLRCoeff = new int[m+2];
			LRCoeff = new double[m+2];
			for (n=0; n < nLRCoeff; n++)
			{
				pLRCoeff[n] = PB[n];
				LRCoeff[n] = CB[n];
			}
			delete[] PB;
			delete[] CB;
		}
		else
		{
			pLRCoeff = new int[2];
			LRCoeff = new double[2];
		}
	}
	L += Uinf;
	pLRCoeff[m] = pnC;
	pLRCoeff[m + 1] = pnC + 2;
	x = pow(Ra, pnC + 1);
	D = dAdR - dLdR;
	F = D + double(pnC) / Ra * (A-L);
	//printf("Ra=%e, pnC=%d, x=%g, D=%g, F=%g\n", Ra, pnC, x, D, F);
	LRCoeff[m] = x / double(pnC) * (D - 0.5 * double(pnC + 2) * F);
	LRCoeff[m + 1] = 0.5 * Ra * Ra * x * F;
	nLRCoeff = m + 2;
	Lock->unlock();
}

void AnaPot::cdConnectSR()
{
	if (nPotCoeff == 0) return;
	double z, iR, d;
	double X, x;
	double dXdR;
	int n;
	QMutexLocker Locker(Lock);
	for (d = 0.0; d >= 0.0; Ri += 0.01)
	{
		if (Ri >= Ra) return;
		z = Ri + b * Rm;
		X = 1.0;
		x = (Ri - Rm) / z;
		dXdR = (1.0 - x) / z;
		for (n=0, d = 0.0; n < nPotCoeff; n++, X*=x) d += double(n+1) * PotCoeff[n] * X * dXdR;
	}
	Ri -= 0.01;
	//printf("Ri=%g, d=%g\n", Ri, d);
	iR = pow(Ri, -1.0 - iExp);
	iCoeff = d / (-iExp * iR);
	iOff = Point(Ri, 2) - iCoeff * iR * Ri;
}

void AnaPot::sConnect()
{
	Lock->lock();
	Tm += Point(Ra, 3) - Point(Ra, 2);
	iOff += Point(Ri, 2) - Point(Ri, 1);
	Lock->unlock();
}

void AnaPot::shift(double E)
{
	Uinf += E;
	iOff += E;
	Tm += E;
}

double AnaPot::getUinf()
{
	return Uinf;
}

int AnaPot::cAdRows()
{
    int n, AdRows;
	for (n = AdRows = 0; n < nLRCoeff; n++) if (LRCFree[n] && pLRCoeff[n] >= 12) AdRows++;
	return AdRows;
}

void AnaPot::calc_hCi_cP_EQS(int NEQ, double **EQS, double *Sig, double *EDiff, double hCi_cP)
{
    int n, m, k, l, i, j, N = nPotCoeff + nLRCoeff;
	double CC, F, z, x, X, dxdR;
	z = Ra + b * Rm;
	x = (Ra - Rm) / z;
	dxdR = (1.0 - x) / z;
	for (n=i=0; i < nLRCoeff; i++) if (pLRCoeff[i] >= 12 && LRCFree[i]) n++; 
	for (m=i=0; i < nLRCoeff; i++) if (pLRCoeff[i] >= 12 && LRCFree[i])
	{
		Sig[j = NEQ - n + m++] = 1.0 / ((CC = pow(LRCoeff[i-1] / LRCoeff[i-2], 3.0) * LRCoeff[i-3]) * hCi_cP);
		EDiff[j] = CC - LRCoeff[i];
		if (m==n)
		{
			F = pow(Ra, pLRCoeff[i] + 1) / double(pLRCoeff[i]);
			for (k=l=0, X = F * dxdR; k < nPotCoeff; k++, X*=x) if (SMap[l] == k) EQS[j][l++] = double(k+1) * X;
			for (k=0, X = F * (x = 1.0 / Ra); (l < NumFreePar ? SMap[l] < N : false); k++, X*=x) 
				if (pLRCoeff[SMap[l] - nPotCoeff] == k) EQS[j][l++] = double(-k) * X;
			if (ExcIntFree) EQS[j][l++] = (gamma * x - beta) * pow(Ra, gamma) * exp(-beta * Ra);
		}
		else l=0;
		while (l < NumFreePar) EQS[j][l++] = 0.0;
		for (l=0; (l < NumFreePar ? SMap[l] - nPotCoeff < i-3 : false); l++) ;
		if (SMap[l] - nPotCoeff == i-3) EQS[j][l++] -= CC / LRCoeff[i-3];
		if (SMap[l] - nPotCoeff == i-2) EQS[j][l++] += 3.0 * CC / LRCoeff[i-2];
		if (SMap[l] - nPotCoeff == i-1) EQS[j][l++] -= 3.0 * CC / LRCoeff[i-1];
		if (SMap[l] - nPotCoeff == i) EQS[j][l] = 1.0;
	}
}

double AnaPot::calcPotCoeff(int N, double *R, double *U, double *Sig, int n, double nRm, double nb)
{
	int i;
	double *UC = new double[N], FQSv, FQSn, TmC, PotCoeffC[n];
	b = nb;
	Rm = nRm;
	Lock->lock();
	if (Ri == 0.0) Ri = R[0];
	if (Ra == 0.0) Ra = R[N-1];
	if (nPotCoeff > 0) delete[] PotCoeff;
	PotCoeff = new double[nPotCoeff = n];
	FQSv = 2.0 * (FQSn = fitPotCoeff(N, R, U, n, Rm, b, PotCoeff, Tm, Sig));
	while (FQSn < 0.9999 * FQSv)
	{
		FQSv = FQSn;
		for (i=0; i<N; i++) UC[i] = U[i] - Point(R[i], 2);
		FQSn = fitPotCoeff(N, R, UC, n, Rm, b, PotCoeffC, TmC, Sig);
		for (i=0, Tm += TmC; i<n; i++) PotCoeff[i] += PotCoeffC[i];
	}
	delete[] UC;
	Lock->unlock();
	return FQSn;
	
	/*int i, j, k, m=2*n+1;
	double x, F[m], **M;
	M = Create(n + 1, n + 2);

	for (i=0; i<=n; i++) for (j=0; j<n+2; j++) M[i][j] = 0.0;
	for (i=0, F[0]=1.0; i<N; i++)
	{
		x = (R[i] - Rm) / (R[i] + b * Rm);
		for (j=1; j<m; j++) F[j] = F[j-1] * x;
		for (j=0; j<=n; j++) 
		{
			for (k=0; k<=n; k++) M[j][k] += F[k+j];
			M[j][n+1] += F[j] * U[i];
		}
	}
	SolvLinEqS(M, n+1, 1e-15);
	Tm = M[0][n+1];
	for (i=0; i<n; i++) PotCoeff[i] = M[i+1][n+1];
	Destroy(M, n + 1);*/
	
	//printf("Tm=%f\n", Tm);
	//for (i=0; i<n; i++) printf("a[%d]=%f\n", i, PotCoeff[i]);
}

void AnaPot::fitAdCorrToRmb()
{
	if (NumFreeAdCorr > 0)
	{
		int NPoints = 10000, n, m, a, SStart = nPotCoeff + nLRCoeff - (ExcIntFree ? 0 : 1), SMap[NAdCorr];
		double RMax = 20.0, *adCorrC = new double[NPoints], **EQS = Create(NPoints, NAdCorr), C[NAdCorr];
		double *Sig = new double[NPoints], *AD = new double[NPoints], FQS, oFQS;
		for (n=a=0, m = SStart; n < NAdCorr; n++, m++) if (adCorrFree[n]) SMap[a++] = m;
		for (n=0; n < NPoints; n++) 
		{
			adCorrC[n] = 0.0;
			Sig[n] = 0.01;
		}
		calcAdCorrPot(adCorrC, adCorrC, Ri, (RMax - Ri) / double(NPoints - 1), NPoints, 1.0, 1.0, NAdCorr, 1, PAdCorr, 0.0, 0.0, 
					  AdCorr_Rm, AdCorr_b, adCorr);
		if (AdCorrWF == 0) AdCorrWF = new double[NAdCorr];
		calcS(NPoints, Ri, RMax);
		for (a=0; a < NAdCorr; a++) C[a] = adCorr[a] / AdCorrWF[a];
		for (n=0; n < NPoints; n++) 
		{
			for (a=0; a < NAdCorr; a++) if (!adCorrFree[a]) adCorrC[n] -= C[a] * S[SStart + a][n];
			for (a=0; a < NumFreeAdCorr; a++) EQS[n][a] = S[SMap[a]][n];
		}
		Destroy(S, numFreeCoeff);
		S=0;
		SvdFit(EQS, C, adCorrC, Sig, NumFreeAdCorr, NPoints, 1e-12);
		for (a=0; a < NumFreeAdCorr; a++) adCorr[SMap[a] - SStart] = C[a];
		for (n=0, FQS = 0.0; n < NPoints; n++)
		{
			for (a=0, AD[n] = adCorrC[n]; a < NumFreeAdCorr; a++) AD[n] -= C[a] * EQS[n][a];
			FQS += AD[n] * AD[n];
		}
		printf("AnaPot::fitAdCorrToRmb(): FQS=%f\n", FQS *= 1e4);
		oFQS = 2.0 * FQS;
		while (FQS < 0.9999 *oFQS && FQS > 1.0)
		{
			oFQS = FQS;
			SvdFit(EQS, C, AD, Sig, NAdCorr, NPoints, 1e-12);
			for (a=0; a < NumFreeAdCorr; a++) adCorr[SMap[a] - SStart] += C[a];
			for (n=0, FQS = 0.0; n < NPoints; n++)
			{
				for (a=0, AD[n] = adCorrC[n]; a < NumFreeAdCorr; a++) AD[n] -= adCorr[SMap[a] - SStart] * EQS[n][a];
				FQS += AD[n] * AD[n];
			}
			printf("AnaPot::fitAdCorrToRmb(): FQS=%f\n", FQS *= 1e4);
		}
		for (a=0; a < NumFreeAdCorr; a++) adCorr[SMap[a] - SStart] *= AdCorrWF[SMap[a] - SStart];
		delete[] adCorrC;
		delete[] Sig;
		delete[] AD;
		Destroy(EQS, NPoints);
	}
	AdCorr_b = b;
	AdCorr_Rm = Rm;
}

double AnaPot::fitPotential(int NPoints, double R0, double h, double* U, double* Sig, int NCoeff, int NLRC, int* pLRC, 
							double nRi, double nRa, double nRm, double nb, double nUinf, double niExp)
{
	int i, j, NC = NCoeff + NLRC - 1;
	double **EQS = Create(NPoints, NC), *UD = new double[NPoints];
	double oFQS = 1e99, FQS, C[NC], R;
	bool NewLRC = false;
	if (SMap != 0)
	{
		delete[] SMap;
		SMap = 0;
	}
	if (NCoeff != nPotCoeff)
	{
		if (PotCoeff != 0)
		{
			delete[] PotCoeff;
			delete[] CoeffFree;
		}
		PotCoeff = new double[nPotCoeff = NCoeff];
		CoeffFree = new bool[NCoeff];
	}
	if (nLRCoeff != NLRC)
	{
		if (LRCoeff != 0) 
		{
			delete[] LRCoeff;
			delete[] pLRCoeff;
			delete[] LRCFree;
		}
		LRCoeff = new double[nLRCoeff = NLRC];
		LRCFree = new bool[NLRC];
		pLRCoeff = pLRC;
		NewLRC = true;
	}
	else for (i=0; i < NLRC; i++) if (pLRCoeff[i] != pLRC[i]) 
	{
		NewLRC = true;
		pLRCoeff[i] = pLRC[i];
	}
	for (i=0; i < NCoeff; i++) 
	{
		PotCoeff[i] = 0.0;
		CoeffFree[i] = true;
	}
	if (NewLRC) 
	{
		for (i=0; i < NLRC; i++) 
		{
			LRCoeff[i] = 0.0;
			LRCFree[i] = true;
		}
		numFreeLRC = NLRC - 1;
		NumFreePar = numFreeCoeff = NCoeff + numFreeLRC;
		SMap = new int[numFreeCoeff];
		for (i=0; i < numFreeCoeff; i++) SMap[i] = i;
	}
	else 
	{
		SMap = new int[NCoeff + NLRC - 1];
		for (j=0; j < NCoeff; j++) SMap[j] = j;
		for (i = numFreeLRC = 0; i < NLRC - 1; i++) if (LRCFree[i]) 
		{
			numFreeLRC++;
			SMap[j++] = NCoeff + i;
			LRCoeff[i] = 0.0;
		}
		NumFreePar = numFreeCoeff = NCoeff + numFreeLRC;
	}
	Ri = nRi;
	Ra = nRa;
	Rm = nRm;
	b = nb;
	Uinf = nUinf;
	iExp = niExp;
	calcS(NPoints, R0, R0 + h * double(NPoints - 1));
	double WLRC[NLRC - 1];
	for (i=0; i < NLRC - 1; i++) if (!LRCFree[i]) WLRC[i] = LRCoeff[i] / SWF[NCoeff + i];
	for (i=0; i < NPoints; i++) 
	{
		UD[i] = U[i];
		for (j=0; j < numFreeCoeff; j++) EQS[i][j] = S[SMap[j]][i];
		for (j=0; j < NLRC - 1; j++) if (!LRCFree[j]) UD[i] -= S[NCoeff + j][i] * WLRC[j];
	}
	Destroy(S, NC);
	S=0;
	SvdFit(EQS, C, UD, Sig, numFreeCoeff, NPoints, 1e-12);
	updatePotential(C);
	for (i=0, FQS = 0.0, R = R0; i < NPoints; i++, R+=h)
	{
		UD[i] = U[i] - Point(R);
		FQS += UD[i] * UD[i] / (Sig[i] * Sig[i]);
	}
	printf("FQS = %f\n", FQS);
	//printf("C6=%f\n", C[20] * SWF[20]);
	while (FQS < 0.9999 * oFQS)
	{
		SvdFit(EQS, C, UD, Sig, numFreeCoeff, NPoints, 1e-12);
		oFQS = FQS;
		updatePotential(C);
		for (i=0, FQS = 0.0, R = R0; i < NPoints; i++, R+=h)
		{
			UD[i] = U[i] - Point(R);
			FQS += UD[i] * UD[i] / (Sig[i] * Sig[i]);
		}
		printf("FQS = %f\n", FQS);
	}
	//printf("C6=%f\n", C[20] * SWF[20]);
	delete[] UD;
	Destroy(EQS, NPoints);
	printf("b=%f, FQS=%f\n", b, FQS);
	return FQS;
}

void AnaPot::calcS(int numPoints, double Rmin, double Rmax)
{
	int n, m, k, NOC = nPotCoeff + nLRCoeff - 1; 
	if (ExcIntFree) NOC++;
	NumS = NFitCoeff = NOC + NAdCorr;
	for (n = nLRCoeff - 1; (n>=0 ? !LRCFree[n] : false); n--) ;
	int PLLRC = pLRCoeff[(n>=0 ? n : nLRCoeff - 1)];
	double R, AO[nPotCoeff], AF[nPotCoeff], IF[nPotCoeff], IO[nPotCoeff], LRCO[nLRCoeff - 1], LRCF[nLRCoeff - 1];
	double LLRCO = -pow(Ra, -PLLRC), x = 0.0, xF = 0.0, DRa = 1.0 / (Ra + b * Rm), XA = (Ra - Rm) * DRa, XFA = 1.0;
	double dRi = 1.0 / (Ri + b * Rm), XI = (Ri - Rm) * dRi, XFI = 1.0, LLRCd = -Ra / (double(PLLRC) * LLRCO);
	double ICO = pow(Ri, -iExp), dICd = -Ri / (iExp * ICO), dRa = 1.0 / Ra, dXIdRi = (1.0 - XI) * dRi;
	double dXAdRa = (1.0 - XA) * DRa, ExcIntF = 0.0, ExcIntO = 0.0, h = (Rmax - Rmin) / double(numPoints - 1);
	if (SWF != 0) delete[] SWF;
	SWF = new double[NOC];
	S = Create(NFitCoeff, numPoints);
	for (n=0; n < nPotCoeff; n++)
	{
		AF[n] = double(n+1) * XFA * dXAdRa * LLRCd;
		AO[n] = AF[n] * LLRCO - (XFA *= XA);
		IF[n] = XFI * dICd * double(n+1) * dXIdRi;
		IO[n] = AO[n] + (XFI *= XI) - ICO * IF[n];
	}
	for (n=k=0, R=1.0; n < nLRCoeff - 1; n++, k++)
	{
		while (k < pLRCoeff[n])
		{
			R *= dRa;
			k++;
		}
		LRCO[n] = -R;
		LRCF[n] = -double(k) * (R *= dRa) * LLRCd;
		LRCO[n] += (LRCF[n] * LLRCO);
	}
	if (ExcIntFree)
	{
		ExcIntO = -pow(Ra, gamma) * exp(-beta * Ra);
		ExcIntF = -(gamma * dRa - beta) * ExcIntO * LLRCd;
		ExcIntO += ExcIntF * LLRCO;
	}
	for (n=0; n < NOC; n++) SWF[n] = 0.0;
	for (n=0; n < NAdCorr; n++) AdCorrWF[n] = 0.0;
	for (n=0, R = Rmin; n < numPoints; n++, R+=h)
	{
		if (R <= Ri)
		{
			x = pow(R, -iExp);
			for (k=0; k < nPotCoeff; k++) SWF[k] += fabs(S[k][n] = x * IF[k] + IO[k]);
			for (m=0; m < nLRCoeff - 1; m++, k++) SWF[k] += fabs(S[k][n] = LRCO[m]);
		}
		else if (R < Ra)
		{
			x = (R - Rm) / (R + b * Rm);
			for (k=0, xF = x; k < nPotCoeff; k++, xF *= x) SWF[k] += fabs(S[k][n] = AO[k] + xF);
			for (m=0; m < nLRCoeff - 1; m++, k++) SWF[k] += fabs(S[k][n] = LRCO[m]);
		}
		else
		{
			for (m=1, S[nPotCoeff][n] = -pow(DRa = 1.0 / R, k = pLRCoeff[0]); m < nLRCoeff - 1; m++)
				for (S[nPotCoeff + m][n] = S[nPotCoeff + m - 1][n]; k < pLRCoeff[m]; k++) 
					S[nPotCoeff + m][n] *= DRa;
			for (xF = S[nPotCoeff + m - 1][n]; k < pLRCoeff[m]; k++) xF *= DRa;
			for (k=0; k < nPotCoeff; k++) SWF[k] += fabs(S[k][n] = AF[k] * xF);
			for (m=0; m < nLRCoeff - 1; m++, k++) SWF[k] += fabs(S[k][n] += (LRCF[m] * xF));
		}
		if (ExcIntFree)
		{
			if (R < Ra) S[k][n] = ExcIntO;
			else S[k][n] = -pow(R, gamma) * exp(-beta * R) + ExcIntF * xF;
			SWF[k] += fabs(S[k][n]);
			k++;
		}
		if (NAdCorr > 0)
		{
			if (NAdCorr > 1) x = (R - Rm) / (R + b * Rm);
			AdCorrWF[0] += fabs(S[k++][n] = pow(2.0 * Rm / (R + Rm), PAdCorr));
			for (m=1; m < NAdCorr; m++, k++) AdCorrWF[m] += fabs(S[k][n] = S[k-1][n] * x);
		}
	}
	for (n=0; n < NOC; n++) SWF[n] = 1.0 / SWF[n];
	for (n=0; n < NAdCorr; n++) AdCorrWF[n] = 1.0 / AdCorrWF[n];
	for (n=0; n < numPoints; n++)
	{
		for (k=0; k < NOC; k++) S[k][n] *= SWF[k];
		for (m=0; m < NAdCorr; m++, k++) S[k][n] *= AdCorrWF[m];
	}
}

void AnaPot::createSMap()
{
	int n, m;
	for (n = NumFreeAdCorr = 0; n < NAdCorr; n++) if (adCorrFree[n]) NumFreeAdCorr++;
	NumFreePar = numFreeCoeff + NumFreeAdCorr;
	NumFreeLRC = numFreeLRC;
	if (ExcIntFree)
	{
		NumFreePar++;
		NumFreeLRC++;
	}
	SMap = new int[NumFreePar];
	for (n=m=0; n < nPotCoeff; n++) if (CoeffFree[n]) SMap[m++] = n;
	for (n=0; n < nLRCoeff - 1; n++) if (LRCFree[n]) SMap[m++] = n + nPotCoeff;
	if (!LRCFree[n] && SMap[m-1] >= nPotCoeff) 
	{
		m--;
		NumFreeLRC--;
		NumFreePar--;
	}
	if (ExcIntFree) SMap[m++] = n + nPotCoeff;
	for (n=0; n < NAdCorr; n++) if (adCorrFree[n]) SMap[m++] = n + nPotCoeff + nLRCoeff - (ExcIntFree ? 0 : 1);
}

PotentialData* AnaPot::getPotentialData()
{
    int n;
	PotentialData *R = new PotentialData();
	R->PotType = analyticalPotential;
	Lock->lock();
	R->nPotCoeff = nPotCoeff;
	R->nLRCoeff = nLRCoeff;
	R->LRCoeff = new double[nLRCoeff];
	R->pLRCoeff = new int[nLRCoeff];
	R->LRCFree = new bool[nLRCoeff];
	for (n=0; n < nLRCoeff; n++)
	{
		R->LRCoeff[n] = LRCoeff[n];
		R->pLRCoeff[n] = pLRCoeff[n];
		R->LRCFree[n] = LRCFree[n];
	}
	R->NAdCorr = NAdCorr;
	R->TAdCorr = TAdCorr;
	R->PAdCorr = PAdCorr;
	R->NSpinRGamma = NSpinRGamma;
	R->Ri = Ri;
	R->Ra = Ra;
	R->iExp = iExp;
	R->iOff = iOff;
	R->iCoeff = iCoeff;
	R->b = b;
	R->Rm = Rm;
	R->Tm = Tm;
	R->De = Uinf;
	R->PotCoeff = new double[nPotCoeff];
	R->CFree = new bool[nLRCoeff];
	for (n=0; n < nPotCoeff; n++) 
	{
		R->PotCoeff[n] = PotCoeff[n];
		R->CFree[n] = CoeffFree[n];
	}
	R->A_ex = A_ex;
	R->beta = beta;
	R->gamma = gamma;
	R->q_e = q_e;
	R->q_f = q_f;
	if (NAdCorr > 0)
	{
		R->adCorr = new double[NAdCorr];
		R->adCorrFree = new bool[NAdCorr];
		for (n=0; n < NAdCorr; n++) 
		{
			R->adCorr[n] = adCorr[n];
			R->adCorrFree[n] = adCorrFree[n];
		}
	}
	R->RIso1AdCorr = RIso1AdCorr;
	R->RIso2AdCorr = RIso2AdCorr;
	R->SpinRGamma = (NSpinRGamma > 0 ? new double[NSpinRGamma] : 0);
	for (n=0; n < NSpinRGamma; n++) R->SpinRGamma[n] = SpinRGamma[n];
	R->SpinRR1 = SpinRR1;
	R->SpinRR2 = SpinRR2;
	R->EIFree = ExcIntFree;
	R->fitResult = Result;
	R->FQS = FQS;
	Lock->unlock();
	return R;
}

void AnaPot::restorePotential(double* bC)
{
    int n;
	Lock->lock();
	for (n=0; n < nPotCoeff; n++) PotCoeff[n] = bC[n];
	for (n=0; n < nLRCoeff - 1; n++) LRCoeff[n] = bC[nPotCoeff + n];
	for (n=0; n < NAdCorr; n++) adCorr[n] = bC[nPotCoeff + nLRCoeff - 1 + n];
	if (ExcIntFree) A_ex = bC[nPotCoeff + nLRCoeff + n];
	Lock->unlock();
	cdConnectLR1C();
	cdConnectSR();
}

void AnaPot::saveCoefficients(double*& bC)
{
	int n;
	if (bC == 0) bC = new double[nPotCoeff + nLRCoeff - (ExcIntFree ? 0 : 1) + NAdCorr];
	for (n=0; n < nPotCoeff; n++) bC[n] = PotCoeff[n];
	for (n=0; n < nLRCoeff - 1; n++) bC[nPotCoeff + n] = LRCoeff[n];
	for (n=0; n < NAdCorr; n++) bC[nPotCoeff + nLRCoeff + n - 1] = adCorr[n];
	if (ExcIntFree) bC[nPotCoeff + nLRCoeff + n] = A_ex;
}

void AnaPot::updatePotential(double *CD)
{
	Lock->lock();
	int n, m, N = NumFreePar, a;
	for (n=0; n<N; n++)
	{
		m = (SMap != 0 ? SMap[n] : n);
		if (m < nPotCoeff) PotCoeff[m] += SWF[m] * CD[n];
		else if (m < nPotCoeff + nLRCoeff - 1) LRCoeff[m - nPotCoeff] += SWF[m] * CD[n];
		else if (ExcIntFree && m < nPotCoeff + nLRCoeff) A_ex += SWF[m] * CD[n];
		else 
		{
			a = m - nPotCoeff - nLRCoeff + (ExcIntFree ? 0 : 1);
			adCorr[a] += AdCorrWF[a] * CD[n];
		}
	}
	Lock->unlock();
	cdConnectLR1C();
	cdConnectSR();
	//LRCoeff[nLRCoeff - 1] = 0.0;
}


MTTPot::MTTPot(PotFit *Fit) : PotWorker(Fit, ModifiedTangToenniesPotential)
{
	A = LRC = C = CLRC = 0;
	NA = NLRC = NC = NCLRC = 0;
	alpha = beta = gamma = Uinf = Calpha = Cbeta = CInf = B = alpha1 = 0.0;
	pLRC = pCLRC = 0;
}

MTTPot::MTTPot(const MTTPot& Co) : PotWorker(Co)
{
	int n;
	if ((NA = Co.NA) > 0) for (n=0, A = new double[NA]; n < NA; n++) A[n] = Co.A[n];
	else A = 0;
	if ((NLRC = Co.NLRC) > 0) 
		for (n=0, LRC = new double[NLRC], pLRC = new int[NLRC]; n < NLRC; n++)
	{
		LRC[n] = Co.LRC[n];
		pLRC[n] = Co.pLRC[n];
	}
	B = Co.B;
	alpha1 = Co.alpha1;
	alpha = Co.alpha;
	beta = Co.beta;
	gamma = Co.gamma;
	Uinf = Co.Uinf;
	if ((NC = Co.NC) > 0) for (n=0, C = new double[NC]; n < NC; n++) C[n] = Co.C[n];
	else C = 0;
	if ((NCLRC = Co.NCLRC) > 0) 
		for (n=0, CLRC = new double[NCLRC], pCLRC = new int[NCLRC]; n < NCLRC; n++)
	{
		CLRC[n] = Co.CLRC[n];
		pCLRC[n] = Co.pCLRC[n];
	}
	Calpha = Co.Calpha;
	Cbeta = Co.Cbeta;
	CInf = Co.CInf;
}

MTTPot::~MTTPot()
{
	if (NA > 0) delete[] A;
	if (NLRC > 0)
	{
		delete[] LRC;
		delete[] pLRC;
	}
	if (NC > 0) delete[] C;
	if (NCLRC > 0)
	{
		delete[] CLRC;
		delete[] pCLRC;
	}
}

void MTTPot::getCoefficients(double &rB, double &ralpha1, int& rNA, double*& rA, double& rAlpha, double& rBeta, 
							 double& rGamma, int& rNLRC, int*& rpLRC, double*& rLRC, 
							 double &rUinf, int &rNC, double *&rC, double &rCalpha, 
							 double &rCbeta, int &rNCLRC, int *&rpCLRC, double *&rCLRC, 
							 double &rCInf)
{
	int n;
	if ((rNA = NA) > 0) for (n=0, rA = new double[NA]; n < NA; n++) rA[n] = A[n];
	else rA = 0;
	rB = B;
	ralpha1 = alpha1;
	rAlpha = alpha;
	rBeta = beta;
	rGamma = gamma;
	if ((rNLRC = NLRC) > 0) 
		for (n=0, rLRC = new double[NLRC], rpLRC = new int[NLRC]; n < NLRC; n++)
	{
		rpLRC[n] = pLRC[n];
		rLRC[n] = LRC[n];
	}
	else 
	{
		rLRC = 0;
		rpLRC = 0;
	}
	rUinf = Uinf;
	if ((rNC = NC) > 0) for (n=0, rC = new double[NC]; n < NC; n++) rC[n] = C[n];
	else rC = 0;
	rCalpha = Calpha;
	rCbeta = Cbeta;
	if ((rNCLRC = NCLRC) > 0)
		for (n=0, rCLRC = new double[NCLRC], rpCLRC = new int[NCLRC]; n < NCLRC; n++)
	{
		rpCLRC[n] = pCLRC[n];
		rLRC[n] = CLRC[n];
	}
	rCInf = CInf;
}

double MTTPot::GetDe(int c)
{
	double R, V;
	getMinimum(R, V, c);
	switch (c)
	{
		case 0:
			return Uinf - V;
		case -1:
			return Uinf - V - CInf;
		case 1:
			return Uinf - V + 2.0 * CInf;
			break;
	}
	return 0.0;
}

void MTTPot::getMinimum(double& R, double& V, int c)
{
	double MD = 1e-5, CMin = Uinf - 1e-4, MinR = -1.0, LMin = Uinf, S, Ri = rmin, Ra = rmax;
	Lock->lock();
	for (S=1.0; LMin - CMin > MD; S *= 0.1)
	{
		for (R = Ri, LMin = CMin; R <= Ra; R+=S) if ((V = Point(R, c)) < CMin)
		{
			CMin = V;
			MinR = R;
		}
		Ri = R - S;
		Ra = R + S;
	}
	R = MinR;
	V = CMin;
	Lock->unlock();
}

PotentialData* MTTPot::getPotentialData()
{
	int n;
	PotentialData *R = new PotentialData();
	Lock->lock();
	R->NA = NA;
	R->NLRC = NLRC;
	if (NLRC > 0)
	{
		R->pLRC = new int[NLRC];
		R->LRC = new double[NLRC];
		for (n=0; n < NLRC; n++)
		{
			R->pLRC[n] = pLRC[n];
			R->LRC[n] = LRC[n];
		}
	}
	R->NSA = NC;
	R->NSLRC = NCLRC;
	R->alpha = alpha;
	R->beta = beta;
	R->gamma = gamma;
	if (NA > 0)
	{
		R->A = new double[NA];
		for (n=0; n < NA; n++) R->A[n] = A[n];
	}
	R->Uinf = Uinf;
	R->Salpha = Calpha;
	R->Sbeta = Cbeta;
	if (NC > 0)
	{
		R->SA = new double[NC];
		for (n=0; n < NC; n++) R->SA[n] = C[n];
	}
	if (NCLRC > 0)
	{
		R->pSLRC = new int[NCLRC];
		R->SLRC = new double[NCLRC];
		for (n=0; n < NCLRC; n++)
		{
			R->pSLRC[n] = pCLRC[n];
			R->SLRC[n] = CLRC[n];
		}
	}
	R->Sinf = CInf;
	R->B = B;
	R->alpha1 = alpha1;
	Lock->unlock();
	return R;
}

double MTTPot::Point(double r, int c)
{
	double bR = beta * r, V = 0.0, FS = 1.0, dR = 1.0 / r, RP = 1.0, F = 1.0, BRP = 1.0;
	double expBR = exp(-bR);
	int n, m, m1, m2 = 0;
	for (n = NA - 1; n >= 0; n--)
	{
		V *= r;
		V += A[n];
	}
	V *= exp((-alpha - gamma * r) * r);
	V += Uinf + B * exp(-alpha1 * r);
	for (n=0; n < NLRC; n++)
	{
		m1 = m2;
		m2 = pLRC[n];
		for (m = m1 + 1; m <= m2; m++)
		{
			BRP *= bR;
			F *= m;
			FS += BRP / F;
			RP *= dR;
		}
		V -= LRC[n] * RP * (1.0 - FS * expBR);
	}
	if (c != 0)
	{
		bR = Cbeta * r, FS = 0.0, RP = 1.0, F = 1.0, BRP = 1.0;
		expBR = exp(-bR);
		for (n = NC - 1; n >= 0; n--)
		{
			FS *= r;
			FS += C[n];
		}
		FS *= exp(-Calpha * r);
		if (c==-1) V -= FS + CInf;
		else V += 2.0 * (FS + CInf);
		for (n = m2 = 0, FS = 1.0; n < NCLRC; n++)
		{
			m1 = m2;
			m2 = pCLRC[n];
			for (m = m1 + 1; m <= m2; m++)
			{
				BRP *= bR;
				F *= m;
				FS += BRP / F;
				RP *= dR;
			}
			if (c==-1) V += CLRC[n] * RP * (1.0 - FS * expBR);
			else V -= 2.0 * CLRC[n] * RP * (1.0 - FS * expBR);
		}
	}
	return V;
}

void MTTPot::getLRCoeff(double& R, int& numCoefficients, int*& Exponents, double*& Coefficients)
{
    R = 0.0;
	numCoefficients = NC;
	Exponents = pCLRC;
	Coefficients = C;
}

double MTTPot::GetUinf(int c)
{
	switch (c)
	{
		case 0:
			return Uinf;
		case -1:
			return Uinf - CInf;
		case 1:
			return Uinf + 2.0 * CInf;
			break;
	}
	return 0.0;
}

MTTPot* MTTPot::scalePotential(double newRe, double newDe)
{
	double Re, De;
	int n;
	getMinimum(Re, De);
	Lock->lock();
	De = Uinf - De;
	double XSF = Re / newRe, YSF = newDe / De;
	MTTPot *RPot = new MTTPot(0);
	RPot->alpha = XSF * alpha;
	RPot->alpha1 = XSF * alpha1;
	RPot->B = YSF * B;
	RPot->gamma = XSF * XSF * gamma;
	RPot->A = new double[RPot->NA = NA];
	for (n=0; n < NA; n++) RPot->A[n] = YSF * pow(XSF, n) * A[n];
	RPot->LRC = new double[RPot->NLRC = NLRC];
	RPot->pLRC = new int[NLRC];
	for (n=0; n < NLRC; n++)
	{
		RPot->pLRC[n] = pLRC[n];
		RPot->LRC[n] = YSF * pow(XSF, -pLRC[n]) * LRC[n];
	}
	RPot->beta = XSF * beta;
	RPot->Uinf = Uinf;
	RPot->Calpha = XSF * Calpha;
	if ((RPot->NC = NC) > 0)
	{
		RPot->C = new double[NC];
		for (n=0; n < NC; n++) RPot->C[n] = YSF * pow(XSF, n) * C[n];
	}
	else RPot->C = 0;
	if ((RPot->NCLRC = NCLRC) > 0)
	{
		RPot->CLRC = new double[NCLRC];
		RPot->pCLRC = new int[NCLRC];
		for (n=0; n < NCLRC; n++)
		{
			RPot->pCLRC[n] = pCLRC[n];
			RPot->CLRC[n] = YSF * pow(XSF, -pCLRC[n]) * CLRC[n];
		}
	}
	else
	{
		RPot->pCLRC = 0;
		RPot->CLRC = 0;
	}
	RPot->Cbeta = XSF * beta;
	RPot->CInf = CInf;
	Lock->unlock();
	return RPot;
}

void MTTPot::setCoefficients(double nB, double nalpha1, int nNA, double* nA, double nAlpha, double nBeta, 
							 double nGamma, int nNLRC, int* npLRC, double* nLRC, 
							 double nUinf, int nNC, double *nC, double nCalpha, 
							 double nCbeta, int nNCLRC, int *npCLRC, double *nCLRC, 
							 double nCInf)
{
	if (NA > 0) delete[] A;
	if (NLRC > 0)
	{
		delete[] LRC;
		delete[] pLRC;
	}
	if (NC > 0) delete[] C;
	if (NCLRC > 0)
	{
		delete[] CLRC;
		delete[] pCLRC;
	}
	B = nB;
	alpha1 = nalpha1;
	NA = nNA;
	A = nA;
	alpha = nAlpha;
	beta = nBeta;
	gamma = nGamma;
	NLRC = nNLRC;
	pLRC = npLRC;
	LRC = nLRC;
	Uinf = nUinf;
	NC = nNC;
	C = nC;
	Calpha = nCalpha;
	Cbeta = nCbeta;
	NCLRC = nNCLRC;
	pCLRC = npCLRC;
	CLRC = nCLRC;
	CInf = nCInf;
}


MLRPot::MLRPot(PotFit *Fit) : PotWorker(Fit, MorseLongRangePotential)
{
	Ci = beta = dCiLR = 0;
	pLRC = 0;
	p = q = NLRC = Nbeta = 0;
	De = Re = Rep = Rref = Rrefp = Rrefq = URe = betainf = TAs = dULRdRe = 0.0;
}

MLRPot::MLRPot(const MLRPot& C) : PotWorker(C)
{
	int n;
	if (C.NLRC > 0)
	{
		Ci = new double[NLRC = C.NLRC];
		pLRC = new int[NLRC];
		dCiLR = new double[NLRC];
		for (n=0; n < NLRC; n++) 
		{
			Ci[n] = C.Ci[n];
			pLRC[n] = C.pLRC[n];
			dCiLR[n] = C.dCiLR[n];
		}
	}
	else
	{
		Ci = 0;
		pLRC = 0;
		NLRC = 0;
	}
	if (C.Nbeta > 0)
	{
		beta = new double[Nbeta = C.Nbeta];
		for (n=0; n < Nbeta; n++) beta[n] = C.beta[n];
	}
	else
	{
		beta = 0;
		Nbeta = 0;
	}
	p = C.p;
	q = C.q;
	TAs = C.TAs;
	De = C.De;
	Re = C.Re;
	Rep = C.Rep;
	Rref = C.Rref;
	Rrefp = C.Rrefp;
	Rrefq = C.Rrefq;
	URe = C.URe;
	betainf = C.betainf;
	dULRdRe = C.dULRdRe;
}

MLRPot::~MLRPot()
{
	if (Ci != 0) 
	{
		delete[] Ci;
		delete[] pLRC;
		delete[] dCiLR;
	}
	if (beta != 0) delete[] beta;
}

void MLRPot::getCoefficients(int& rp, int& rq, double& rTAs, double& rDe, double& rRe, 
							 double& rRref, int& rNbeta, double*& rbeta, int& rNLRC, 
							 int *&rpLRC, double*& rLRC)
{
	int n;
	if (NLRC > 0)
	{
		rLRC = new double[rNLRC = NLRC];
		rpLRC = new int[NLRC];
		for (n=0; n < NLRC; n++) 
		{
			rLRC[n] = Ci[n];
			rpLRC[n] = pLRC[n];
		}
	}
	else
	{
		rLRC = 0;
		rpLRC = 0;
		rNLRC = 0;
	}
	if (Nbeta > 0)
	{
		rbeta = new double[rNbeta = Nbeta];
		for (n=0; n < Nbeta; n++) rbeta[n] = beta[n];
	}
	else
	{
		rbeta = 0;
		rNbeta = 0;
	}
	rp = p;
	rq = q;
	rTAs = TAs + De;
	rDe = De;
	rRe = Re;
	rRref = Rref;
}

void MLRPot::getDerivatives(double r, double& E, double* dbeta, double* dC, double& dRe, double& dDe)
{
	int n;
	double Rq = pow(r, q), zDe = 2.0 * De, UExp, myp;
	double Rp = Rq * pow(r, p-q), Rs;
	double yqRef = (Rq - Rrefq) / (Rq + Rrefq), ypRef = (Rp - Rrefp) / (Rp + Rrefp), yep = (Rp - Rep) / (Rs = Rp + Rep);
	double ULR, b = 0.0, mypRef = (1.0 - ypRef), ULRRe, Exp; 
	for (n=1, ULR = Ci[0] * (dC[0] = pow(r, -pLRC[0])); n < NLRC; n++)
	{
		dC[n] = dC[n-1] * pow(r, pLRC[n-1] - pLRC[n]);
		ULR += Ci[n] * dC[n];
	}
	for (n = Nbeta - 1; n>=0; n--)
	{
		b *= yqRef;
		b += beta[n];
	}
	b *= mypRef;
	b += ypRef * betainf;
	double zUM, mUExp = 1.0 - (UExp = (ULRRe = ULR * URe) * (Exp = exp(-b * yep)));
	E = TAs + De * mUExp * mUExp;
	double db = (zUM = zDe * UExp * mUExp) * yep * mypRef;
	for (n=1, dbeta[0] = db; n < Nbeta; n++) dbeta[n] = yqRef * dbeta[n-1];
	double pExp = zDe * mUExp * Exp * URe, pyepRef = ULRRe * (1.0 - (myp = yep * ypRef));
	for (n=0; n < NLRC; n++) dC[n] = pExp * (dCiLR[n] * pyepRef - dC[n]);
	dRe = zUM * (dULRdRe * (myp - 1.0) - b * double(p) * Rep / (Re * Rs) * (1.0 + yep)); 
	dDe = mUExp * mUExp + zUM * myp / De;
}

void MLRPot::getLRCoeff(double& R, int& numCoefficients, int*& Exponents, double*& Coefficients)
{
    R = 0.0;
	numCoefficients = NLRC;
	Exponents = pLRC;
	Coefficients = Ci;
}

void MLRPot::getNC(int& N, int& NC, int& nLRC)
{
	N = (NC = Nbeta) + (nLRC = NLRC) + 3; 
}

double MLRPot::GetDe()
{
	return De;
}

void MLRPot::getMinimum(double& R, double& T)
{
	Lock->lock();
	R = Re;
	T = TAs - De;
	Lock->unlock();
}

PotentialData* MLRPot::getPotentialData()
{
	int n;
	PotentialData *R = new PotentialData();
	Lock->lock();
	R->p = p;
	R->q = q;
	R->NCi = NLRC;
	R->Nb = Nbeta;
	R->pCi = new int[NLRC];
	R->NAdCorr = NAdCorr;
	R->TAdCorr = TAdCorr;
	R->PAdCorr = PAdCorr;
	R->bA = new double[Nbeta];
	for (n=0; n < Nbeta; n++) R->bA[n] = beta[n];
	R->Ci = new double[NLRC];
	for (n=0; n < NLRC; n++)
	{
		R->Ci[n] = LRC[n];
		R->pCi[n] = pLRC[n];
	}
	R->Re = Re;
	R->Rref = Rref;
	R->T = TAs;
	R->De = De;
	if (NAdCorr > 0)
	{
		R->adCorr = new double[NAdCorr];
		for (n=0; n < NAdCorr; n++) R->adCorr[n] = adCorr[n];
	}
	R->RIso1AdCorr = RIso1AdCorr;
	R->RIso2AdCorr = RIso2AdCorr;
	R->AdCorr_b = AdCorr_b;
	R->AdCorr_Rm = AdCorr_Rm;
	Lock->unlock();
	return R;
}

double MLRPot::Point(double r, int)
{
	int n;
	double Rq = pow(r, q);
	double Rp = Rq * pow(r, p-q);
	double yqRef = (Rq - Rrefq) / (Rq + Rrefq), ypRef = (Rp - Rrefp) / (Rp + Rrefp);
	double ULR = 0.0, b = 0.0;
	for (n = NLRC - 1; n>=0; n--)
	{
		ULR += Ci[n];
		ULR *= pow(r, (n>0 ? pLRC[n-1] - pLRC[n] : -pLRC[n]));
	}
	for (n = Nbeta - 1; n>=0; n--)
	{
		b *= yqRef;
		b += beta[n];
	}
	b *= (1.0 - ypRef);
	b += ypRef * betainf;
	b = 1.0 - ULR * URe * exp(- b * (Rp - Rep) / (Rp + Rep));
	//printf("exp=%e, ", pr);
	return TAs + De * b * b;
}

double MLRPot::getUinf()
{
	return TAs;
}

MLRPot* MLRPot::scalePotential(double newRe, double newDe)
{
	Lock->lock();
	MLRPot *rPot = new MLRPot(0);
	int n, *rpLRC = new int[NLRC];
	double RF = newRe / Re, DF = newDe / De;
	double *rbeta = new double[Nbeta], *rCi = new double[NLRC];
	for (n=0; n < Nbeta; n++) rbeta[n] = beta[n];
	for (n=0; n < NLRC; n++) rCi[n] = Ci[n] * DF * pow(RF, -(rpLRC[n] = pLRC[n]));
	rPot->setCoefficients(p, q, TAs * DF, newDe, newRe, Rref * RF, Nbeta, rbeta, NLRC, 
						  rpLRC, rCi);
	Lock->unlock();
	return rPot;
}

void MLRPot::setCoefficients(int np, int nq, double nTAs, double nDe, double nRe, double nRref,
							 int nNbeta, double* nbeta, int nNLRC, int* npLRC, double* nLRC)
{
	p = np;
	q = nq;
	TAs = nTAs - nDe;
	De = nDe;
	Re = nRe;
	Rref = nRref;
	Nbeta = nNbeta;
	if (beta != 0) delete[] beta;
	beta =  nbeta;
	if (nNLRC != NLRC) 
	{
	  delete[] dCiLR;
	  dCiLR = new double[nNLRC];
	}
	NLRC = nNLRC;
	if (pLRC != 0) delete[] pLRC;
	if (Ci != 0) delete[] Ci;
	Ci = nLRC;
	pLRC = npLRC;
	calcRIndP();
}

void MLRPot::calcRIndP()
{
	int n;
	for (n=0, URe = dULRdRe = 0.0; n < NLRC; n++) 
	{
		URe += Ci[n] * (dCiLR[n] = pow(Re, -pLRC[n]));
		dULRdRe += pLRC[n] * Ci[n] * dCiLR[n] / Re;
	}
	dULRdRe *= (URe = 1.0 / URe);
	Rep = pow(Re, p);
	Rrefp = pow(Rref, p);
	Rrefq = pow(Rref, q);
	betainf = log(2.0 * De * URe);
}

void MLRPot::shift(double Energy)
{
	TAs += Energy;
}

void MLRPot::test()
{
	QFile F("test.dat");
	F.open(QIODevice::WriteOnly);
	QTextStream S(&F);
	double B, V, R, E, dBeta[Nbeta], dC[NLRC], dRe, dDe, ds, lE;
	int n;
	for (R = 0.5 * Re; R <= 5 * Re; R += (R < Re ? 0.1 * Re : Re))
	{
		S << "R = " << QString::number(R, 'f', 2) << ":\n";
		for (n=0; n < Nbeta; n++)
		{
			S << "beta[" << QString::number(n) << "]:\n";
			ds = 100.0 / (B = beta[n]);
			for (V = 0.9 * B, lE = 0.0; (V > 0.0 ? V <= 1.1 * B : V >= 1.1 * B); V += 0.01 * B, lE = E)
			{
				beta[n] = V;
				calcRIndP();
				getDerivatives(R, E, dBeta, dC, dRe, dDe);
				S << "V=" << QString::number(V, 'g', 3) << ", E=" << QString::number(E, 'g', 7)
				  << ", dE/dbeta=" << QString::number(dBeta[n], 'g', 7) << " ~ " 
				  << (lE != 0.0 ? QString::number((E - lE) * ds, 'g', 7) : "?") << "\n";
			}
			beta[n] = B;
		}
		for (n=0; n < NLRC; n++)
		{
			S << "Ci[" << QString::number(n) << "]:\n";
			ds = 100 / Ci[n];
			for (V = 0.9 * (B = Ci[n]), lE = 0.0; (V > 0.0 ? V <= 1.1 * B : V >= 1.1 * B); V += 0.01 * B, lE = E)
			{
				Ci[n] = V;
				calcRIndP();
				getDerivatives(R, E, dBeta, dC, dRe, dDe);
				S << "V=" << QString::number(V, 'g', 3) << ", E=" << QString::number(E, 'g', 7)
				  << ", dE/dC=" << QString::number(dC[n], 'g', 7) << " ~ " 
				  << (lE != 0.0 ? QString::number((E - lE) * ds, 'g', 7) : "?") << "\n";
			}
			Ci[n] = B;
		}
		S << "Re:\n";
		ds = 100 / Re;
		for (V = 0.9 * (B = Re), lE = 0.0; V <= 1.1 * B; V += 0.01 * B, lE = E)
		{
			Re = V;
			calcRIndP();
			getDerivatives(R, E, dBeta, dC, dRe, dDe);
			S << "V=" << QString::number(V, 'g', 3) << ", E=" << QString::number(E, 'g', 7)
			  << ", dE/dRe=" << QString::number(dRe, 'g', 7) << " ~ " 
			  << (lE != 0.0 ? QString::number((E - lE) * ds, 'g', 7) : "?") << "\n";
		}
		Re = B;
		S << "De:\n";
		ds = 100 / De;
		for (V = 0.9 * (B = De), lE = 0.0; V <= 1.1 * B; V += 0.01 * B, lE = E)
		{
			De = V;
			calcRIndP();
			getDerivatives(R, E, dBeta, dC, dRe, dDe);
			S << "V=" << QString::number(V, 'g', 3) << ", E=" << QString::number(E, 'g', 7)
			  << ", dE/dDe=" << QString::number(dDe, 'g', 7) << " ~ " 
			  << (lE != 0.0 ? QString::number((E - lE) * ds, 'g', 7) : "?") << "\n";
		}
		De = B;
	}
	calcRIndP();
}

bool ELevel::isGT(ELevel L)
{	
	if (Iso > L.Iso) return true;
	if (Iso < L.Iso) return false;
	if (J > L.J) return true;
	if (J < L.J) return false;
	if (v > L.v) return true;
	return false;
}

bool ELevel::isGT(int I, int cJ, int cv)
{	
	if (Iso > I) return true;
	if (Iso < I) return false;
	if (J > cJ) return true;
	if (J < cJ) return false;
	if (v > cv) return true;
	return false;
}

CoefficientDialog::CoefficientDialog(QWidget *parent, QStringList &Data) : QDialog(parent)
{
	QGridLayout *Layout = new QGridLayout(this);
	Add = new QPushButton("Add", this);
	Layout->addWidget(Add, 1, 1);
	connect(Add, SIGNAL(clicked()), this, SLOT(add()));
	Cancel = new QPushButton("Cancel", this);
	Layout->addWidget(Cancel, 4, 1);
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
	CList = new QListWidget(this);
	CList->addItems(Data);
	Layout->addWidget(CList, 1, 0, 2, 1);
	Edit = new QLineEdit(this);
	Layout->addWidget(Edit, 0, 1);
	Label = new QLabel("Name/R [A]:", this);
	Layout->addWidget(Label, 0, 0);
	OK = new QPushButton("OK", this);
	Layout->addWidget(OK, 4, 0);
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	Remove = new QPushButton("Remove", this);
	Layout->addWidget(Remove, 2, 1);
	connect(Remove, SIGNAL(clicked()), this, SLOT(Remove()));
	Layout->setRowMinimumHeight(3, 10);
}

void CoefficientDialog::add()
{
	QString C = Edit->text();
	if (!C.isEmpty()) CList->addItem(C);
}

QStringList CoefficientDialog::getResults()
{
	int n, N = CList->count();
	QStringList Result;
	for (n=0; n<N; n++) Result << CList->item(n)->text();
	return Result;
}

void CoefficientDialog::remove()
{
	int n = CList->currentRow();
	if (n<0) return;
	QListWidgetItem *CI = CList->takeItem(n);
	delete CI;
}

CMLRPotDialog::CMLRPotDialog(QStringList *nCVL, QWidget* parent): QDialog(parent)
{
	CVL = nCVL;
	QGridLayout *L = new QGridLayout(this);
	L->addWidget(new QLabel("p", this), 0, 0);
	L->addWidget(pE = new QLineEdit("5", this), 0, 1);
	L->addWidget(new QLabel("q", this), 0, 2);
	L->addWidget(qE = new QLineEdit("3", this), 0, 3);
	L->addWidget(new QLabel("R_ref [A]", this), 0, 4);
	L->addWidget(RrefE = new QLineEdit("5.5", this), 0, 5);
	L->addWidget(new QLabel("N beta coeff.:", this), 0, 6);
	L->addWidget(NCE = new QLineEdit("10", this), 0, 7);
	L->setRowMinimumHeight(1, 20);
	L->addWidget(new QLabel("Long range coefficients:", this), 2, 0, 1, 8);
	L->addWidget(LRC = new QListWidget(this), 3, 0, 3, 4);
	L->addWidget(addB = new QPushButton("<- add coefficient", this), 3, 4, 1, 2);
	L->addWidget(new QLabel("new LRC:", this), 3, 6);
	L->addWidget(LRCB = new QComboBox(this), 3, 7);
	LRCB->addItems(QStringList() << "C3" << "C6" << "C8" << "C10" << "C12");
	LRCB->setEditable(false);
	L->addWidget(delB = new QPushButton("<- delete coeff.", this), 4, 4, 1, 2);
	L->addWidget(setB = new QPushButton("<- set value", this), 5, 4, 1, 2);
	L->addWidget(new QLabel("Value [cm^-1/A^n]:", this), 5, 6);
	L->addWidget(CValE = new QLineEdit("1.525e7", this), 5, 7);
	L->setRowMinimumHeight(6, 20);
	L->addWidget(new QLabel("Fit range:", this), 7, 0, 1, 8);
	QGridLayout *L2 = new QGridLayout();
	L2->addWidget(new QLabel("R_min [A]:", this), 0, 0);
	L2->addWidget(mRE = new QLineEdit("3.0", this), 0, 1);
	L2->addWidget(new QLabel("R_max [A]:", this), 0, 2);
	L2->addWidget(MRE = new QLineEdit("30,0", this), 0, 3);
	L2->addWidget(new QLabel("Num points:", this), 0, 4);
	L2->addWidget(NPE = new QLineEdit("10000", this), 0, 5);
	L->addLayout(L2, 8, 0, 1, 8);
	L->setRowMinimumHeight(9, 20);
	QGridLayout *L3 = new QGridLayout();
	L3->addWidget(OK = new QPushButton("OK", this), 0, 0);
	L3->addWidget(Cancel = new QPushButton("Cancel", this), 0, 2);
	L->addLayout(L3, 10, 0, 1, 8);
	connect(addB, SIGNAL(clicked()), this, SLOT(addC()));
	connect(delB, SIGNAL(clicked()), this, SLOT(delC()));
	connect(setB, SIGNAL(clicked()), this, SLOT(setV()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
}

void CMLRPotDialog::addC()
{
	int n = LRCB->currentIndex();
	if ((*CVL)[n].isEmpty()) LRC->addItem(LRCB->currentText());
	else LRC->addItem(LRCB->currentText() + " = " + (*CVL)[n]);
}

void CMLRPotDialog::delC()
{
	if (LRC->currentRow() >= 0) delete LRC->takeItem(LRC->currentRow());
}

void CMLRPotDialog::setV()
{
	if (LRC->currentRow() < 0) return;
	QString T = LRC->currentItem()->text();
	int n = T.indexOf(" = ");
	if (n >= 0) T = T.left(n);
	if (CValE->text().isEmpty()) T += " = " + CValE->text();
	LRC->currentItem()->setText(T);
}


CoupledSineWaveFunc::CoupledSineWaveFunc(MainWindow *MW, Molecule *mol) 
					: MDIChild(SimpleDiagWindow, MW, "SineWaveFuncs (*.SWF)", ".SWF")
{
	Data = 0;
	Iso = SC = 0;
	R = 0;
	nChan = nCoeff = nI = nJ = nv = 0;
	States = 0;
	E = 0;
	MC = 0;
	Mol = mol;
	ASC = 0;
	Basis = 0;
	RMin = RMax = 0.0;
	nPoints = 0;
	vA = CA = 0;
	MapFuncs = 0;
}

CoupledSineWaveFunc::~CoupledSineWaveFunc()
{
	DestroyData();
	if (MapFuncs != 0) delete MapFuncs;
}

void CoupledSineWaveFunc::DestroyData()
{
	int I, J, v, C;
	if (Data != 0)
	{
		for (I=0; I < nI; I++)
		{
			for (J=0; J < nJ; J++)
			{
				for (v=0; v < nv; v++) if (Data[I][J][v] != 0)
				{
					for (C=0; C < nChan; C++)  delete[] Data[I][J][v][C];
					delete[] Data[I][J][v];
					delete[] MC[I][J][v];
				}
				delete[] Data[I][J];
				delete[] MC[I][J];
				delete[] E[I][J];
			}
			delete[] Data[I];
			delete[] MC[I];
			delete[] E[I];
		}
		delete[] Data;
		Data = 0;
		delete[] MC;
		delete[] E;
	}
	if (R != 0)
	{
		delete[] R;
		R=0;
	}
	if (Iso != 0)
	{
		delete[] Iso;
		Iso = 0;
	}
	if (SC != 0)
	{
		delete[] SC;
		delete[] States;
		SC = 0;
	}
	if (Basis != 0) 
	{
		Destroy(Basis, nPoints);
		Basis = 0;
	}
	if (vA != 0) 
	{
		Destroy(vA, nI, nJ);
		vA = 0;
	}
	if (CA != 0) 
	{
		Destroy(CA, nI, nJ);
		CA = 0;
	}
}

double**** CoupledSineWaveFunc::getEnergy()
{
	double ****rE = Create(1, nI, nv, nJ);
	int I, v, J;
	for (I=0; I < nI; I++) for (v=0; v < nv; v++) for (J=0; J < nJ; J++) rE[0][I][v][J] = E[I][J][v];
	return rE;
}

void CoupledSineWaveFunc::CalcvCA()
{
	int v, I, J, n, c[nChan], bc;
	double MMC;
	if (vA != 0) Destroy(vA, nI, nJ);
	if (CA != 0) Destroy(CA, nI, nJ);
	vA = CreateInt(nI, nJ, nv);
	CA = CreateInt(nI, nJ, nv);
	for (I=0; I < nI; I++) for (J=0; J < nJ; J++)
	{
		for (n=0; n < nChan; n++) c[n] = 0;
		for (v=0; v < nv; v++) if (Data[I][J][v] != 0)
		{
			for (n=1, MMC = MC[I][J][v][0], bc=0; n < nChan; n++) 
				if (MC[I][J][v][n] > MMC)
			{
				MMC = MC[I][J][v][n];
				bc = n;
			}
			vA[I][J][v] = c[bc]++;
			CA[I][J][v] = bc;
		}
	}
}

void CoupledSineWaveFunc::getIso(int& N, int*&I)
{
	N = nI;
	I = Iso;
}

int CoupledSineWaveFunc::getNChannels()
{
	return nChan;
}

int CoupledSineWaveFunc::getNLevels(int I, int J)
{
	if (I >= nI && J >= nJ) return 0;
	int n, r;
	for (n=r=0; r < nv; n++) if (Data[I][J][n] != 0) r++;
	return r;
}

void CoupledSineWaveFunc::getStateComp(int Channel, ElState*& State, int& Comp)
{
	State = States[Channel];
	Comp = SC[Channel];
}

void CoupledSineWaveFunc::getv(int I, int J, int& N, int*& v)
{
	if (I >= nI && J >= nJ)
	{
		N=0;
		v=0;
		return;
	}
	int n;
	for (n=N=0; n < nv; n++) if (Data[I][J][n] != 0) N++;
	v = new int[N];
	for (n=N=0; n < nv; n++) if (Data[I][J][n] != 0) v[N++] = n;
}

int CoupledSineWaveFunc::getv(int I, int J, double SE)
{
	int n;
	if (I<0 || I >= nI || J<0 || J >= nJ) return -1;
	for (n=0; n < nv; n++) if (fabs(SE - E[I][J][n]) < 1e-3) return n;
	return -1;
}

bool CoupledSineWaveFunc::isJA(int I, int J)
{
	int n;
	if (I<0 || I >= nI || J<0 || J >= nJ) return false;
	for (n=0; n < nv; n++) if (Data[I][J][n] != 0) return true;
	return false;
}

bool CoupledSineWaveFunc::getWaveFunc(int I, int J, int v, int Channel, double RStart, double RStop, int NPoints, double* WF, 
									  double& rE, int& MMC)
{
	if (I >= nI || J >= nJ || v >= nv) return false;
	if (Data[I][J][v] == 0) return false;
	int n, m, j;
	double M;
	double h = (RStop - RStart) / (NPoints - 1), r, x, F = 0.5 / double(nCoeff), F1 = 0.5 * M_PI / double(nCoeff);
	double F2 = double(2 * nCoeff - 1) * F1;
	//if (RStart != RMin || RStop != RMax || NPoints != nPoints) CalcBasis(RStart, RStop, NPoints); 
	for (n=0; n < NPoints; n++) WF[n] = 0.0;
	for (n=0, r = RStart; r < R[0] && n < NPoints; r+=h, n++) ;
	for (m=0; r <= R[nCoeff - 1] && n < NPoints; r+=h, n++)
	{
		while (R[m] < r) m++;
		x = MapFuncs->getPoint(r);
		for (j=0; j < nCoeff; j++) 
			WF[n] += (sin(F2 * (x - double(j))) / sin(F1 * (x - double(j)))
				   - sin(F2 * (x + double(j))) / sin(F1 * (x + double(j)))) * Data[I][J][v][Channel][j];
		WF[n] *= F * sqrt(MapFuncs->getDerivative(r));
	}
	rE = E[I][J][v];
	for (MMC = 0, n=1, M = MC[I][J][v][0]; n < nChan; n++) if (MC[I][J][v][n] > M)
	{
		MMC = n;
		M = MC[I][J][v][n];
	}
	return true;
}

void CoupledSineWaveFunc::CalcBasis(double RStart, double RStop, int NPoints)
{
	double h = (RStop - RStart) / (NPoints - 1), r, x, F = 0.5 / double(nCoeff), F1 = 0.5 * M_PI / double(nCoeff);
	double F2 = double(2 * nCoeff - 1) * F1;
	int n, m, j;
	if (Basis != 0) Destroy(Basis, nPoints);
	Basis = Create(nPoints = NPoints, nCoeff);
	for (n=0; n < NPoints; n++) for (j=0; j < nCoeff; j++) Basis[n][j] = 0.0;
	for (n=0, r = RMin = RStart, RMax = RStop; r < R[0] && n < NPoints; r+=h, n++) ;
	for (m=0; r <= R[nCoeff - 1] && n < NPoints; r+=h, n++)
	{
		while (R[m] < r) m++;
		x = MapFuncs->getPoint(r);
		for (j=0; j < nCoeff; j++) 
			Basis[n][j] = (sin(F2 * (x - double(j))) / sin(F1 * (x - double(j)))
							- sin(F2 * (x + double(j))) / sin(F1 * (x + double(j)))) * F * sqrt(MapFuncs->getDerivative(r));
	}
}

bool CoupledSineWaveFunc::getWaveFuncs(int I, int v, int *Channels, int NChannels, int NJ, int *J, 
									   double RStart, double RStop, int NPoints, 
									   double*** WF, double* rE, int *va, int *ca)
{
	if (I >= nI || v >= nv || J[NJ-1] > nJ) return false;
	//if (RStart != RMin || RStop != RMax || NPoints != nPoints) CalcBasis(RStart, RStop, NPoints);
	int n, m, j, i, c;
	double M;
	double h = (RStop - RStart) / (NPoints - 1), r, x, F = 0.5 / double(nCoeff), F1 = 0.5 * M_PI / double(nCoeff);
	double F2 = double(2 * nCoeff - 1) * F1, TF;
	if (vA == 0 || CA == 0) CalcvCA();
	for (n=0; n < NPoints; n++) for (i=0; i < NJ; i++) for (c=0; c < NChannels; c++) WF[i][c][n] = 0.0;
	for (n=0, r = RStart; r < R[0] && n < NPoints; r+=h, n++) ;
	for (m=0; r <= R[nCoeff - 1] && n < NPoints; r+=h, n++)
	{
		while (R[m] < r) m++;
		x = MapFuncs->getPoint(r);
		for (j=0; j < nCoeff; j++)
		{
			M = sin(F2 * (x - double(j))) / sin(F1 * (x - double(j)))
							- sin(F2 * (x + double(j))) / sin(F1 * (x + double(j)));
			for (i=0; i < NJ; i++) for (c=0; c < NChannels; c++) 
				if (Data[I][J[i]][v] != 0) WF[i][c][n] += Data[I][J[i]][v][Channels[c]][j] * M;
		}
		TF = F * sqrt(MapFuncs->getDerivative(r));
		for (i=0; i < NJ; i++) for (c=0; c < NChannels; c++) WF[i][c][n] *= TF;
	}
	for (i=0; i < NJ; i++) 
	{
		rE[i] = E[I][J[i]][v];
		va[i] = vA[I][J[i]][v];
		ca[i] = CA[I][J[i]][v];
	}
	return true;
}

bool CoupledSineWaveFunc::getWaveFuncs(int I, int J, int Channel, double RStart, double RStop, int NPoints, double** WF, 
									   double* rE)
{
	if (I >= nI || J >= nJ) return false;
	int n, j, m, Nv, v, c;
	double M, MS;
	double h = (RStop - RStart) / (NPoints - 1), r, x, F = 0.5 / double(nCoeff), F1 = 0.5 * M_PI / double(nCoeff);
	double F2 = double(2 * nCoeff - 1) * F1;
	//if (RStart != RMin || RStop != RMax || NPoints != nPoints) CalcBasis(RStart, RStop, NPoints);
	for (v = Nv = 0; v < nv; v++) if (Data[I][J][v] != 0) Nv = v + 1;
	if (Nv == 0) return false;
	double Sq[Nv], **WFt = Create(Nv, nChan);
	for (v=0; v < Nv; v++) Sq[v] = 0.0;
	for (n=0, r = RStart; r < R[0] && n < NPoints; r+=h, n++) ;
	for (m=0; r <= R[nCoeff - 1] && n < NPoints; r+=h, n++)
	{
		for (v=0; v < Nv; v++) for (c=0; c < nChan; c++) WFt[v][c] = 0.0;
		while (R[m] < r) m++;
		x = MapFuncs->getPoint(r);
		for (j=0; j < nCoeff; j++) for (c=0; c < nChan; c++)
		{
			M = sin(F2 * (x - double(j))) / sin(F1 * (x - double(j)))
							- sin(F2 * (x + double(j))) / sin(F1 * (x + double(j)));
			for (v=0; v < Nv; v++) if (Data[I][J][v] != 0) 
				WFt[v][c] += Data[I][J][v][c][j] * M;
		}
		MS = sqrt(MapFuncs->getDerivative(r));
		for (v=0; v < Nv; v++) 
		{
			for (c=0; c < nChan; c++) 
			{
				WFt[v][c] *= F * MS;
				Sq[v] += WFt[v][c] * WFt[v][c];
			}
			WF[v][n] = WFt[v][Channel];
		}
	}
	for (v=0; v < Nv; v++) 
	{
		rE[v] = E[I][J][v];
		for (Sq[v] = 1.0 / sqrt(Sq[v]), n=0; n < NPoints; n++) WF[v][n] *= Sq[v];
	}
	Destroy(WFt, Nv);
	return true;
}

bool CoupledSineWaveFunc::getWaveFuncs(int I, int J, double RStart, double RStop, int NPoints, double*** WF, double* rE)
{
	if (I >= nI || J >= nJ) return false;
	//if (RStart != RMin || RStop != RMax || NPoints != nPoints) CalcBasis(RStart, RStop, NPoints);
	int n, m, j, Nv, v, Channel;
	double M;
	double h = (RStop - RStart) / (NPoints - 1), r, x, F = 0.5 / double(nCoeff), F1 = 0.5 * M_PI / double(nCoeff);
	double F2 = double(2 * nCoeff - 1) * F1;
	for (v = Nv = 0; v < nv; v++) if (Data[I][J][v] != 0) Nv = v + 1;
	if (Nv == 0) return false;
	for (n=0; n < NPoints; n++) for (v=0; v < Nv; v++) for (Channel = 0; Channel < nChan; Channel++)
		WF[Channel][v][n] = 0.0;
	for (n=0, r = RStart; r < R[0] && n < NPoints; r+=h, n++) ;
	for (m=0; r <= R[nCoeff - 1] && n < NPoints; r+=h, n++)
	{
		while (R[m] < r) m++;
		x = MapFuncs->getPoint(r);
		for (j=0; j < nCoeff; j++) 
		{
			M = sin(F2 * (x - double(j))) / sin(F1 * (x - double(j)))
							- sin(F2 * (x + double(j))) / sin(F1 * (x + double(j)));
			for (v=0; v < Nv; v++) if (Data[I][J][v] != 0) for (Channel = 0; Channel < nChan; Channel++) 
				WF[Channel][v][n] += Data[I][J][v][Channel][j] * M;
		}
		for (Channel = 0; Channel < nChan; Channel++) for (v=0; v < Nv; v++) 
			WF[Channel][v][n] *= F * sqrt(MapFuncs->getDerivative(r));
	}
	for (v=0; v < Nv; v++) rE[v] = E[I][J][v];
	return true;
}

bool CoupledSineWaveFunc::readData(QString FN)
{
    if (Mol == 0) 
	{
		printf("CoupledSineWaveFunc::readData error: Can't read any data without being assinged to a molecule!");
		return false;
	}
	int n, m, N, I, J, v, c;
	SplinePoint *points;
	QFile Datei(FN);
	if (!read(&Datei)) return false;
	QTextStream S(&Datei);
	QStringList L = S.readLine().split(' ', QString::SkipEmptyParts);
	if (L.count() > 0 ? L[0].indexOf("name:", 0, Qt::CaseInsensitive) == -1 : true) return false;
	if (L.count() >= 2) setName(L[1]);
	if ((L = S.readLine().split(' ', QString::SkipEmptyParts)).count() > 0 ? 
		L[0].indexOf("source:", 0, Qt::CaseInsensitive) == -1 : true) return false;
	if (L.count() >= 2) Source = L[1];
	DestroyData();
	if ((nChan = (L = S.readLine().split(' ', QString::SkipEmptyParts)).count() / 2) == 0) return false;
	for (n=0, SC = new int[nChan], States = new ElState*[nChan]; n < nChan; n++)
	{
		States[n] = Mol->getState(L[2*n]);
		SC[n] = L[2*n+1].toInt();
	}
	if ((nI = (L = S.readLine().split(' ', QString::SkipEmptyParts)).count()) == 0) return false;
	for (n=0, Iso = new int[nI]; n < nI; n++) Iso[n] = L[n].toInt();
	if ((nCoeff = (L = S.readLine().split(' ', QString::SkipEmptyParts)).count()) == 0) return false;
	for (n=0, R = new double[nCoeff], points = new SplinePoint[nCoeff]; n < nCoeff; n++) 
	{
		points[n].y = double(n);
		points[n].x = R[n] = L[n].toDouble();
	}
	if (MapFuncs == 0) MapFuncs = new NaturalSpline(points, nCoeff);
	else MapFuncs->setPoints(points, nCoeff);
	if ((L = S.readLine().split(' ', QString::SkipEmptyParts)).count() < 3) return false;
	if ((nJ = L[0].toInt()) < 0 || (nv = L[1].toInt()) < 0) return false;
	for (I=0, Data = new double****[nI], MC = new double***[nI], E = new double**[nI]; I < nI; I++)
		for (J=0, Data[I] = new double***[nJ], MC[I] = new double**[nJ], E[I] = new double*[nJ]; J < nJ; J++)
			for (v=0, Data[I][J] = new double**[nv], MC[I][J] = new double*[nv], E[I][J] = new double[nv]; v < nv; v++)
				Data[I][J][v] = 0;
	for (n=0, N = L[2].toInt(); n<N; n++)
	{
		if (S.atEnd()) break;
		if ((L = S.readLine().split(' ', QString::SkipEmptyParts)).count() < 4) break;
		if ((I = L[0].toInt()) >= nI || (J = L[1].toInt()) >= nJ || (v = L[2].toInt()) >= nv) break;
		if (I < 0 || J < 0 || v < 0) break;
		E[I][J][v] = L[3].toDouble();
		if ((L = S.readLine().split(' ', QString::SkipEmptyParts)).count() < nChan) break;
		for (c=0, MC[I][J][v] = new double[nChan], Data[I][J][v] = new double*[nChan]; c < nChan; c++)
		{
			MC[I][J][v][c] = L[c].toDouble();
			Data[I][J][v][c] = new double[nCoeff];
		}
		for (m=0; m < nCoeff; m++)
		{
			if ((L = S.readLine().split(' ', QString::SkipEmptyParts)).count() < nChan) break;
			for (c=0; c < nChan; c++) Data[I][J][v][c][m] = L[c].toDouble();
		}
		if (m < nCoeff) break;
	}
	if (n<N) return false;
	Saved();
	return true;
}

void CoupledSineWaveFunc::setData(int NI, int NJ, int Nv, int NChan, int NCoeff, ElState **St, int *C, int* I, double* RS, 
								  double***** data, double ****MCo, double ***En)
{
	DestroyData();
	int n;
	SplinePoint *points = new SplinePoint[NCoeff];
	for (n=0; n < NCoeff; n++)
	{
		points[n].y = double(n);
		points[n].x = RS[n];
	}
	if (MapFuncs == 0) MapFuncs = new NaturalSpline(points, NCoeff);
	else MapFuncs->setPoints(points, NCoeff);
	nI = NI;
	nJ = NJ;
	nv = Nv;
	nChan = NChan;
	nCoeff = NCoeff;
	States = St;
	SC = C;
	Iso = I;
	R = RS;
	Data = data;
	MC = MCo;
	E = En;
	Changed();
	setImported();
}

bool CoupledSineWaveFunc::writeData(QString FN)
{
    int I, J, v, c, n;
	QFile Datei(FN);
	if (!write(&Datei)) return false;
	QTextStream S(&Datei);
	S << "Name: " << getName() << "\n";
	S << "Source: " << Source << "\n";
	for (n=0; n < nChan; n++) S << States[n]->getName() << " " << QString::number(SC[n]) << (n < nChan - 1 ? " " : "\n");
	for (n=0; n < nI; n++) S << QString::number(Iso[n]) << (n < nI - 1 ? " " : "\n");
	for (n=0; n < nCoeff; n++) S << QString::number(R[n], 'f', 7) << (n < nCoeff - 1 ? " " : "\n");
	for (I=n=0; I < nI; I++) for (J=0; J < nJ; J++) for (v=0; v < nv; v++) if (Data[I][J][v] != 0) n++;
	S << QString::number(nJ) << " " << QString::number(nv) << " " << QString::number(n) << "\n";
	for (I=0; I < nI; I++) for (J=0; J < nJ; J++) for (v=0; v < nv; v++) if (Data[I][J][v] != 0)
	{
		S << QString::number(I) << " " << QString::number(J) << " " << QString::number(v) << " " 
			<< QString::number(E[I][J][v], 'f', 4) << "\n";
		for (c=0; c < nChan; c++) S << QString::number(MC[I][J][v][c], 'f', 5) << (c+1 < nChan ? " " : "\n");
		for (n=0; n < nCoeff; n++) for (c=0; c < nChan; c++) 
			S << QString::number(Data[I][J][v][c][n], 'E', 7) << (c+1 < nChan ? " " : "\n");
	}
	Saved();
	return true;
}

CSWFImportDialog::CSWFImportDialog(MainWindow* MW, QString InitDir): QDialog(MW)
{
	int n, N = MW->getNumMolecules();
	mw = MW;
	ST = CT = 0;
	setWindowTitle("MolSpektAnalysis");
	QGridLayout *L = new QGridLayout(this);
	L->addWidget(new QLabel("Import wave functions:", this), 0, 0, 1, 5);
	L->setRowMinimumHeight(1, 20);
	L->addWidget(new QLabel("Molecule:", this), 2, 0);
	L->addWidget(MolBox = new QComboBox(this), 2, 1);
	L->addWidget(new QLabel("Potential file:", this), 4, 0);
	L->addWidget(PotFile = new QLineEdit(InitDir, this), 4, 1);
	L->addWidget(SelectPotFile = new QPushButton("...", this), 4, 2);
	L->addWidget(new QLabel("File T5_E_Bv:", this), 5, 0);
	L->addWidget(T5_E_Bv = new QLineEdit(InitDir, this), 5, 1);
	L->addWidget(SelectT5_E_Bv = new QPushButton("...", this), 5, 2);
	L->addWidget(new QLabel("Electronic states:", this), 2, 3, 1, 2);
	L->addWidget(SList = new QListWidget(this), 3, 3, 3, 1);
	L->addWidget(SBox = new QComboBox(this), 3, 4);
	L->addWidget(AddState = new QPushButton("<- Add", this), 4, 4);
	L->addWidget(RemoveState = new QPushButton("<- Remove", this), 5, 4);
	L->setRowMinimumHeight(6, 20);
	L->addWidget(new QLabel("Data directories:", this), 7, 0, 1, 3);
	L->addWidget(new QLabel("Isotopologue:", this), 7, 3);
	L->addWidget(IsoBox = new QComboBox(this), 7, 4);
	L->addWidget(new QLabel("Directory for isotopologue:", this), 8, 0, 1, 3);
	L->addWidget(IsoDir = new QLineEdit(InitDir, this), 8, 3);
	L->addWidget(SelectIsoDir = new QPushButton("...", this), 8, 4);
	L->addWidget(IsoList = new QListWidget(this), 9, 0, 3, 4);
	L->addWidget(AddIso = new QPushButton("<- Add", this), 9, 4);
	L->setRowMinimumHeight(10, 20);
	L->addWidget(RemoveIso = new QPushButton("<- Remove", this), 11, 4);
	L->setRowMinimumHeight(12, 20);
	L->addWidget(OK = new QPushButton("OK", this), 13, 0, 1, 2);
	L->addWidget(Cancel = new QPushButton("Cancel", this), 13, 3, 1, 2);
	MolBox->setEditable(false);
	IsoBox->setEditable(false);
	SBox->setEditable(false);
	for (n=0; n<N; n++) MolBox->addItem(MW->getMolecule(n)->getName());
	connect(MolBox, SIGNAL(currentIndexChanged(int)), this, SLOT(molChanged(int)));
	connect(SelectPotFile, SIGNAL(clicked()), this, SLOT(selectPotFile()));
	connect(SelectT5_E_Bv, SIGNAL(clicked()), this, SLOT(selectT5_E_Bv()));
	connect(SelectIsoDir, SIGNAL(clicked()), this, SLOT(selectIsoDir()));
	connect(AddIso, SIGNAL(clicked()), this, SLOT(addIso()));
	connect(AddState, SIGNAL(clicked()), this, SLOT(addState()));
	connect(RemoveIso, SIGNAL(clicked()), this, SLOT(removeIso()));
	connect(RemoveState, SIGNAL(clicked()), this, SLOT(removeState()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
	molChanged(0);
}

void CSWFImportDialog::addIso()
{
	QString Iso = IsoBox->currentText(), File = IsoDir->text();
	QString LI = (IsoList->count() > 0 ? IsoList->item(IsoList->count() - 1)->text() : "");
	QDir Dir(File);
	int n;
	if (!Iso.isEmpty() && ((n = LI.indexOf(' ')) > 0 ? File != LI.right(LI.length() - n - 1) : true) && Dir.exists())
	{
		n = IsoBox->currentIndex();
		IsoList->addItem(Iso + ' ' + File);
		IsoBox->removeItem(n);
		IL2.append(IL1.takeAt(n));
	}
	else QMessageBox::information(this, "MolSpektAnalysis", 
		"Error: You can add an isotopologue only if a valid isotopologue is choosen and a valid directory, which has to be different compared to the ones already choosen for other isotopologues!");
}

void CSWFImportDialog::addState()
{
	int n;
	if (!SBox->currentText().isEmpty())
	{
		SList->addItem(SBox->currentText());
		SBox->removeItem(n = SBox->currentIndex());
		SL2.append(SL1.takeAt(n));
	}
}

void CSWFImportDialog::getData(Molecule*& Mol, ElState**& States, int *&Component, int& NStates, int*& Iso, QString*& IsoDirs, 
							   int& NIso, QString &PotentialFile, QString &T5_E_BvFile)
{
	int n, m;
	QString IsoText;
	Mol = mol;
	States = new ElState*[NStates = SL2.count()];
	Component = new int[NStates];
	for (n=0; n < NStates; n++) 
	{
		States[n] = Mol->getStateP(ST[SL2[n]]);
		Component[n] = CT[SL2[n]];
	}
	Iso = new int[NIso = IL2.count()];
	IsoDirs = new QString[NIso];
	for (n=0; n < NIso; n++)
	{
		Iso[n] = IL2[n];
		IsoText = IsoList->item(n)->text();
		m = IsoText.indexOf(' ');
		IsoDirs[n] = IsoText.right(IsoText.length() - m - 1);
	}
	PotentialFile = PotFile->text();
	T5_E_BvFile = T5_E_Bv->text();
}

void CSWFImportDialog::molChanged(int i)
{
	int n, m, N, NC, M;
	IsoBox->clear();
	IsoList->clear();
	IL1.clear();
	IL2.clear();
	SL1.clear();
	SL2.clear();
	SList->clear();
	SBox->clear();
	mol = mw->getMolecule(i);
	if (mol == 0) return;
	IsoTab *Iso = mol->getIso();
	for (n = NC = 0, N = mol->getNumStates(); n < N; n++) 
		NC += (int(2.0 * mol->getStateP(n)->getS()) + 1) * (mol->getStateP(n)->getLambda() == 0 ? 1 : 2);
	if (ST != 0)
	{
		delete[] ST;
		delete[] CT;
	}
	ST = new int[NC];
	CT = new int[NC];
	for (n = NC = 0; n<N; n++)
	{
		M = int(2.0 * mol->getStateP(n)->getS()) + 1;
		if (M==1)
		{
			if (mol->getStateP(n)->getLambda() == 0)
			{
				ST[NC] = n;
				CT[NC] = 0;
				SL1.append(NC++);
				SBox->addItem(mol->getState(n));
			}
			else
			{
				ST[NC] = n;
				CT[NC] = 0;
				SL1.append(NC++);
				SBox->addItem(mol->getState(n) + " e");
				ST[NC] = n;
				CT[NC] = 1;
				SL1.append(NC++);
				SBox->addItem(mol->getState(n) + " f");
			}
		}
		else for (m=0; m<M; m++)
		{
			if (mol->getStateP(n)->getLambda() == 0)
			{
				ST[NC] = n;
				CT[NC] = m;
				SL1.append(NC++);
				SBox->addItem(mol->getState(n) + ' ' + QString::number(m));
			}
			else
			{
				ST[NC] = n;
				CT[NC] = 2*m;
				SL1.append(NC++);
				SBox->addItem(mol->getState(n) + ' ' + QString::number(m) + 'e');
				ST[NC] = n;
				CT[NC] = 2*m+1;
				SL1.append(NC++);
				SBox->addItem(mol->getState(n) + ' ' + QString::number(m) + 'f');
			}
		}
	}
	for (n=0; n < Iso->numIso; n++)
	{
		IL1.append(n);
		IsoBox->addItem(Iso->getIsoName(n));
	}
	delete Iso;
}

void CSWFImportDialog::removeIso()
{
	QString Text = IsoList->currentItem()->text();
	int n = IsoList->currentRow();
	if (Text.isEmpty()) return;
	IL1.append(IL2.takeAt(n));
	delete IsoList->takeItem(n);
	IsoBox->addItem(Text.left(Text.indexOf(' ')));
}

void CSWFImportDialog::removeState()
{
	int n = SList->currentRow();
	if (n < 0) return;
	QListWidgetItem *Item = SList->takeItem(n);
	SBox->addItem(Item->text());
	delete Item;
	SL1.append(SL2.takeAt(n));
}

void CSWFImportDialog::selectIsoDir()
{
	QString Dir = QFileDialog::getExistingDirectory(this, "Select directory with wave functions for the isotopologue " 
													+ IsoBox->currentText(), IsoDir->text());
	if (!Dir.isEmpty()) IsoDir->setText(Dir);
}

void CSWFImportDialog::selectPotFile()
{
	QString File = QFileDialog::getOpenFileName(this, "Please select the potential file", PotFile->text(), "Potential files (*)");
	if (!File.isEmpty()) PotFile->setText(File);
}

void CSWFImportDialog::selectT5_E_Bv()
{
	QString File = QFileDialog::getOpenFileName(this, "Please select the file T5_E_Bv", T5_E_Bv->text(), "T5_E_Bv (T5_E_Bv)");
	if (!File.isEmpty()) T5_E_Bv->setText(File);
}


MonteCarloSimDialog::MonteCarloSimDialog(MainWindow* MW) : QDialog(MW)
{
	QGridLayout *L = new QGridLayout(this);
	QPushButton *FDB = new QPushButton("...", this), *OK = new QPushButton("OK", this), *Cancel = new QPushButton("Cancel", this);
	QDoubleValidator *Val;
	setWindowTitle("Monte Carlo Simulation");
	L->addWidget(new QLabel("Number of iterations:", this), 0, 0);
	L->addWidget(NItE = new QLineEdit("1000", this), 0, 1, 1, 2);
	L->addWidget(new QLabel("Number of parallel fits:", this), 1, 0);
	L->addWidget(NParFitsE = new QLineEdit("8", this), 1, 1, 1, 2);
	L->addWidget(new QLabel("Result directory:", this), 2, 0);
	L->addWidget(DirectoryE = new QLineEdit(MW->getDir(), this), 2, 1);
	L->addWidget(FDB, 2, 2);
	L->addWidget(new QLabel("Uncertainty factor:", this), 3, 0);
	L->addWidget(UncFactE = new QLineEdit("1.0", this), 3, 1, 1, 2);
	L->setRowMinimumHeight(4, 20);
	L->addWidget(OK, 5, 0);
	L->addWidget(Cancel, 5, 1, 1, 2);
	NItE->setValidator(new QIntValidator(1, 10000, NItE));
	NParFitsE->setValidator(new QIntValidator(1, 100, NParFitsE));
	UncFactE->setValidator(Val = new QDoubleValidator(UncFactE));
	Val->setBottom(1e-300);
	connect(FDB, SIGNAL(clicked()), this, SLOT(showFileDialog()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
}

bool MonteCarloSimDialog::exec(QString& Directory, int& NIterations, int& NParFit, double &UncFact)
{
	if (QDialog::exec() == QDialog::Rejected) return false;
	QFileInfo FI(Directory = DirectoryE->text());
	if (!FI.isDir())
	{
		QMessageBox::information(this, "MolSpektAnalysis", "Error: " + Directory + " is not an existing directory!");
		return exec(Directory, NIterations, NParFit, UncFact);
	}
	if (Directory.right(1) != "\\" && Directory.right(1) != "/") Directory += DIRSEP;
	NParFit = NParFitsE->text().toInt();
	NIterations = NItE->text().toInt();
	UncFact = UncFactE->text().toDouble();
	return true;
}

void MonteCarloSimDialog::showFileDialog()
{
	QString Dir = QFileDialog::getExistingDirectory(this, "Select directory for results", DirectoryE->text());
	if (!Dir.isEmpty()) DirectoryE->setText(Dir);
}


CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::CreateAnaPotSeriesFromMCSSplinePotSeriesDialog(MainWindow* parent)
												: QDialog(parent), MW(parent)
{
	int n;
	QGridLayout *L = new QGridLayout(this);
	QPushButton *OK = new QPushButton("OK", this), *Cancel = new QPushButton("Cancel", this);
	QPushButton *FDDB = new QPushButton("...", this), *APDB = new QPushButton("...", this);
	setWindowTitle("Create AnaPot series from MCS SplinePot series");
	L->addWidget(improve = new QCheckBox("Improve existing analytical potentials", this), 0, 0, 1, 3);
	L->addWidget(new QLabel("Dir with SplinePots:", this), 1, 0);
	L->addWidget(SplinePotDirE = new QLineEdit(MW->getDir(MDIChild::PotData), this), 1, 1);
	L->addWidget(SPDB = new QPushButton("...", this), 1, 2);
	L->addWidget(new QLabel("Dir with FitData sets:", this), 2, 0);
	L->addWidget(FDDirE = new QLineEdit(MW->getDir(MDIChild::FitDataSet), this), 2, 1);
	L->addWidget(FDDB, 2, 2);
	L->addWidget(new QLabel("Dir for AnaPots:", this), 3, 0);
	L->addWidget(AnaPotDirE = new QLineEdit(MW->getDir(MDIChild::PotData), this), 3, 1);
	L->addWidget(APDB, 3, 2);
	L->addWidget(new QLabel("Molecule:", this), 4, 0);
	L->addWidget(MoleculeBox = new QComboBox(this), 4, 1, 1, 2);
	L->addWidget(new QLabel("Electronic state:", this), 5, 0);
	L->addWidget(StateBox = new QComboBox(this), 5, 1, 1, 2);
	L->addWidget(new QLabel("Fit precision:", this), 6, 0);
	L->addWidget(PrecE = new QLineEdit("1.0", this), 6, 1, 1, 2);
	L->addWidget(new QLabel("Max N fit iterations:", this), 7, 0);
	L->addWidget(MaxIt = new QComboBox(this), 7, 1, 1, 2);
	L->addWidget(useSVD = new QCheckBox("Use SVD", this), 8, 0);
	L->addWidget(useLM = new QCheckBox("Use Levenb.Mrq.", this), 8, 1, 1, 2);
	L->setRowMinimumHeight(9, 20);
	L->addWidget(OK, 10, 0);
	L->addWidget(Cancel, 10, 1, 1, 2);
	for (n=0; n < MW->getNumMolecules(); n++) MoleculeBox->addItem(MW->getMolecule(n)->getName());
	MoleculeBox->setEditable(false);
	StateBox->setEditable(false);
	PrecE->setValidator(new QDoubleValidator(1e-10, 1e10, 10, PrecE));
	improve->setChecked(false);
	MaxIt->addItem("No limit");
	for (n=1; n<=10; n++) MaxIt->addItem(QString::number(n));
	useLM->setChecked(true);
	useSVD->setChecked(true);
	connect(improve, SIGNAL(toggled(bool)), this, SLOT(improveChanged(bool)));
	connect(MaxIt, SIGNAL(editTextChanged(QString)), this, SLOT(MaxItChanged(QString)));
	connect(SPDB, SIGNAL(clicked()), this, SLOT(showSplinePotDirDialog()));
	connect(FDDB, SIGNAL(clicked()), this, SLOT(showFDDirDialog()));
	connect(APDB, SIGNAL(clicked()), this, SLOT(showAnaPotDirDialog()));
	connect(MoleculeBox, SIGNAL(currentIndexChanged(int)), this, SLOT(molChanged(int)));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
	molChanged(0);
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::getData(QString& MolFN, QString& StateN, QString& FDDir,
			QString& SPDir, QString& APDir, bool& improveAnaPots, bool& UseSvd, bool& UseLeveMarq, int& MaxIter, 
			double& Prec)
{
	MolFN = MW->getMolecule(MoleculeBox->currentIndex())->getFileName();
	StateN = StateBox->currentText();
	FDDir = FDDirE->text();
	SPDir = SplinePotDirE->text();
	APDir = AnaPotDirE->text();
	improveAnaPots = improve->isChecked();
	UseSvd = useSVD->isChecked();
	UseLeveMarq = useLM->isChecked();
	MaxIter = MaxIt->currentText().toInt();
	Prec = PrecE->text().toDouble();
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::improveChanged(bool checked)
{
	SplinePotDirE->setEnabled(!checked);
	SPDB->setEnabled(!checked);
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::MaxItChanged(const QString& text)
{
	if (text.toInt() <= 0) MaxIt->setCurrentIndex(0);
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::molChanged(int i)
{
	int n;
	Molecule *mol = MW->getMolecule(i);
	StateBox->clear();
	for (n=0; n < mol->getNumStates(); n++) StateBox->addItem(mol->getState(n));
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::showAnaPotDirDialog()
{
	QString Dir = QFileDialog::getExistingDirectory(this, "Select directory for AnaPots", AnaPotDirE->text());
	if (!Dir.isEmpty()) AnaPotDirE->setText(Dir);
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::showFDDirDialog()
{
	QString Dir = QFileDialog::getExistingDirectory(this, "Select directory with FitData sets", FDDirE->text());
	if (!Dir.isEmpty()) FDDirE->setText(Dir);
}

void CreateAnaPotSeriesFromMCSSplinePotSeriesDialog::showSplinePotDirDialog()
{
	QString Dir = QFileDialog::getExistingDirectory(this, "Select directory with SplinePots", SplinePotDirE->text());
	if (!Dir.isEmpty()) SplinePotDirE->setText(Dir);
}


FitAnaPotDialog::FitAnaPotDialog(bool improve, double Ri, double Ra, double iExp, int NC, int NLRC, int* pLRC, double hCi_cP, 
								 PotentialType CurrentPotential, QWidget* parent) : QDialog(parent)
{
	int n;
	QPushButton *OK = new QPushButton("OK", this), *Cancel = new QPushButton("Cancel", this);
	QGridLayout *L = new QGridLayout(this);
	setWindowTitle("Fit potential");
	L->addWidget(new QLabel("Fit potential of type:", this), 0, 0);
	L->addWidget(TypeBox = new QComboBox(this), 0, 1);
	TypeBox->addItems(QStringList() << "spline potential" << "analytical potential");
	TypeBox->setEditable(false);
	switch (CurPotential = CurrentPotential)
	{
		case SplinePotential:
			TypeBox->setCurrentIndex(0);
			break;
		case analyticalPotential:
		default:
			TypeBox->setCurrentIndex(1);
			break;
	}
	L->addWidget(Improve = new QCheckBox("Fit current potential", this), 1, 0, 1, 2);
	Improve->setChecked(improve);
	Improve->setEnabled(improve);
	L->addWidget(UseSVD = new QCheckBox("Use Singular Value Decomposition", this), 2, 0, 1, 2);
	UseSVD->setChecked(true);
	L->addWidget(UseLeveMrq = new QCheckBox("Use Levenberg-Marquardt", this), 3, 0, 1, 2);
	UseLeveMrq->setChecked(true);
	L->addWidget(AdjustTEWF = new QCheckBox("Adjust TE weight factors", this), 4, 0, 1, 2);
	AdjustTEWF->setChecked(false);
	L->addWidget(new QLabel("Ri:", this), 5, 0);
	L->addWidget(RiE = new QLineEdit(QString::number(Ri, 'f', 3), this), 5, 1);
	L->addWidget(new QLabel("Ra:", this), 6, 0);
	L->addWidget(RaE = new QLineEdit(QString::number(Ra, 'f', 1), this), 6, 1);
	L->addWidget(new QLabel("iExp:", this), 7, 0);
	L->addWidget(iExpE = new QLineEdit(QString::number(iExp, 'g', 12), this), 7, 1);
	iExpE->setValidator(new QDoubleValidator(2.0, 20.0, 12, iExpE));
	L->addWidget(new QLabel("Number of coefficients:", this), 8, 0);
	L->addWidget(NCE = new QLineEdit(QString::number(NC >= 4 ? NC : 15), this), 8, 1);
	NCE->setValidator(new QIntValidator(4, 40, NCE));
	L->setRowMinimumHeight(9, 20);
	L->addWidget(new QLabel("Long range coefficients:", this), 10, 0, 1, 2);
	L->addWidget(CoeffList = new QListWidget(this), 11, 0, 3, 1);
	for (n=0; n < NLRC; n++) CoeffList->addItem('C' + QString::number(pLRC[n]));
	L->addWidget(CoeffBox = new QComboBox(this), 11, 1);
	CoeffBox->addItems(QStringList() << "C3" << "C6" << "C8" << "C10" << "C12");
	CoeffBox->setEditable(true);
	L->addWidget(Add = new QPushButton("<- Add", this), 12, 1);
	L->addWidget(Remove = new QPushButton("Remove", this), 13, 1);
	L->addWidget(new QLabel("high Ci rel. prec.:", this), 14, 0);
	L->addWidget(hCi_cPE = new QLineEdit(QString::number(hCi_cP, 'g', 4), this), 14, 1);
	if (hCi_cP < 0.0) hCi_cPE->setEnabled(false);
	L->setRowMinimumHeight(15, 20);
	L->addWidget(new QLabel("Fit range:", this), 16, 0);
	L->addWidget(new QLabel("R min:", this), 17, 0);
	L->addWidget(RminE = new QLineEdit(QString::number(Ri - 1.0, 'f', 3), this), 17, 1);
	L->addWidget(new QLabel("R max:", this), 18, 0);
	L->addWidget(RmaxE = new QLineEdit("50.0", this), 18, 1);
	L->addWidget(new QLabel("Number of points:", this), 19, 0);
	L->addWidget(NumPE = new QLineEdit("10000", this), 19, 1);
	L->addWidget(new QLabel("b min:", this), 20, 0);
	L->addWidget(bminE = new QLineEdit("-0.1", this), 20, 1);
	L->addWidget(new QLabel("b max:", this), 21, 0);
	L->addWidget(bmaxE = new QLineEdit("-0.59", this), 21, 1);
	L->addWidget(new QLabel("b step:", this), 22, 0);
	L->addWidget(bstepE = new QLineEdit("-0.01", this), 22, 1);
	L->setRowMinimumHeight(23, 20);
	L->addWidget(OK, 24, 0);
	L->addWidget(Cancel, 24, 1);
	connect(TypeBox, SIGNAL(currentIndexChanged(int)), this, SLOT(typeChanged(int)));
	connect(Improve, SIGNAL(toggled(bool)), this, SLOT(improveChanged(bool)));
	connect(Add, SIGNAL(clicked()), this, SLOT(addCoeff()));
	connect(Remove, SIGNAL(clicked()), this, SLOT(removeCoeff()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
	improveChanged(improve);
	typeChanged(TypeBox->currentIndex());
}

void FitAnaPotDialog::addCoeff()
{
	QString C = CoeffBox->currentText();
	if (C[0] == 'c' || C[0] == 'C') C = C.right(C.length() - 1);
	int n, c = C.toInt();
	if (c>0)
	{
		C = 'C' + QString::number(c);
		for (n=0; (n < CoeffList->count() ? CoeffList->item(n)->text() != C : false); n++) ;
		if (n == CoeffList->count()) CoeffList->addItem(C);
	}
}

bool FitAnaPotDialog::exec(bool &improve, bool &useLeveMrq, bool &useSVD, double& Ri, double& Ra, double& iExp, int& NC, int& NLRC, 
						   int*& pLRC, double& Rmin, double& Rmax, int& NumP, double &bmin, double &bmax, double &bstep, bool &ATEWF,
						   double &hCi_cP, PotentialType &NewPotType)
{
	if (QDialog::exec() == QDialog::Rejected) return false;
	improve = Improve->isChecked();
	useLeveMrq = UseLeveMrq->isChecked();
	useSVD = UseSVD->isChecked();
	ATEWF = AdjustTEWF->isChecked();
	hCi_cP = hCi_cPE->text().toDouble();
	if (improve) return true;
	int n;
	QString C;
	Ri = RiE->text().toDouble();
	Ra = RaE->text().toDouble();
	iExp = iExpE->text().toDouble();
	NC = NCE->text().toDouble();
	pLRC = new int[NLRC = CoeffList->count()];
	for (n=0; n < NLRC; n++) 
	{
		C = CoeffList->item(n)->text();
		pLRC[n] = C.right(C.length() - 1).toInt();
	}
	Rmin = RminE->text().toDouble();
	Rmax = RmaxE->text().toDouble();
	NumP = NumPE->text().toInt();
	bmin = bminE->text().toDouble();
	bmax = bmaxE->text().toDouble();
	bstep = bstepE->text().toDouble();
	NewPotType = (TypeBox->currentIndex() == 0 ? SplinePotential : analyticalPotential);
	return true;
}

void FitAnaPotDialog::improveChanged(bool State)
{
	Add->setEnabled(!State);
	CoeffBox->setEnabled(!State);
	CoeffList->setEnabled(!State);
	NCE->setEnabled(!State);
	NumPE->setEnabled(!State);
	RaE->setEnabled(!State);
	Remove->setEnabled(!State);
	RiE->setEnabled(!State);
	RmaxE->setEnabled(!State);
	RminE->setEnabled(!State);
	iExpE->setEnabled(!State);
	bminE->setEnabled(!State);
	bmaxE->setEnabled(!State);
	bstepE->setEnabled(!State);
}

void FitAnaPotDialog::removeCoeff()
{
	QListWidgetItem *I = CoeffList->takeItem(CoeffList->currentRow());
	delete I;
}

void FitAnaPotDialog::typeChanged(int Index)
{
	if ((CurPotential == SplinePotential && Index == 0) || (CurPotential == analyticalPotential && Index == 1)) 
		Improve->setEnabled(true);
	else
	{
		if (Improve->isChecked()) Improve->setChecked(false);
		Improve->setEnabled(false);
	}
}


PotFit::PotFit()
{
	robustWeighting = false;
	UseSVD = true;
	threshhold = 1e-10;
	useLevenbergMarquardt = true;
	MaxIt = -1;
	adjustTEWeightFact = false;
	hCi_cP = 0.0;
	Worker = 0;
}

void PotFit::FitAnaPot(bool nrobustWeighting, bool nUseSVD, double nthreshhold, 
						 bool nuseLevenbergMarquardt, int nMaxIt, bool nadjustTEWeightFact, double nhCi_cP)
{
	robustWeighting = nrobustWeighting;
	UseSVD = nUseSVD;
	threshhold = nthreshhold;
	useLevenbergMarquardt = nuseLevenbergMarquardt;
	MaxIt = nMaxIt;
	adjustTEWeightFact = nadjustTEWeightFact;
	hCi_cP = nhCi_cP;
	FitPot = analyticalPotential;
	restart = false;
	ToDo = PotentialFit;
}

void PotFit::run()
{
    int NumWFPoints = NumPoints;
    setTerminationEnabled(true);
	if (Worker != 0) 
	{
		switch (ToDo)
		{
			case PotentialFit:
				switch (FitPot)
				{
					case analyticalPotential:
						Worker->FitAnaPot(robustWeighting, UseSVD, threshhold, useLevenbergMarquardt, MaxIt, adjustTEWeightFact, hCi_cP, restart);
						break;
					case SplinePotential:
						Worker->FitSplinePot(MaxIt, threshhold);
						break;
					default:
						printf("Potentials of type %d currently do not be fitted in a separate thread!", FitPot);
						break;
				}
				break;
			case MonteCarloSimulation:
				Worker->MonteCarloSimIteration(MCST, MCSUncFact);
				break;
			case calcFQS:
                Worker->getFQS(NumWFPoints, false, 0, true, SFQSRad);
				break;
		}
	}
}

void PotFit::setWorker(PotWorker* nWorker)
{
	Worker = nWorker;
}

PotFitLogWindow::PotFitLogWindow(QWidget*) : MDIChild()
{
	QGridLayout *L = new QGridLayout(this);
	fitFinished = fitRunning = fitStopped = false;
	Stop = new QPushButton("Stop fit", this);
	Cancel = new QPushButton("Cancel fit", this);
	Close = new QPushButton("Close window", this);
	Stop->setEnabled(false);
	Cancel->setEnabled(false);
	Close->setEnabled(false);
	L->addWidget(Stop, 0, 0);
	L->addWidget(Cancel, 0, 1);
	L->addWidget(Close, 0, 2);
	L->addWidget(Log = new QPlainTextEdit(this), 1, 0, 1, 3);
	Log->setReadOnly(true);
	L->setRowStretch(1, 1);
	connect(Stop, SIGNAL(clicked()), this, SLOT(stopFit()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(cancelFit()));
	connect(Close, SIGNAL(clicked()), this, SLOT(close()));
}

void PotFitLogWindow::addTextRow(QString Text)
{
	Log->appendPlainText(Text);
}

void PotFitLogWindow::cancelFit()
{
	if (QMessageBox::question(this, "MolSpektAnalysis", "Do you really want to kill the running fit?", 
							  QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes)
	{
		emit CancelFit(false);
		Stop->setEnabled(false);
		Cancel->setEnabled(false);
	}
}

bool PotFitLogWindow::askForQuit()
{
	if (!fitFinished && !fitRunning && !fitStopped)
	{
		QMessageBox::information(this, "MolSpektAnalysis", "The window cannot be closed while a fit is starting!");
		return false;
	}
	if (!fitRunning) return true;
	if (QMessageBox::information(this, "MolSpektAnalysis", "Closing this window kills the running fit!",
								 QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Ok)
	{
		emit CancelFit(true);
		Stop->setEnabled(false);
		Cancel->setEnabled(false);
		Close->setEnabled(false);
		Stop->setText("Stop fit");
		return true;
	}
	else return false;
}

void PotFitLogWindow::FitFinished()
{
	fitRunning = false;
	fitFinished = true;
	Stop->setText("Start fit");
	Cancel->setEnabled(false);
	Log->appendPlainText("Fit finished");
}

void PotFitLogWindow::FitStopped()
{
	fitRunning = false;
	fitStopped = true;
	Cancel->setEnabled(false);
	Stop->setEnabled(true);
	Log->appendPlainText("Fit stopped");
}

void PotFitLogWindow::FitRunning()
{
	if (fitStopped)
	{
		Log->appendPlainText("Fit restarted");
		fitStopped = false;
	}
	fitFinished = false;
	fitRunning = true;
	Stop->setEnabled(true);
	Cancel->setEnabled(true);
	Close->setEnabled(true);
}

void PotFitLogWindow::FitTerminated()
{
	fitRunning = false;
	fitFinished = true;
	Stop->setText("Start fit");
	Cancel->setEnabled(false);
	Log->appendPlainText("Fit terminated!");
}

void PotFitLogWindow::stopFit()
{
	if (Stop->text() == "Stop fit")
	{
		emit StopFit();
		Stop->setText("Restart fit");
	}
	else if (Stop->text() == "Restart fit")
	{
		emit RestartFit();
		Stop->setEnabled(false);
	}
	else 
	{
		emit StartFit();
		fitFinished = fitRunning = fitStopped = false;
		Stop->setEnabled(false);
		Cancel->setEnabled(false);
		Close->setEnabled(false);
		Stop->setText("Stop fit");
	}
}


ThreadControl::ThreadControl(MainWindow* mw, ElState* nSt, int NumIterations, int NumParFits)
{
	int n;
	MW = mw;
	St = nSt;
	MFD = 0;
	NFC = 0;
	AdTabRows = 0;
	NumIt = NumIterations;
	MMParF = MaxParF = NumParFits;
	NumParF = ItFinished = NumTermF = 0;
	NumResultTabColumns = 0;
	PotI = new int[MaxParF];
	Pots = new Potential*[MaxParF];
	for (n=0; n < MaxParF; n++)
	{
		PotI[n] = -1;
		Pots[n] = 0;
	}
	ItI = new int[NumIterations];
	for (n=0; n < NumIterations; n++) ItI[n] = -1;
	QGridLayout *L = new QGridLayout(this);
	L->addWidget(new QLabel("Number of iterations:", this), 0, 0);
	L->addWidget(NumItE = new QLineEdit(QString::number(NumIt), this), 0, 1);
	L->addWidget(new QLabel("Number of parallel fits:", this), 1, 0);
	L->addWidget(NumParFE = new QLineEdit(QString::number(MaxParF), this), 1, 1);
	L->setRowMinimumHeight(2, 20);
	L->addWidget(Progress = new QProgressBar(this), 3, 0, 1, 2);
	Progress->setTextVisible(true);
	Progress->setRange(0, NumIt);
	Progress->setValue(0);
	Progress->setFormat("%v iterations finished");
	NumItE->setValidator(NumItEValid = new QIntValidator(0, 100000, NumItE));
	NumParFE->setValidator(NumParFEValid = new QIntValidator(0, 1000, NumParFE));
	connect(NumItE, SIGNAL(editingFinished()), this, SLOT(NItChanged()));
	connect(NumParFE, SIGNAL(editingFinished()), this, SLOT(NParFChanged()));
	ResultTab = MW->createTableWindow();
	ResultTab->show();
}

ThreadControl::~ThreadControl()
{
	delete[] Pots;
	delete[] PotI;
	delete[] ItI;
	if (NFC > 0) delete[] FCRows;
}

void ThreadControl::CalcFinished(int ThreadNum)
{
	PotI[ThreadNum] = -1;
	NumParF--;
	NumTermF++;
	Progress->setValue(++ItFinished);
	if (ItFinished + NumParF < NumIt && NumParF < MaxParF) StartCalcs();
	else if (ItFinished == NumIt) AllCalculationsFinished();
}

void ThreadControl::AllCalculationsFinished()
{
	QMessageBox::information(this, "MolSpektAnalysis", "The calculations are finished.");
}

void ThreadControl::CalcTerminated(int ThreadNum)
{
	MaxParF--;
	NumParF--;
	NumTermF++;
	ItI[PotI[ThreadNum]] = -1;
	PotI[ThreadNum] = -1;
	NumParFE->setText(QString::number(MaxParF));
}

void ThreadControl::closeEvent(QCloseEvent* event)
{
    if (NumParF > 0 ? QMessageBox::information(this, "MolSpektAnalysis", 
			"Closing this window will make you unable to directly control the running calculations", 
			QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel : false) 
		event->ignore();
	else 
	{
		QString Name;
		int i, j, N;
		for (i=0; i < MMParF; i++)
		{
			if (Pots[i] == 0) continue;
			for (j=0, N = St->getNumPotentials(), Name = Pots[i]->getName(); j<N; j++)
				if (St->getPotentialName(j) == Name) St->removePotential(j);
		}
		event->accept();
	}
}

void ThreadControl::NItChanged()
{
	int n, nIt = NumItE->text().toInt(), *nItI;
	ResultTab->setTabDimensions(nIt + AdTabRows, NumResultTabColumns);
	if (nIt > NumIt)
	{
		nItI = new int[nIt];
		for (n=0; n < NumIt; n++) nItI[n] = ItI[n];
		while (n < nIt) nItI[n++] = -1;
		delete[] ItI;
		ItI = nItI;
	}
	NumIt = nIt;
	Progress->setMaximum(NumIt);
	if (NumParF < MaxParF && ItFinished + NumParF < NumIt) StartCalcs(); 
}

void ThreadControl::NParFChanged()
{
	MaxParF = NumParFE->text().toInt();
	if (MaxParF > MMParF)
	{
		Potential** nPots = new Potential*[MaxParF];
		int n, *nPotI = new int[MaxParF];
		for (n=0; n < MMParF; n++)
		{
			nPots[n] = Pots[n];
			nPotI[n] = PotI[n];
		}
		while (n < MaxParF)
		{
			nPots[n] = 0;
			nPotI[n++] = -1;
		}
		delete[] Pots;
		delete[] PotI;
		Pots = nPots;
		PotI = nPotI;
		MMParF = MaxParF;
	}
	if (NumParF < MaxParF && ItFinished + NumParF < NumIt) StartCalcs(); 
}

void ThreadControl::StartCalcs()
{
	int n, p;
	while (NumParF < MaxParF && ItFinished + NumParF < NumIt)
	{
		for (p=0; PotI[p] != -1; p++) ;
		for (n=0; ItI[n] != -1; n++) ;
		ItI[n] = p;
		PotI[p] = n;
		StartCalc(p, n);
		if (Pots[p] == 0) 
		{
			MaxParF = p;
			NumParFE->setText(QString::number(p));
			break;
		}
		NumParF++;
	}
	NumItEValid->setBottom(ItFinished + NumParF);
}

void ThreadControl::StartCalc(int, int)
{

}


MonteCarloSimControl::MonteCarloSimControl(MainWindow* mw, Potential* Pot, QString PotDir, int NumIterations, int NumParFits, 
                                           double UncFact, int NumWFPoints) : ThreadControl(mw, Pot->getElState(), NumIterations, NumParFits)
{
	int n;
	double X, S, SF = 32768 / 2506.62827;
	QStringList HL;
	setWindowTitle("MCS Control");
	UncF = UncFact;
    Pot->getFQS(NumWFPoints);
	MFD = St->getFitData();
	MCSFD = MW->CreateFitData();
	MCSFD->setName("MCSFD_" + Pot->getName());
	MCSFD->setSource(MFD->getName() + " from " + MFD->getSource());
	MCSFD->setvMax(MFD->getvMax());
	MCSFD->setJMax(MFD->getJMax());
	St->setMainFitData(MCSFD);
	PotName = Pot->getName();
	PotFileName = Pot->getFName();
	ResultDir = PotDir;
	NumPDRows = NumPDColumns = 0;
	ResultTab->setWindowTitle("Results Monte Carlo sim " + Pot->getName());
	PotData = Pot->getData(NumPDRows, NumPDColumns);
	Pot->getFixedCoefficients(NFC, FCRows);
	ResultTab->setTabDimensions(NumIt, NumResultTabColumns = NumPDRows + 1);
	for (n=0; n < NumPDRows; n++) HL << PotData[n][0];
	HL << "FQS";
	ResultTab->setHorizontalHeader(HL);
	MCSTab = new double[32768];
	for (n=1, X = -10.0, S = 0.0; X<=0.0; X += 0.001)
	{
		S += SF * exp(-0.5*X*X);
		while (int(S) < n)
		{
			X += 0.001;
			S += SF * exp(-0.5*X*X);
		}
		while (n <= int(S)) MCSTab[n++] = X;
	}
	for (MCSTab[0] = MCSTab[1] - 1.0 ; n < 32768; n++) MCSTab[n] = -MCSTab[32767 - n];
	StartCalcs();
}

MonteCarloSimControl::~MonteCarloSimControl()
{
	Destroy(PotData, NumPDRows);
	delete[] MCSTab;
}

void MonteCarloSimControl::FitFinished(int ThreadNum, double FQS)
{
	QString File = ResultDir + PotFileName.left(PotFileName.lastIndexOf('.')) + "MCS" + QString::number(PotI[ThreadNum]) + ".pot";
	int NumR = 0, NumC = 0, n;
	QString **ResData = Pots[ThreadNum]->getData(NumR, NumC), RData[NumPDRows + 1];
	for (n=0; n < NumR; n++) RData[n] = ResData[n][1];
	RData[n] = QString::number(FQS, 'f', 6);
	ResultTab->setRowData(PotI[ThreadNum], RData);
	Pots[ThreadNum]->writeData(File);
	Destroy(ResData, NumR);
	CalcFinished(ThreadNum);
}

void MonteCarloSimControl::StartCalc(int p, int n)
{
    int NL; // i=0, j=0;
	TableLine *Lines;
	if (Pots[p] == 0)
	{
		Pots[p] = MW->CreatePotential(p);
		if (Pots[p] == 0) return;
		St->addPotential(Pots[p]);
		Pots[p]->show();
		connect(Pots[p], SIGNAL(FitFinished(int,double)), this, SLOT(FitFinished(int,double)));
		connect(Pots[p], SIGNAL(FitTerminated(int)), this, SLOT(CalcTerminated(int)));
	}
	Pots[p]->setName(PotName + "MCS" + QString::number(n));
		/*FileName = File + QString::number(n) + ".pot";
		Test.setFileName(FileName);
		if (Test.exists())
		{
			Pots[p]->readData(FileName);
			if (FD == 0) 
			{
				FD = new FitData;
				St->setMainFitData(FD);
			}
			FD->readData(File + QString::number(n) + ".fdat");
			Pots[p]->CalcFQS();
			FD->getData(Lines, NL);
			for (i=0, FQS = 0.0; i < NL; i++) FQS += Lines[i].DevR * Lines[i].DevR;
			delete[] Lines;
			FQS = Pots[p]->getFQS();
			QString RData[NumPDRows + 1], **ResData = Pots[p]->getData(i, j);
			for (i=0; i < NumPDRows; i++) RData[i] = ResData[i][1];
			RData[i] = QString::number(FQS, 'f', 6);
			Destroy(ResData, i);
			ResultTab->setRowData(n, RData);
			Progress->setValue(++ItFinished);
			continue;
		}
		else 
		{
			St->setMainFitData(MFDN);*/
	MFD->getData(Lines, NL);
	MCSFD->setData(Lines, NL);
	Pots[p]->setData(PotData, NumPDRows, NumPDColumns);
	if (NFC > 0) Pots[p]->VaryCoefficients(false, NFC, FCRows);
	Pots[p]->MonteCarloSimIteration(MCSTab, UncF);
}


SFQSCalcControl::SFQSCalcControl(MainWindow* MW, ElState* St, Potential *Pot, QString PotDir, QString FDatDir, double SFQSRad, int NumParFits)
					: ThreadControl(MW, St, 0, NumParFits)
{
	QDir Dir(PDir = PotDir);
	int NRows = 0, NCols = 0, c;
	QString **PotData = Pot->getData(NRows, NCols);
	QStringList HeaderLabels;
	setWindowTitle("FQS Calc Control");
	OPot = Pot;
	FDir = FDatDir;
	Rad = SFQSRad;
	SFQSU = 0;
	AdTabRows = 1;
	if (PDir.right(1) != "\\" && PDir.right(1) != "/") PDir += DIRSEP;
	PotList = Dir.entryList(QStringList("*.pot"), QDir::Files);
	if (!FDir.isEmpty())
	{
		if (FDir.right(1) != "\\" && FDir.right(1) != "/") FDir += DIRSEP;
		MFD = MW->CreateFitData();
		MFD->setName("MCSFDat");
	}
	ResultTab->setWindowTitle("Results Monte Carlo sim " + Pot->getName());
	for (c=0; c < NRows; c++) HeaderLabels << PotData[c][0];
	HeaderLabels << "SFQS" << "FQS";
	ResultTab->setTabDimensions(PotList.count() + 1, NumResultTabColumns = NRows + 2);
	ResultTab->setHorizontalHeader(HeaderLabels);
	Destroy(PotData, NRows);
	connect(Pot, SIGNAL(SFQScalculated(int,double,double)), this, SLOT(CalcFinished(int,double,double)));
	Pot->calcFQS_SFQS(0, SFQSRad);
}

SFQSCalcControl::~SFQSCalcControl()
{
	int I, c, J, NI = OPot->getIsoT()->numIso, NC = int(St->getS() * 2.0) + 1;
	if (SFQSU != 0)
	{
		for (c=0; c < NC; c++)
		{
			for (I=0; I < NI; I++)
			{
				for (J=0; J < mJ[I]; J++) delete[] SFQSU[c][I][J];
				delete[] SFQSU[c][I];
			}
			delete[] SFQSU[c];
		}
		delete[] SFQSU;
		for (I=0; I < NI; I++) delete[] mv[I];
		delete[] mv;
		delete[] mJ;
	}
}

void SFQSCalcControl::CalcFinished(int ThreadNum, double SFQS, double FQS)
{
	int n, r = (ThreadNum >= 0 ? PotI[ThreadNum] + 1 : 0), NRows = 0, NCols = 0;
	Potential *Pot = (ThreadNum >= 0 ? Pots[ThreadNum] : OPot);
	QString **PotData = Pot->getData(NRows, NCols), Res[NRows + 2];
	for (n=0; n < NRows; n++) Res[n] = PotData[n][1];
	Res[n] = QString::number(SFQS, 'g', 10);
	Res[n+1] = QString::number(FQS, 'g', 12);
	ResultTab->setRowData(r, Res);
	Destroy(PotData, NRows);
	if (ThreadNum >= 0) ThreadControl::CalcFinished(ThreadNum);
	else 
	{
		Pot->getSFQSU(SFQSU, mJ, mv);
		NumItE->setText(QString::number(PotList.count()));
		NItChanged();
	}
}

void SFQSCalcControl::StartCalc(int p, int n)
{
    if (Pots[p] == 0)
	{
		Pots[p] = MW->CreatePotential(p);
		Pots[p]->show();
		connect(Pots[p], SIGNAL(FitTerminated(int)), this, SLOT(CalcTerminated(int)));
		connect(Pots[p], SIGNAL(SFQScalculated(int,double,double)), this, SLOT(CalcFinished(int,double,double)));
	}
	St->removePotential(Pots[p]);
	Pots[p]->readData(PDir + PotList[n]);
	if (!FDir.isEmpty())
	{
		MFD->readData(FDir + PotList[n].left(PotList[n].length() - 3) + "fdat");
		if (St->getFitData() != MFD) St->setMainFitData(MFD);
	}
	St->addPotential(Pots[p]);
	Pots[p]->calcFQS_SFQS(SFQSU);
}


SFQSCalcDialog::SFQSCalcDialog(MainWindow* mw, QString Dir) : QDialog(mw)
{
	int n, N = (MW = mw)->getNumMolecules();
	setWindowTitle("Settings for calculation of FQS series:");
	QGridLayout *L = new QGridLayout(this);
	L->addWidget(new QLabel("Molecule:", this), 0, 0);
	L->addWidget(MolBox = new QComboBox(this), 0, 1, 1, 2);
	for (n=0; n<N; n++) MolBox->addItem(MW->getMolecule(n)->getName());
	MolBox->setEditable(false);
	L->addWidget(new QLabel("El. State:", this), 1, 0);
	L->addWidget(StateBox = new QComboBox(this), 1, 1, 1, 2);
	StateBox->setEditable(false);
	L->addWidget(new QLabel("Origin pot. of series:", this), 2, 0);
	L->addWidget(PotBox = new QComboBox(this), 2, 1, 1, 2);
	PotBox->setEditable(false);
	L->addWidget(new QLabel("Dir of potentials:", this), 3, 0);
	L->addWidget(PotDirE = new QLineEdit(Dir, this), 3, 1);
	L->addWidget(sPotDirB = new QPushButton("...", this), 3, 2);
	L->addWidget(new QLabel("Dir of fit datasets:", this), 4, 0);
	L->addWidget(FDatDirE = new QLineEdit(Dir, this), 4, 1);
	L->addWidget(sFDatDirB = new QPushButton("...", this), 4, 2);
	L->addWidget(new QLabel("Out.Turn.Point f. SFQS:", this), 5, 0);
	L->addWidget(SFQSRadE = new QLineEdit("12.0", this), 5, 1, 1, 2);
	L->addWidget(new QLabel("Num. paralell calc.:", this), 6, 0);
	L->addWidget(NumParFitsE = new QLineEdit("8", this), 6, 1, 1, 2);
	SFQSRadE->setValidator(new QDoubleValidator(SFQSRadE));
	NumParFitsE->setValidator(new QIntValidator(1, 100, NumParFitsE));
	L->setRowMinimumHeight(7, 20);
	L->addWidget(OK = new QPushButton("OK", this), 8, 0);
	L->addWidget(Cancel = new QPushButton("Cancel", this), 8, 1, 1, 2);
	connect(MolBox, SIGNAL(currentIndexChanged(int)), this, SLOT(MolChanged(int)));
	connect(StateBox, SIGNAL(currentIndexChanged(int)), this, SLOT(StateChanged(int)));
	connect(sPotDirB, SIGNAL(clicked()), this, SLOT(sPotDir()));
	connect(sFDatDirB, SIGNAL(clicked()), this, SLOT(sFDatDir()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
	MolChanged(0);
}

void SFQSCalcDialog::exec(ElState*& St, Potential*& OPot, QString& PotDir, QString& FDatDir, double& SFQSRad, int& NumParFits)
{
	if (QDialog::exec() == QDialog::Accepted)
	{
		St = MW->getMolecule(MolBox->currentIndex())->getStateP(StateBox->currentIndex());
		OPot = St->getPotential(PotBox->currentIndex());
		PotDir = PotDirE->text();
		FDatDir = FDatDirE->text();
		SFQSRad = SFQSRadE->text().toDouble();
		NumParFits = NumParFitsE->text().toInt();
	}
	else
	{
		St = 0;
		SFQSRad = 0.0;
		NumParFits = 0;
	}
}

void SFQSCalcDialog::MolChanged(int n)
{
	Molecule *Mol = MW->getMolecule(n);
	int N = Mol->getNumStates();
	StateBox->clear();
	for (n=0; n<N; n++) StateBox->addItem(Mol->getState(n));
	StateChanged(0);
}

void SFQSCalcDialog::sFDatDir()
{
	QString Res = QFileDialog::getExistingDirectory(this, "MolSpektAnalysis", FDatDirE->text());
	if (!Res.isEmpty()) FDatDirE->setText(Res);
}

void SFQSCalcDialog::sPotDir()
{
	QString Res = QFileDialog::getExistingDirectory(this, "MolSpektAnalysis", PotDirE->text());
	if (!Res.isEmpty()) PotDirE->setText(Res);
}

void SFQSCalcDialog::StateChanged(int n)
{
	ElState *St = (n>=0 ? MW->getMolecule(MolBox->currentIndex())->getStateP(n) : 0);
	PotBox->clear();
	if (St != 0) for (n=0; n < St->getNumPotentials(); n++) PotBox->addItem(St->getPotentialName(n));
}


SplinePot::SplinePot(PotFit* Fit): PotWorker(Fit, SplinePotential)
{

}

SplinePot::SplinePot(const SplinePot& C): PotWorker(C)
{

}


CreateAnaPotSeriesControl::CreateAnaPotSeriesControl(MainWindow* MW, ElState* St, Potential* OPot, FitData* fDat, QString SPDir, QString FDDir, QString APDir, 
													 double Threshold, int NumParFits, int maxFitIterations, int nStart, bool ImproveAnaPots, bool useSvd, bool useLeveMarq) 
															: ImprovePotSeriesControl(MW, St, OPot, fDat, SPDir, Threshold, 0)
{
	setWindowTitle("Create AnaPot Series Control");
	ResultTab->setWindowTitle("Results of creating of AnaPot series");
	FitDataDir = FDDir;
	ResultPotDir = APDir;
	MaxFitIterations = maxFitIterations;
	NStart = nStart;
	improveAnaPots = ImproveAnaPots;
	UseSvd = useSvd;
	UseLeveMarq = useLeveMarq;
	FDat = fDat;
	PotN = St->getNumPotentials();
	FDN = St->getNumFitDataSets();
	MFD = St->getFitDataFileName();
	if (improveAnaPots) SPDir = APDir;
	FDat = (FDDir.isEmpty() ? 0 : MW->CreateFitData());
	if (APDir.right(1) != "/" && APDir.right(1) != "\\") APDir += '/';
	if (FDat != 0 && FDDir.right(1) != "/" && FDDir.right(1) != "\\") FDDir += '/';
	LogFile = new QFile(APDir + "CreateAnaPots.log");
	LogFile->open(QIODevice::WriteOnly);
	LogStream = new QTextStream(LogFile);
	*LogStream << "CreateAnaPotSeriesFromMCSSplinePotSeries(\"" << St->getMolecule()->getFileName() << "\", \"" << St->getName() << "\", \"" 
			   << FDDir << "\", \"" << SPDir << "\", \"" << APDir << (improveAnaPots ? "\", true, " : "\", false, ")
			   << (UseSvd ? "true, " : "false, ") << (UseLeveMarq ? "true, " : "false , ") << QString::number(MaxFitIterations) 
			   << ", " << QString::number(Threshold, 'g', 5) << ")\n";
	LogStream->flush();
	NumParFE->setText(QString::number(NumParFits));
	NParFChanged();
}

CreateAnaPotSeriesControl::~CreateAnaPotSeriesControl()
{
	delete LogStream;
	delete LogFile;
}

void CreateAnaPotSeriesControl::AllCalculationsFinished()
{
    *LogStream << "Series finished!!!\n";
	St->removePotential(PotN);
	if (FDat != 0 && (UseLeveMarq || UseSvd))
	{
		St->setMainFitData(MFD);
		St->removeFitData(FDN);
	}
	ThreadControl::AllCalculationsFinished();
}

QString CreateAnaPotSeriesControl::getWriteFileName(QString CurrentFileName)
{
    return ImprovePotSeriesControl::getWriteFileName(CurrentFileName);
}

void CreateAnaPotSeriesControl::StartCalc(int p, int n)
{
    PreparePot(p, n);
	QDir SplinePotDir(PotDir);
	QString SFile, CFile = SplinePotDir.absoluteFilePath(PotList[n]);
	if (!improveAnaPots)
	{
		SFile = ResultPotDir + PotList[n].left(PotList[n].lastIndexOf('.')) + "AnaPot.pot";
		QFile File(SFile);
		if (File.exists()) 
		{
			CalcFinished(p);
			return;
		}
	}
	if (improveAnaPots) SFile = CFile;
	printf("Potential%d: %s\n", n, CFile.toLatin1().data());
	if (FDat != 0 && (UseLeveMarq || UseSvd))
	{
		if (improveAnaPots)
		{
			CFile = FitDataDir + PotList[n].left(PotList[n].indexOf("AnaPot.pot")) + ".fdat";
			QFile File(CFile);
			if (!File.exists()) CFile = FitDataDir + PotList[n].left(PotList[n].lastIndexOf('.') + 1) + "fdat";
		}
		else CFile = FitDataDir + PotList[n].left(PotList[n].lastIndexOf('.') + 1) + "fdat";
		if (!FDat->readData(CFile))
		{
			*LogStream << "Error reading file \"" + CFile + "\"!";
			CalcFinished(p);
			return;
		}
		if (St->getFitData() != FDat) St->setMainFitData(FDat);
	}
	if (!improveAnaPots)
	{
		Pots[p]->setFileName(SFile);
		Pots[p]->setName(Pots[p]->getName() + "AnaPot");
		Pots[p]->setSource("SplinePotential from " + Pots[p]->getSource());
		Pots[p]->FitAnaPot(false, false, UseSvd, threshold, true, false, UseLeveMarq);
	}
	else if (UseLeveMarq || UseSvd) Pots[p]->FitAnaPot(false, false, UseSvd, threshold, false, false, UseLeveMarq, MaxFitIterations);
}

void CreateAnaPotSeriesControl::CalcFinished(int ThreadNum)
{
    QString SFile = Pots[ThreadNum]->getFileName();
	int n = ItFinished;
	ThreadControl::CalcFinished(ThreadNum);
	*LogStream << "Run" << QString::number(n) << " with file \"" << SFile << "\" finished!\n";
	LogStream->flush();
}


ImprovePotSeriesControl::ImprovePotSeriesControl(MainWindow* MW, ElState* St, Potential* OPot, FitData* FDat, QString potDir, 
												 double Threshold, int NumParFits) : ThreadControl(MW, St, 0, NumParFits)
{
	QDir Dir(PotDir = potDir);
	int NRows = 0, NCols = 0, c;
	threshold = Threshold;
	if (PotDir.right(1) != "\\" && PotDir.right(1) != "/") PotDir += DIRSEP;
	PotList = Dir.entryList(QStringList("*.pot"), QDir::Files);
	if (PotList.count() == 0)
	{
		CalcFinished(0);
		return;
	}
	if (OPot != 0) OPot->getFixedCoefficients(NFC, FCRows);
	else
	{
		OPot = Pots[0] = MW->CreatePotential(0);
		OPot->readData(PotDir + PotList[0]);
		Pots[0]->show();
		connect(Pots[0], SIGNAL(FitTerminated(int)), this, SLOT(CalcTerminated(int)));
		connect(Pots[0], SIGNAL(FitFinished(int, double)), this, SLOT(FitFinished(int,double)));
	}
	UseLeveMarq = OPot->getPotType() != SplinePotential;
	QString **PotData = OPot->getData(NRows, NCols);
	QStringList HeaderLabels;
	setWindowTitle("Improve Potential Series Control");
	AdTabRows = 0;
	ResultTab->setWindowTitle("Results of Potential Series Improvement");
	for (c=0; c < NRows; c++) HeaderLabels << PotData[c][0];
	HeaderLabels << "FQS" << "FQS_Bad" << "NBad" << "NBadPAL" << "Sigma";
	ResultTab->setTabDimensions(PotList.count() + 1, NumResultTabColumns = NRows + 5);
	ResultTab->setHorizontalHeader(HeaderLabels);
	Destroy(PotData, NRows);
	if (St->getFitData() != FDat) St->setMainFitData(FDat);
	NumItE->setText(QString::number(PotList.count()));
	NItChanged();
}

void ImprovePotSeriesControl::FitFinished(int ThreadNum, double FQS)
{
	int n, r = PotI[ThreadNum], NRows = 0, NCols = 0;
	Potential *Pot = Pots[ThreadNum];
	QString **PotData = Pot->getData(NRows, NCols), Res[NRows + 5], FileName = Pot->getFileName();
	FitResult FRes = Pot->getFitResult();
	for (n=0; n < NRows; n++) Res[n] = PotData[n][1];
	Res[n] = QString::number(FQS, 'g', 10);
	Res[n+1] = QString::number(FRes.FQS_Bad, 'g', 10);
	Res[n+2] = QString::number(FRes.NBad);
	Res[n+3] = QString::number(FRes.NBadPAL);
	Res[n+4] = QString::number(FRes.Sigma, 'g', 10);
	ResultTab->setRowData(r, Res);
	Destroy(PotData, NRows);
	Pot->writeData(getWriteFileName(FileName));
	CalcFinished(ThreadNum);
}

QString ImprovePotSeriesControl::getWriteFileName(QString CurrentFileName)
{
	return CurrentFileName.left(CurrentFileName.length() - 4) + "_i.pot";
}

void ImprovePotSeriesControl::StartCalc(int p, int n)
{
    PreparePot(p, n);
	Pots[p]->FitAnaPot(false, false, false, threshold, false, false, true, -1, false, 0.0, true);
}

void ImprovePotSeriesControl::PreparePot(int p, int n)
{
	if (Pots[p] == 0)
	{
		Pots[p] = MW->CreatePotential(p);
		Pots[p]->show();
		connect(Pots[p], SIGNAL(FitTerminated(int)), this, SLOT(CalcTerminated(int)));
		connect(Pots[p], SIGNAL(FitFinished(int,double)), this, SLOT(FitFinished(int,double)));
	}
	St->removePotential(Pots[p]);
	Pots[p]->readData(PotDir + PotList[n]);
	if (NFC > 0) Pots[p]->VaryCoefficients(false, NFC, FCRows);
	St->addPotential(Pots[p]);
}


ImprovePotSeriesDialog::ImprovePotSeriesDialog(MainWindow* MW, QString Dir, Potential* OPot): QDialog()
{
	QGridLayout *L = new QGridLayout(this);
	QPushButton *sPotDirB = new QPushButton("...", this), *OK = new QPushButton("OK", this), *Cancel = new QPushButton("Cancel", this);
	QDoubleValidator *ThV;
	Molecule *TM;
	int n, N = (mw = MW)->getNumMolecules();
	Mol = (OPot != 0 ? OPot->getMolecule() : 0);
	St = (OPot != 0 ? OPot->getElState() : 0);
	setWindowTitle("Settings for Improvement of Pot Series");
	L->addWidget(new QLabel("Molecule:", this), 0, 0);
	L->addWidget(MoleculeB = new QComboBox(this), 0, 1, 1, 2);
	for (n=0; n<N; n++) 
	{
		MoleculeB->addItem((TM = MW->getMolecule(n))->getName());
		if (Mol == TM) MoleculeB->setCurrentIndex(n);
	}
	MoleculeB->setEditable(false);
	L->addWidget(new QLabel("Electronic State:", this), 1, 0);
	L->addWidget(StateB = new QComboBox(this), 1, 1, 1, 2);
	StateB->setEditable(false);
	L->addWidget(new QLabel("Fit Data set:", this), 2, 0);
	L->addWidget(FitDataB = new QComboBox(this), 2, 1, 1, 2);
	FitDataB->setEditable(false);
	L->addWidget(new QLabel("Directory of Potentials:"), 3, 0);
	L->addWidget(PotDirE = new QLineEdit(Dir, this), 3, 1);
	L->addWidget(sPotDirB, 3, 2);
	L->addWidget(new QLabel("Threshol:", this), 4, 0);
	L->addWidget(ThresholdE = new QLineEdit((OPot != 0 ? (OPot->getPotType() == SplinePotential ? "0.0" : "1.0") : "1.0"), this), 
				 4, 1, 1, 2);
	ThresholdE->setValidator(ThV = new QDoubleValidator(ThresholdE));
	ThV->setBottom(0.0);
	L->addWidget(new QLabel("Number of parallel fits:", this), 5, 0);
	L->addWidget(NumParFitsE = new QLineEdit("8", this), 5, 1, 1, 2);
	NumParFitsE->setValidator(new QIntValidator(1, 100, NumParFitsE));
	L->setRowMinimumHeight(6, 20);
	L->addWidget(OK, 7, 0);
	L->addWidget(Cancel, 7, 1, 1, 2);
	connect(MoleculeB, SIGNAL(currentIndexChanged(int)), this, SLOT(MolChanged(int)));
	connect(StateB, SIGNAL(currentIndexChanged(int)), this, SLOT(StateChanged(int)));
	connect(sPotDirB, SIGNAL(clicked()), this, SLOT(sPotDir()));
	connect(OK, SIGNAL(clicked()), this, SLOT(accept()));
	connect(Cancel, SIGNAL(clicked()), this, SLOT(reject()));
	MolChanged(0);
}

bool ImprovePotSeriesDialog::exec(ElState*& St, FitData*& FDat, QString& PotDir, double& Threshold, int &NumParFits)
{
	if (QDialog::exec() == QDialog::Rejected) return false;
	Molecule *Mol = mw->getMolecule(MoleculeB->currentIndex());
	St = Mol->getStateP(StateB->currentIndex());
	if (St == 0)
	{
		QMessageBox::information(this, "MolSpektAnalysis", "You have to select a molecule where fit data is available!");
		return exec(St, FDat, PotDir, Threshold, NumParFits);
	}
	FDat = St->getFitData(FitDataB->currentIndex());
	if (FDat == 0)
	{
		QMessageBox::information(this, "MolSpektAnalysis", "You have to select an electronic state where fit data is available!");
		return exec(St, FDat, PotDir, Threshold, NumParFits);
	}
	QDir D(PotDir = PotDirE->text());
	QStringList L = D.entryList(QStringList("*.pot"), QDir::Files);
	if (L.count() == 0)
	{
		QMessageBox::information(this, "MolSpektAnalysis", "You have to select a containing potential files (\"*.pot\")!");
		return exec(St, FDat, PotDir, Threshold, NumParFits);
	}
	Threshold = ThresholdE->text().toDouble();
	NumParFits = NumParFitsE->text().toInt();
	return true;
}

void ImprovePotSeriesDialog::MolChanged(int i)
{
	Mol = mw->getMolecule(i);
	int n, N = Mol->getNumStates();
	StateB->clear();
	for (n=0; n<N; n++) 
	{
		StateB->addItem(Mol->getState(n));
		if (Mol->getStateP(n) == St) StateB->setCurrentIndex(n);
	}
}

void ImprovePotSeriesDialog::sPotDir()
{
	QString Dir = QFileDialog::getExistingDirectory(this, "Select a directory containing potentials (\"*.pot\")", PotDirE->text());
	if (!Dir.isEmpty()) PotDirE->setText(Dir);
}

void ImprovePotSeriesDialog::StateChanged(int i)
{
	St = Mol->getStateP(i);
	FitDataB->clear();
	if (St != 0)
	{
		int n, N = St->getNumFitDataSets();
		for (n=0; n<N; n++) FitDataB->addItem(St->getFitDataName(n));
		FitDataB->setCurrentIndex(St->getMainFitDataNum());
	}
}


PotentialData::PotentialData()
{
	nPotCoeff = nLRCoeff = 0;
    pLRCoeff = 0;
	NAdCorr = TAdCorr = PAdCorr = NSpinRGamma = 0;
	Ri = Ra = Ro = iExp = iOff = iCoeff = b = Rm = Tm = De = 0.0;
	PotCoeff = LRCoeff = 0;
	A_ex = beta = gamma = q_e = q_f = 0.0;
	adCorr = 0;
	RIso1AdCorr = RIso2AdCorr = 0.0;
	SpinRGamma = 0;
	SpinRR1 = SpinRR2 = 0.0;
	CFree = LRCFree = 0;
	EIFree = true;
	adCorrFree = 0;
	shortRVariable = true;
	A = C = 0;
	Uinf = iA = iO = alpha = 0.0;
	LRC = 0;
	Salpha = Sbeta = 0.0;
	SA = SLRC = 0;
	Sinf = B = alpha1 = 0.0;
	NA = N = NLRC = 0;
	PLRC = 0;
	p = q = NCi = Nb = 0;
	pCi = 0;
	pLRC = 0;
	NSA = NSLRC = 0;
	pSLRC = 0;
	bA = Ci = 0;
	Re = Rref = T = De = 0.0;
	AdCorr_b = AdCorr_Rm = 0.0;
	points = 0;
	PotType = NoPotential;
}

PotentialData::~PotentialData()
{
	if (pLRCoeff != 0) delete[] pLRCoeff;
	if (PotCoeff != 0) delete[] PotCoeff;
	if (LRCoeff != 0) delete[] LRCoeff;
	if (adCorr != 0) delete[] adCorr;
	if (SpinRGamma != 0) delete[] SpinRGamma;
	if (CFree != 0) delete[] CFree;
	if (LRCFree != 0) delete[] LRCFree;
	if (adCorrFree != 0) delete[] adCorrFree;
	if (A != 0) delete[] A;
	if (C != 0) delete[] C;
	if (LRC != 0) delete[] LRC;
	if (SA != 0) delete[] SA;
	if (SLRC != 0) delete[] SLRC;
	if (PLRC != 0) delete[] PLRC;
	if (pCi != 0) delete[] pCi;
	if (pLRC != 0) delete[] pLRC;
	if (pSLRC != 0) delete[] pSLRC;
	if (bA != 0) delete[] bA;
	if (Ci != 0) delete[] Ci;
	if (points !=0 ) delete[] points;
}

